/*
   common memory allocating functions:
    - any allocation rounded up to 64k internally
    - (re)allocated size is zero filled
    - incorrect block structures will cause immediate panic
    - all information stored in a single block at the start of the working area
      (previous variant used block memory to store free block headers - this
       allowed to catch overwrites, but incompatible with paging)
*/
#include "qslocal.h"

// memtable & memftable - arrays with 64k step
u32t                *memtable; // size of block if used block starts here
                               // FFFFsize - free block size (*64k)
                               // zero     - in the middle of free block
                               // -1       - in the middle of allocated block
u8t               *memheapres; // bit array - r/o blocks, available to heap mgr
process_context    **memowner; // pointers to process cintext or 0
u16t               *memftable; // free block lists (block size is index)
free_block          *memflink; /* prev/next in free list (for each block),
                                  4 chars info signature - for used blocks */
u32t                *memvaddr;

extern u16t         headblock;
extern u16t   physmem_entries; // number of entries in physmem
extern physmem_block  physmem[PHYSMEM_TABLE_SIZE];

static u32t          availmem; // total used memory (including arrays & holes)
u32t                memblocks; // the same in 64k blocks
extern u32t         phmembase; // base address of used memory

void* memreserve(u32t physaddr, u32t length, u32t *reason, int ro);
void  hlp_memprint(printf_function pfn);

#define RESV_SIGN (MAKEID4('r','e','s','v'))

void memmgr_init(void) {
   u32t     ii, sz;

   availmem   = physmem[physmem_entries-1].startaddr +
                physmem[physmem_entries-1].blocklen - phmembase & 0xFFFF0000;
   memblocks  = availmem>>16;

   memtable   = (u32t*)phmembase;
   memftable  = (u16t*)(memtable + memblocks);
   memflink   = (free_block*)(memftable + memblocks);
   memowner   = (process_context**)(memflink + memblocks);
   memheapres = (u8t*)(memowner + memblocks);
   memvaddr   = 0;
   // memtable + memowner + memflink + memftable + memheapres
   sz = memblocks*(4*2+sizeof(free_block)+2) + (memblocks>>3);
   // init memory tables
   memset(memtable,0,sz);
   // save own tables as used block (free lists assume block 0 as never free)
   memtable[0] = sz;
   sz = Round64k(sz);
   ii = sz>>16;
   while (--ii) memtable[ii] = FFFF;
   // create one free memory entry (links zeroed by memset above)
   ii = availmem-sz>>16;
   memftable[ii] = sz>>=16;
   memtable [sz] = FMARK|ii;
   // mark block as r/o
   memheapres[0] = 1;
   // information string
   memflink[0].sign = 0x2A2A2A; // ***
   // fixups (call uses hlp_memgetsize(), so tables should be initilized first)
   push_fixup(&memtable, 4, &memftable, &memflink, &memowner, &memheapres);
   /* mark all "holes" in memory as r/o reserved blocks.
      arrays above are safe because meminit.c checks for at least 8Mb for us */
   for (ii=headblock+1; ii<physmem_entries; ii++) {
      u32t epos = physmem[ii-1].startaddr + physmem[ii-1].blocklen,
           bpos = physmem[ii].startaddr;
      bpos  = Round64k(bpos);
      epos &= 0xFFFF0000;
      if (bpos>epos) memreserve(epos, bpos-epos, 0, 1);
   }
}

static void halterr(u32t type, u32t addr, u32t info, char here, u32t caller) {
   mt_swlock();
   hlp_memprint(0);
   if (!mod_secondary) exit_pm32(QERR_MCBERROR); else
      mod_secondary->mempanic(type, addr, info, here, caller);
}

/*  here: 'a' - alloc, 'r' - realloc, 'f' - free, 'i' - isfree, 'm' - avail,
          'b' - blocksize */
static void mterr(void *blk, u32t ii, char here, u32t caller) {
   log_printf("used/free: %08X mt:%08X %c %08X\n", blk, memtable[ii], here, caller);
   halterr(MEMHLT_MEMTABLE, (u32t)blk, ii, here, caller);
}

// trying to free or realloc read-only block
static int hrchk(u32t pos, char here, u32t caller) {
   if (memheapres[pos>>3]&1<<(pos&7)) {
      u32t addr = phmembase + (pos<<16);
      log_printf("r/o: %08X %c %08X\n", addr, here, caller);
      if (here!='f') halterr(MEMHLT_REALLOCRO, addr, 0, here, caller);
      return 1;
   }
   return 0;
}

static void addrcheck(void *addr, char here, u32t caller) {
   if ((u32t)addr<phmembase || (u32t)addr>=phmembase+availmem ||
      ((u32t)addr-phmembase & 0xFFFF))
   {
      log_printf("bad addr: %08X %c %08X\n", addr, here, caller);
      halterr(MEMHLT_WRONGADDR, (u32t)addr, 0, here, caller);
   }
}

static void fbcheck(u32t pos, u64t size64k, char here, u32t caller) {
   u32t bsv = memtable[pos];
   // only non-zero size64k is used
   if (bsv<FMARK || bsv==FFFF || size64k && (u16t)bsv!=size64k) {
      u32t addr = phmembase + (pos<<16);
      log_printf("free size: %08X %08X %X %c %08X\n", addr, bsv, size64k, here,
                 caller);
      halterr(MEMHLT_FBSIZE, addr, size64k, here, caller);
   }
}

// add block to list of free blocks with the same size
static void fb_add(u32t ps, u32t size64k, char here, u32t caller) {
   u16t phdr = memftable[size64k];
   // verify/link current top block of the same size
   if (phdr) {
      fbcheck(phdr, size64k, here, caller);
      memflink[phdr].prev = ps;
   }
   memtable [ps] = FMARK|size64k;
   memflink [ps].next = phdr;
   memflink [ps].prev = 0;
   memftable[size64k] = ps;
}

// unlink block from list of free blocks
static void fb_del(u32t ps, u32t size64k, char here, u32t caller) {
   u16t next = memflink[ps].next,
        prev = memflink[ps].prev;

   fbcheck(ps, size64k, here, caller);
   if (!size64k) size64k = (u16t)memtable[ps];
   // check both prev & next in list if available
   if (prev) {
      fbcheck(prev, size64k, here, caller);
      memflink[prev].next = next;
   } else
      memftable[size64k] = next;
   if (next) {
      fbcheck(next, size64k, here, caller);
      memflink[next].prev = prev;
   }
   memtable[ps] = 0;
   // zero links too (just to be safe)
   memflink[ps].sign = 0;
}

// allocate memory
static void *alloc_worker(u32t size, u32t caller, u32t fl, u32t sig,
                          process_context *opq)
{
   u32t       *mt = memtable, sz, ii, ps, fbsize;
   u16t      *mft = memftable;
   void   *result = 0;
   s8t      align = fl&QSMA_MAXADDR?1:(fl&QSMA_MINADDR?-1:0);
#ifndef EFI_BUILD
   /* BIOS host:
      * force max addr if <24Mb of memory
      * else force min addr (because holes in memory can put the entire
        QSINIT to 4th Gb) */
   if (lowmem) align=1; else
      if (!align) align=-1;
#endif
   ii=sz=Round64k(size)>>16;

   mt_swlock();
   // verify it at least a bit!
   if (opq && opq->size!=sizeof(process_context)) {
      log_printf("bad PQ: %08X %08X\n", opq, caller);
      halterr(MEMHLT_BADPQ, (u32t)opq, 0, 'a', caller);
   }
   /* max/min addr allocation cause walking over ALL free blocks
      (and checking of signatures) */
   while (ii<memblocks) {
      if (mft[ii]) {
         u32t nps = mft[ii];
         while (nps) {
            // check free block
            fbcheck(nps, ii, 'a', caller);
            // get any or follow align ...
            if (!result || align && !Xor(nps<ps,align<0)) {
               result = (u8t*)phmembase + (nps<<16);
               ps     = nps;
               fbsize = ii;
               if (!align) break;
            }
            nps = memflink[nps].next;
         }
         if (!align) break;
      }
      ii++;
   }
   if (result) {
      u8t      *hptr;
      // unlink it from the free block list
      fb_del(ps, fbsize, 'a', caller);
      // create a free block at the start of free space
      if (align>0 && fbsize>sz) {
         fb_add(ps, fbsize-sz, 'a', caller);
         // fix position & result pointer
         ps += fbsize-sz;
         result = (u8t*)result + (fbsize-sz<<16);
      } else
      // create a free block of smaller size at the end of free space
      if (fbsize!=sz) fb_add(ps+sz, fbsize-sz, 'a', caller);
      // save user size
      mt[ps]       = size;
      memowner[ps] = 0;
      memflink[ps].sign = sig;
      // update const ptr flag
      hptr = memheapres+(ps>>3);
      if (fl&QSMA_READONLY) *hptr|=1<<(ps&7); else {
         *hptr&=~(1<<(ps&7));
         // set "local" owner for non-ro block only
         memowner[ps] = opq;
      }
      // mark blocks in size array
      ii = ps;
      while (--sz)
         if (mt[++ii]) mterr(result,ii,'a',caller); else mt[ii]=FFFF;
   }
#ifndef EFI_BUILD
   // convert to address in VM area
   if (result && VMA())
      if (fl & (QSMA_PHYS|QSMA_LOCAL)) memvaddr[ps] = 0; else {
         u32t vaddr = vm_reserve((u32t)result, size, fl);
         if (!vaddr) {
            hlp_memfree(result);
            result = 0;
         } else {
            memvaddr[ps] = vaddr;
            result = (u32t*)vaddr;
         }
      }
#endif
   mt_swunlock();
   return result;
}

static void* alloc_common(u32t size, u32t caller, const char *sig, u32t flags) {
   void* rc = 0;
   if (!size) return 0;
   if (size<availmem) {
      u32t   sv = 0, step = 0;
      process_context *pq = 0;

      if (flags & QSMA_CONTEXT) {
         pq = (process_context*)sig;
         flags &= ~QSMA_CONTEXT;
      } else
      if (sig) sv = *(u32t*)sig;

      while (step<2) {
         /* try to notify someone about low memory, at least CACHE should turn
            itself OFF */
         if (step++ && mod_secondary) mod_secondary->sys_notifyexec(SECB_LOWMEM,0);

         rc = alloc_worker(size, caller, flags, sv, pq?pq:
                           flags&QSMA_LOCAL?mt_exechooks.mtcb_ctxmem:0);
         if (rc) break;
      }
   }
   if (!rc && (flags&QSMA_RETERR)==0) exit_pm32(QERR_NOMEMORY);
   // zero-fill requested part only
   if (rc && (flags&QSMA_NOCLEAR)==0) memset(rc,0,size);
   return rc;
}

void* _std hlp_memallocsig(u32t size, const char *sig, u32t flags) {
   return alloc_common(size, CALLER(size), sig, flags);
}

void* hlp_memalloc(u32t size, u32t flags) {
   return alloc_common(size, CALLER(size), 0, flags);
}

static void free_common(void *addr, void *paddr, u32t caller) {
   u32t    sz,ii,stp,pos;
   u32t      *mt = memtable; // to avoid many 32bit data segment offsets
   u16t     *mft = memftable;
   // invalid address?
   addrcheck(paddr,'f',caller);
   // position
   pos = (u32t)paddr-phmembase>>16;
   // lock mt
   mt_swlock();
   // deny all 3 cases: free header and 0/FFFF in the middle of the block
   if (!mt[pos] || mt[pos]>=FMARK) mterr(paddr,pos,'f',caller);
   // ignore free of r/o block
   if (!hrchk(pos,'f',caller)) {
      // size in blocks
      sz  = Round64k(mt[pos])>>16;
      // begin of feature free block
      stp = pos;
      while (stp--)
         if (mt[stp]) break;
      // two or more free 64k or a single entry checked by FMARK
      if (stp<pos-1 || mt[stp]>FMARK && mt[stp]!=FFFF)
         fb_del(stp, pos-stp, 'f', caller); else stp++;
      // check and clear size array
      if (sz>1)
         if (memchrnd(mt+pos+1,FFFF,sz-1)) mterr(paddr,pos,'f',caller);
      memsetd(mt+pos,0,sz);
#ifndef EFI_BUILD
      // free vm entry
      if (memvaddr) {
         if (VMA() && memvaddr[pos]) {
            // it`s better to panic here
            if (paddr!=addr && (u32t)addr!=memvaddr[pos]) {
               log_printf("vm mismatch: %08X %08X<>%08X\n", paddr, addr, memvaddr[pos]);
               halterr(MEMHLT_VMADDR, (u32t)addr, (u32t)paddr, 'f', caller);
            }
            vm_release(memvaddr[pos]);
         }
         memvaddr[pos] = 0;
      }
#endif
      // add a free block after this
      if (pos+sz < memblocks)
         if (mt[pos+sz] > FMARK) { // this can only be a block header
            u32t fbsize = mt[pos+sz] & 0xFFFF;
            fb_del(pos+sz, 0, 'f', caller);
            sz += fbsize;
         }
      // add block to list
      fb_add(stp, pos-stp+sz, 'f', caller);
   }
   mt_swunlock();
}

/// free memory. addr must point to the begin of allocated block
void hlp_memfree(void *addr) {
   void  *paddr = addr;
#ifndef EFI_BUILD
   if (VMA() && (u32t)addr>=VMBASE) {
      u32t pa = vm_vtophys((u32t)addr);
      if (pa) paddr = (void*)pa;
   }
#endif
   free_common(addr, paddr, CALLER(addr));
}

qserr hlp_memfreephys(u64t addr) {
   if (addr>=_4GBLL) {
      if (mod_secondary) return mod_secondary->memfreephys(addr);
      return E_SYS_INVPARM;
   }
   free_common((void*)addr, (void*)addr, CALLER(addr));
   return 0;
}

// query block size by pointer
u32t hlp_memgetsize(void *addr) {
   u32t     *mt = memtable; // avoid 32bit offsets
   u32t  caller = CALLER(addr), pos;
#ifndef EFI_BUILD
   if (VMA() && (u32t)addr>=VMBASE)
      if ((pos = vm_vtophys((u32t)addr))) addr = (void*)pos;
#endif
   // invalid address?
   addrcheck(addr,'b',caller);
   // position
   pos = (u32t)addr-phmembase>>16;
   // lock
   mt_swlock();
   // why is was here?
   //hrchk(pos,'b',caller);
   pos = mt[pos];
   // unlock
   mt_swunlock();
   return pos;
}

// realloc region of memory. return new/old address.
void* hlp_memrealloc(void *addr, u32t newsize) {
   u32t      *mt = memtable; // to avoid many 32bit data segment offsets
   u16t     *mft = memftable;
   u32t   caller = CALLER(addr), sz, nsz, pos;
   void   *paddr = addr;     // physical address
#ifndef EFI_BUILD
   if (VMA() && (u32t)addr>=VMBASE)
      if ((pos = vm_vtophys((u32t)addr))) paddr = (void*)pos;
#endif
   // invalid address?
   addrcheck(paddr,'r',caller);
   if (!newsize) newsize++;
   // position
   pos = (u32t)paddr-phmembase>>16;
   // mt lock
   mt_swlock();
   // check for r/o
   hrchk(pos,'r',caller);
   // size in blocks
   sz  = Round64k(mt[pos])>>16;
   nsz = Round64k(newsize)>>16;
   if (sz==nsz) { // realloc does not require table changes
      if (newsize>mt[pos]) memset((u8t*)addr+mt[pos], 0, newsize-mt[pos]);
      mt[pos] = newsize;
   } else
   if (nsz<sz) { // new size is smaller
      mt[pos] = newsize;
      memsetd(mt+pos+nsz, 0, sz-nsz);
      // merge with the next free block
      if (pos+sz < memblocks)
         if (mt[pos+sz] > FMARK) {
            u32t fbsize = mt[pos+sz] & 0xFFFF;
            fb_del(pos+sz, 0, 'r', caller);
            sz += fbsize;
         }
      // a free block at the end
      fb_add(pos+nsz, sz-nsz, 'r', caller);
#ifndef EFI_BUILD
      // it should never return 0 in this case
      if (paddr!=addr)
         addr = (void*)(memvaddr[pos] = vm_realloc((u32t)addr, newsize));
#endif
   } else {
      u8t cmway = 1;
      if (pos+sz < memblocks)
         if (mt[pos+sz] > FMARK) {
            u32t fbsize = mt[pos+sz] & 0xFFFF;
            if (sz+fbsize >= nsz) { // free block behind processed
               cmway = 0;
#ifndef EFI_BUILD
               // realloc VM area address
               if (paddr!=addr) {
                  u32t na = vm_realloc((u32t)addr, newsize);
                  if (!na) cmway = 1; else addr = (void*)(memvaddr[pos] = na);
               }
#endif
               if (!cmway) {
                  fb_del(pos+sz, 0, 'r', caller);
                  if (sz+fbsize > nsz) // save remainder
                     fb_add(pos+nsz, sz+fbsize-nsz, 'r', caller);
                  // mark blocks as used
                  memsetd(mt+pos+sz, FFFF, nsz-sz);
                  // zero fill new bytes
                  memset((u8t*)addr+mt[pos], 0, newsize-mt[pos]);
                  mt[pos] = newsize;
               }
            }
         }
      // common allocating via alloc/free
      if (cmway) {
         u32t  fl = 0;
         void *rc;
         // make sure that block will be allocated in the same area ...
         if (!memowner[pos]) fl|=QSMA_LOCAL; else
            if (VMA() && paddr==addr) fl|=QSMA_PHYS;
         rc = alloc_worker(newsize, caller, fl, memflink[pos].sign, memowner[pos]);
         // unlock it here (nothing unsafe below)
         mt_swunlock();
         if (rc) {
            memcpy(rc, addr, mt[pos]);
            // zero additional space
            if (mt[pos]<newsize) memset((u8t*)rc+mt[pos], 0, newsize-mt[pos]);
            hlp_memfree(addr);
         }
         return rc;
      }
   }
   mt_swunlock();
   return addr;
}

/** query list of constant blocks (internal use only).
    reporting only blocks with, at least, 4k of free size, any other still
    marked as r/o (to prevent free), but not visible to memmgr */
u32t hlp_memqconst(u32t *array, u32t pairsize) {
   u32t  *mt = memtable, ii,
         cnt = 0;
   mt_swlock();
   for (ii=0; ii<memblocks && cnt<pairsize; ii++)
      if ((memheapres[ii>>3]&1<<(ii&7)) && mt[ii] && mt[ii]<FMARK)
         if (mt[ii]<_64KB-_4KB) {
            u32t addr = VMA() ? memvaddr[ii] : 0;
            *array++  = addr ? addr : phmembase + (ii<<16);
            *array++  = mt[ii];
            cnt++;
         }
   mt_swunlock();
   return cnt;
}

void* memreserve(u32t physaddr, u32t length, u32t *reason, int ro) {
   if (physaddr<phmembase || physaddr>=phmembase+availmem) {
      if (reason) *reason = QSMR_OUTOFRANGE;
      return 0;
   } else {
      u32t     *mt = memtable;
      u16t    *mft = memftable;
      u32t  caller = CALLER(physaddr), sz, pos, paorg,
            whyerr = QSMR_SUCCESS;
      // truncate length (must never occur)
      if (physaddr+length > phmembase+availmem) {
         length = phmembase+availmem-physaddr;
         whyerr = QSMR_TRUNCATED;
      }
      // round addr down to allocator step
      sz = (u32t)physaddr-phmembase & 0xFFFF;
      length  += sz;
      physaddr-= sz;
      paorg    = physaddr;
      // invalid address?
      addrcheck((void*)physaddr,'i',caller);

      pos = physaddr-phmembase>>16;
      sz  = Round64k(length)>>16;

      mt_swlock();
      // check that entire length is free (first pos can have a free blk header)
      if (sz>1 && memchrnd(mt+pos+1,0,sz-1) || mt[pos] && mt[pos]<=FMARK ||
          mt[pos]==FFFF)
      {
         whyerr   = QSMR_USED;
         physaddr = 0;
      } else {
         // mark this block as allocated
         u32t  ii = pos, fbsize;
         // find the start of free block
         while (!mt[ii]) ii--;
         fbsize = mt[ii] & 0xFFFF;
         fb_del(ii, 0, 'i', caller);
         // free block before reserved space
         if (pos-ii) fb_add(ii, pos-ii, 'i', caller);
         // free block after reserved space
         if (pos+sz<ii+fbsize) fb_add(pos+sz, ii+fbsize-(pos+sz), 'i', caller);
         // used memory
         mt[pos] = length;
         if (ro) memheapres[pos>>3] |= 1<<(pos&7);
         memflink[pos].sign = RESV_SIGN;
         while (--sz)
            if (mt[++pos]) mterr((void*)physaddr, pos, 'i', caller);
               else mt[pos] = FFFF;
      }
      mt_swunlock();
#ifndef INITDEBUG
      if (whyerr)
#endif
         log_printf("resrvd: %08X %u %u!\n", paorg, length, whyerr);

      if (reason) *reason = whyerr;
      return (void*)physaddr;
   }
}

void* _std hlp_memreserve(u32t physaddr, u32t length, u32t *reason) {
   return memreserve(physaddr, length, reason, 0);
}

// process exit callback
void mem_procexit(process_context *pq) {
   auto u32t* mt = memtable;
   u32t       ii;
   mt_swlock();
   for (ii=0; ii<memblocks; ii++)
      if (mt[ii] && mt[ii]<FMARK)
         if (memowner[ii])
            if (memowner[ii]->pid==pq->pid)
               if (VMA() && memvaddr[ii]) hlp_memfree((void*)memvaddr[ii]);
                  else hlp_memfree((u8t*)phmembase + (ii<<16));
   mt_swunlock();
}

// print memory contol table contents (debug)
void hlp_memprint(printf_function pfn) {
#if 1
   if (mod_secondary)
      mod_secondary->memprint(pfn, memtable, memftable, memheapres, memflink,
         memblocks, memowner, memvaddr, phmembase);
#else // moved to start
   u32t      *mt = memtable, ii;
   u16t     *mft = memftable;
   u8t      *mhr = memheapres;
   char   sigstr[5];
   sigstr[4] = 0;
   if (!pfn) pfn = log_printf;
   pfn("<=====QS memory dump=====>\n");
   pfn("Total : %d kb\n", memblocks<<6);
   for (ii=0; ii<memblocks; ii++) {
      u32t addr = phmembase + (ii<<16);
      if (mt[ii] && mt[ii]<FMARK) {
         *(u32t*)&sigstr = memflink[ii].sign;
         pfn("%5d. %08X - %c %-4s  %7d kb (%d)\n", ii, addr,
            mhr[ii>>3]&1<<(ii&7) ? 'R' : ' ', sigstr, Round64k(mt[ii])>>10, mt[ii]);
      } else
      if (mt[ii]>=FMARK && mt[ii]!=FFFF)
         pfn("%5d. %08X - <free>  %7d kb\n", ii, addr, (mt[ii]&0xFFFF)<<6);
   }
   pfn("%5d. %08X - end\n", memblocks, phmembase+(memblocks<<16));

   pfn("Free table:\n");
   for (ii=0; ii<memblocks; ii++) {
      if (mft[ii]) {
         u16t fb = mft[ii];
         pfn("%u kb (%u):", ii<<6, ii);
         while (fb) {
            u32t mtv = mt[fb];
            pfn(" %08X", phmembase + (fb<<16));
            // is it a free block header?
            if (mtv<=FMARK || mtv==FFFF) pfn("(S:%08X!)", mtv);
            // and check a block size
            mtv &= 0xFFFF;
            if (mtv!=ii) pfn("(%u!)", mtv);

            fb = memflink[fb].next;
         }
         pfn("\n");
      }
   }
#endif
}

u32t _std hlp_memavail(u32t *maxblock, u32t *total, u32t flags) {
   u16t *mft = memftable;
   u32t   rc = 0,
          ii = memblocks,
      caller = CALLER(maxblock), mb = 0;

   mt_swlock();
   while (ii--)
      if (mft[ii]) {
         u16t   fb = mft[ii];
         u32t size = ii<<16;
         if (!mb) mb = size;
         while (fb) {
            // check for damaged free block header
            fbcheck(fb, ii, 'm', caller);
            rc+= size;
            fb = memflink[fb].next;
         }
      }
   if (maxblock) *maxblock = mb;
   if (total) *total = availmem;
#ifndef EFI_BUILD
   // if VM area requested - return the smallest values of both it and above
   if (VMA() && !(flags&(QSMA_LOCAL|QSMA_PHYS))) {
      u32t vmmax, vmtotal,
            vmav = vm_avail(&vmmax, &vmtotal);
      if (vmav<rc) rc = vmav;
      if (maxblock && vmmax<*maxblock) *maxblock = vmmax;
      if (total && vmtotal<*total) *total = vmtotal;
   }
#endif
   mt_swunlock();
   return rc;
}

u32t _std hlp_memused(u32t pid, u32t *usedlow, u32t *usedhigh) {
   u32t ii, size = 0, ul = FFFF, uh = 0;

   mt_swlock();
   for (ii=0; ii<memblocks; ii++) {
      u32t mti = memtable[ii], b_addr, b_size;
      if (mti && mti<FMARK) {
         if (pid!=FFFF) {
            if (Xor(pid,memowner[ii])) continue;
            if (pid && memowner[ii]->pid!=pid) continue; 
         }
         b_addr = (u32t)memtable + (ii<<16);
         b_size = Round64k(mti);
         size  += b_size;
         // update highest/lowest used address
         if (ul>b_addr) ul = b_addr;
         if (uh<b_addr+b_size-1) uh = b_addr + b_size - 1;
         
      }
   }
   mt_swunlock();
   // limits
   if (size) {
      if (usedlow)  *usedlow  = ul;
      if (usedhigh) *usedhigh = uh;
   }
   return size;
}

#ifndef EFI_BUILD
void memmgr_vminit(u32t *fxtab) {
   u32t  *mt = memtable, ii,
        *mvt = memvaddr;
   /* assign addresses for all allocated blocks except reserved memory blocks,
      all of it should be memory holes at the moment and they can take A LOT */
   for (ii=0; ii<memblocks; ii++)
      if (mt[ii] && mt[ii]<FMARK && memflink[ii].sign!=RESV_SIGN)
         if (!(mvt[ii] = vm_reserve(phmembase+(ii<<16), Round64k(mt[ii]), 0))) {
            log_printf("memmgr blk %u (%X)\n", ii, mt[ii]);
            vm_panic();
         }
   // and save them to fixup table ...
   while (*fxtab) {
      u32t addr = *(u32t*)(*fxtab++);
      // invalid address?
      addrcheck((void*)addr,'v',0);
      // position
      ii = addr-phmembase>>16;

      if (!mvt[ii]) {
         log_printf("fix pos %X not found\n", addr);
         vm_panic();
      }
      *fxtab++ = mvt[ii] - addr;
      // skip sub-addresses ...
      while (*fxtab++!=FFFF) ;
   }
   // impossible situation, but better to be safe ...
   if (phmembase + availmem > VMBASE)
      memreserve(VMBASE, availmem-VMBASE, 0, 1);
}
#endif
