//
// QSINIT
// init time paging support
// -------------------------------------------------------------
// - 128Mb area at the top of 4Gb is used for virtual addresses
// - anything else mapped plain and leaved for START to handle ...
// - address allocation step is 64k (same with membase.c)
//
#include "qslocal.h"
#include "qsbinfmt.h"
#include "pagedesc.h"
#include "cpudef.h"

#define VMBLOCKS  (INIT_SYS_AREA>>16)    // 2048 blocks of 64k in 128Mb 
#define PDSTEP2M  (1<<PD2M_ADDRSHL)
#define PDSTEP4M  (1<<PD4M_ADDRSHL)

extern
MKBIN_HEADER  *pbin_header;
extern u32t       highbase;     // phys.base of 32-bit QSINIT`s object
extern u8t      memhiavail;     // memory above 4Gb is available
extern u32t      memblocks,
                 *memvaddr,
                 highbasev,
                 qs16basev,
                 phmembase;     // base address of used memory
static u8t          npmode = 0; // QPGM_VPAE/V32 (set _before_ paging, for alloc funcs)
static u32t         GLOBAL;     // PT_GLOBAL bit (or 0)

#define VMB_FREE   0    // zero value is assumed in code!
#define VMB_USED   1

typedef struct {
   union {
      struct {
         u16t        type;  ///< VMB_*
         u16t        size;  ///< block size or 0 if in the middle of it
         union {
            struct {
               u16t  prev;  ///< previous in free block list
               u16t  next;  ///< next in free block list
            };
            u32t     attr;
         };
      };
      u64t          umark;  ///< VMB_USED/VMB_FREE in the middle of used block
   };
} vmloc;

static vmloc     *vmtable;  ///< array of vm block info
static u16t     *vmftable;  ///< first free block index (vmftable[VMBLOCKS])
static u8t         *pdmem,  ///< page dirs
                   *pdvma;  ///< page tables for VM area

void memmgr_vminit(u32t *fxtab);
// add FAT micro-FSD fixup pointers if available
void fat_fixup    (void);
// actual code moving call
void _std relocate(u32t *fxtab, u32t base16, u32t base32);

// print memory contol table contents (debug)
void vm_print(printf_function pfn) {
#if 0
   vmloc    *vmt = vmtable;
   u16t    *vmft = vmftable;
   u32t          ii, cnt;
   char        sigstr[8];
   sigstr[4] = 0;
   if (!pfn) pfn = log_printf;
   pfn("<===== VM init dump =====>\n");
   for (ii=0; ii<VMBLOCKS; ii++) {
      vmloc *vme = vmt+ii;
      if (vme->size) {
         u32t addr = VMBASE+(ii<<16);

         if (vme->type==VMB_USED) {
            u32t   pa = vm_vtophys(addr);
            char attr[4];
            *(u32t*)&attr = 0;

            if (!(vme->attr&PT_PRESENT)) {
               attr[0] = 'N';
               attr[1] = 'P';
            } else {
               attr[0] = vme->attr&PT_WRITE?'W':'R';
               attr[1] = vme->attr&PT_R3ACCESS?'U':'S';
               if (vme->attr&PT_GLOBAL) attr[2] = 'G';
            }
            pfn("%4u. %08X/%08X - %7u kb %-3s (%u)\n", ii, addr, pa,
               vme->size<<6, attr, vme->size);
         } else
         if (vme->type!=VMB_FREE) pfn("free header destroyd!!!\n");
            else pfn("%4u. %08X          - %7u kb <free>\n", ii, addr,
               vme->size<<6);
      }
   }
   pfn("Free table:\n");
   for (ii=0; ii<VMBLOCKS; ii++)
      if (vmft[ii]) {
         u32t nps = vmft[ii];
         pfn("%u kb (%u):", ii<<6, ii);
         while (nps) {
            vmloc *vme = vmt+nps;
            pfn(" %08X", VMBASE+(nps<<16));
            if (vme->size!=ii) pfn("(%u!)", vme->size);
            if (vme->type!=VMB_FREE) pfn("(t:%04X!)", vme->type);
            nps = vme->next;
         }
         pfn("\n");
      }
#endif
}

void vm_panic(void) {
   vm_print(0);
   exit_pm32(QERR_MCBERROR);
}

static void vm_szerr(u32t ii, u32t mustbe, char here, u32t caller) {
   log_printf("vm size (%s): %u vmt:%4hb expected:%04X %c %08X\n", 
      vmtable[ii].type==VMB_FREE?"free":"used", ii, vmtable+ii, mustbe, 
         here, caller);
   vm_panic();
}

static void vm_mterr(u32t ii, char here, u32t caller) {
   log_printf("vm used/free: %u vmt:%4hb %c %08X\n", ii, vmtable+ii, here, caller);
   vm_panic();
}

static void vm_checkaddr(u32t addr, char here, u32t caller) {
   if (addr<VMBASE || (addr&0xFFFF)) {
      log_printf("vm bad addr: %08X %c %08X\n", addr, here, caller);
      vm_panic();
   }
}

// add block to the list of free blocks of the same size
static void vm_fbadd(u32t ii, u32t size64k, char here, u32t caller) {
   vmloc   *vmt = vmtable;
   u16t  *vfmte = vmftable + size64k;

   vmt[ii].type = VMB_FREE;
   vmt[ii].size = size64k;
   vmt[ii].prev = 0;
   vmt[ii].next = *vfmte;

   if (*vfmte) {
      if (vmt[*vfmte].type!=VMB_FREE) vm_mterr(*vfmte, here, caller);
      vmt[*vfmte].prev = ii;
   }
   *vfmte = ii;
   // fill other entries with zero ...
   if (size64k>1) memsetq(&vmt[ii+1].umark, 0, size64k-1);
}

// unlink from the list of free blocks
static void vm_fbdel(u32t ii, u32t size64k, char here, u32t caller) {
   u16t   pfb, nfb;
   vmloc *vmt = vmtable; // to avoid a lot of 32bit data segment offsets

   if (vmt[ii].type!=VMB_FREE) vm_mterr(ii, here, caller);
   // if no size specified - no check
   if (!size64k) size64k = vmt[ii].size; else
      if (vmt[ii].size!=size64k) vm_szerr(ii, size64k, here, caller);
   pfb = vmt[ii].prev;
   nfb = vmt[ii].next;
   if (pfb) vmt[pfb].next = nfb; else vmftable[size64k] = nfb;
   if (nfb) vmt[nfb].prev = pfb;
   // just to be safe
   vmt[ii].prev = 0;
   vmt[ii].next = 0;
}

/// update page tables
static void setpages(u32t addr, u32t npages, u32t base, u32t attr) {
   u32t page;
   if (addr<VMBASE) vm_checkaddr(addr, 'p', CALLER(addr));
   page = addr - VMBASE >> 12;
   // fit in range
   if (page + npages > INIT_SYS_AREA>>12) {
      log_printf("wrong set area: %04X %u!\n", page, npages);
      npages = (INIT_SYS_AREA>>12) - page;
   }
   base &= PT_ADDRMASK;
   base |= attr;

   while (npages) {
      if (npmode==QPGM_VPAE) ((u64t*)pdvma)[page] = base;
         else ((u32t*)pdvma)[page] = base;
      // this set only when page mode active
      if (VMA()) invpage(addr);

      if (--npages==0) break;
      if (base & PT_PRESENT) base += PAGESIZE;
      addr += PAGESIZE;
      page++;
   }
}

/** allocate virtual memory area.
    @param  n64k       Number of 64k blocks to allocate
    @param  caller     Caller address for the error dump
    @param  fl         Subset of QSMA_* flags (min/max address request)
    @param  physbase   Physical address of the region to map
    @return address of allocated area or 0 */
static u32t vm_worker(u32t n64k, u32t caller, u32t fl, u32t physbase) {
   vmloc    *vmt = vmtable; // to avoid many 32bit data segment offsets
   u16t    *vmft = vmftable;
   u32t   result = 0, ps, fbsize, nps,
              ii = n64k;
   s8t     align = fl&QSMA_MAXADDR?1:(fl&QSMA_MINADDR?-1:0);

   mt_swlock();
   /* max/min addr allocation cause walking over ALL free blocks
      (and checking of signatures) */
   while (ii<VMBLOCKS) {
      if (vmft[ii]) {
         nps = vmft[ii];
         while (nps) {
            // is it marked as free in vmtable array?
            if (vmt[nps].type!=VMB_FREE) vm_mterr(nps, 'a', caller);

            if (!result || align && !Xor(nps<ps,align<0)) {
               result = VMBASE + (nps<<16);
               ps     = nps;
               fbsize = ii;
               if (!align) break;
            }
            nps = vmt[nps].next;
         }
         if (!align) break;
      }
      ii++;
   }
   if (result) {
      // unlink it from the free block list
      vm_fbdel(ps, fbsize, '1', caller);
      // create a free block at the start of free space
      if (align>0 && fbsize>n64k) {
         vm_fbadd(ps, fbsize-n64k, '2', caller);
         // fix position & result pointer
         ps    += fbsize-n64k;
         result+= fbsize-n64k<<16;
      } else
      // create a free block of smaller size at the end of free space
      if (fbsize!=n64k) vm_fbadd(ps+n64k, fbsize-n64k, '3', caller);
      // set size/type
      vmt[ps].size = n64k;
      vmt[ps].type = VMB_USED;
      vmt[ps].attr = GLOBAL|PT_PRESENT|PT_WRITE;

      setpages(VMBASE + (ps<<16), n64k<<4, physbase, vmt[ps].attr);
      // mark blocks in the size array (and verify that free entries are == 0)
      while (--n64k)
         if (vmt[++ps].umark) vm_mterr(ps, '4', caller);
            else vmt[ps].umark = VMB_USED;
   }
   mt_swunlock();
   return result;
}

/// return physical address for virtual in VM area
u32t vm_vtophys(u32t vmaddr) {
   u32t page;
   if (!npmode) return vmaddr;
   if (vmaddr<VMBASE) return 0;

   page = vmaddr - VMBASE >> 12;

   if (npmode==QPGM_VPAE) {
      u64t pa = ((u64t*)pdvma)[page];
      // should not occurs on this stage
      if (pa>>32) {
         log_printf("64-bit addr? (%LX)\n", pa);
         return 0;
      }
      page = (u32t)pa;
   } else
      page = ((u32t*)pdvma)[page];

   if (page & PT_PRESENT) {
      page &= PT_ADDRMASK;
      page += vmaddr & PAGEMASK;
   } else
      page = 0;
   return page;
}

u32t vm_reserve(u32t physbase, u32t size, u32t flags) {
   if (!size) return 0;
   size = Round64k(size) >> 16;
   return vm_worker(size, CALLER(physbase), flags, physbase);
}

/// free memory. addr must point to the begin of allocated block
void vm_release(u32t addr) {
   vmloc    *vmt = vmtable; // to avoid many 32bit data segment offsets
   u32t   caller = CALLER(addr), sz, pos, stp;
   // invalid address?
   vm_checkaddr(addr,'f',caller);
   // position
   pos = addr-VMBASE >> 16;

   mt_swlock();
   // check type
   if (vmt[pos].type!=VMB_USED || !vmt[pos].size) vm_mterr(pos, '5', caller);
   // size in blocks
   sz  = vmt[pos].size;
   /* check if a free block before us (stp cannot reach 0 because 1st block is
      always used) */
   stp = pos;
   while (stp--)
      if (vmt[stp].umark!=VMB_FREE) break;
   if (vmt[stp].type==VMB_FREE) vm_fbdel(stp, pos-stp, '6', caller);
      else stp = pos;
   // check vmtable entires
   if (sz>1)
      if (memchrnq((u64t*)(vmt+pos+1),VMB_USED,sz-1)) vm_mterr(pos, '7', caller);
   // remove page mapping
   setpages(VMBASE + (pos<<16), sz<<4, 0, 0);
   // check next block
   if (pos+sz < VMBLOCKS) {
      vmloc *next = vmt + (pos+sz);
      if (next->type==VMB_FREE) {
         if (!next->size) vm_mterr(pos, '8', caller);
         vm_fbdel(pos+sz, 0, '9', caller);
         sz += next->size;
         next->umark = 0;
      }
   }
   // add block to list
   vm_fbadd(stp, pos - stp + sz, 'f', caller);

   mt_swunlock();
}

/** realloc region of memory. return new/old address.
    since real realloc in hlp_memrealloc() is really rare, we just implemented
    here a dumb alloc/free logic ... */
u32t vm_realloc(u32t addr, u32t newsize) {
   vmloc    *vmt = vmtable;
   u32t   caller = CALLER(addr), pos, sz, nsz;
   // invalid address?
   vm_checkaddr(addr,'r',caller);
   if (!newsize) newsize++;
   // position
   pos = addr-VMBASE >> 16;
   // mt lock
   mt_swlock();
   // size in blocks
   sz  = vmt[pos].size;
   nsz = Round64k(newsize)>>16;
   // check type
   if (vmt[pos].type!=VMB_USED || !sz) vm_mterr(pos, 'r', caller);

   if (sz>nsz) {
      vmt[pos].size = nsz;
      vm_fbadd(pos+nsz, sz-nsz, 'r', caller);
      // remove page mapping
      setpages(VMBASE + (pos+nsz<<16), sz-nsz<<4, 0, 0);
   } else
   if (sz<nsz) {
      u32t base = vm_vtophys(VMBASE + (pos<<16));
      // unable to get phys base?
      if (!base) addr = 0; else {
         // free it first and then alloc! but this causes a panic if failed
         vm_fbadd(pos, sz, 'r', caller);
         setpages(VMBASE + (pos<<16), sz<<4, 0, 0);
         // alloc a new area
         addr = vm_worker(nsz, caller, QSMA_MAXADDR, base);
         if (!addr) {
            log_printf("vm realloc failed: %04X %u!\n", pos, newsize);
            vm_panic();
         }
      }
   }
   mt_swunlock();
   return addr;
}

u32t vm_avail(u32t *maxblock, u32t *total) {
   vmloc  *vmt = vmtable;
   u16t  *vmft = vmftable;
   u32t     rc = 0,
            ii = VMBLOCKS,
        caller = CALLER(maxblock), mb = 0;
   mt_swlock();
   while (ii--)
      if (vmft[ii]) {
         u16t   fb = vmft[ii];
         u32t size = ii<<16;
         if (!mb) mb = size;
         while (fb) {
            // check for damaged free block header
            if (vmt[fb].type!=VMB_FREE) vm_mterr(fb, 'm', caller);
            rc+= size;
            fb = vmt[fb].next;
         }
      }
   mt_swunlock();
   if (maxblock) *maxblock = mb;
   if (total) *total = INIT_SYS_AREA;
   return rc;
}

// ------------------------------------------------------------------------
#define   FIXUP_LIST_SIZE  32   // only 16 is actually used now ...
static u32t vm_fixup_list[FIXUP_LIST_SIZE];

void __cdecl push_fixup(void *addr, u32t nsub, ...) {
   u32t    **moffs = (u32t**)(&nsub + 1),  // additional pointers in this block
              pblk = *(u32t*)addr,
            mbsize = hlp_memgetsize((void*)pblk), ii;
   static u32t pos = 0;
   // an assert
   if (pos + 3 + nsub >= FIXUP_LIST_SIZE) exit_pm32(QERR_SOFTFAULT);
   // addr, diff, nsub subaddrs, FFFF; addr ...; 0
   vm_fixup_list[pos++] = (u32t)addr;
   vm_fixup_list[pos++] = 0;

   for (ii=0; ii<nsub; ii++)
      if (*moffs[ii] - pblk >= mbsize) {
         log_printf("invalid push_fixup: %08X %08X!\n", pblk, *moffs[ii]);
      } else
         vm_fixup_list[pos++] = (u32t)moffs[ii];
   vm_fixup_list[pos++] = FFFF;
   vm_fixup_list[pos]   = 0;
   //log_printf("fixup pos = %u\n", pos);
}

// ------------------------------------------------------------------------
/** some system changes *before* page mode activation.
    also called from pag_enable() in START */
void _std pag_preinit(void) {
   u32t  id[4];
   // just a bit of order ...
   if (mt_exechooks.mtcb_pgmode) return;
   // change reversed FLAT data selector to plain mode
   if (1/*!safeMode*/) {
      struct desctab_s sd;
      u16t  flatds = get_flatcs() + 8;
      u32t   fbase = 0;
      sd.d_limit   = 0xFFFF;
      sd.d_loaddr  = fbase;
      sd.d_hiaddr  = fbase>>16;
      sd.d_access  = D_DATA0;
      sd.d_attr    = D_DBIG|D_GRAN4K|D_AVAIL|0x0F;
      sd.d_extaddr = fbase>>24;

      sys_seldesc(flatds, &sd);
   }
   // make sure that PAT:0 is WB cache type
   if (hlp_getcpuid(1,id)) 
      if (id[3] & CPUID_FI2_PAT) {
         u32t  patlo, pathi;
         hlp_readmsr(MSR_IA32_PAT, &patlo, &pathi);
         if ((patlo&7)!=MSR_PAT_WB) {
            log_it(2, "Warning! PAT:0 type was %d\n", patlo&7);
         
            patlo = patlo&~7|MSR_PAT_WB;
            hlp_writemsr(MSR_IA32_PAT, patlo, pathi);
         }
      }
}

// check and prepare page mode ...
static u32t prepare_pages(void) {
   u32t  *cptr, ii, size, idbits[4], vcr3, vcr4, newbase16, pg0copy;
   vmloc  *vme;
   // 486??
   if (!hlp_getcpuid(1,idbits)) return 0;
   /* use 32-bit paging if no memory above 4Gb (this includes all VMs, PAE
      is deadly for many of them) */
   if (!memhiavail && (idbits[3]&CPUID_FI2_PSE) && vmmode!=2) npmode = QPGM_V32; else
      if (idbits[3] & CPUID_FI2_PAE) npmode = QPGM_VPAE; else {
         log_printf("CPU w/o PAE/PSE?\n");
         return 0;
      }
   GLOBAL = idbits[3] & CPUID_FI2_PGE ? PT_GLOBAL : 0;
   // verify that this module fixups are still alive
   size   = pbin_header->pmTotal - pbin_header->pmStored;
   cptr   = (u32t*)(highbase + pbin_header->pmSize + size);

   vcr4   = getcr4();
   if (npmode==QPGM_VPAE) vcr4 = vcr4&~CPU_CR4_PSE|CPU_CR4_PAE;
      else vcr4 = vcr4&~CPU_CR4_PAE|CPU_CR4_PSE;
   if (GLOBAL) vcr4|=CPU_CR4_PGE; else vcr4&=~CPU_CR4_PGE;

   for (ii=0; ii<4; ii++)
      if (cptr[ii]!=MKBIN_SIGN+1) {
         log_printf("fixup data is broken!\n");
         return 0;
      }
   // add FAT fixups just here since pointers in it can change ...
   fat_fixup();

   size      = (sizeof(u16t) + sizeof(vmloc)) * VMBLOCKS;
   vmftable  = hlp_memallocsig(size, "pgi", 0);
   vmtable   = (vmloc*)(vmftable + VMBLOCKS);
   /* add own table as a 0 block (to make free lists work!) and free block itself
      see also 2 setpages() below for these allocations */
   size      = Round64k(size)>>16;
   vme       = vmtable;
   vme->size = size;
   vme->type = VMB_USED;
   vme->attr = PT_PRESENT | PT_WRITE | GLOBAL;
   for (ii=1; ii<size; ii++) vme[ii].umark = VMB_USED;
   vme += size;
   vme->size = VMBLOCKS - size - (GUARD0_AREA >> 16);
   // a single free table entry
   vmftable[vme->size] = size;
   // 4Mb below 0 are off use ...
   vme += vme->size;
   vme->size = GUARD0_AREA >> 16;
   vme->type = VMB_USED;

   if (npmode==QPGM_VPAE) {
      // 4 page dirs + page for 1st mb + pages for VM area
      ii   = 4*PAGESIZE + PAGESIZE;
      size = ii + (INIT_SYS_AREA >> PD2M_ADDRSHL)*PAGESIZE + 32;
   } else {
      // page dir + page for 1st mb + pages for VM area
      ii   = PAGESIZE + PAGESIZE;
      size = ii + (INIT_SYS_AREA >> PD4M_ADDRSHL)*PAGESIZE;
   }
   pdmem    = (u8t*)hlp_memallocsig(size + memblocks * 4, "pdir", QSMA_READONLY);
   pdvma    = pdmem + ii;
   memvaddr = (u32t*)(pdmem + size);
   push_fixup(&pdmem, 2, &pdvma, &memvaddr);
   /* FLAT page table mapping (except VM area).
      1st Mb uses own page table because it has a penalty on Intel if mapped
      via big page. VM area is zeroed by hlp_memalloc() */
   if (npmode==QPGM_VPAE) {
      /*** PAE pages ***/
      u64t *pde = (u64t*)pdmem, lp, *pdpt;
      *pde++ = (u32t)pdmem + 4*PAGESIZE | PT_PRESENT | PT_WRITE;
      // loop big pages from 2Mb to VM area
      lp = PD_BIGPAGE | PT_PRESENT | PT_WRITE | PDSTEP2M | GLOBAL;
      for (ii=0; ii<512*4 - 1 - (INIT_SYS_AREA >> PD2M_ADDRSHL); ii++) {
         *pde++ = lp;
         lp    += PDSTEP2M;
      }
      // VM area page tables
      for (ii=0; ii< INIT_SYS_AREA >> PD2M_ADDRSHL; ii++)
         *pde++ = (u32t)pdvma + ii*PAGESIZE | PT_PRESENT | PT_WRITE;
      // first 2Mb mapping
      pde   = (u64t*)((u32t)pdmem + 4*PAGESIZE);
      lp    = PT_PRESENT | GLOBAL;
      for (ii=0; ii<512; ii++) {
         *pde++ = lp;
         lp    += PAGESIZE;
         if (!ii) lp|=PT_WRITE;
      }
      // PDPT
      vcr3  = (u32t)memvaddr - 32;
      pdpt  = (u64t*)vcr3;
      // fill 4 PDPTEs
      for (ii=0; ii<4; ii++)
         pdpt[ii] = PT_PRESENT | (u32t)pdmem + (ii<<PAGESHIFT);
   } else {
      /*** 32-bit pages ***/
      u32t *pde = (u32t*)pdmem, lp;
      *pde++ = (u32t)pdmem + PAGESIZE | PT_PRESENT | PT_WRITE;
      // loop big pages from 4Mb to VM area
      lp = PD_BIGPAGE | PT_PRESENT | PT_WRITE | PDSTEP4M | GLOBAL;
      for (ii=0; ii<1024 - 1 - (INIT_SYS_AREA >> PD4M_ADDRSHL); ii++) {
         *pde++ = lp;
         lp    += PDSTEP4M;
      }
      // VM area page tables
      for (ii=0; ii< INIT_SYS_AREA >> PD4M_ADDRSHL; ii++)
         *pde++ = (u32t)pdvma + ii*PAGESIZE | PT_PRESENT | PT_WRITE;
      // first 4Mb mapping
      pde   = (u32t*)((u32t)pdmem + PAGESIZE);
      lp    = PT_PRESENT | GLOBAL;
      for (ii=0; ii<1024; ii++) {
         *pde++ = lp;
         lp    += PAGESIZE;
         if (!ii) lp|=PT_WRITE;
      }
      vcr3  = (u32t)pdmem;
   }
   // set pages for 2 areas preallocated above
   setpages((u32t)-GUARD0_AREA, GUARD0_AREA>>12, 0, 0);
   setpages(VMBASE, vmtable[0].size<<4, (u32t)vmftable, vmtable[0].attr);

   // normal vm_reserve() can be called from this point ...

   highbasev = vm_reserve(highbase, phmembase-highbase, QSMA_MAXADDR);
   // manual fixup (this entire 64k aligned block is allocated in real mode)
   if (minifsd_ptr)
      minifsd_ptr = (void*)((u32t)minifsd_ptr + highbasev - highbase);
   /* immediate update of qs16basev is a bad idea since some real mode calls
      will be broken! */
   newbase16 = vm_reserve(qs16base, _64KB, QSMA_MAXADDR);
   // new disk buffer address (this can be unaligned, at least in PXE boot)
   DiskBufPM = vm_reserve(DiskBufPM, _64KB, QSMA_MAXADDR) + (DiskBufPM & PAGEMASK);
   // map writable page 0 just after disk buffer data (see page0_fptr below)
   pg0copy   = PAGEROUND(DiskBufPM + DISKBUF_SIZE);
   setpages(pg0copy, 1, 0, GLOBAL|PT_PRESENT|PT_WRITE);

   /* allocate addresses for all blocks in the memory manager, 
      hlp_memalloc() should not be used below this line! */
   memmgr_vminit(vm_fixup_list);

   pag_preinit();

   log_printf("relocating (%s) ... %08X->%08X, %08X->%08X\n",
      npmode==QPGM_VPAE?"PAE":"32-bit", highbase, highbasev, qs16base, newbase16);
   vm_print(0);

   __asm { cli }
   // NO LOG PRINTING between logrmbuf update and "mov cr0, eax" below
   if (logbufseg>rm16code && logbufseg-rm16code<0x1000)
      logrmbuf = (u8t*)(newbase16 + (logbufseg-rm16code<<PARASHIFT));
   else
   if (logbufseg>DiskBufRM_Seg && logbufseg-DiskBufRM_Seg<0x1000)
      logrmbuf = (u8t*)(DiskBufPM + (logbufseg-DiskBufRM_Seg<<PARASHIFT));
   else // should never occurs
      log_printf("log seg %04X!?\n", logbufseg);

   sys_setcr3(vcr3,vcr4);
   /* set page mode, reload ss from ds (SELZERO->SEL32DATA) and update
      writable page 0 pointer */
   __asm {
      mov  eax, cr0
      or   eax, CPU_CR0_PG
      mov  cr0, eax

      mov  ax, ds
      mov  ss, ax
      mov  word ptr page0_fptr+4, ax

      mov  eax, pg0copy
      mov  dword ptr page0_fptr, eax
   }

   return newbase16;
}

/** relocating and switching code (and all memory manager blocks) to
   the selected page mode (PAE or 32-bit).

   This function is a special case, it should not use any local copies of
   ANY pointers since relocate() really relocates THIS code and data and
   STACK to the top of 4Gb.

   It marked as "aborts;" so it called via jmp and cannot use return.
*/
void init_pages(void) {
   // non-zero result means CLI and page mode!
   u32t newbase16 = prepare_pages();
   if (newbase16) {
      // DO IT!
      relocate(vm_fixup_list, newbase16, highbasev);
      // global page dir pointer
      mt_exechooks.mtcb_vpdir   = pdvma;
      mt_exechooks.mtcb_pgmode  = npmode;
      // this affects real mode calls
      qs16basev = newbase16;
      __asm { sti }
   }
   call_start();
}
