#include "qsbase.h"
#include "qslog.h"
#include "stdlib.h"
#include "errno.h"
#include "vioext.h"
#include "pci.h"
#include "qcl/cplib.h"

static const char *classname[PCI_CB_DSP+1] = {"Legacy", "Storage", "Network",
   "Display", "Multimedia", "Memory", "Bridge", "Comm", "Peripheral", "Input",
   "Docking", "Processor", "Sb", "Wireless", "Io", "Sat", "Encrypt", "Dsp"};

static void shl_printpci(int nopause) {
   pci_location dev;
   int           ok;
   cmd_printseq("PCI devices:", nopause?-1:0, 0);

   ok = hlp_pcigetnext(&dev,1,1);
   while (ok) {
      char   sbuf[128], *hmsg;
      u8t  gclass = dev.classcode>>8;
      snprintf(sbuf, 128, "_PCI%06X",(u32t)dev.classcode<<8|dev.progface);
      hmsg = cmd_shellgetmsg(sbuf);

      snprintf(sbuf, 128, "\t%d.%d.%d\t- %04X:%04X class %02X.%02X.%02X - %s",
         dev.bus, dev.slot, dev.func, dev.vendorid, dev.deviceid, gclass,
            dev.classcode&0xFF, dev.progface, hmsg?hmsg:(gclass>PCI_CB_DSP?
               "Unknown":classname[gclass]));
      if (hmsg) free(hmsg);
      // this function expand tabs
      if (cmd_printtext(sbuf,nopause?-1:0,0)) break;

      ok = hlp_pcigetnext(&dev,0,1);
   }
}

static char *flag_list(char *hmsg, const char *fmtstr, u16t flags) {
   int  dcii;
   for (dcii=0; dcii<16; dcii++) {
      if ((flags & 1<<dcii)) {
         char keyname[32], *cname;
         snprintf(keyname, 32, fmtstr, 1<<dcii);
         cname = cmd_shellgetmsg(keyname);
         if (cname) {
            int plen = hmsg ? strlen(hmsg) : 0;
            if (plen && hmsg[plen-1]!='(') hmsg = strcat_dyn(hmsg, ", ");
            hmsg = strcat_dyn(hmsg, cname);
            free(cname);
         }
      }
   }
   return hmsg;
}

u32t _std shl_pci(const char *cmd, str_list *args) {
   int rc=-1, nopause=0, tolog=0;
   if (args->count>0) {
      static char *argstr   = "/np|/log";
      static short argval[] = {  1,   2,-1,  8};
      char            *arg0 = 0;
      // help overrides any other keys
      if (str_findkey(args, "/?", 0)) {
         cmd_shellhelp(cmd,CLR_HELP);
         return 0;
      }
      args = str_parseargs(args, 0, SPA_RETLIST, argstr, argval, &nopause, &tolog);

      // process command
      if (args->count>0) {
         strupr(arg0 = args->item[0]);
         cmd_printseq(0, 1+tolog, 0);
      }
      if (args->count==1) {
         if (strcmp(arg0,"VIEW")==0) {
            shl_printpci(nopause);
            rc = 0;
         }
      } else
      if (args->count>0) {
         pci_location d;
         int success = 0, batch = 0;
         char  *arg1 = args->item[1];
         strupr(arg1);

         if (strcmp(arg1,"ALL")==0) batch = 1; else
         if (strccnt(arg1,':')==1) {  // ven:dev
            str_list *dstr = str_split(arg1,":");
            u32t     venid = strtoul(dstr->item[0],0,16),
                     devid = strtoul(dstr->item[1],0,16);
            int        idx = arg0[0]=='W'?5:2;  // index of "index" parameter
            free(dstr);

            if (args->count-1==idx && isdigit(args->item[idx][0])) {
               idx = str2long(args->item[idx]);
               // remove optional to get constant number of parameters
               args->count--;
            } else idx = 0;

            if (idx>=0)
               if (hlp_pcifind(venid, devid, idx, &d, 1)) success = 1; else
                  rc = ENODEV;
         } else
         if (strccnt(arg1,'.')==2) {  // bus.slot.func
            str_list *dstr = str_split(arg1,".");
            d.bus  = str2ulong(dstr->item[0]);
            d.slot = str2ulong(dstr->item[1]);
            d.func = str2ulong(dstr->item[2]);
            free(dstr);

            if (d.func>=8 || !hlp_pciexist(&d)) rc = ENODEV; else
               success = 1;
         }

         if (args->count==2 && (strcmp(arg0,"VIEW")==0 || strcmp(arg0,"DUMP")==0)) {
            // enable ansi for this scope
            u32t ast_save = vio_setansi(1);

            if (batch) success = hlp_pcigetnext(&d,1,1);

            if (!success) {
               rc = 0;
               cmd_printf("PCI is not available\n");
            } else
            while (success) {
               static const char *hdrname[3] = {"normal device", "bridge", "cardbus"};
               char    sbuf[128], *hmsg;
               u8t   gclass = d.classcode>>8,
                       hdrt = d.header&0x7F;
               u32t  ubreak;
               u16t  devcmd;

               rc = EINTR;
               if (cmd_printseq(" Bus %d Slot %d Function %d", nopause?-1:0,
                  VIO_COLOR_LWHITE, d.bus, d.slot, d.func)) break;

               if (cmd_printseq(" Vendor %04X:%04X, type %d (%s)", nopause?-1:0, 0,
                  d.vendorid, d.deviceid, hdrt, hdrt>2?"invalid":hdrname[hdrt]))
                     break;

               snprintf(sbuf, 128, "_PCI%06X",(u32t)d.classcode<<8|d.progface);
               hmsg   = cmd_shellgetmsg(sbuf);
               ubreak = cmd_printseq(" Class %02X.%02X.%02X : %s", nopause?-1:0, 0,
                  gclass, d.classcode&0xFF, d.progface, hmsg?hmsg:
                     (gclass>PCI_CB_DSP?"Unknown":classname[gclass]));
               if (hmsg) free(hmsg);
               if (ubreak) break;

               if (hdrt==0) {
                  u32t devsub = hlp_pciread(d.bus, d.slot, d.func,
                                            PCI_SUBSYSTEM_VENDOR_ID, 4);
                  if (cmd_printseq(" Subsystem Vendor %04X, Subsystem ID %04X",
                     nopause?-1:0, 0, devsub&0xFFFF, devsub>>16)) break;
               }

               devcmd = hlp_pciread(d.bus, d.slot, d.func, PCI_COMMAND, 2);
               // use dyn buffer for to be safe
               hmsg   = sprintf_dyn(" Command %04hX (", devcmd);
               // no i/o and memory access - warn it!
               if ((devcmd&(PCI_CMD_IO|PCI_CMD_MEMORY))==0)
                  hmsg = strcat_dyn(hmsg, ANSI_LRED "Bus Access Disabled" ANSI_RESET);

               if (devcmd) hmsg = flag_list(hmsg, "_PCICMD_%04X", devcmd);
               hmsg = strcat_dyn(hmsg, ")");

               if ((hlp_pciread(d.bus, d.slot, d.func, PCI_HEADER_TYPE, 1) & 0x7F)==
                  PCI_HEADER_BRIDGE)
               {
                  char *amsg;
                  u32t  ms = (hlp_pciread(d.bus, d.slot, d.func, PCI_MEMORY_BASE, 2)
                                          & PCI_MEM_RANGE_MASK) << 16,
                        me = (hlp_pciread(d.bus, d.slot, d.func, PCI_MEMORY_LIMIT, 2)
                                          & PCI_MEM_RANGE_MASK) << 16;
                  u16t mpt = hlp_pciread(d.bus, d.slot, d.func, PCI_PREF_MEMORY_BASE, 2);
                  u64t mps = ((u32t)mpt & PCI_PREF_RANGE_MASK) << 16,
                       mpe = (hlp_pciread(d.bus, d.slot, d.func, PCI_PREF_MEMORY_LIMIT, 2)
                                          & PCI_PREF_RANGE_MASK) << 16;
                  u16t iot = hlp_pciread(d.bus, d.slot, d.func, PCI_IO_BASE, 1);
                  u32t ios = ((u32t)iot & PCI_IO_1K_RANGE_MASK) << 8,
                       ioe = ((u32t)hlp_pciread(d.bus, d.slot, d.func, PCI_IO_LIMIT, 1)
                                          & PCI_IO_1K_RANGE_MASK) << 8 | 0xFFF;
                  int  mpx = (mpt & PCI_PREF_RANGE_TYPE_MASK) == PCI_PREF_RANGE_TYPE_64,
                       iox = (iot & PCI_IO_RANGE_TYPE_MASK) == PCI_IO_RANGE_TYPE_32;
                  u16t  bc = hlp_pciread(d.bus, d.slot, d.func, PCI_BRIDGE_CONTROL, 2);
                  char *bs = flag_list(0, "_PCIBC_%04X", bc);
                  // zero if no memory ranges
                  if (me) me |= 0xFFFFF;
                  if (mpe) mpe |= 0xFFFFF;

                  if (mpx) {
                     mps |= (u64t)hlp_pciread(d.bus, d.slot, d.func,
                                              PCI_PREF_BASE_UPPER32, 4) << 32;
                     mpe |= (u64t)hlp_pciread(d.bus, d.slot, d.func,
                                              PCI_PREF_LIMIT_UPPER32, 4) << 32;
                  }
                  if (iox) {
                     ios |= hlp_pciread(d.bus, d.slot, d.func, PCI_IO_BASE_UPPER16, 2) << 16;
                     ioe |= hlp_pciread(d.bus, d.slot, d.func, PCI_IO_LIMIT_UPPER16, 2) << 16;
                  }

                  amsg = sprintf_dyn("\n"
                     "  Bus: primary %u, secondary %u, subordinate %u\n"
                     "  Bridge control: %04X (%s)\n"
                     "  IO range%s: %08X-%08X\n"
                     "  Memory range  : %08X-%08X\n"
                     "  Prefetchable range%s: %010LX-%010LX",
                     hlp_pciread(d.bus, d.slot, d.func, PCI_PRIMARY_BUS, 1),
                     hlp_pciread(d.bus, d.slot, d.func, PCI_SECONDARY_BUS, 1),
                     hlp_pciread(d.bus, d.slot, d.func, PCI_SUBORDINATE_BUS, 1),
                     bc, bs ? bs : "-", iox ? " (32-bit)" : "      ",
                     ios, ioe, ms, me, mpx ? " (64-bit)" : "", mps, mpe);
                  if (bs) free(bs);

                  hmsg   = strcat_dyn(hmsg, amsg);
                  free(amsg);
               }

               ubreak = cmd_printseq(hmsg, nopause?-1:0, 0);
               free(hmsg); hmsg=0;
               if (ubreak) break;

               if (arg0[0]=='D') {
                  u32t  buf[4];
                  u64t  mem;
                  int    ii;
                  for (ii=0; ii<64; ii++) {
                     buf[ii&3] = hlp_pciread(d.bus, d.slot, d.func, ii<<2, 4);
                     if ((ii&3)==3)
                        if (cmd_printseq("%03X: %16b", nopause?-1:0, 0,
                           (ii&~3)<<2, &buf)) break;
                  }
                  // user break in line above, return EINTR here
                  if (ii<64) break;
                  // dump pci-e space if available
                  mem = hlp_pcigetmem(0, d.bus, d.slot, d.func);
                  if (mem) {
                     u32t *mm = malloc_th(4096);
                     int   ok = 0;
                     if (hlp_hosttype()==QSHT_EFI && mem>=_4GBLL)
                        ok = sys_memhicopy((u32t)mm, mem, 4096);
                     else {
                        void *mp = pag_physmap(mem, 4096, 0);
                        if (mp) {
                           ok = 1;
                           memcpy(mm, mp, 4096);
                           pag_physunmap(mp);
                        }
                     }
                     // wrong address or PCI device (all FF at 0x100)
                     if (ok)
                        if (mm[0]!=((u32t)d.deviceid<<16|d.vendorid) ||
                           mm[64]==FFFF) ok = 0;
                     if (ok)
                        for (ii=0; ii<240; ii++)
                           if (cmd_printseq("%03X: %16b", nopause?-1:0, 0,
                              ii+16<<4, mm+(ii+16<<2))) break;
                     free(mm);
                     // user break in the loop above
                     if (ok && ii<240) break;
                  }
               } else {
                  u64t bases[6], sizes[6];
                  u32t bcnt = hlp_pcigetbase(&d, bases, sizes), ii;
                  if (bcnt>0) {
                     if (cmd_printseq(" Addresses:", 0, 0)) break;

                     for (ii=0; ii<bcnt; ii++) {
                        if ((bases[ii]&PCI_ADDR_SPACE)==PCI_ADDR_SPACE_IO) {
                           u32t start = bases[ii]&PCI_ADDR_IO_MASK;
                           snprintf(sbuf, 128, "  %d. Port: %04X-%04X (%d)", ii+1,
                              start, start+(u32t)sizes[ii]-1,(u32t)sizes[ii]);
                        } else {
                           u64t start = bases[ii]&~((u64t)0xF);
                           snprintf(sbuf, 128, "  %d. Memory: %010LX-%010LX (%u kb)%s%s",
                              ii+1, start, start+sizes[ii]-1, (u32t)(sizes[ii]>>10),
                                 bases[ii]&PCI_ADDR_MEM_PREFETCH?", Prefetch":"",
                                    (bases[ii]&PCI_ADDR_MEMTYPE_MASK)==PCI_ADDR_MEMTYPE_64?
                                       ", 64-bit":"");
                        }
                        if (cmd_printseq(sbuf, nopause?-1:0, 0)) break;
                     }
                     // user break in line above, return EINTR here
                     if (ii<bcnt) break;
                  }
                  // interrupt #
                  ii = hlp_pciread(d.bus,d.slot,d.func,PCI_INTERRUPT_LINE,2);
                  if (ii>>8)
                     if (cmd_printseq(" Interrupt %d, INT #%c", nopause?-1:0, 0,
                        ii&0xFF, '@'+(ii>>8))) break;
               }
               // next device?
               success = batch ? hlp_pcigetnext(&d,0,1) : 0;
               // set no error
               rc = 0;
            }
            vio_setansi(ast_save);
         } else
         if (args->count==5 && strcmp(arg0,"WRITE")==0) {
            if (success) {
               u32t offs = str2ulong(args->item[2]),
                     val = str2ulong(args->item[3]),
                      sz = str2ulong(args->item[4]);

               if (isdigit(args->item[2][0]) && isdigit(args->item[3][0]) &&
                  isdigit(args->item[4][0]) && (sz==1 || sz==2 || sz==4))
               {
                  hlp_pciewrite(0, d.bus, d.slot, d.func, offs, sz, val);
                  rc = 0;
               }
            }
         }
      }
      // free args clone, returned by str_parseargs() above
      free(args);
   }
   if (rc<0) {
      rc = EINVAL;
      if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   }
   return rc;
}

u32t _std shl_log(const char *cmd, str_list *args) {
   int rc=-1, nopause=0, usedate=0, usetime=1, usecolor=1, level=-1, useflags=0;

   if (args->count>0) {
      static char *argstr   = "f|d|np|nt|nc|l|l:0|l:1|l:2|l:3";
      static short argval[] = {1,1, 1, 0, 0,3,  0,  1,  2,  3};
      // help overrides any other keys
      if (str_findkey(args, "/?", 0)) {
         cmd_shellhelp(cmd,CLR_HELP);
         return 0;
      }
      args = str_parseargs(args, 0, SPA_RETLIST|SPA_NOSLASH, argstr, argval,
                   &useflags, &usedate, &nopause, &usetime, &usecolor,
                   &level, &level, &level, &level, &level);
      // process command
      if (args->count>0 && level>=-1 && level<=3) {
         char *arg0 = args->item[0];
         strupr(arg0);
         // init "paused" output
         cmd_printseq(0, 1, 0);

         if (strcmp(arg0,"SAVE")==0 || strcmp(arg0,"TYPE")==0 || strcmp(arg0,"LIST")==0) {
            int is_save = arg0[0]=='S';

            if (!is_save || args->count==2) {
               u32t flags = LOGTF_DOSTEXT;
               char   *cp;

               if (usedate) flags|=LOGTF_DATE;
               if (usetime) flags|=LOGTF_TIME;
               if (useflags) flags|=LOGTF_FLAGS;
               if (level>=0) flags|=LOGTF_LEVEL|level; else
               if (usecolor) flags|=LOGTF_LEVEL|LOG_GARBAGE;

               cp = log_gettext(flags);

               if (is_save) {
                  FILE *lf = fopen(args->item[1], "wb");
                  if (!lf) rc = errno; else {
                     u32t len = strlen(cp);
                     if (fwrite(cp,1,len,lf)!=len) rc = EIO; else rc = 0;
                     if (fclose(lf) && !rc) rc = errno;
                  }
                  free(cp);
               } else {
                  u32t      ii;
                  str_list *ft = str_settext(cp, 0, 0);
                  free(cp);

                  for (ii=0; ii<ft->count; ii++) {
                     char    *cs = ft->item[ii];
                     u8t   color = 0;

                     if (usecolor) {
                        char *cp = strchr(cs,'[');
                        if (cp && cp-cs<19) {
                           char *lp = cp;
                           while (!isdigit(*lp) && *lp) lp++;
                           switch (*lp) {
                              case '0': color = VIO_COLOR_YELLOW; break;
                              case '1': color = VIO_COLOR_LWHITE; break;
                              case '2': color = VIO_COLOR_WHITE; break;
                              case '3': color = VIO_COLOR_CYAN; break;
                           }
                           if (level<0) {
                              char *ep = strchr(cp,']');
                              if (ep && ep-cp<=4) memmove(cp,ep+1,strlen(ep));
                           }
                        }
                     }
                     if (cmd_printseq(cs, nopause?-1:0, color)) { rc = EINTR; break; }
                  }
                  free(ft);
                  if (rc<0) rc = 0;
               }
            }
         } else
         if (strcmp(arg0,"CLEAR")==0) {
            log_clear();
            rc = 0;
         }
      }
      // free args clone, returned by str_parseargs() above
      free(args);
   }
   if (rc<0) rc = EINVAL;
   if (rc && rc!=EINTR) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}

static int print_comp(char *out, int col, int quiet, int len, u64t ofs, u8t *src, u8t *cmp) {
   int state = 0, jj;

   sprintf(out, "%s%08LX: ", col?ANSI_RESET:"", ofs);
   for (jj=0; jj<len; jj++) {
      char hex[8];
      sprintf(hex, "%02X ", src[jj]);
      if (col && state!=cmp[jj]) {
         strcat(out, cmp[jj]?ANSI_LRED:ANSI_RESET);
         state = cmp[jj];
      }
      strcat(out, hex);
   }
   return cmd_printseq(out, 0, 0);
}

u32t _std shl_comp(const char *cmd, str_list *args) {
   int rc=-1, nopause=0, quiet=0, icase=0, elvlvl = 2,
       usecolor=isatty(fileno(stdout));

   if (args->count>0) {
      static char *argstr   = "q|i|np|nc";
      static short argval[] = {1,1, 1, 0};
      // help overrides any other keys
      if (str_findkey(args, "/?", 0)) {
         cmd_shellhelp(cmd,CLR_HELP);
         return 0;
      }
      args = str_parseargs(args, 0, SPA_RETLIST|SPA_NOSLASH, argstr, argval,
                   &quiet, &icase, &nopause, &usecolor);
      // drop ANSI logic if quiet
      if (quiet) usecolor = 0;
      // process command
      if (args->count>=2) {
         char  *f1n = 0,
               *f2n = 0;
         u32t    ii;
         u64t  ofs1 = 0, ofs2 = 0, len = 0;
         for (ii=0; ii<args->count; ii++) {
            if (args->item[ii][0]=='/' || args->item[ii][0]=='-')
               if (strnicmp(args->item[ii]+1,"O1:",3)==0) {
                  ofs1 = str2uint64(args->item[ii]+4);
                  continue;
               } else
               if (strnicmp(args->item[ii]+1,"O2:",3)==0) {
                  ofs2 = str2uint64(args->item[ii]+4);
                  continue;
               } else
               if (strnicmp(args->item[ii]+1,"LEN:",4)==0) {
                  len = str2uint64(args->item[ii]+5);
                  continue;
               }
            if (!f1n) f1n = args->item[ii]; else
               if (!f2n) f2n = args->item[ii]; else { rc = E2BIG; break; }
         }
         if (!f1n || !f2n) rc = EINVAL;
         if (rc<0) {
            io_handle  fh1 = 0,
                       fh2 = 0;
            u64t     size1, size2;
            int     errsrc = 0;  // file # where error occurs
            qserr      err;
            // init "paused" output
            cmd_printseq(0, nopause?PRNSEQ_INITNOPAUSE:PRNSEQ_INIT, 0);
            // open file 1 and verify/set offset
            err = io_open(f1n, IOFM_OPEN_EXISTING|IOFM_READ|IOFM_SHARE|IOFM_NODEV,
                          &fh1, 0);
            if (!err) err = io_size(fh1, &size1);
            if (!err && ofs1>=size1) err = E_SYS_SEEKERR;
            if (!err && ofs1)
               if (io_seek(fh1, ofs1, IO_SEEK_SET)==FFFF64) err = E_SYS_SEEKERR;
            // the same for the second ...
            if (!err) {
               err = io_open(f2n, IOFM_OPEN_EXISTING|IOFM_READ|IOFM_SHARE|IOFM_NODEV,
                             &fh2, 0);
               errsrc++;
               if (!err) err = io_size(fh2, &size2);
               if (!err && ofs2>=size2) err = E_SYS_SEEKERR;
               if (!err && ofs2)
                  if (io_seek(fh2, ofs2, IO_SEEK_SET)==FFFF64) err = E_SYS_SEEKERR;
            }
            while (!err) {
               int trunc = -1;
               if (!len) {
                  if (size1-ofs1 != size2-ofs2) {
                     if (!quiet)
                        printf("%s size mismatch!\n", !ofs1 && !ofs2 ? "File" : "Data");
                     quiet = 1;
                     break;
                  }
                  len = size1 - ofs1;
               } else {
                  if (ofs1+len > size1) { len = size1 - ofs1; trunc = 0; }
                  if (ofs2+len > size2) { len = size2 - ofs2; trunc = 1; }
               }

               if (!quiet) {
                  cmd_printf("Comparing ");
                  if (len) cmd_printf("%Lu bytes ", len);
                  if (ofs1) cmd_printf("from offset 0x%LX ", ofs1);
                  cmd_printf("%s%s and ", len||ofs1? "in ":"", f1n);
                  if (ofs2) cmd_printf("from offset 0x%LX in ", ofs2);
                  cmd_printf("%s:\n", f2n);
                  if (trunc>=0) cmd_printf("  (length truncated by size of %s)\n",
                     trunc ? f2n : f1n);
               }
               // main action ...
               if (!err) {
                  qs_cpconvert cpl = 0;
                  u16t       svcol;
                  u64t       ready = 0, tls;
                  u32t      buflen = len > _64KB ? _64KB : (u32t)len;
                  char       *buf1 = malloc_th(buflen),
                             *buf2 = malloc_th(buflen),
                              *out = malloc_th(512);
                  if (icase) cpl = NEW(qs_cpconvert);
                  // force ANSI for this thread
                  if (usecolor) {
                     svcol = vio_getcolor();
                     tls   = mt_tlsget(QTLS_FORCEANSI);
                     mt_tlsset(QTLS_FORCEANSI, 2);
                  }

                  while (len) {
                     u32t  rlen = len > _64KB ? _64KB : (u32t)len, ii, jj,
                          lines = Round16(rlen)>>4;

                     if (io_read(fh1, buf1, rlen)!=rlen)
                        { err = E_DSK_ERRREAD; errsrc = 0; break; }
                     if (io_read(fh2, buf2, rlen)!=rlen)
                        { err = E_DSK_ERRREAD; errsrc = 1; break; }

                     for (ii=0; ii<lines; ii++) {
                        u8t   cmp[16], sum = 0,
                             *cb1 = buf1 + (ii<<4),
                             *cb2 = buf2 + (ii<<4);
                        u32t slen = ii==lines-1 ? rlen - (ii<<4) : 16;

                        for (jj=0; jj<slen; jj++) {
                           if (!icase) cmp[jj] = cb1[jj]!=cb2[jj]; else
                              cmp[jj] = !cpl ? toupper(cb1[jj])!=toupper(cb2[jj])
                                 : cpl->toupper(cb1[jj])!=cpl->toupper(cb2[jj]);
                           sum+=cmp[jj];
                        }
                        if (sum) {
                           elvlvl = 1;
                           // one mismatch is enough in quiet mode
                           if (quiet) break;

                           if (cmd_printseq("", 0, 0) ||
                              print_comp(out, usecolor, quiet, slen, ofs1 +
                                 ready + (ii<<4), cb1, cmp) ||
                              print_comp(out, usecolor, quiet, slen, ofs2 +
                                 ready + (ii<<4), cb2, cmp))
                                 { rc = EINTR; break; }
                        }
                     }
                     ready += rlen;
                     len   -= rlen;
                     // exit loop if quiet and mismatch found ...
                     if (quiet && elvlvl==1) len = 0;
                  }
                  if (!len) rc = 0;

                  if (!err && rc<=0 && elvlvl==2) {
                     if (!quiet) cmd_printf("No differences found.\n");
                     elvlvl = 0;
                  } 
                  if (cpl) DELETE(cpl);
                  free(buf2); free(buf1); free(out);
                  if (usecolor) {
                     mt_tlsset(QTLS_FORCEANSI, tls);
                     vio_setcolor(svcol);
                  }
               }

               break;
            }
            if (err) {
               rc = qserr2errno(err);
               if (!quiet) {
                  char *fn = sprintf_dyn("%s :", errsrc?f2n:f1n);
                  cmd_shellerr(EMSG_QS, err, fn);
                  free(fn);
                  // skip error output below
                  quiet = 1;
               }
            }

            if (fh2) io_close(fh2);
            if (fh1) io_close(fh1);
         }
      }
      // free args clone, returned by str_parseargs() above
      free(args);
   }
   if (rc<0) rc = EINVAL;
   if (rc && rc!=EINTR && !quiet) cmd_shellerr(EMSG_CLIB,rc,0);
   // return errorlevel in any case (quiet and not)
   return elvlvl;
}
