//
// QSINIT "fsfat" module
// fs class implementation
//
#include "fsfat.h"
#include "qslog.h"
#include "qsstor.h"
#include "parttab.h"
#include "qsxcpt.h"
#include "qcl/sys/qsvolapi.h"
#include "ff.h"
#include "wchar.h"

// FatFs >= R0.13 defines
#ifndef FF_VOLUMES
#define FF_VOLUMES _VOLUMES
#define FF_MAX_LFN _MAX_LFN
#endif

#define VFAT_SIGN        0x54414656   ///< VFAT string
#define VFATDIR_SIGN     0x45444656   ///< VFDE string
#define VFATFH_SIGN      0x48464656   ///< VFFH string

typedef struct {
   u32t           sign;
   int             vol;     ///< -1 or 0..9
   qs_sysvolume   self;     ///< ptr to self instance
   int        mount_ok;     ///< FatFs mount was ok
   int           exfat;     ///< this is ExFAT partition
} volume_data;

typedef struct {
   u32t           sign;
   DIR             dir;
   FILINFO          fi;
} dir_enum_data;

typedef struct {
   u32t           sign;
   FIL              fh;
   int              ro;     ///< r/o mode
   c16t             fp[QS_MAXPATH+1];
} file_handle;

static FATFS   *extdrv[FF_VOLUMES];

#define instance_ret(type,inst,err)      \
   type *inst = (type*)data;             \
   if (inst->sign!=VFAT_SIGN) return err;

#define instance_void(type,inst)         \
   type *inst = (type*)data;             \
   if (inst->sign!=VFAT_SIGN) return;

static qserr errcvt(FRESULT err) {
   static u32t codes[FR_INVALID_PARAMETER+1] = { 0, E_DSK_ERRREAD, E_SYS_INVPARM,
      E_DSK_NOTREADY, E_SYS_NOFILE, E_SYS_NOPATH, E_SYS_INVNAME, E_SYS_ACCESS,
      E_SYS_EXIST, E_SYS_INVOBJECT, E_DSK_WP, E_SYS_INVPARM, E_DSK_NOTMOUNTED,
      E_DSK_UNKFS, E_SYS_INCOMPPARAMS, E_SYS_TIMEOUT, E_SYS_SHARE, E_SYS_NOMEM,
      E_SYS_FILES, E_SYS_INVPARM};
   return err>FR_INVALID_PARAMETER? E_SYS_INVOBJECT : codes[err];
}

/// convert read err to write err
static qserr wrfunc(qserr err) {
   return err==E_DSK_ERRREAD?E_DSK_ERRWRITE:err;
}

static qserr _exicc fat_initialize(EXI_DATA, u8t vol, u32t flags,
                                   void *bootsec, const char *opts)
{
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   if (vol>=FF_VOLUMES) return E_DSK_BADVOLNAME;

   if (extdrv[vol]) {
      vol_data         *vdta = _extvol+vol;
      FATFS            *fdta = extdrv[vol];
      qserr              err = E_DSK_NOTMOUNTED;
      struct Boot_Record *br = (struct Boot_Record*)bootsec;
      FRESULT           mres;
      u32t           *fsname = 0;
      c16t             dp[8];
      // FatFs mount
      swprintf(dp, 8, L"%c:\\", 'A'+vol);
      mres = f_mount(fdta, dp, 1);

#if 0
      if (mres!=FR_OK) log_printf("%s fatfs mount err %u\n", dp, mres);
#endif
      if (br->BR_BPB.BPB_SecPerFAT)
         if (br->BR_BPB.BPB_RootEntries) fsname = (u32t*)br->BR_EBPB.EBPB_FSType;
            else fsname = (u32t*)((struct Boot_RecordF32*)br)->BR_F32_EBPB.EBPB_FSType;
      // if it is not FAT really - fix error code to acceptable
      if (mres!=FR_OK && (!fsname || (*fsname&0xFFFFFF)!=0x544146) &&
         strnicmp(br->BR_OEM,"EXFAT",5)) mres = FR_NO_FILESYSTEM;
      // have a FAT
      if (mres==FR_OK) {
         vdta->flags|= VDTA_FAT;
         vd->vol     = vol;
      }
      err = errcvt(mres);

      if (!err && (vdta->flags&VDTA_FAT)) {
         u8t vfs = fdta->fs_type;
         strcpy(vdta->fsname, vfs<=FS_FAT16?"FAT":(vfs==FS_EXFAT?"exFAT":"FAT32"));
         vdta->clsize  = fdta->csize;
         vdta->cltotal = (vdta->length - fdta->database)/fdta->csize;
         vdta->serial  = 0;
         vd->mount_ok  = 1;
         vd->exfat     = vfs==FS_EXFAT;
      }
      return err;
   }
   return E_SYS_INVOBJECT;
}

/// unmount volume
static qserr _exicc fat_finalize(EXI_DATA) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   if (vd->vol<0) return E_DSK_NOTMOUNTED; else {
      FRESULT    res = 0;
      vol_data *vdta = _extvol+vd->vol;

      if (vd->mount_ok) {
         c16t     dp[8];
         swprintf(dp, 8, L"%c:\\", 'A'+vd->vol);
         res = f_mount(0, dp, 0);
      }
      vdta->flags &= ~VDTA_FAT;
      vd->vol      = -1;
      vd->mount_ok = 0;
      return errcvt(res);
   }
}

static int _exicc fat_drive(EXI_DATA) {
   instance_ret(volume_data, vd, -1);
   return vd->vol;
}

static int fat_chkpath(volume_data *vd, const c16t *fp) {
   if (fp[1]!=':' || fp[0]-'A'!=vd->vol) return 0;
   return 1;
}

static qserr _exicc fat_open(EXI_DATA, const c16t *name, u32t mode,
                             io_handle_int *pfh, u32t *action)
{
   FRESULT       res;
   static u32t mf[4] = { FA_OPEN_EXISTING, FA_OPEN_ALWAYS, FA_CREATE_NEW,
                         FA_CREATE_ALWAYS };
   file_handle   *fh = 0;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   if (mode>SFOM_CREATE_ALWAYS) return E_SYS_INVPARM;
   if (!pfh || !name || !action) return E_SYS_ZEROPTR;
   if (wcslen(name)>QS_MAXPATH) return E_SYS_LONGNAME;

   if (!fat_chkpath(vd, name)) return E_SYS_INVPARM;
   *pfh  = 0;
   fh    = (file_handle*)malloc(sizeof(file_handle));
   res   = f_open(&fh->fh, name, FA_READ|FA_WRITE|mf[mode]);
   if (res==FR_WRITE_PROTECTED) {
      fh->ro = 1;
      /* patched FatFs code returns "disk write protected" instead
         of "no access" err in case of r/o attr on file. This causes
         reopening in r/o mode. On success we`ll have this file as
         r/o over system - until the last close */
      res    = f_open(&fh->fh, name, FA_READ|mf[mode]);
      if (res==FR_WRITE_PROTECTED) res = FR_DENIED;
   } else
   if (!res) fh->ro = 0;
   // used only in finfo
   if (!res) {
      wcsncpy(fh->fp, name, QS_MAXPATH+1);
      fh->fp[QS_MAXPATH] = 0;
   }
   if (res) { free(fh); return errcvt(res); }

   // FatFs set it if file was actually created
   if ((fh->fh.flag&FA_CREATE_ALWAYS)) *action = IOFN_CREATED; else
      if ((mode&FA_CREATE_ALWAYS)) *action = IOFN_TRUNCATED; else
         *action = IOFN_EXISTED;

   fh->sign = VFATFH_SIGN;
   *pfh     = (io_handle_int)fh;

   return fh->ro?E_SYS_READONLY:0;
}

static qserr _exicc fat_read(EXI_DATA, io_handle_int file, u64t pos,
                             void *buffer, u32t *size)
{
   file_handle   *fh = (file_handle*)file;
   FRESULT       res = 0;
   UINT           br = 0;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   if (!size) return E_SYS_ZEROPTR;
   if (!buffer || fh->sign!=VFATFH_SIGN) { *size = 0; return E_SYS_INVPARM; }
   if (pos>=fh->fh.obj.objsize) { *size = 0; return E_SYS_EOF; }
   if (!*size) return 0;

   if (pos!=fh->fh.fptr) res = f_lseek(&fh->fh, pos);
   if (!res) res = f_read(&fh->fh, buffer, *size, &br);
   *size = br;

   return errcvt(res);
}

static qserr _exicc fat_write(EXI_DATA, io_handle_int file, u64t pos,
                              void *buffer, u32t *size)
{
   file_handle   *fh = (file_handle*)file;
   FRESULT       res = 0;
   UINT           bw = 0;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   if (!buffer || !size) return E_SYS_ZEROPTR;
   if (fh->sign!=VFATFH_SIGN) return E_SYS_INVPARM;
   if (!vd->exfat && (pos>=FFFF || pos+*size>=FFFF)) return E_SYS_FSLIMIT;
   if (fh->ro) return E_SYS_ACCESS;
   if (!*size) return 0;

   if (pos!=fh->fh.fptr) res = f_lseek(&fh->fh, pos);
   if (!res) res = f_write(&fh->fh, buffer, *size, &bw);
   // check after FatFs - it can leave a swine for us, reporting no error
   if (!res && bw<*size) res = FR_NOT_READY;
   *size = bw;
   return wrfunc(errcvt(res));
}

static qserr _exicc fat_flush(EXI_DATA, io_handle_int file) {
   file_handle   *fh = (file_handle*)file;
//   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (fh->sign!=VFATFH_SIGN) return E_SYS_INVPARM;

   return wrfunc(errcvt(f_sync(&fh->fh)));
}

static qserr _exicc fat_close(EXI_DATA, io_handle_int file) {
   FRESULT       res;
   file_handle   *fh = (file_handle*)file;
//   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (fh->sign!=VFATFH_SIGN) return E_SYS_INVPARM;

   res = f_close(&fh->fh);
   fh->sign = 0;
   free(fh);
   return errcvt(res);
}

static qserr _exicc fat_size(EXI_DATA, io_handle_int file, u64t *size) {
   file_handle   *fh = (file_handle*)file;
//   instance_ret(volume_data, vd, 0);
   if (fh->sign!=VFATFH_SIGN) return E_SYS_INVPARM;
   if (!size) return E_SYS_ZEROPTR;
   *size = fh->fh.obj.objsize;
   return 0;
}

static qserr _exicc fat_setsize(EXI_DATA, io_handle_int file, u64t newsize) {
   FRESULT       res = FR_INT_ERR;
   file_handle   *fh = (file_handle*)file;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   if (fh->sign!=VFATFH_SIGN) return E_SYS_INVPARM;
   if (fh->ro) return E_SYS_ACCESS;
   if (!vd->exfat && newsize>=FFFF) return E_SYS_FSLIMIT;
   // empty/new file? try to use f_expand()
   if (newsize && !fh->fh.obj.sclust)
      res = f_expand(&fh->fh, newsize, 1);
   // default way or failed f_expand()
   if (res) {
      res = 0;
      if (newsize!=fh->fh.fptr) res = f_lseek(&fh->fh, newsize);
      if (!res && fh->fh.fptr!=newsize) return E_DSK_DISKFULL;
      if (!res && newsize<fh->fh.obj.objsize) res = f_truncate(&fh->fh);
   }
   if (!res) res = f_sync(&fh->fh);

   return errcvt(res);
}

/* !function assumes that offsets in io_direntry_info & io_handle_info are equal!
   name=-1 - io_handle_info, name=0 - normal name, name=1 - short name
   also short name altname is always returned in altname (if !=0) */
static qserr fat_pathinfo_int(const c16t *cpath, io_handle_info *info,
                              c16t *name, c16t *altname)
{
   FRESULT  res;
   FILINFO  *fi;
   // f_stat() does not support root
   if (cpath[2]=='\\' && cpath[3]==0) {
      io_timetoio(&info->wtime, ERROR_TIME_T);
      info->atime  = info->wtime;
      info->ctime  = info->wtime;
      info->attrs  = IOFA_DIR;
      info->fileno = 0;
      info->size   = 0;
      info->vsize  = 0;
      return 0;
   }
   // do not use stack for such large alloc
   fi = (FILINFO*)calloc_th(1, sizeof(FILINFO));
   if (!fi) return E_SYS_NOMEM;
   res = f_stat(cpath, fi);
   if (res==FR_OK) {
      time_t  tmv;
      info->attrs  = fi->fattrib;
      info->fileno = 0;
      info->size   = fi->fsize;
      info->vsize  = fi->fsize;

      tmv = dostimetotime((u32t)fi->fdate<<16|fi->ftime);
      io_timetoio(&info->atime, tmv);
      io_timetoio(&info->wtime, tmv);
      tmv = dostimetotime((u32t)fi->cdate<<16|fi->ctime);
      io_timetoio(&info->ctime, tmv);

      if (name) {
         wcsncpy(name, fi->fname, FF_MAX_LFN + 1);
         name[FF_MAX_LFN] = 0;
      }
      if (fi->fname[0] && fi->altname[0]) {
         info->attrs |= IOFA_ALTNAME;
         if (altname) {
            wcsncpy(altname, fi->altname, 13);
            altname[12] = 0;
         }
      } else
      if (altname) *altname = 0;
   }
   free(fi);
   return res?errcvt(res):0;
}

static qserr _exicc fat_pathinfo(EXI_DATA, const c16t *path, io_handle_info *info,
                                 c16t *name, c16t *altname)
{
   qserr    res;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (!path || !info) return E_SYS_ZEROPTR;

   if (!fat_chkpath(vd, path)) return E_SYS_INVPARM;
   return fat_pathinfo_int(path, info, name, altname);
}

static qserr _exicc fat_finfo(EXI_DATA, io_handle_int file, io_handle_info *info)
{
   file_handle *fh = (file_handle*)file;
//   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (fh->sign!=VFATFH_SIGN) return E_SYS_INVPARM;

   return fat_pathinfo_int(fh->fp, info, 0, 0);
}

static qserr _exicc fat_setinfo(EXI_DATA, const c16t *path, io_handle_info *info,
                                u32t flags)
{
   FRESULT       res;
   qserr         err = 0;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (!path || !info) return E_SYS_ZEROPTR;

   if (!fat_chkpath(vd, path)) return E_SYS_INVPARM;

   if (flags&IOSI_ATTRS) {
      res = f_chmod(path, info->attrs, AM_RDO|AM_ARC|AM_SYS|AM_HID);
      if (res!=FR_OK) err = errcvt(res);
   }
   if (!err && (flags&(IOSI_CTIME|IOSI_WTIME|IOSI_ATIME))) {
      io_handle_info   ci;
      u32t             dt;
      io_time  *pct, *pwt;
      FILINFO          fi;
      // combine two times into one
      if ((flags&IOSI_WTIME) && (flags&IOSI_ATIME)==0) pwt = &info->wtime;
         else pwt = flags&IOSI_ATIME?&info->atime:0;
      pct = flags&IOSI_CTIME?&info->ctime:0;
      // read missing one
      if (!pct || !pwt) {
         qserr rc = fat_pathinfo(data, 0, path, &ci, 0, 0);
         if (rc) return rc;
         if (!pct) pct = &ci.ctime;
         if (!pwt) pwt = &ci.atime;
      }
      // a bit of paranoya
      fi.fname[0] = 0;

      dt = timetodostime(io_iototime(pwt));
      fi.fdate = dt>>16;
      fi.ftime = (WORD)dt;
      dt = timetodostime(io_iototime(pct));
      fi.cdate = dt>>16;
      fi.ctime = (WORD)dt;

      res = f_utime(path, &fi);
      if (res!=FR_OK) err = errcvt(res);
   }
   return err;
}

static qserr _exicc fat_setexattr(EXI_DATA, const c16t *path, const c16t *aname, void *buffer, u32t size) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   return E_SYS_UNSUPPORTED;
}

static qserr _exicc fat_getexattr(EXI_DATA, const c16t *path, const c16t *aname, void *buffer, u32t *size) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   return E_SYS_UNSUPPORTED;
}

static str_list* _exicc fat_lstexattr(EXI_DATA, const c16t *path) {
   instance_ret(volume_data, vd, 0);
   return 0;
}

// warning! it acts as rmdir too!!
static qserr _exicc fat_remove(EXI_DATA, const c16t *path) {
   qserr          rc;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (!path) return E_SYS_ZEROPTR;

   if (!fat_chkpath(vd, path)) return E_SYS_INVPARM;
   return errcvt(f_unlink(path));
}

static qserr _exicc fat_renmove(EXI_DATA, const c16t *oldpath, const c16t *newpath) {
   qserr          rc;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (!oldpath || !newpath) return E_SYS_ZEROPTR;

   if (!fat_chkpath(vd, oldpath)) return E_SYS_INVPARM;
   if (!fat_chkpath(vd, newpath)) rc = E_SYS_INVPARM; else
      rc  = errcvt(f_rename(oldpath,newpath));

   return rc;
}

static qserr _exicc fat_makepath(EXI_DATA, const c16t *path) {
   FRESULT    res;
   c16t  *fp, *cp;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   if (!fat_chkpath(vd, path)) return E_SYS_INVPARM;
   fp = _wcsdup(path);
   // create path.
   cp = wcschr(fp,'\\');
   while (cp) {
      *cp  = 0;
      res   = f_mkdir(fp);
      *cp++= '\\';
      if (res!=FR_EXIST && res!=FR_OK) break;
      cp = wcschr(cp,'\\');
   }
   // final target
   if (!cp) res = f_mkdir(fp);
   free(fp);
   return errcvt(res);
}

static qserr _exicc fat_diropen(EXI_DATA, const c16t *path, dir_handle_int *pdh) {
   dir_enum_data  *dd;
   qserr           rc;
   FRESULT        res;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (!path || !pdh) return E_SYS_ZEROPTR;
   *pdh = 0;
   // dir handle storage
   dd   = (dir_enum_data*)malloc(sizeof(dir_enum_data));
   if (!dd) return E_SYS_NOMEM;
   if (wcslen(path)>QS_MAXPATH) rc = E_SYS_LONGNAME; else {
      if (!fat_chkpath(vd, path)) rc = E_SYS_INVPARM; else {
         rc = errcvt(f_opendir(&dd->dir, path));
         if (!rc) {
            dd->sign = VFATDIR_SIGN;
            *pdh     = (dir_handle_int)dd;
         }
      }
   }
   if (rc && dd) free(dd);
   return rc;
}

static qserr _exicc fat_dirnext(EXI_DATA, dir_handle_int dh, io_direntry_infow *de,
                                int altname)
{
   dir_enum_data *dd = (dir_enum_data*)dh;
   FRESULT       res;
   time_t        tmv;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (!dd || !de) return E_SYS_ZEROPTR;
   if (dd->sign!=VFATDIR_SIGN) return E_SYS_INVPARM;

   dd->fi.fname[0]   = 0;
   dd->fi.altname[0] = 0;
   res        = f_readdir(&dd->dir,&dd->fi);
   // no file or error?
   if (res!=FR_OK || !dd->fi.fname[0]) {
      if (res==FR_OK) res=FR_NO_FILE;
      return errcvt(res);
   }
   de->attrs  = dd->fi.fattrib;
   de->fileno = 0;
   de->size   = dd->fi.fsize;
   de->vsize  = de->size;

   if (dd->fi.fname[0] && (!altname || !dd->fi.altname[0])) {
      wcsncpy(de->name, dd->fi.fname, FF_MAX_LFN + 1);
   } else {
      wcsncpy(de->name, dd->fi.altname, 13);
      de->name[12] = 0;
   }
   de->name[FF_MAX_LFN] = 0;
   // short name available
   if (dd->fi.fname[0] && dd->fi.altname[0]) de->attrs |= IOFA_ALTNAME;

   tmv = dostimetotime((u32t)dd->fi.fdate<<16|dd->fi.ftime);
   io_timetoio(&de->atime, tmv);
   io_timetoio(&de->wtime, tmv);
   tmv = dostimetotime((u32t)dd->fi.cdate<<16|dd->fi.ctime);
   io_timetoio(&de->ctime, tmv);

   return 0;
}

static qserr _exicc fat_dirclose(EXI_DATA, dir_handle_int dh) {
   dir_enum_data *dd = (dir_enum_data*)dh;
   FRESULT       res;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (!dd) return E_SYS_ZEROPTR;
   if (dd->sign!=VFATDIR_SIGN) return E_SYS_INVPARM;

   dd->sign = 0;
   res = f_closedir(&dd->dir);
   free(dd);

   return errcvt(res);
}

static qserr _exicc fat_volinfo(EXI_DATA, disk_volume_data *info, int fast) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   if (!info) return E_SYS_ZEROPTR;
   memset(info, 0, sizeof(disk_volume_data));

   if (vd->vol>=0) {
      vol_data  *vdta = _extvol+vd->vol;
      FATFS     *fdta = extdrv[vd->vol];
      // mounted?
      if (vdta->flags&VDTA_ON) {
         info->StartSector  = vdta->start;
         info->TotalSectors = vdta->length;
         info->Disk         = vdta->disk;
         info->SectorSize   = vdta->sectorsize;
         // was it really mounted?
         if ((vdta->flags&VDTA_FAT)!=0) {
            FATFS *pfat = 0;
            DWORD nclst = 0;
            c16t     cp[3];
            cp[0] = 'A'+vd->vol; cp[1]=':'; cp[2]=0;

            info->DataSector  = fdta->database;
            info->FsVer       = fdta->fs_type;
            info->FSizeLim    = info->FsVer==FST_EXFAT ? fdta->csize*0x7FFFFFFDLL : FFFF;
            // update info to actual one
            if (!fast) {
               if (f_getfree(cp,&nclst,&pfat)==FR_OK) vdta->clfree = nclst;
               if (!vdta->serial) {
                  c16t vlabel[32];
                  vlabel[0] = 0;
                  f_getlabel(cp, vlabel, &vdta->serial);
                  if (vlabel[0])
                     if (!hlp_wtoa(vlabel, vdta->label, sizeof(vdta->label), 0, '_'))
                        vlabel[0] = 0;
                  // no label or conversion error (should never occurs)
                  if (!vlabel[0]) vdta->label[0] = 0;
               }
            }
         } else {
            info->DataSector  = 0;
            info->FSizeLim    = 0;
         }
         info->ClBad     = vdta->badclus;
         info->SerialNum = vdta->serial;
         info->ClSize    = vdta->clsize;
         info->ClAvail   = vdta->clfree;
         info->ClTotal   = vdta->cltotal;
         strcpy(info->Label,vdta->label);
         strcpy(info->FsName,vdta->fsname);

         return 0;
      }
   }
   return E_DSK_NOTMOUNTED;
}

static qserr _exicc fat_setlabel(EXI_DATA, const c16t *label) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   if (vd->vol>=0) {
      vol_data  *vdta = _extvol+vd->vol;
      FATFS     *fdta = extdrv[vd->vol];

      // mounted & FAT?
      if (vd->vol<=DISK_LDR || (vdta->flags&VDTA_ON)!=0) {
         FRESULT  res;
         c16t      lb[14];
         lb[0] = vd->vol+'A';
         lb[1] = ':';
         if (label) {
            wcsncpy(lb+2,label,12);
            lb[13] = 0;
         } else
            lb[2] = 0;

         res = f_setlabel(lb);
         // drop file label cache to force its reading in volinfo
         if (res==FR_OK) vdta->serial = 0;
         return errcvt(res);
      }
   }
   return E_DSK_NOTMOUNTED;
}

static u32t _exicc fat_avail(EXI_DATA) {
   instance_ret(volume_data, vd, 0);
   return vd->exfat?0:SFAF_ALTNAME;
}

static qserr _exicc fat_service(EXI_DATA, u32t func, void *info) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   return E_SYS_UNSUPPORTED;
}

static void _std fat_init(void *instance, void *userdata) {
   volume_data *vd = (volume_data*)userdata;
   vd->sign        = VFAT_SIGN;
   vd->vol         = -1;
   vd->self        = (qs_sysvolume)instance;
   vd->mount_ok    = 0;
}

static void _std fat_done(void *instance, void *userdata) {
   volume_data *vd = (volume_data*)userdata;
   if (vd->sign==VFAT_SIGN)
      if (vd->vol>=0) log_printf("Volume %d fini???\n", vd->vol);
   memset(vd, 0, sizeof(volume_data));
}

static void *m_list[] = { fat_initialize, fat_finalize, fat_avail, fat_volinfo,
   fat_setlabel, fat_drive, fat_open, fat_read, fat_write, fat_flush,
   fat_close, fat_size, fat_setsize, fat_finfo, fat_setexattr, fat_getexattr,
   fat_lstexattr, fat_remove, fat_renmove, fat_makepath, fat_remove,
   fat_pathinfo, fat_setinfo, fat_diropen, fat_dirnext, fat_dirclose ,fat_service
};

u32t register_fatio(void) {
   u32t   ii, cid;
   for (ii=0; ii<FF_VOLUMES; ii++) extdrv[ii] = (FATFS*)malloc(sizeof(FATFS));
   // register private(unnamed) class
   cid = exi_register(FS_CLASS_NAME_PFX "fat", m_list, sizeof(m_list)/sizeof(void*),
                      sizeof(volume_data), EXIC_PRIVATE, fat_init, fat_done, 0);
   // register fs
   if (cid) {
      qserr res = io_registerfs(cid);
      if (res) {
         log_printf("fatfs: regfs err %08X\n", res);
         exi_unregister(cid);
         cid = 0;
      }
   }
   return cid;
}
