//
// QSINIT "fslib" module
// ISOFS filesystem r/o support
// ------------------------------------------------------------
// * joliet names stored in UNICODE, regular volumes - as ascii.
// * if name violates ISO in non-joliet volume (asc codes >=128) - it
//   tranlated to unicode using the currect selected codepage
//
#include "fslib.h"
#include "sys/fs/isofs.h"

#define ISO_SIGN         0x4F494443   ///< CDIO string
#define ISODIR_SIGN      0x45444443   ///< CDDE string
#define ISOFH_SIGN       0x48464443   ///< CDFH string

#define ISO_DEBUG        0            ///< debug message level

struct _dir_data;

/// directory entry data (32 bytes)
typedef struct {
   u64t               size;  ///< file size
   struct _dir_data    *dd;  ///< directory data (if available & directory is)
   cc_loc             *loc;
   io_time           ctime;
   union {
      char           *name;  ///< iso: ascii name
      wchar_t       *wname;  ///< joliet: unicode name
   };
   u8t               attrs;
   u8t               nmlen;
   u16t               res2;
} dir_entry;

struct _dir_data {
   u32t               sign;
   u32t            entries;
   struct _dir_data   *pdd;  ///< parent directory data
   u32t          badname:1;  ///< directory has some zero hash entries
   u32t          pindex:31;  ///< index in the parent directory
   u32t              usage;  ///< usage counter
   u32t              *hash;  ///< name hash array
   dir_entry         de[1];
};

typedef struct _dir_data dir_data;

typedef struct {
   u32t               sign;
   u32t               vdsk;  ///< 0 or volume|QDSK_VOLUME
   qs_sysvolume       self;  ///< ptr to self instance
   int            mount_ok;  ///< mount was ok
   int              sierra;
   int              joliet;
   iso_desc            *pd;
   dir_data          *root;
   u32t           vdendpos;
   u32t              nclus;  ///< disk length from the header
} volume_data;

typedef struct {
   u32t               sign;
   dir_data            *dd;
   u32t                pos;
} dir_enum_data;

typedef struct {
   u32t               sign;
   dir_entry           *de;
   dir_data            *dd;
   qs_filecc            fd;
   c16t                 fp[1];
} file_handle;

#define instance_ret(type,inst,err)      \
   type *inst = (type*)data;             \
   if (inst->sign!=ISO_SIGN) return err;

#define instance_void(type,inst)         \
   type *inst = (type*)data;             \
   if (inst->sign!=ISO_SIGN) return;

/** try to free directory and all its children.
    call free only for directories with zero usage count.
    @return 1 on success */
static int iso_free_dir(dir_data *dd) {
   u32t ii, rc = 1;
   if (!dd) return 1;

   for (ii=0; ii<dd->entries; ii++) {
      dir_entry *de = dd->de + ii;
      if (de->attrs&IOFA_DIR) {
         int rrc = iso_free_dir(de->dd);
         if (rrc) de->dd = 0;
         rc &= rrc;
      }
   }
   if (rc)
      if (dd->usage) rc = 0; else {
         dd->sign = 0;
         free(dd);
      }
   return rc;
}

static u32t _std data_cache_notify(u32t code, void *usr) {
   volume_data *vd = (volume_data*)usr;
   if (vd->sign!=ISO_SIGN) return DCNR_GONE;

   if (vd->root && code==DCN_MEM) {
      log_it(3, "vol %X mem free\n", vd->vdsk);
      if (iso_free_dir(vd->root)) vd->root = 0;
   }
   return DCNR_SAFE;
}

/// convert Sierra header to ISO-9660
static void sierra_to_iso(iso_desc *pd) {
   hsg_desc *sd = (hsg_desc*)pd;

   memmove(&pd->type,      &sd->type,      offsetof(iso_desc, ptab_le));
   memmove(&pd->ptab_le,   &sd->ptab_le,   8);
   memmove(&pd->ptab_be,   &sd->ptab_be,   8);
   memmove(&pd->root,      &sd->root,      sizeof(iso_dirrec));
   memmove(&pd->volset_id, &sd->volset_id, 128*4);

   pd->fsver = sd->fsver;
   // field shift swaps direction here, so just wipe them ...
   memset(&pd->cpr_id, 0, offsetof(iso_desc, fsver) - offsetof(iso_desc, cpr_id));
}

static qserr _exicc iso_initialize(EXI_DATA, u8t vol, u32t flags,
                                   void *bootsec, const char *opts)
{
   vol_data         *vdta = _extvol+vol;
   qserr              err = E_DSK_UNKFS;
   struct Boot_Record *br = (struct Boot_Record*)bootsec;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   vd->mount_ok = 0;
   vd->root     = 0;

   if (vdta->sectorsize!=ISO_SECTOR_SIZE) err = E_DSK_SSIZE; else
   if (vdta->length) {
      iso_desc *pd = malloc_th(ISO_SECTOR_SIZE);
      hsg_desc *sd = (hsg_desc*)pd;
      u32t  ii, si = 16;
      char  *uopts = opts ? strupr(strdup(opts)) : 0;
      int   sierra = -1,
          nojoliet = uopts ? (strstr(uopts,"-NOJOLIET") ? 1 : 0) : 0;
      vd->vdsk     = vol|QDSK_VOLUME;

      if (uopts) free(uopts);

      // log_it(3, "vol %X, cdfs try\n", vol);

      while (1) {
         if (!hlp_diskread(vol|QDSK_VOLUME, si, 1, pd)) {
            if (err) err = E_DSK_ERRREAD;
            break;
         }
         if (sierra<0) sierra = memcmp(sd->id,ISO_SIERRA_ID,5) ? 0 : 1;

         if (!sierra && memcmp(pd->id,ISO_STANDARD_ID,5) ||
              sierra && memcmp(sd->id,ISO_SIERRA_ID,5))
         {
            if (!err) vd->vdendpos = si;
            break;
         } else
         if (!sierra && pd->type==ISO_VD_END || sierra && sd->type==ISO_VD_END) {
            vd->vdendpos = si;
            break;
         } else
         if (!sierra && pd->type==ISO_VD_PRIMARY ||
              sierra && sd->type==ISO_VD_PRIMARY ||
             !sierra && !nojoliet && pd->type==ISO_VD_SUPPLEMENTARY &&
            (!memcmp(&pd->escape, "%/@", 3) || !memcmp(&pd->escape, "%/C", 3) ||
             !memcmp(&pd->escape, "%/E", 3)))
         {
            if (sierra) sierra_to_iso(pd);
            // we have a descriptor already, just check the secondary
            if (!err) {
               // PRIMARY should be first, but if not - it just will be ignored
               if (pd->type==ISO_VD_SUPPLEMENTARY)
                  // volume size match to the PRIMARY
                  if (vdta->cltotal==pd->vspacesize.le) {
                     // keep PRIMARY label and only use the directory data
                     free(vd->pd);
                     mem_shareblock(pd);
                     vd->pd     = pd;
                     vd->joliet = 1;
                     log_it(3, "joliet root %X %u\n", pd->root.extent.le,
                        pd->root.size.le);
                     // alloc a new one for next reads
                     pd = malloc_th(ISO_SECTOR_SIZE);
                  }
            } else
            // volume size should match the header
            if (vdta->length<pd->vspacesize.le) {
               err = E_DSK_MOUNTERR;
               break;
            } else {
               vdta->cltotal = pd->vspacesize.le;
               // get mount time as VSN
               vdta->serial  = tm_getdate();
               // how can place secondary header first?
               vd->joliet    = pd->type==ISO_VD_SUPPLEMENTARY;

               if (vd->joliet) {
                  /* use dumb conversion for the volume label, anyway a primary
                     should be used because should always be the first */
                  for (ii=0; ii<sizeof(pd->volid)/2; ii++) {
                     u16t ch = bswap16(((u16t*)&pd->volid)[ii]);
                     vdta->label[ii] = ch<' ' || ch>=0x7F ? ' ' : ch;
                  }
                  vdta->label[16] = 0;
               } else {
                  for (ii=0; ii<sizeof(pd->volid); ii++)
                     if (pd->volid[ii]<' ' || pd->volid[ii]>=0x7F)
                        pd->volid[ii] = ' ';
                  memcpy(vdta->label, pd->volid, 32);
                  vdta->label[32] = 0;
               }
               trimright(vdta->label, " ");
            
               err = 0;
               mem_shareblock(pd);
               vd->pd     = pd;
               vd->sierra = sierra;
               log_it(3, "vol %X, root %X %u%s\n", vd->vdsk, pd->root.extent.le,
                  pd->root.size.le, sierra?" (High Sierra)" : "");
#if ISO_DEBUG>0
               log_it(3, "   vsetsize %u\n", pd->vsetsize.le);
               log_it(3, "   blksize  %u\n", pd->blocksize.le);
               log_it(3, "   ptabsize %u\n", pd->ptab_size.le);
               log_it(3, "   ptabpos  %X\n", pd->ptab_le);
               log_it(3, "   root usz %u\n", pd->root.unitsize);
               log_it(3, "   root flg %X\n", REC_FLAGS(&pd->root, sierra));
#endif
               // alloc a new one for next reads
               pd = malloc_th(ISO_SECTOR_SIZE);
            }
         }
         si++;
      }
      free(pd);
      log_it(3, "vol %X, cdfs rc=%x\n", vol, err);
   }

   if (!err) {
      vdta->clsize = 1;
      vd->mount_ok = 1;
      strcpy(vdta->fsname, "CDFS");
      sys_dcenew(vd, DCF_NOTIMER, 0, data_cache_notify);
   } else
      vd->vdsk = 0;
   return err;
}

static void iso_freedata(volume_data *vd) {
   if (vd->root) {
      if (!iso_free_dir(vd->root))
         log_printf("Vol %X free dir failed!\n", vd->vdsk);
      vd->root = 0;
   }
   if (vd->pd) { free(vd->pd); vd->pd = 0; }

   sys_dcedel(vd);

   vd->vdsk     = 0;
   vd->mount_ok = 0;
}

/// unmount volume
static qserr _exicc iso_finalize(EXI_DATA) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   if (!vd->vdsk) return E_DSK_NOTMOUNTED; else {
      iso_freedata(vd);
      return 0;
   }
}

static int _exicc iso_drive(EXI_DATA) {
   instance_ret(volume_data, vd, -1);
   return vd->vdsk ? vd->vdsk&QDSK_DISKMASK : -1;
}

static void getname_a(dir_entry *de, c16t *dst) {
   u32t ii;
   for (ii=0; ii<de->nmlen; ii++) {
      u16t ch = ff_convert(de->name[ii],1);
      dst[ii] = ch?ch:'?';
   }
   dst[de->nmlen] = 0;
}

c16t char_to_wu(char cc) {
   return (u8t)cc>=0x80 ? towupper(ff_convert((u8t)cc,1)) : toupper(cc);
}

/** calculate name hash for UPPERcased name.
    @param   name         name to calc
    @param   len          # of chars from name to use (or 0)
    @return hash value or 0 if at least one char >=128 found */
static u32t calc_hash_a(char *name, u32t len) {
   u32t res = 0;
   if (!len) len = FFFF;
   while (*name && len--) {
      char ch = *name++;
      if (ch>=128) return 0;
      res += toupper(ch);
      res = (res>>2) + (res<<30);
   }
   return res;
}

// for both char & wchar
#define FIX_NAME(len, name)                                         \
   if (len>2)                                                       \
      if (name[len-1]=='1' && name[len-2]==';') name[len-=2] = 0;   \
   if (len>1)                                                       \
      if (name[len-1]=='.') name[--len] = 0;

/** read entire directory into the linear list.
    @param   vd           volume data
    @param   sector       starting sector
    @param   dirlen       length in sectors
    @param   [out] dlist  directory data (in the single heap block) in module
                          owned heap block
    @return error code or 0 */
static qserr read_dir(volume_data *vd, u32t sector, u32t dirlen, dir_data **dlist) {
   dir_data    *dd = 0;
   qserr       res = 0;
   char  *namedata,
          *namechk;
   void   *readbuf;
   cc_loc  *locptr;
   u32t         ii, idx, absidx, mbsize,
           namelen = 6,  // ".." string (*2 for joliet)
              fcnt = 0,  // # of files (multi-extent counted as one)
            extcnt = 0;  // real # of files
#if ISO_DEBUG>0
   log_it(3, "read_dir(,%X,%u)\n", sector, dirlen);
#endif
   // just drop cases that looks invalid (dir size > 8Mb)
   if ((u64t)dirlen<<ISO_SECTOR_BITS > _8MB) return E_DSK_FSSTRUCT;

   readbuf = hlp_memalloc(dirlen<<ISO_SECTOR_BITS, QSMA_LOCAL|QSMA_NOCLEAR|QSMA_RETERR);
   if (!readbuf) return E_SYS_NOMEM;
   // read the entire directory!
   if (hlp_diskread(vd->vdsk, sector, dirlen, readbuf) != dirlen) {
      hlp_memfree(readbuf);
      return E_DSK_ERRREAD;
   }
   // calculate memory usage
   for (ii=0; ii<dirlen; ii++) {
      u8t       *sptr = (u8t*)readbuf + (ii<<ISO_SECTOR_BITS);
      iso_dirrec  *de = (iso_dirrec*)sptr, *pde = 0;
      u32t        rem = ISO_SECTOR_SIZE, mextlen;
      char      *mext = 0;

      while (rem && de->length) {
         // skip . & ..
         if (absidx<=1) ; else
         // check that name length fits in recorded size
         if (de->namelen + FIELDOFFSET(iso_dirrec,name) > (u32t)de->length) {
            log_it(3, "vol %X, invalid dir record in sector %X ofs %X\n",
               vd->vdsk, sector, (u8t*)de - sptr);
            // set a break for the second pass
            de->length = 0;
            break;
         } else {
            // convert Sierra format record
            if (vd->sierra) {
               de->flags    = de->date.gmt;
               de->date.gmt = 0;
            }

            if (mext) {
               if (mextlen == de->namelen && !memcmp(de->name, mext, mextlen)) {
                  // multi-extent is valid
               } else {
                  mext = 0;
                  // simplify life for the second pass
                  pde->flags &= ~ISODF_MULTI;
               }
            }
            if (!mext) {
               fcnt++;
               if (vd->joliet)
                  // length in bytes + \0
                  namelen += de->namelen + 2;
               else
                  namelen += de->namelen + 1;
               // first slice of the multi-extent file
               if (de->flags & ISODF_MULTI) {
                  mext    = de->name;
                  mextlen = de->namelen;
               }
            }
            extcnt++;
         }
         pde  = de;
         rem -= de->length;
         if (rem) de = (iso_dirrec*)((u8t*)de + de->length);
         absidx++;
      }
   }
#if ISO_DEBUG>0
   log_it(3, "read_dir(,%X,) - name len %u, %u files\n", sector, namelen, fcnt);
#endif
   // not a root - add entry for ..
   if (sector!=vd->pd->root.extent.le) fcnt++;
   // allocate single block for whole directory
   mbsize      = sizeof(dir_data) + sizeof(dir_entry)*fcnt + 4*fcnt +
                 namelen + sizeof(cc_loc)*(extcnt+fcnt);
   dd          = (dir_data*)malloc(mbsize);
   dd->sign    = ISODIR_SIGN;
   dd->usage   = 0;
   dd->pdd     = 0;
   dd->pindex  = 0;
   dd->badname = 0;
   dd->hash    = (u32t*)((u8t*)(dd+1) + sizeof(dir_entry)*(fcnt-1));
   locptr      = (cc_loc*)(dd->hash + fcnt);
   namedata    = (char*)(locptr + (extcnt+fcnt));
   namechk     = namedata;
   idx         = 0;

   for (ii=0, idx=0, absidx=0; !res && ii<dirlen; ii++) {
      u8t       *sptr = (u8t*)readbuf + (ii<<ISO_SECTOR_BITS);
      iso_dirrec  *de = (iso_dirrec*)sptr;
      u32t        rem = ISO_SECTOR_SIZE;
      int        mext = 0;
      // first entry should be .
      if (absidx==0 && ((de->flags&ISODF_DIR)==0 || de->namelen>1)) {
         res = E_DSK_FSSTRUCT;
         break;
      }
      // per sector loop
      while (rem && de->length) {
         u32t szinsec = Round2k(de->size.le) >> ISO_SECTOR_BITS;
         // skip . & .. (in the root)
         if (absidx>1 || absidx==1 && sector!=vd->pd->root.extent.le) {
            dir_entry *rde;
            // name compared in first pass (above)
            if (mext) {
               rde = dd->de + (idx - 1);
               // can we merge the extents?
               if (rde->loc->cluster + rde->loc->len == de->extent.le) {
                  rde->loc->len  += szinsec;
               } else {
                  locptr->cluster = de->extent.le;
                  locptr->len     = szinsec;
                  locptr++;
               }
               rde->size += de->size.le;
            } else {
               rde = dd->de + idx;
               // verify .. record
               if (absidx==1 && (de->name[0]!=1 || (de->flags&ISODF_DIR)==0)) {
                  res = E_DSK_FSSTRUCT;
                  break;
               }
               rde->size  = de->size.le;
               rde->attrs = (de->flags & ISODF_HIDDEN ? IOFA_HIDDEN : 0) |
                            (de->flags & ISODF_DIR ? IOFA_DIR : 0);
               rde->dd    = 0;
               rde->ctime.dsec = 0;
               rde->ctime.sec  = de->date.sec;
               rde->ctime.min  = de->date.min;
               rde->ctime.hour = de->date.hour;
               rde->ctime.day  = de->date.day;
               rde->ctime.mon  = de->date.month - 1;
               rde->ctime.year = de->date.year + (1900 - 1781);
               if (!rde->size || absidx==1) rde->loc = 0; else {
                  rde->loc        = locptr;
                  locptr->cluster = de->extent.le;
                  locptr->len     = szinsec;
                  locptr++;
               }

               if (vd->joliet) {
                  wchar_t *np = (wchar_t*)namedata;
                  u32t     ni;

                  if (absidx==1) {
                     rde->nmlen = 2;
                     np[0] = '.'; np[1] = '.'; np[2] = 0;
                  } else {
                     rde->nmlen = de->namelen >> 1;
                     // BE->LE
                     for (ni=0; ni<rde->nmlen; ni++)
                        np[ni] = bswap16(((u16t*)&de->name)[ni]);
                     np[rde->nmlen] = 0;
                     FIX_NAME(rde->nmlen, np);
                  }
                  rde->wname = np;
                  dd->hash[idx] = calc_hash_w(np, rde->nmlen);
                  namedata += rde->nmlen + 1 << 1;
               } else {
                  if (absidx==1) {
                     rde->nmlen = 2;
                     strcpy(namedata, "..");
                  } else {
                     rde->nmlen = de->namelen;
                     memcpy(namedata, &de->name, de->namelen);
                     namedata[de->namelen] = 0;
                     FIX_NAME(rde->nmlen, namedata);
                  }
                  rde->name = namedata;
                  dd->hash[idx] = calc_hash_a(namedata, 0);
                  /* at least one entry in the directory has incorrect ISO
                     name (char codes >=128), it needs to be converted to
                     Unicode using the current code page. */
                  if (!dd->hash[idx]) dd->badname = 1;
                  
                  namedata += rde->nmlen + 1;
               }
               idx++;
            }
            // multi-extent file (first/next slice)
            mext = de->flags & ISODF_MULTI ? 1 : 0;
            // fin zeroes only if file is really has an extents
            if (!mext && rde->loc) {
               locptr->cluster = 0;
               locptr->len     = 0;
               locptr++;
            }
         }
         rem -= de->length;
         if (rem) de = (iso_dirrec*)((u8t*)de + de->length);
         absidx++;
      }
   }
   if (!res) {
      u32t alen   = namedata-namechk;
      dd->entries = idx;
      *dlist      = dd;
      // i can`t catch this :(
      if (alen > namelen) {
         log_it(0, "read_dir(,%X,) - bad length (%u>%u) (%X)!!!\n", sector,
            alen, namelen, dd);
         mem_checkmgr();
      } else
      /* if at least one free heap block (16+16) is available - cut it.
         mem_cut() guarantee that memory pointer is not changed */ 
      if (namelen-alen >= 32) {
         mem_cut(dd, mbsize-(namelen-alen));
#if ISO_DEBUG>0
         log_it(3, "read_dir(,%X,) - cut %u<=%u (%X)\n", sector, alen, namelen, dd);
#endif
      }
#if ISO_DEBUG>0
      if (dd->badname)
         log_it(3, "read_dir(,%X,) - ISO names violation\n", sector);
#endif
   } else
      if (dd) free(dd);

   hlp_memfree(readbuf);

   return res;
}

static void usage_modify(dir_data *dd, int incr) {
   while (dd) {
      if (incr) dd->usage++; else dd->usage--;
      dd = dd->pdd;
   }
}

/** walk over tree to the selected path.
    @param   vd           volume data
    @param   path         full path (accepts '\' only)
    @param   [out] dd     directory data
    @param   [out] index  0 means directory read, else it return index of
                          dir_entry in parent (even for directory!) and parent
                          directory in dd.
    @return error code or 0 */
static qserr follow_path(volume_data *vd, const c16t *path, dir_data **dd, int *index) {
   qserr     res = 0;
   dir_data *cdd = 0;
   c16t      *cp = (c16t*)path+2;
   // path MUST be full
   if (path[1]!=':' || path[0]-'A'!=(vd->vdsk&QDSK_DISKMASK)) return E_SYS_SOFTFAULT;
   // cache root
   if (!vd->root)
      res = read_dir(vd, vd->pd->root.extent.le, Round2k(vd->pd->root.size.le)>>
         ISO_SECTOR_BITS, &vd->root);
   // is it root dir?
   if (!res) {
      while (*cp=='\\') cp++;
      if (*cp==0) {
         if (index) return E_SYS_INVNAME;
         *dd = vd->root;
         return 0;
      }
      cdd = vd->root;
      // directory read
      do {
         c16t   *ce = cp;
         u32t   idx = 0, nhash, nmlen, *npos;

         while (*ce && *ce!='\\') ce++;
         nmlen = ce-cp;
         nhash = calc_hash_w(cp,nmlen);

         // loop over directory
         do {
            u32t   *pnpos = cdd->hash + idx;
            dir_entry *de;

            npos = memchrd(pnpos, nhash, cdd->entries - idx);
            if (!npos) res = E_SYS_NOPATH; else {
               idx  = npos - cdd->hash;
               de   = cdd->de + idx;
               
               if (nmlen==de->nmlen) {
                  c16t *cpp = cp;
               
                  if (vd->joliet) {
                     wchar_t *fnp = de->wname;
                     while (cpp!=ce)
                        if (towupper(*cpp++)!=towupper(*fnp++)) break;
                  } else {
                     char *fnp = de->name;
                     // failed conversion in char_to_wu() will return zero
                     while (cpp!=ce)
                        if (towupper(*cpp++)!=char_to_wu(*fnp++)) break;
                  }
                  // name match!
                  if (cpp==ce) {
#if ISO_DEBUG>0
                     // print if all this logic with zero hash had success ;)
                     if (!nhash)
                        log_it(3, "follow_path(%S) - bad name match!\n", path);
#endif
                     cp = ce;
                     while (*cp=='\\') cp++;
                     // reach the end?
                     if (!*cp && index) *index = idx; else
                     if ((de->attrs&IOFA_DIR)==0) {
                        // disable the second pass for this case
                        nhash = 0;
                        res = E_SYS_NOPATH;
                     } else
                     if (de->dd) cdd = de->dd; else {
                        dir_data  *cdn;
               
                        res = read_dir(vd, de->loc->cluster, de->loc->len, &cdn);
                        // next level
                        if (!res) {
                           de->dd      = cdn;
                           cdn->pdd    = cdd;
                           cdn->pindex = idx;
                           cdd = cdn;
                        }
                     }
                     if (!*cp && !res) *dd = cdd;
                     break;
                  }
               }
               if (++idx>=cdd->entries) res = E_SYS_NOPATH;
            }
            /* not found: try again, checking all names with chr codes >=128,
               current system codepage may cause another match for them */
            if (res==E_SYS_NOPATH && cdd->badname && nhash) {
               nhash = 0;
               idx   = 0;
               res   = 0;
            }
         } while (!res);
      } while (!res && *cp);
   }
#if ISO_DEBUG>0
   log_it(3, "follow_path(%S, %X, %i) = %X\n", path, *dd, index?(int)index:-1, res);
#endif
   return res;
}

static qserr _exicc iso_open(EXI_DATA, const c16t *name, u32t mode,
                             io_handle_int *pfh, u32t *action)
{
   file_handle   *fh = 0;
   dir_data      *dd;
   qserr         res;
   int         index;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   if (mode>SFOM_CREATE_ALWAYS) return E_SYS_INVPARM;
   if (!pfh || !name || !action) return E_SYS_ZEROPTR;
   if (wcslen(name)>QS_MAXPATH) return E_SYS_LONGNAME;
   // deny any file overwrite
   if (mode>SFOM_OPEN_ALWAYS) return E_SYS_ACCESS;

   *pfh  = 0;
   res   = follow_path(vd, name, &dd, &index);

   if (!res) {
      dir_entry *de = &dd->de[index];
      if (de->attrs&IOFA_DIR) res = E_SYS_ACCESS; else {
         qs_filecc fd = NEW_G(qs_filecc);
         // setup file cache in r/o mode
         res = fd->setdisk(0,0,0,0, vd->vdsk, 0, vd->pd->vspacesize.le, ISO_SECTOR_BITS);

         if (!res) {
            io_handle_info  fi;
            fi.attrs = de->attrs;
            fi.size  = de->size;
            fi.vsize = de->size;
            res = fd->init(&fi, fi.size?de->loc:0, FCCF_READONLY);

            if (!res) {
               fh = (file_handle*)malloc(sizeof(file_handle)+wcslen(name)*sizeof(c16t));
               fh->sign = ISOFH_SIGN;
               fh->de   = de;
               fh->dd   = dd;
               fh->fd   = fd;
               wcscpy(fh->fp, name);
               usage_modify(dd,1);
               *pfh = (io_handle_int)fh;
            }
         }
         if (res) DELETE(fd);
      }
   }
   if (!res) *action = IOFN_EXISTED;
   // forced read-only
   return res?res:E_SYS_READONLY;
}

static qserr _exicc iso_read(EXI_DATA, io_handle_int file, u64t pos,
                             void *buffer, u32t *size)
{
   file_handle   *fh = (file_handle*)file;
   if (!size) return E_SYS_ZEROPTR;
   if (!buffer || fh->sign!=ISOFH_SIGN) { *size = 0; return E_SYS_INVPARM; }
   if (pos>=fh->de->size) { *size = 0; return E_SYS_EOF; }
   if (!*size) return 0;
   return fh->fd->read(pos, buffer, *size, size);
}

static qserr _exicc iso_write(EXI_DATA, io_handle_int file, u64t pos,
                              void *buffer, u32t *size)
{
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   return E_SYS_ACCESS;
}

static qserr _exicc iso_flush(EXI_DATA, io_handle_int file) {
   file_handle   *fh = (file_handle*)file;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (fh->sign!=ISOFH_SIGN) return E_SYS_INVPARM;
   return E_SYS_ACCESS;
}

static qserr _exicc iso_close(EXI_DATA, io_handle_int file) {
   file_handle   *fh = (file_handle*)file;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (fh->sign!=ISOFH_SIGN) return E_SYS_INVPARM;
   fh->fd->close();
   DELETE(fh->fd);
   usage_modify(fh->dd,0);
   fh->sign = 0;
   free(fh);
   return 0;
}

static qserr _exicc iso_size(EXI_DATA, io_handle_int file, u64t *size) {
   file_handle *fh = (file_handle*)file;
   if (fh->sign!=ISOFH_SIGN) return E_SYS_INVPARM;
   if (!size) return E_SYS_ZEROPTR;
   *size = fh->de->size;
   return 0;
}

static qserr _exicc iso_setsize(EXI_DATA, io_handle_int file, u64t newsize) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   return E_DSK_WP;
}

static qserr _exicc iso_finfo(EXI_DATA, io_handle_int file, io_handle_info *info)
{
   file_handle *fh = (file_handle*)file;
   if (fh->sign!=ISOFH_SIGN) return E_SYS_INVPARM;
   if (!info) return E_SYS_ZEROPTR;
   // a bit faster than function below
   info->attrs   = fh->de->attrs;
   info->fileno  = 0;
   info->size    = fh->de->size;
   info->vsize   = fh->de->size;
   info->atime   = fh->de->ctime;
   info->wtime   = fh->de->ctime;
   info->ctime   = fh->de->ctime;
   return 0;
}

static qserr _exicc iso_pathinfo(EXI_DATA, const c16t *path, io_handle_info *info,
                                 c16t *name, c16t *altname)
{
   qserr         res;
   int         index;
   dir_data      *dd;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (!path || !info) return E_SYS_ZEROPTR;

   res  = follow_path(vd, path, &dd, &index);
   if (!res) {
      dir_entry *de = &dd->de[index];
      info->attrs   = de->attrs;
      info->fileno  = 0;
      info->size    = de->size;
      info->vsize   = de->size;
      info->atime   = de->ctime;
      info->wtime   = de->ctime;
      info->ctime   = de->ctime;
      if (altname) *altname = 0;
      if (name)
         if (vd->joliet) wcscpy(name, de->wname); else
            getname_a(de, name);
   }
   return res;
}

static qserr _exicc iso_setinfo(EXI_DATA, const c16t *path, io_handle_info *info, u32t flags) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (!path || !info) return E_SYS_ZEROPTR;
   return E_DSK_WP;
}
		
static qserr _exicc iso_setexattr(EXI_DATA, const c16t *path, const c16t *aname, void *buffer, u32t size) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   return E_DSK_WP;
}

static qserr _exicc iso_getexattr(EXI_DATA, const c16t *path, const c16t *aname, void *buffer, u32t *size) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   return E_SYS_UNSUPPORTED;
}

static str_list* _exicc iso_lstexattr(EXI_DATA, const c16t *path) {
   instance_ret(volume_data, vd, 0);
   return 0;
}

// warning! it acts as rmdir too!!
static qserr _exicc iso_remove(EXI_DATA, const c16t *path) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (!path) return E_SYS_ZEROPTR;
   return E_DSK_WP;
}

static qserr _exicc iso_renmove(EXI_DATA, const c16t *oldpath, const c16t *newpath) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (!oldpath || !newpath) return E_SYS_ZEROPTR;
   return E_DSK_WP;
}

static qserr _exicc iso_makepath(EXI_DATA, const c16t *path) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   return E_DSK_WP;
}

static qserr _exicc iso_diropen(EXI_DATA, const c16t *path, dir_handle_int *pdh) {
   dir_enum_data  *ed;
   dir_data       *dd;
   qserr          res;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   if (!path || !pdh) return E_SYS_ZEROPTR;
   *pdh = 0;
   if (wcslen(path)>QS_MAXPATH) return E_SYS_LONGNAME;
   // dir handle storage
   ed = (dir_enum_data*)malloc(sizeof(dir_enum_data));
   if (!ed) return E_SYS_NOMEM;

   res = follow_path(vd, path, &dd, 0);
   if (!res) {
      ed->sign = ISODIR_SIGN;
      ed->dd   = dd;
      ed->pos  = 0;
      *pdh     = (dir_handle_int)ed;
      usage_modify(dd,1);
   }
   if (res && ed) free(ed);
   return res;
}

static qserr _exicc iso_dirnext(EXI_DATA, dir_handle_int dh,
                                io_direntry_infow *info, int altname) {
   dir_enum_data *ed = (dir_enum_data*)dh;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (!ed || !info) return E_SYS_ZEROPTR;
   if (ed->sign!=ISODIR_SIGN) return E_SYS_INVPARM;

   if (ed->pos>=ed->dd->entries) return E_SYS_NOFILE; else {
      dir_entry *de = &ed->dd->de[ed->pos++];
      info->attrs   = de->attrs;
      info->fileno  = 0;
      info->size    = de->size;
      info->vsize   = de->size;
      info->atime   = de->ctime;
      info->wtime   = de->ctime;
      info->ctime   = de->ctime;
      // cvt name to current codepage
      if (vd->joliet) wcscpy(info->name, de->wname); else
         getname_a(de, info->name);
   }
   return 0;
}

static qserr _exicc iso_dirclose(EXI_DATA, dir_handle_int dh) {
   dir_enum_data *ed = (dir_enum_data*)dh;
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   if (!ed) return E_SYS_ZEROPTR;
   if (ed->sign!=ISODIR_SIGN) return E_SYS_INVPARM;

   ed->sign = 0;
   usage_modify(ed->dd,0);
   free(ed);
   return 0;
}

static qserr _exicc iso_volinfo(EXI_DATA, disk_volume_data *info, int fast) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);

   if (!info) return E_SYS_ZEROPTR;
   memset(info, 0, sizeof(disk_volume_data));

   if (vd->vdsk) {
      vol_data *vdta = _extvol + (vd->vdsk&QDSK_DISKMASK);
      // mounted?
      if (vdta->flags&VDTA_ON) {
         info->StartSector  = vdta->start;
         info->TotalSectors = vdta->length;
         info->Disk         = vdta->disk;
         info->SectorSize   = vdta->sectorsize;
         info->FSizeLim     = _2KB * _4GBLL - _32KB - _2KB;
         info->ClBad        = 0;
         info->SerialNum    = vdta->serial;
         info->ClSize       = vdta->clsize;
         //info->ClAvail      = 0;
         info->ClTotal      = vdta->cltotal;
         info->FsVer        = FST_OTHER;
         info->InfoFlags    = VIF_READONLY;
         //info->Reserved     = 0;
         info->DataSector   = vd->vdendpos + 1;
         //info->OemData      = 0;
         strcpy(info->Label,vdta->label);
         strcpy(info->FsName,vdta->fsname);
         return 0;
      }
   }
   return E_DSK_NOTMOUNTED;
}

static qserr _exicc iso_setlabel(EXI_DATA, const char *label) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   return vd->vdsk ? E_DSK_WP : E_DSK_NOTMOUNTED;
}

static qserr _exicc iso_service(EXI_DATA, u32t func, void *info) {
   instance_ret(volume_data, vd, E_SYS_INVOBJECT);
   return E_SYS_UNSUPPORTED;
}

static u32t _exicc iso_avail(EXI_DATA) { return 0; }

static void _std iso_init(void *instance, void *userdata) {
   volume_data *vd = (volume_data*)userdata;
   vd->sign        = ISO_SIGN;
   vd->vdsk        = 0;
   vd->self        = (qs_sysvolume)instance;
}

static void _std iso_done(void *instance, void *userdata) {
   volume_data *vd = (volume_data*)userdata;
   if (vd->sign==ISO_SIGN)
      if (vd->vdsk) log_printf("Vol %X fini???\n", vd->vdsk);
   iso_freedata(vd);
   memset(vd, 0, sizeof(volume_data));
}

static void *m_list[] = { iso_initialize, iso_finalize, iso_avail,
   iso_volinfo, iso_setlabel, iso_drive, iso_open, iso_read, iso_write,
   iso_flush, iso_close, iso_size, iso_setsize, iso_finfo, iso_setexattr,
   iso_getexattr, iso_lstexattr, iso_remove, iso_renmove, iso_makepath,
   iso_remove, iso_pathinfo, iso_setinfo, iso_diropen, iso_dirnext,
   iso_dirclose, iso_service };

u32t register_isofs(void) {
   u32t cid = exi_register(FS_CLASS_NAME_PFX "iso9660", m_list,
                           sizeof(m_list)/sizeof(void*), sizeof(volume_data),
                           EXIC_PRIVATE, iso_init, iso_done, 0);
   if (cid) {
      qserr res = io_registerfs(cid);
      if (res) {
         log_printf("isofs: regfs err %X!\n", res);
         exi_unregister(cid);
         cid = 0;
      }
   }
   return cid;
}
