//
// QSINIT API
// 32bit direct emu
//
#ifndef QSINIT_DIRECTH
#define QSINIT_DIRECTH

#ifdef __cplusplus
extern "C" {
#endif
#include "qstypes.h"
#include "time.h"

#pragma pack(1)

// file attribute constants for d_attr field
#define _A_NORMAL         0x00  // Normal file - read/write permitted
#define _A_RDONLY         0x01  // Read-only file
#define _A_HIDDEN         0x02  // Hidden file
#define _A_SYSTEM         0x04  // System file
#define _A_VOLID          0x08  // Volume-ID entry
#define _A_SUBDIR         0x10  // Subdirectory
#define _A_ARCH           0x20  // Archive file
// special bits in d_attr field
#define _A_NAMEERR      0x8000  // Error in conversion unicode name to ascii
#define _A_NAMELENERR   0x4000  // Directory name too long (_dos_readtree())
#define _A_DIRNMATCH    0x2000  // Directory name doesn`t match pattern! (_dos_readtree())

#define NAME_MAX        QS_MAXPATH

struct dirent {
   u16t           d_attr;       ///< file attribute
   u16t       d_reserved;       ///< reserved space (align)
   void       *d_sysdata;       ///< internal data
   u64t           d_size;       ///< file size
   /** note that time value in these 8 bytes depends on QTLS_DIRENTTIME
       TLS variable:
       QDETIME_UNIX (0, default): d_wtime and d_ctime with unixtime are
                            returned in dirent from all libc functions in
                            this thread
       QDETIME_WRITE  (1) : file write time returned in d_time
       QDETIME_CREATE (2) : file creation time returned in d_time
       QDETIME_ACCESS (3) : file access time returned in d_time */
   union {
      struct {
         time_t  d_wtime;       ///< file time
         time_t  d_ctime;       ///< file creation time
      };
      timeq_t     d_time;
   };
   char      *d_openpath;       ///< path specified to opendir (with trailing \)
   struct 
   dirent      *d_subdir,       ///< subdirectory data (_dos_readtree())
                 *d_next;       ///< next in the list (_dos_readtree(),0=end)
   char  d_name[NAME_MAX+1];    ///< file name
};
typedef struct dirent dir_t;

#define QDETIME_UNIX      0
#define QDETIME_WRITE     1
#define QDETIME_CREATE    2
#define QDETIME_ACCESS    3

struct _wdirent {
   u16t           d_attr;       ///< file attribute
   u16t       d_reserved;       ///< this struct size (_dos_readtreew())
   void       *d_sysdata;       ///< internal data
   u64t           d_size;       ///< file size
   /// the same logic as in dirent above
   union {
      struct {
         time_t  d_wtime;       ///< file time
         time_t  d_ctime;       ///< file creation time
      };
      timeq_t     d_time;
   };
   c16t      *d_openpath;       ///< path specified to opendir (with trailing \)
   struct 
   _wdirent    *d_subdir,       ///< subdirectory data (_dos_readtreew())
                 *d_next;       ///< next in the list (_dos_readtree(),0=end)
   c16t  d_name[NAME_MAX+1];    ///< file name
};
typedef struct _wdirent dirw_t;

int    _std chdir (const char *path);
int    _std mkdir (const char *path);
char*  _std getcwd(char *buf, size_t size);
/// drive A:==0, B:==1, etc
char*  _std _getdcwd(int drive, char *buf, size_t size);
c16t*  _std _wgetdcwd(int drive, c16t *buf, size_t size);

/** open a directory.
    Note that both opendir() and _dos_findfirst() actually use unicode file
enumeration API, so each name that failed conversion from Unicode to the
current codepage will have _A_NAMEERR bit set in the file attributes field.
All failed characters will be replaced to '?'.

    @param  dir     Directory path.
    @return dir_t struct for enumeration or 0 if error (see errno) */
dir_t* _std opendir(const char* dir);
dir_t* _std readdir(dir_t*);
int    _std closedir(dir_t*);

dirw_t*_std _wopendir(const c16t* dir);
dirw_t*_std _wreaddir(dirw_t*);
int    _std _wclosedir(dirw_t*);

int    _std rmdir(const char *path);

qserr  _std _dos_findfirst(const char *path, u16t attributes, dir_t *buffer);
qserr  _std _dos_findnext(dir_t *buffer);
qserr  _std _dos_findclose(dir_t *buffer);

qserr  _std _wdos_findfirst(const c16t *path, u16t attributes, dirw_t *buffer);
qserr  _std _wdos_findnext(dirw_t *buffer);
qserr  _std _wdos_findclose(dirw_t *buffer);

void   _std _dos_getdrive(unsigned *drive);
void   _std _dos_setdrive(unsigned drive, unsigned *total);

/** tiny stat function replacement.
    @param  path    File path.
    @param  buffer  Stat data.
    @return 0 on success or QS error code */
qserr  _std _dos_stat(const char *path, dir_t *buffer);
qserr  _std _dos_statw(const c16t *path, dirw_t *buffer);

struct diskfree_t {
   u32t     total_clusters;     ///< total clusters on disk
   u32t     avail_clusters;     ///< available clusters on disk
   u32t     sectors_per_cluster;///< sectors per cluster
   u32t     bytes_per_sector;   ///< bytes per sector
};
typedef struct diskfree_t diskfree_t;

/** get disk free space.
    @param [in]  drive      Disk number (0 - default, 1 for A:/0:, 2 for B:/1:, etc.
    @param [out] diskspace  Return data.
    @return zero on success. */
qserr  _std _dos_getdiskfree(unsigned drive, diskfree_t *diskspace);

/** callback for _dos_readtree.
    @param fp           file info. d_openpath field point to file`s directory
    @param cbinfo       user data for callback proc
    @return 1 to continue processing, 0 to skip this entry and -1
            to stop call and return 0 from _dos_readtree() */
typedef int _std (*_dos_readtree_cb) (dir_t *fp, void *cbinfo);
/// callback for _dos_readtreew (all the same with _dos_readtree()).
typedef int _std (*_dos_readtreew_cb)(dirw_t *fp, void *cbinfo);

/** read directory tree.
    @attention list, returned by this function is a thread owned memory. So
               exiting a thread, which called this function will release the
               entire list.

    @param dir          Directory.
    @param mask         File mask, can be 0 (assume *.*).<br>
                        note: directories returned without pattern match,
                        but for all unmatched entries d_attr will have
                        _A_DIRNMATCH set. "." is never returned, ".."
                        returned but has _A_DIRNMATCH set when mask is *.*
                        or * to prevent circular references.
    @param [out] info   Pointer to the first entry of the file list. Next
                        entry addressed via d_next with d_next==0 in the last
                        entry. In _A_SUBDIR entries d_subdir field points to
                        the subdirectory data in the same format. The caller
                        must expect 0 in both info in d_subdir if list is empty
                        or error occured.<br>
                        This list must be freed by single _dos_freetree() call
                        with the top dir *info value.
    @param subdirs      Scan all subdirectories flag.
    @param utf8         Return Unicode names in UTF-8 format. Note that UTF-8
                        name string can be longer than QS_MAXPATH.
    @param [out] dircnt Number of found subdirectories (ptr can be 0).
    @param callback     Callback to process data while reading.
                        One of info or callback can be 0.
    @param cbinfo       User data for callback proc.
    @return total number of found files (with subdirectories). Function resets
            errno value if everything is ok. */
u32t   _std _dos_readtree(const char *dir, const char *mask,
                          dir_t **info, int subdirs, int utf8, u32t *dircnt,
                          _dos_readtree_cb callback, void *cbinfo);

/** utf-16 version of _dos_readtree().
    use _dos_freetree() to free returned tree data */
u32t   _std _dos_readtreew(const c16t *dir, const c16t *mask,
                          dirw_t **info, int subdirs, u32t *dircnt,
                          _dos_readtreew_cb callback, void *cbinfo);

/** free directory tree data.
    @param tree         Tree data (dir_t* or dirw_t*) returned by
                        _dos_readtree() or _dos_readtreew()
    @return 0 or QS error code */
qserr  _std _dos_freetree(void *tree);

/** set file attributes.
    @param path         File path.
    @param attributes   File attributes.
    @return 0 on success or QS error code */
qserr  _std _dos_setfileattr(const char *path, unsigned attributes);

/** get file attributes.
    @param       path         File path.
    @param [out] attributes   File attributes.
    @return 0 on success or QS error code */
qserr  _std _dos_getfileattr(const char *path, unsigned *attributes);

#define _DT_MODIFY     0x0001   ///< query/set file modification time
#define _DT_CREATE     0x0002   ///< query/set file creation time

/** set file/dir time.
    @param       path         File path.
    @param       dostime      Time in DOS format.
    @param       type         Time type (_DT_* constant), OR-ed flags
                              can be used for the same time for some values.
    @return 0 on success or QS error code */
qserr  _std _dos_setfiletime(const char *path, u32t dostime, u32t type);

/** get file/dir time.
    @param       path         File path.
    @param [out] dostime      Time in DOS format.
    @return 0 on success or QS error code */
qserr  _std _dos_getfiletime(const char *path, u32t *dostime, u32t type);

/** check string for a exist directory name.
    @param dir          Directory name.
    @return 1 if string is existing directory name, else 0 */
u32t   _std hlp_isdir(const char *dir);
/// utf-16 version of hlp_isdir()
u32t   _std hlp_isdirw(const c16t *dir);

#pragma pack()

#ifdef __cplusplus
}
#endif

#endif // QSINIT_DIRECTH

