//
// QSINIT
// help file functions
// shared class implementation
//
#ifndef QSINIT_HELPSHARED
#define QSINIT_HELPSHARED

#include "qstypes.h"
#include "qsclass.h"
#include "qshelp.h"

#ifdef __cplusplus
extern "C" {
#endif

/// shared class for help access without static linking of HELP.DLL
typedef struct qs_helpctrl_s {
   /** open help file.
       @param  name        Path to a file
       @param  [out] hf    Help file handle
       @return error code or 0 */
   qserr      _exicc (*open)    (const char *name, qsh_file *hf);
   
   /** close help file.
       Note that this is actually a notification call, all open files
       stay loaded until help_release() call.
       @param  hf          Help file handle
       @return error code or 0 */
   qserr      _exicc (*close)   (qsh_file hf);
   
   /** free help file cache.
       All unused help files are closed during this call.
       The help file is assumed as unused if help_close() is called or process
       that called help_open() is already terminated. */
   void       _exicc (*release) (void);
   
   /** query list of topics in a help file.
       @param  hf          Help file handle
       @return help file information in a single application owned heap block
               (must be released by caller) or 0 if invalid handle */
   qsh_info*  _exicc (*info)    (qsh_file hf);

   /// query topic id by name 
   u32t       _exicc (*topicid) (qsh_file hf, const char *topic);

   /** query color scheme by name 
    @param  hf          Help file handle
    @param  topic       Topic name (0 returns index topic if available)
    @return color scheme id or negative value if error or not found */
   s32t       _exicc (*schemeid)(qsh_file hf, const char *scheme);
   
   /** convert a topic to the screen data buffer (text/attr pair).
       @param  hf          Help file handle
       @param  scheme      Color scheme index (0 for default)
       @param  topic       Topic id
       @param  width       Output width in characters
       @param  [out] err   Error code value (optional)
       @return output data in a single application owned heap block (must be
               released by caller) or 0 if error (no topic or invalid
               parameters) */
   qsh_topic* _exicc (*draw)    (qsh_file hf, u32t scheme, u32t topic,
                                 u32t width, qserr *err);

   /** convert a topic to the ANSI text sequence.
       @param  hf          Help file handle
       @param  scheme      Color scheme index (0 for default)
       @param  topic       Topic id
       @param  outx        Start column in the result ANSI text (i.e. x offset
                           at the start of each line)
       @param  width       Output width in characters
       @param  [out] err   Error code value (optional)
       @return application owned heap block or 0 if error (no topic or invalid
               parameters). */
   char*      _exicc (*drawansi)(qsh_file hf, u32t scheme, u32t topic,
                                 u32t xpos, u32t width, qserr *err);

   /** high level "show help" function.
       Function shows a help window.
       
       Note that "winwidth" cannot be smaller than specified in the help file
       "minwidth" value.
       HELPSIZE environment variable is used for default size calculation.
   
       @param  name        Help file name (path or name to search in HELPPATH)
                           or qsh_file if HSFL_HANDLE flag is set in options,
                           if the first character is ? then it acts as a link
                           in common form "?file/topic/mark" and both topic and
                           mark parameters are ignored.
       @param  topic       Topic to open (use 0 for the index topic)
       @param  mark        Optional mark name in the topic to pos on it (can be 0)
       @param  winx        Help window x position (default if < 0)
       @param  winy        Help window y position (default if < 0)
       @param  winwidth    Help window width (default if 0)
       @param  winheight   Help window height (default if 0)
       @param  flags       Options (HSFL_*)
       @return error code or 0 */
   qserr      _exicc (*show)    (const void *name, const char *topic,
                                 const char *mark, int winx, int winy,
                                 u32t winwidth, u32t winheight, u32t flags);

   /** helper for a link navigation code.
       Function searches for a link in the visible area of the help topic.
   
       @param  td          Help topic data
       @param  prev        Selected link (index in td->llist, can be -1 for HLINK_ANY)
       @param  direction   Where to go - HLINK_* value
       @param  ypos        Visible area starts on this topic line
       @param  ysize       Visible area height
       @return link index (first in the group) or -1 if not found. */
   int        _exicc (*findlink)(qsh_topic *td, int prev, u32t direction,
                                 u32t ypos, u32t ysize);
} _qs_helpctrl, *qs_helpctrl;

#ifdef __cplusplus
}
#endif
#endif // QSINIT_HELPSHARED
