//
// QSINIT API
// low level volume api
//
#ifndef QSINIT_VOLIO_CLASS
#define QSINIT_VOLIO_CLASS

#include "qstypes.h"
#include "qsclass.h"
#include "qsutil.h"
#include "qsio.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef u32t   io_handle_int;
typedef u32t  dir_handle_int;

/** Low level volume api, provided by fs implementation.
    This code is, actually, internal thing and should not be used
    on a user level (even if you can acquire such ptr in some way ;). */
typedef struct qs_sysvolume_s {
   /** initialize it.
       Note, that function should not modify anything in FS, especially
       unknown one.
       @param   vol      Volume (0==A, 1==B and so on)
       @param   flags    No flags now (just 0)
       @param   bootsec  Boot sector data
       @param   opts     Mount options (any text data). If class name started
                         with FS_CLASS_NAME_PFX then it will receive the value
                         of environment string MOUNT_OPTIONS_suffix (or 0). So
                         "qs_fs_hpfs" correponds to MOUNT_OPTIONS_HPFS string.
       @return error code */
   qserr     _exicc (*init)     (u8t vol, u32t flags, void *bootsec,
                                 const char *opts);
   /// unmount volume
   qserr     _exicc (*done)     (void);
   /// return information about optional functions support (see SFAF_*)
   u32t      _exicc (*avail)    (void);
   /** query volume info.
       note, that info->FsVer should contain 0 if FS was not recognized, else
       any non-zero value (ex. FAT type)
       @param   info     Returning information
       @param   fast     Fast version flag (1/0). Fast call returns zero or
                         wrong information in serial number, label and free
                         space fields.
       @return error code */
   qserr     _exicc (*volinfo)  (disk_volume_data *info, int fast);
   qserr     _exicc (*setlabel) (const c16t *label);
   /** mounted drive letter.
       @return -1 if volume is not mounted, else volume number (0..x) */
   int       _exicc (*drive)    (void);

   /** open a file.
       If file has r/o attribute or filesystem is r/o function returns 
       E_SYS_READONLY and opens file in r/o mode.
       After r/o attribute reset while file is open there are two possible
       ways: file still r/o for all OR this file handle can be used for writes
       w/o any problem. To signal the second case - avail() should return
       SFAF_NPRO bit on.

       @param   [in]  name    File path
       @param   [in]  mode    File mode (file is always opened once per system,
                              in write mode)
       @param   [out] pfh     Ptr to file handle
       @param   [out] action  Actual action was performed on open.
       @return error code. Note, what E_SYS_READONLY means SUCCESS, but
               the file will deny any write activity. */
   qserr     _exicc (*open)     (const c16t *name, u32t mode, io_handle_int *pfh, u32t *action);
   qserr     _exicc (*read)     (io_handle_int fh, u64t pos, void *buffer, u32t *size);
   qserr     _exicc (*write)    (io_handle_int fh, u64t pos, void *buffer, u32t *size);
   qserr     _exicc (*flush)    (io_handle_int fh);
   qserr     _exicc (*close)    (io_handle_int fh);
   qserr     _exicc (*size)     (io_handle_int fh, u64t *size);
   qserr     _exicc (*setsize)  (io_handle_int fh, u64t newsize);

   /** query file info.
       @param   fh            File handle
       @param   [out] info    File information      
       @return error code or 0 */
   qserr     _exicc (*finfo)    (io_handle_int fh, io_handle_info *info);

   qserr     _exicc (*setexattr)(const c16t *path, const c16t *aname, void *data, u32t size);
   qserr     _exicc (*getexattr)(const c16t *path, const c16t *aname, void *buffer, u32t *size);
   str_list  _exicc (*lstexattr)(const c16t *path);
   qserr     _exicc (*remove)   (const c16t *path);
   qserr     _exicc (*renmove)  (const c16t *oldname, const c16t *newname);

   qserr     _exicc (*makepath) (const c16t *path);
   qserr     _exicc (*rmdir)    (const c16t *path);
   /** query path info.
       both ptrs to name (real name) and altname (short name) can be 0. */
   qserr     _exicc (*pathinfo) (const c16t *path, io_handle_info *info,
                                 c16t *name, c16t *altname);
   qserr     _exicc (*setinfo)  (const c16t *path, io_handle_info *info, u32t flags);

   qserr     _exicc (*diropen)  (const c16t *path, dir_handle_int *pdh);
   qserr     _exicc (*dirnext)  (dir_handle_int dh, io_direntry_infow *de, int altname);
   qserr     _exicc (*dirclose) (dir_handle_int dh);
   /// "reserved" function (also to force non-utf16 fs libs fail on init)
   qserr     _exicc (*service)  (u32t func, void *data);
} _qs_sysvolume, *qs_sysvolume;

/// @name io_open() mode value
//@{
#define SFOM_OPEN_EXISTING      0     ///< opens the file, fails if the file is not existing
#define SFOM_OPEN_ALWAYS        1     ///< opens existing file, else create new
#define SFOM_CREATE_NEW         2     ///< creates a new file, fails if file is existing
#define SFOM_CREATE_ALWAYS      3     ///< creates a new file, existing file will be overwritten
//@}

/// @name avail() flags
//@{
#define SFAF_ALTNAME     0x000001     ///< alternative names supported
#define SFAF_VFS         0x000002     ///< virtual FS
#define SFAF_NPRO        0x000004     ///< r/o open is not persistent (see open() above)
//@}

#define FS_CLASS_NAME_PFX "qs_fs_"    // recommended ;)

/** register qs_sysvolume compatible class as a new filesystem handler.
    Function forces remounting of all unrecognized volumes (if avail()
    has no SFAF_VFS bit).
    @return error code (too many FS handlers or wrong number of functions in
            the supplied class) or 0 on success. */
qserr _std io_registerfs(u32t classid);

#ifdef __cplusplus
}
#endif
#endif // QSINIT_VOLIO_CLASS
