//
// QSINIT
// help file functions
// note that header can be used in SysView`s PC build
//
#ifndef QSINIT_HELPDATA
#define QSINIT_HELPDATA

#include "qstypes.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef void *qsh_file;

/* multilined links are grouped in the qsh_topic.llist: the group head has
   filled "back" field (llist[0] is always a head), all next group entries
   has back = 0xFFFF */
typedef struct {
   u16t       xpos;
   u16t       ypos;
   u16t        len;
   u16t       back;    ///< start of the previous group or 0xFFFF if inside multiline link
} qsh_link;

typedef struct {
   u32t      width;    ///< width of the text
   u32t      lines;    ///< # of lines in the text
   u16t*      text;    ///< text and attr, width * lines
   u32t     nlinks;    ///< # of links
   u32t     nmarks;    ///< # of marks
   char     **link;    ///< ptr to link values array
   qsh_link *llist;    ///< link highlight list
   char     **mark;    ///< ptr to mark names array
   u16t     *mlist;    ///< mark line list
   char     *topic;    ///< topic name (title from <n> tag)
   u16t     selcol;    ///< a color for the selected link (currect scheme value)
   u16t     reserv;    ///< field for align
} qsh_topic;

typedef struct {
   u32t    ntopics;    ///< # of topics in this file
   u32t   nschemes;    ///< # of color schemes in this file
   char**     item;    ///< array of topics (topic id is an index in this array + 1)
   char**   scheme;    ///< array of scheme names
   char     *title;    ///< help file name ("title")
   u32t      index;    ///< topic id with the index topic or 0
   u32t   minwidth;    ///< min window width for this file
} qsh_info;

/** open help file.
    @param  name        Help file name (path or name to search in HELPPATH)
    @param  [out] hf    Help file handle
    @return error code or 0 */
qserr      _std help_open    (const char *name, qsh_file *hf);

/** close help file.
    Note that this is actually a notification call, all open files
    stay loaded until help_release() call.
    @param  hf          Help file handle
    @return error code or 0 */
qserr      _std help_close   (qsh_file hf);

/** free help file cache.
    All unused help files are closed during this call.
    The help file is assumed as unused if help_close() is called or process
    that called help_open() is already terminated. */
void       _std help_release (void);

/** query list of topics in a help file.
    @param  hf          Help file handle
    @return help file information in a single application owned heap block
            (must be released by caller) or 0 if invalid handle */
qsh_info*  _std help_info    (qsh_file hf);

/** query topic id by name 
    @param  hf          Help file handle
    @param  topic       Topic name (0 returns index topic if available)
    @return topic id or 0 */
u32t       _std help_topicid (qsh_file hf, const char *topic);

/** query color scheme by name 
    @param  hf          Help file handle
    @param  topic       Topic name (0 returns index topic if available)
    @return color scheme id or negative value if error or not found */
s32t       _std help_schemeid(qsh_file hf, const char *scheme);

/** convert a topic to the screen data buffer (text/attr pair).
    @param  hf          Help file handle
    @param  scheme      Color scheme index (0 for default)
    @param  topic       Topic id
    @param  width       Output width in characters
    @param  [out] err   Error code value (optional)
    @return output data in a single application owned heap block (must be
            released by caller) or 0 if error (no topic or invalid parameters) */
qsh_topic* _std help_draw    (qsh_file hf, u32t scheme, u32t topic, u32t width,
                              qserr *err);

/** convert a topic to the ANSI text sequence.
    @param  hf          Help file handle
    @param  scheme      Color scheme index (0 for default)
    @param  topic       Topic id
    @param  outx        Start column in the result ANSI text (i.e. x offset
                        at the start of each line)
    @param  width       Output width in characters
    @param  [out] err   Error code value (optional)
    @return application owned heap block or 0 if error (no topic or invalid
            parameters). */
char*      _std help_drawansi(qsh_file hf, u32t scheme, u32t topic, u32t xpos,
                              u32t width, qserr *err);

/** high level "show help" function.
    Function shows a help window.
    
    Note that "winwidth" cannot be smaller than specified in the help file
    "minwidth" value.
    HELPSIZE environment variable is used for default size calculation.

    @param  name        Help file name (path or name to search in HELPPATH)
                        or qsh_file if HSFL_HANDLE flag is set in options,
                        if the first character is ? then it acts as a link
                        in common form "?file/topic/mark" and both topic and
                        mark parameters are ignored.
    @param  topic       Topic to open (use 0 for the index topic)
    @param  mark        Optional mark name in the topic to pos on it (can be 0)
    @param  winx        Help window x position (default if < 0)
    @param  winy        Help window y position (default if < 0)
    @param  winwidth    Help window width (default if 0)
    @param  winheight   Help window height (default if 0)
    @param  flags       Options (HSFL_*)
    @return error code or 0 */
qserr      _std help_show    (const void *name, const char *topic,
                              const char *mark, int winx, int winy,
                              u32t winwidth, u32t winheight, u32t flags);

/// @name help_show() options
//@{
#define HSFL_QUIET      0x0001    ///< be quiet on errors (no error messagebox)
#define HSFL_HANDLE     0x0002    ///< name parameter is qsh_file
#define HSFL_DEFINDEX   0x0004    ///< show index topic if requested not found
#define HSFL_NOF10      0x0008    ///< disable F10 key (exiting with topic drawing)
//@}

/** helper for a link navigation code.
    Function searches for a link in the visible area of the help topic.

    @param  td          Help topic data
    @param  prev        Selected link (index in td->llist, can be -1 for HLINK_ANY)
    @param  direction   Where to go - HLINK_* value
    @param  ypos        Visible area starts on this topic line
    @param  ysize       Visible area height
    @return link index (first in the group) or -1 if not found. */
int        _std help_findlink(qsh_topic *td, int prev, u32t direction,
                              u32t ypos, u32t ysize);

/// @name help_findlink() direction value
//@{
#define HLINK_ANY            0    ///< return top link in the visible area
#define HLINK_UP             1    ///< try to find link above "prev" in the visible area
#define HLINK_DOWN           2    ///< try to find link below "prev" in the visible area
#define HLINK_RIGHT          3    ///< try to find link to the right
#define HLINK_LEFT           4    ///< try to find link to the left
//@}

// default sys help file
#define SYS_HELP_FILE    "SYS.QSH"

#ifdef __cplusplus
}
#endif
#endif // QSINIT_HELPDATA
