//
// QSINIT "fslib" module
// ISOFS structs
//
#ifndef QSINIT_ISOFSINFO
#define QSINIT_ISOFSINFO

#ifdef __cplusplus
extern "C" {
#endif

#pragma pack(1)

#define ISO_SECTOR_BITS            (11)
#define ISO_SECTOR_SIZE            (1<<ISO_SECTOR_BITS)

// volume descriptor types
#define ISO_VD_PRIMARY               1
#define ISO_VD_SUPPLEMENTARY         2
#define ISO_VD_END                 255

#define ISO_STANDARD_ID        "CD001"
#define ISO_SIERRA_ID          "CDROM"

typedef struct {
  u8t                year;   ///< since 1900
  u8t               month;   ///< 1..12
  u8t                 day;   ///< 1..31
  u8t                hour;   ///< 0..23
  u8t                 min;   ///< 0..59
  u8t                 sec;   ///< 0..59
  s8t                 gmt;   ///< GMT offset (15 min*x): from -48 (West) to +52 (East)
} iso_date;

typedef struct {
  u16t                 le;
  u16t                 be;
} iso16t;

typedef struct {
  u32t                 le;
  u32t                 be;
} iso32t;

typedef struct {
  u8t              length;
  u8t           ea_length;
  iso32t           extent;
  iso32t             size;
  iso_date           date;
  u8t               flags;
  u8t            unitsize;   ///< file unit size
  u8t          interleave;
  iso16t             vseq;
  u8t             namelen;
  char            name[1];
} iso_dirrec;

#define ISODF_HIDDEN   0x01  ///< hidden file
#define ISODF_DIR      0x02  ///< directory
#define ISODF_ASSC     0x04  ///< "associated file"
#define ISODF_EXT      0x08  ///< ext.attrs present
#define ISODF_POSIX    0x10  ///< ext.attrs with owner and group permissions
#define ISODF_MULTI    0x80  ///< more directory records present for this file

/** get dir entry flags.
    there is no gmt field in Sierra and iso "flags" field is reserved (0) */
#define REC_FLAGS(rec, sierra) ((sierra) ? (u8t)(rec)->date.gmt : (rec)->flags)

/// ISO9660 volume desc
typedef struct {
  u8t                type;
  u8t               id[5];  ///< "CD001"
  u8t             version;
  u8t               flags;  ///< ms joliet only
  u8t           sysid[32];
  u8t           volid[32];
  u8t             res1[8];
  iso32t       vspacesize;
  u8t          escape[32];  ///< ms joliet only
  iso16t         vsetsize;
  iso16t             vseq;
  iso16t        blocksize;  ///< logical block size
  iso32t        ptab_size;  ///< path table size
  u32t            ptab_le;
  u32t        ptab_opt_le;
  u32t            ptab_be;
  u32t        ptab_opt_be;
  iso_dirrec         root;
  u8t      volset_id[128];
  u8t         pub_id[128];
  u8t        prep_id[128];
  u8t         app_id[128];
  u8t          cpr_id[37];
  u8t          abs_id[37];
  u8t          bib_id[37];  ///< file name of bibliography
  u8t          crdate[17];
  u8t         moddate[17];
  u8t         expdate[17];
  u8t         effdate[17];
  u8t               fsver;  ///< file structure version
  u8t                res2;
  u8t        appdata[512];
  u8t           res3[653];
} iso_desc;

/// High Sierra volume desc
typedef struct {
  iso32t              lbn;  ///< logical block number
  u8t                type;  ///< volume type: 1 = standard, 2 = coded
  u8t               id[5];  ///< "CDROM"
  u8t             version;
  u8t               flags;
  u8t           sysid[32];  ///< system identifier
  u8t           volid[32];  ///< volume identifier
  u8t             res1[8];
  iso32t       vspacesize;
  u8t         charset[32];  ///< character set bytes (0 = ASCII)
  iso16t         vsetsize;
  iso16t             vseq;
  iso16t        blocksize;  ///< logical block size
  iso32t        ptab_size;  ///< path table size
  u32t         ptab_le[4];  ///< LBN of 4 path tables (little endian)
  u32t         ptab_be[4];  ///< LBN of 4 path tables (big endian)
  iso_dirrec         root;
  u8t      volset_id[128];  ///< volume set identifier
  u8t         pub_id[128];  ///< publisher identifier
  u8t        prep_id[128];  ///< data preparer identifier
  u8t         app_id[128];  ///< application identifier
  u8t          cpr_id[32];  ///< file name of copyright notice
  u8t          abs_id[32];  ///< file name of abstract
  u8t          crdate[16];  ///< volume creation date and time
  u8t         moddate[16];  ///< volume modification date and time
  u8t         expdate[16];  ///< volume expiration date and time
  u8t         effdate[16];  ///< volume effective date and time
  u8t               fsver;  ///< file structure version
  u8t                res2;
  u8t        appdata[512];
  u8t           res3[680];
} hsg_desc;

#pragma pack()

#ifdef __cplusplus
}
#endif

#endif // QSINIT_ISOFSINFO
