//
// QSINIT "fslib" module
// JFS structs
//
#ifndef QSINIT_JFSINFO
#define QSINIT_JFSINFO

#include "qstypes.h"

#ifdef __cplusplus
extern "C" {
#endif

#define JFS_SUPER_SIGN   0x3153464A // "JFS1"
#define JFS_VERSION               1 // version number

#define JFS_AIX          0x80000000 ///< AIX support
#define JFS_OS2          0x40000000 ///< OS/2 support
#define JFS_DFS          0x20000000 ///< DCE DFS LFS support
#define JFS_LINUX        0x10000000 ///< Linux support

#define JFS_SPARSE       0x00020000 ///< sparse regular file support

#define JFS_GROUPCOMMIT  0x00000100 ///< group (of 1) commit
#define JFS_LAZYCOMMIT   0x00000200 ///< lazy commit
#define JFS_TMPFS        0x00000400 ///< temporary file system - do not log/commit
#define JFS_INLINELOG    0x00000800 ///< inline log within file system

typedef struct {
   u32t        tv_sec;
   u32t       tv_nsec;
} jfs_time;

typedef struct {
   u32t         addr1;
   u32t         addr2;
} jfs_pxd;

#pragma pack(1)
typedef struct {
   u32t          sign;
   u32t       version;   ///< version number
   u64t          size;   ///< aggregate size in hardware/LVM blocks
   u32t         bsize;   ///< aggregate block size in bytes;
   u16t       l2bsize;   ///< log2 of bsize
   u16t     l2bfactor;   ///< log2 (bsize/hardware block size)
   u32t        pbsize;   ///< hardware/LVM block size in bytes
   u16t      l2pbsize;   ///< log2 of pbsize
   u16t     reserved1;   ///< data alignment
   u32t        agsize;   ///< allocation group size in aggr. blocks
   u32t          flag;   ///< aggregate attributes
   u32t         state;   ///< mount/unmount/recovery state
   u32t      compress;   ///< > 0 if data is compressed
   jfs_pxd       ait2;   ///< first extent of secondary aggregate inode table
   jfs_pxd       aim2;   ///< first extent of secondary aggregate inode map
   u32t        logdev;   ///< device address of log
   u32t     logserial;   ///< log serial number at aggregate mount
   jfs_pxd     logpxd;   ///< inline log extent
   jfs_pxd    fsckpxd;   ///< inline fsck work space extent
   jfs_time      time;   ///< time last updated
   u32t    fsckloglen;   ///< Number of filesystem blocks reserved for the fsck service log.
   u8t        fscklog;   /** which fsck service log is most recent:
                             0 => no service log data yet
                             1 => the first one
                             2 => the 2nd one */
   char     fpack[11];   ///< file system volume name
   // extend FS features
   u64t         xsize;   ///< extend size
   jfs_pxd   xfsckpxd;   ///< extend fsckpxd
   jfs_pxd    xlogpxd;   ///< extend logpxd
   // 136 bytes boundary (jfs header lies about 128 here)
   char        attach;   ///< flag: set when aggregate is attached
   u8t   reserved2[7];   ///< reserved - set to 0
   u64t   totalUsable;   ///< # of 1K blocks available to "normal" users.
   u64t       minFree;   ///< # of 1K blocks held in reserve for exclusive use of root.
   u64t      realFree;   ///< # of free 1K blocks can be used by "normal" users.
   u8t     empty[344];
} jfs_superblock;

// file system state (superblock state)
#define FM_CLEAN         0x00000000	///< file system is unmounted and clean
#define FM_MOUNT         0x00000001 ///< file system is mounted cleanly
#define FM_DIRTY         0x00000002 ///< file system is dirty
#define	FM_LOGREDO       0x00000004 ///< log based recovery failed
#define	FM_EXTENDFS      0x00000008 ///< extendfs() in progress

#define pxd_addr(pxd) ((u64t)((pxd)->addr1&0xFF000000)<<8 | (pxd)->addr2)
#define pxd_len(pxd)  ((pxd)->addr1&0xFFFFFF)

#define pxd_len_set(pxd, len32)     \
   (pxd)->addr1 = ((pxd)->addr1 & 0xFF000000) + ((len32) & 0xFFFFFF)

#define pxd_addr_set(pxd, addr64) { \
   BYTEOF((pxd)->addr1,3) = (u8t)((u64t)(addr64) >> 32); \
   (pxd)->addr2 = (u32t)(addr64);   \
}

///< data extent descriptor
typedef struct {
   u8t           flag;
   u8t         res[3];
   u32t          size;   ///< size in byte
   u32t         addr1;
   u32t         addr2;
} jfs_dxd;
 
// dxd_t flags
#define DXD_INDEX              0x80  ///< B+-tree index
#define DXD_INLINE             0x40  ///< in-line data extent
#define DXD_EXTENT             0x20  ///< out-of-line single extent
#define DXD_FILE               0x10  ///< out-of-line file (inode)
#define DXD_CORRUPT            0x08  ///< Inconsistency detected

#define dxd_len_set   pxd_len_set
#define dxd_addr_set  pxd_addr_set

typedef struct {
   u8t           flag;
   u16t         rsvrd;
   u8t           off1;   ///< offset in unit of fsblksize
   u32t          off2;
   u32t         addr1;   ///< address in unit of fsblksize
   u32t         addr2;
} jfs_xad;

#define XTSLOTSIZE               16
#define L2XTSLOTSIZE              4

// possible values for maxentry
#define XTROOTINITSLOT           10
#define XTROOTMAXSLOT            18
#define XTREEMAXSLOT            256
#define XTENTRYSTART              2

#define xad_addr(xad) ((u64t)((xad)->addr1&0xFF000000)<<8 | (xad)->addr2)
#define xad_ofs(xad)  ((u64t)(xad)->off1<<32 | (xad)->off2)
#define xad_len(xad)  ((xad)->addr1&0xFFFFFF)

#define xad_ofs_set(xad, ofs64) {   \
   (xad)->off1 = (u8t)((u64t)(ofs64) >> 32); \
   (xad)->off2 = (u32t)(ofs64);     \
}
#define xad_len_set   pxd_len_set
#define xad_addr_set  pxd_addr_set

typedef union {
   struct xtheader {
      u64t       next;
      u64t       prev;
      u8t        flag;
      u8t        res1;
      s16t  nextindex;
      s16t   maxentry;
      s16t       res2;
      jfs_pxd    self;
   } header;
   jfs_xad        xad[XTREEMAXSLOT];
} jfs_xtreepage;

///< DASD limit information
typedef struct {
   u8t         thresh;   ///< alert threshold (in percent)
   u8t          delta;   ///< alert threshold delta (in percent)
   u8t           res1;
   u8t       limit_hi;   ///< DASD limit (in logical blocks)
   u32t      limit_lo;
   u8t        res2[3];
   u8t        used_hi;   ///< DASD usage (in logical blocks)
   u32t       used_lo;
} jfs_dasd;

#define DTSLOTDATALEN            15

/** directory page slot
    directory entry consists of type dependent head/only segment/slot and
    additional segments/slots linked via next field; end marked with -1 */
typedef struct {
   s8t           next;
   s8t            cnt;
   u16t          name[DTSLOTDATALEN];
} jfs_dtslot;

#define DTIHDRSIZE               10
#define DTIHDRDATALEN            11

/// internal node entry head/only segment
typedef struct {
   jfs_pxd         xd;   ///< child extent descriptor
   s8t           next;
   u8t         namlen;
   u16t          name[DTIHDRDATALEN];
} jfs_idtentry;

#define DTLHDRSIZE                6
#define DTLHDRDATALEN            13

/// leaf node entry head/only segment
typedef struct {
   u32t       inumber;
   s8t           next;
   u8t         namlen;
   u16t          name[DTLHDRDATALEN];
} jfs_ldtentry;

typedef union {
   struct {
      jfs_dasd   dasd;   ///< DASD limit/usage info
      u8t        flag;
      s8t   nextindex;   ///< next free entry in stbl
      s8t     freecnt;   ///< free count
      s8t    freelist;   ///< freelist header
      u32t    idotdot;   ///< parent inode number
      s8t     stbl[8];   ///< sorted entry index table
   } header;
   jfs_dtslot slot[9];
} jfs_dtroot;

#define DTROOTMAXSLOT             9
#define DTENTRYSTART              1

/** directory regular page.
   entry slot array of 32 byte slot

   sorted entry slot index table (stbl):
   contiguous slots at slot specified by stblindex, 1-byte per entry
     512 block:  16 entry tbl (1 slot)
    1024 block:  32 entry tbl (1 slot)
    2048 block:  64 entry tbl (2 slot)
    4096 block: 128 entry tbl (4 slot)

   data area:
     512 block:  16 - 2 =  14 slot
    1024 block:  32 - 2 =  30 slot
    2048 block:  64 - 3 =  61 slot
    4096 block: 128 - 5 = 123 slot

   Index is zero based; index fields refer to slot index except nextindex which
   refers to entry index in stbl;
   end of entry stot list or freelist is marked with -1 */
typedef union {
   struct {
      u64t       next;   ///< next sibling
      u64t       prev;   ///< previous sibling
      u8t        flag;
      s8t   nextindex;   ///< next entry index in stbl
      s8t     freecnt;
      s8t    freelist;   ///< slot index of head of freelist
      u8t     maxslot;   ///< number of slots in page slot[]
      s8t   stblindex;   ///< slot index of start of stbl
      u8t       rsrvd[2];
      jfs_pxd    self;   ///< self pxd
   } header;
   jfs_dtslot    slot[128];
} jfs_dtpage;


typedef struct {
   u8t            res;
   u8t           flag;   ///< 0 if free
   u8t           slot;   ///< slot within leaf page of entry
   u8t          addr1;   ///< upper 8 bits of leaf page address
   u32t         addr2;   ///< low dd of leaf page -OR- index of next entry when this entry was deleted
} jfs_dir_table_slot;

#define BT_TYPE       0x07
#define BT_ROOT       0x01  ///< root page
#define BT_LEAF       0x02  ///< leaf page
#define BT_INTERNAL   0x04  ///< internal page
#define BT_RIGHTMOST  0x10  ///< rightmost page
#define BT_LEFTMOST   0x20  ///< leftmost page

typedef struct {
   u32t   di_inostamp;   ///< stamp to show inode belongs to fileset
   s32t    di_fileset;   ///< fileset number
   u32t     di_number;   ///< inode number, aka file serial number
   u32t        di_gen;   ///< inode generation number
   jfs_pxd   di_ixpxd;   ///< inode extent descriptor
   u64t       di_size;   ///< size
   u64t    di_nblocks;   ///< number of blocks allocated
   u32t      di_nlink;   ///< number of links to the object
   u32t        di_uid;   ///< user id of owner
   u32t        di_gid;   ///< group id of owner
   u32t       di_mode;   ///< attribute, format and permission

   jfs_time  di_atime;   ///< time last data accessed
   jfs_time  di_ctime;   ///< time last status changed
   jfs_time  di_mtime;   ///< time last data modified
   jfs_time  di_otime;   ///< time created

   jfs_dxd     di_acl;   ///< acl descriptor
   jfs_dxd      di_ea;   ///< ea descriptor

   s32t   di_next_idx;   ///< next available dir_table index
   s32t    di_acltype;   ///< Type of ACL

   union {
      /* this table contains the information needed to find a directory
         entry from a 32-bit index. If index is small enough, the table
         is inline, otherwise, an x-tree root overlays this table */
      jfs_dir_table_slot table[12];
      u8t              extdata[96];
   };
   union {
      jfs_dtroot        dtroot;       ///< dtree root
      u8t               xtroot[288];  ///< jfs_xtreepage - 18(16) entries
      struct {
         jfs_dasd         dasd;
         jfs_dxd           dxd;
         union {
             u32t         rdev;
             u8t       symlink[128];
         };
         u8t          inlineea[128];
      } special;
   };
} jfs_dinode;

/// @name dinode.di_mode
//@{
#define IFMT             0x0000F000  ///< mask of file type
#define IFDIR            0x00004000  ///< directory
#define IFREG            0x00008000  ///< regular file
#define IFLNK            0x0000A000  ///< symbolic link

#define IEXEC            0x00000040  ///< execute permission
#define IWRITE           0x00000080  ///< write permission
#define IREAD            0x00000100  ///< read permission

#define IFJOURNAL        0x00010000  ///< journalled file
#define ISPARSE          0x00020000  ///< sparse file enabled
#define INLINEEA         0x00040000  ///< inline EA area free

#define IREADONLY        0x02000000  ///< no write access to file
#define IARCHIVE         0x40000000  ///< archive bit
#define ISYSTEM          0x08000000  ///< system file
#define IHIDDEN          0x04000000  ///< hidden file
#define IRASH            0x4E000000  ///< mask for changeable attributes
#define INEWNAME         0x80000000  ///< non-8.3 filename format
#define IDIRECTORY       0x20000000  ///< directory (shadow of real bit)
#define ATTRSHIFT                25  ///< bit shift to make dos attr bits
#define ISWAPFILE        0x00800000  ///< file open for pager swap space
//@}

/// @name fixed reserved inode number
//@{
#define AGGR_RESERVED_I           0  ///< aggregate inode (reserved)
#define AGGREGATE_I               1  ///< aggregate inode map inode
#define BMAP_I                    2  ///< aggregate block allocation map inode
#define LOG_I                     3  ///< aggregate inline log inode
#define BADBLOCK_I                4  ///< aggregate bad block inode
#define FILESYSTEM_I             16  ///< 1st/only fileset inode
//@}

/// @name per fileset inode
//@{
#define FILESET_RSVD_I            0  ///< fileset inode (reserved)
#define FILESET_EXT_I             1  ///< fileset inode extension
#define ROOT_I                    2  ///< fileset root inode
#define ACL_I                     3  ///< fileset ACL inode
#define FILESET_OBJECT_I          4  ///< 1st fileset inode available for a file/dir/link
//@}

// PSIZE >= file system block size >= physical block size >= DISIZE

/// @name fs fundamental size
//@{
#define PSIZE                  4096  ///< jfs "page" size
#define L2PSIZE                  12  ///< bit shift of PSIZE
#define DISIZE                  512  ///< on-disk inode size
#define L2DISIZE                  9  ///< bit shift of DISIZE
#define IDATASIZE               256  ///< inode inline data size
#define IXATTRSIZE              128  ///< inode inline extended attribute size

#define INOSPERIAG             4096  ///< number of disk inodes per iag
#define L2INOSPERIAG             12  ///< bit shift for # of disk inodes per iag
#define INOSPEREXT               32  ///< number of disk inode per extent
#define L2INOSPEREXT              5
#define IXSIZE   (DISIZE*INOSPEREXT) ///< inode extent size
#define INOSPERPAGE               8  ///< number of disk inodes per 4K page
#define L2INOSPERPAGE             3  ///< bit shift INOSPERPAGE

#define INODE_EXTENT_SIZE    IXSIZE  ///< inode extent size
#define JFS_MAXSIZE    ((u64t)1<<52)
#define MINJFS            0x1000000  ///< Minimum size of a JFS partition

#define MINBLOCKSIZE            512
#define MAXBLOCKSIZE           4096
//@}

/// @name fixed byte offset address
//@{
#define JFS_SUPER1_OFS       0x8000  ///< primary superblock physical offset
#define JFS_AIMAP_OFF        0x9000  ///< control page of volume inode map
#define JFS_AITBL_OFF        0xB000  ///< 1st extent of volume inode table
#define JFS_SUPER2_OFS (JFS_AITBL_OFF+INODE_EXTENT_SIZE)  ///< secondary superblock
#define JFS_BMAP_OFF   (JFS_SUPER2_OFS + PSIZE) ///< bitmap
//@}

#define EXTSPERIAG              128  ///< number of disk inode extent per iag
#define SMAPSZ                    4  ///< number of words per summary map

#define MAXAG                   128  ///< max number of allocation groups
#define L2MAXAG                   7  ///< shift of max num of AG

/// convert inode number to iag number */
#define INOTOIAG(ino)   ((ino) >> L2INOSPERIAG)

/// convert iag number to logical block number of the iag page
#define IAGTOLBLK(iagno,l2nbperpg)  (((iagno) + 1) << (l2nbperpg))

/// starting block number of the 4K page of an inode extent that contains ino
#define INOPBLK(pxd,ino,l2nbperpg) \
  (pxd_addr(pxd) + ((((ino) & INOSPEREXT-1) >> L2INOSPERPAGE) << (l2nbperpg)))

/// inode allocation group page (per 4096 inodes of an AG)
typedef struct {
   u64t       agstart;   ///< starting block of ag
   s32t        iagnum;   ///< inode allocation group number
   s32t    inofreefwd;   ///< ag inode free list forward
   s32t   inofreeback;   ///< ag inode free list back
   s32t    extfreefwd;   ///< ag inode extent free list forward
   s32t   extfreeback;   ///< ag inode extent free list back
   s32t       iagfree;   ///< iag free list
   u32t       inosmap[SMAPSZ]; ///< summary map: 1 bit per inode extent
   u32t       extsmap[SMAPSZ]; ///< sum map of mapwords w/ free extents
   s32t     nfreeinos;   ///< number of free inodes
   s32t     nfreeexts;   ///< number of free extents
   u8t      pad[1976];   ///< pad to 2048 bytes
   u32t          wmap[EXTSPERIAG]; ///< working allocation map
   u32t          pmap[EXTSPERIAG]; ///< persistent allocation map
   jfs_pxd     inoext[EXTSPERIAG]; ///< inode extent addresses
} jfs_iag;

// per AG control information (in inode map control page)
typedef struct {
   s32t       inofree;   ///< free inode list anchor
   s32t       extfree;   ///< free extent list anchor
   u32t       numinos;   ///< number of backed inodes
   u32t       numfree;   ///< number of free inodes
} iagctl_t;

// per fileset/aggregate inode map control page
typedef struct {
   s32t       freeiag;   ///< free iag list anchor
   s32t       nextiag;   ///< next free iag number
   u32t       numinos;   ///< num of backed inodes
   u32t       numfree;   ///< num of free backed inodes
   u32t     nbperiext;   ///< num of blocks per inode extent
   u32t   l2nbperiext;   ///< l2 of in_nbperiext
   u32t     diskblock;
   u32t         maxag;
   u8t            pad[2016];  ///< pad to 2048
   iagctl_t     agctl[MAXAG]; ///< AG control information
} dinomap_t;

#ifdef JFS_LOGFORMAT

/// log superblock
#define LOGSUPER_B                1
#define LOGSTART_B                2
#define LOGMAGIC         0x87654321
#define LOGVERSION                1
#define LOGPSIZE               4096
#define L2LOGPSIZE               12

typedef struct {
   u32t         magic;   ///< log lv identifier
   u32t       version;   ///< version number
   u32t        serial;   ///< log open/mount counter
   u32t          size;   ///< size in number of LOGPSIZE blocks
   u32t         bsize;   ///< logical block size in byte
   u32t       l2bsize;   ///< log2 of bsize

   u32t          flag;   ///< option
   u32t         state;   ///< state - see below

   u32t           end;   ///< address of last log record
   u32t        active[8];  ///< active file systems bit vector
   u32t      reserved[LOGPSIZE/4 - 17];
} logsuper_t;

/* log state */
#define LOGMOUNT                  0  ///< log mounted
#define LOGREDONE                 1  ///< log shutdown
#define LOGWRAP                   2  ///< log wrapped
#define LOGREADERR                3  ///< log read error detected

typedef struct {
   struct { // header
      u32t       page;   ///< log sequence page number
      u16t   reserved;
      u16t        eor;   ///< end-of-log offset of lasrt record write
   } h;
   u32t          data[LOGPSIZE/4 - 4]; ///< log record area
   struct { // trailer
      u32t       page;   ///< normally the same as h.page
      u16t   reserved;
      u16t        eor;   ///< normally the same as h.eor
   } t;
} logpage_t;

#define LOGPHDRSIZE               8  ///< log page header size
#define LOGPTLRSIZE               8  ///< log page trailer size

/// @name log record types
//@{
#define LOG_COMMIT           0x8000
#define LOG_SYNCPT           0x4000
#define LOG_MOUNT            0x2000
#define LOG_REDOPAGE         0x0800
#define LOG_NOREDOPAGE       0x0080
#define LOG_NOREDOINOEXT     0x0040
#define LOG_UPDATEMAP        0x0008
#define LOG_NOREDOFILE       0x0001
//@}

/// @name REDOPAGE/NOREDOPAGE log record data type
#define LOG_INODE            0x0001
#define LOG_XTREE            0x0002
#define LOG_DTREE            0x0004
#define LOG_BTROOT           0x0010
#define LOG_EA               0x0020
#define LOG_ACL              0x0040
#define LOG_DATA             0x0080
#define LOG_NEW              0x0100
#define LOG_EXTEND           0x0200
#define LOG_RELOCATE         0x0400
//@}

/// @name UPDATEMAP log record descriptor type
#define LOG_ALLOCXADLIST     0x0080
#define LOG_ALLOCPXDLIST     0x0040
#define LOG_ALLOCXAD         0x0020
#define LOG_ALLOCPXD         0x0010
#define LOG_FREEXADLIST      0x0008
#define LOG_FREEPXDLIST      0x0004
#define LOG_FREEXAD          0x0002
#define LOG_FREEPXD          0x0001
//@}

typedef struct lrd {
   // type independent area (16)
   u32t        logtid;   ///< log transaction identifier
   s32t     backchain;   ///< ptr to prev record of same transaction
   u16t          type;   ///< record type
   u16t        length;   ///< length of data in record (in byte)
   s32t     aggregate;   ///< file system lv/aggregate
   // type dependent area (20)
   union {
       // COMMIT: transaction commit - no type-dependent information
       // REDOPAGE: apply after-image
       struct {
          u32t     fileset;   ///< fileset number
          u32t       inode;   ///< inode number
          u16t        type;   ///< REDOPAGE record type
          u16t  l2linesize;   ///< log2 of line size
          jfs_pxd      pxd;   ///< on-disk page pxd
       } redopage;
       // NOREDOPAGE: the page is freed
       struct {
          u32t     fileset;   ///< fileset number
          u32t       inode;   ///< inode number
          u16t        type;   ///< NOREDOPAGE record type
          u16t    reserved;   ///< reserved
          jfs_pxd      pxd;   ///< on-disk page pxd
       } noredopage;
       // UPDATEMAP: update block allocation map
       struct {
          u32t     fileset;   ///< fileset number
          u32t       inode;   ///< inode number
          u16t        type;   ///< UPDATEMAP record type
          u16t         nxd;   ///< number of extents
          jfs_pxd      pxd;
       } updatemap;
       // NOREDOINOEXT: the inode extent is freed
       struct {
          u32t     fileset;   ///< fileset number
          u32t      iagnum;   ///< IAG number
          u32t  inoext_idx;   ///< inode extent index
          jfs_pxd      pxd;   ///< on-disk page pxd
       } noredoinoext;
       // SYNCPT: log sync point
       struct {
          s32t        sync;  ///< syncpt address (0 = here)
       } syncpt;
       // MOUNT: file system mount - no type-dependent information;
       // FREEXTENT: free specified extent(s)
       struct {
          u32t        type;  ///< FREEXTENT record type
          u32t     nextent;  ///< number of extents
          // data: PXD or XAD list
       } freextent;
       // NOREDOFILE: this file is freed
       struct {
          u32t     fileset;  ///< fileset number
          u32t       inode;  ///< inode number
       } noredofile;
       // NEWPAGE:
       struct {
          u32t     fileset;  ///< fileset number
          u32t       inode;  ///< inode number
          u32t        type;  ///< NEWPAGE record type
          jfs_pxd      pxd;  ///< on-disk page pxd
       } newpage;
   };
} lrd_t;

#define LOGRDSIZE  (sizeof(struct lrd))

#endif // JFS_LOGFORMAT

#ifdef JFS_DMAP

#define BMAPVERSION               1  ///< version number
#define TREESIZE     (256+64+16+4+1) ///< size of a dmap tree
#define LEAFIND          (64+16+4+1) ///< index of 1st leaf of a dmap tree
#define LPERDMAP                256  ///< num leaves per dmap tree
#define L2LPERDMAP                8  ///< shift number of leaves per dmap tree
#define DBWORD                   32  ///< # of blks covered by a map word
#define L2DBWORD                  5  ///< shift # of blks covered by a mword
#define BUDMIN             L2DBWORD  ///< max free string in a map word
#define BPERDMAP   (LPERDMAP*DBWORD) ///< num of blks per dmap
#define L2BPERDMAP               13  ///< shift of num of blks per dmap
#define CTLTREESIZE (1024+256+64+16+4+1)  ///< size of a dmapctl tree
#define CTLLEAFIND   (256+64+16+4+1) ///< idx of 1st leaf of a dmapctl tree
#define LPERCTL                1024  ///< num of leaves per dmapctl tree
#define L2LPERCTL                10  ///< shift of leaves per dmapctl tree
#define ROOT                      0  ///< index of the root of a tree
#define NOFREE             ((s8t)-1) ///< no free blocks
#define L2MINAGSZ                25  ///< shift of minimum AG size in bytes
#define BMAPBLKNO                 0  ///< lblkno of bmap within the map

// maximum l2 number of disk blocks at the various dmapctl levels.
#define L2MAXL0SIZE     (L2BPERDMAP + 1 * L2LPERCTL)
#define L2MAXL1SIZE     (L2BPERDMAP + 2 * L2LPERCTL)
#define L2MAXL2SIZE     (L2BPERDMAP + 3 * L2LPERCTL)

// maximum number of disk blocks at the various dmapctl levels.
#define MAXL0SIZE          ((u64t)1 << L2MAXL0SIZE)
#define MAXL1SIZE          ((u64t)1 << L2MAXL1SIZE)
#define MAXL2SIZE          ((u64t)1 << L2MAXL2SIZE)

#define MAXMAPSIZE        MAXL2SIZE  ///< maximum aggregate map size

/// determine the maximum free string for four (lower level) nodes of the tree
#define	TREEMAX(cp) ((s8t)(max(max(*(cp),*((cp)+1)),max(*((cp)+2),*((cp)+3)))))

// convert map size to the zero origin dmapctl level of the top dmapctl
#define BMAPSZTOLEV(size)  ((size)<=MAXL0SIZE ? 0 : (size)<=MAXL1SIZE ? 1 : 2)

/// dmap summary tree
typedef struct {
   u32t        nleafs;   ///< number of tree leafs         
   u32t      l2nleafs;   ///< bit shift of tree leafs          
   u32t       leafidx;   ///< index of first tree leaf         
   u32t        height;   ///< height of the tree           
   s8t         budmin;   ///< min l2 tree leaf value to combine    
   s8t          stree[TREESIZE];///< TREESIZE: tree              
   u8t            pad[2];     ///< 2: pad to word boundary         
} dmaptree_t; // 360

/// dmap page per 8k blocks bitmap
typedef struct {
   u32t       nblocks;   ///< num blks covered by this dmap 
   u32t         nfree;   ///< num of free blks in this dmap 
   u64t         start;   ///< starting blkno for this dmap  
   dmaptree_t    tree;   ///< dmap tree                   
   u8t            pad[1672];     ///< pad to 2048 bytes          
   u32t          wmap[LPERDMAP]; ///< bits of the working map    
   u32t          pmap[LPERDMAP]; ///< bits of the persistent map 
} dmap_t;

/// disk map control page per level
typedef struct {
   u32t        nleafs;   ///< number of tree leafs          
   u32t      l2nleafs;   ///< bit shift of tree leafs       
   u32t       leafidx;   ///< index of the first tree leaf  
   u32t        height;   ///< height of tree            
   s8t         budmin;   ///< minimum l2 tree leaf value    
   s8t          stree[CTLTREESIZE]; ///< dmapctl tree    
   u8t            pad[2714];        ///< pad to 4096                
} dmapctl_t;

typedef union {
   dmaptree_t      t1;
   dmapctl_t       t2;
} dmtree_t;
    
// on-disk aggregate disk allocation map descriptor
typedef struct {
   u64t       mapsize;   ///< 8: number of blocks in aggregate  
   u64t         nfree;   ///< 8: num free blks in aggregate map 
   u32t   l2nbperpage;   ///< 4: number of blks per page        
   u32t         numag;   ///< 4: total number of ags            
   u32t      maxlevel;   ///< 4: number of active ags       
   u32t         maxag;   ///< 4: max active alloc group number  
   u32t        agpref;   ///< 4: preferred alloc group (hint)   
   u32t       aglevel;   ///< 4: dmapctl level holding the AG   
   u32t      agheigth;   ///< 4: height in dmapctl of the AG    
   u32t       agwidth;   ///< 4: width in dmapctl of the AG     
   u32t       agstart;   ///< 4: start tree index at AG height  
   u32t      agl2size;   ///< 4: l2 num of blks per alloc group 
   u64t        agfree[MAXAG]; ///< 8*MAXAG: per AG free count     
   u64t        agsize;   ///< 8: num of blks per alloc group    
   s8t     maxfreebud;   ///< 1: max free buddy system      
   u8t            pad[3007];  ///< 3007: pad to 4096                 
} dbmap_t;

#endif // JFS_DMAP

#pragma pack()

#ifdef __cplusplus
}
#endif
#endif // QSINIT_JFSINFO
