//
// QSINIT help module
// help file processing
//
#include "helploc.h"
#ifdef __QSINIT__
#include "qstask.h"
#include "qsio.h"
#endif
#include "sp_ini.h"
#include "qs_rt.h"

#define HDATA_SIGN  0x31504C48
#define HUSER_SIGN  0x32504C48

#define DEF_ATTR           0xF // def text color (white on black)
#define DEF_LINKATTR      0x3F // def link color
#define DEF_LINKSELATTR   0xCF // def selected link color
#define DEF_BADATTR       0xC0 // unknown style
#define DEF_BLACK         0x08 // byte 00 attr replacement
#define LINE_ALLOC_STEP     60
#define OTHER_AXIS_W         4 /* larger value means lower priority of another
                                  axis in arrow navigation */

#define CHAR_VBAR       '\xDD'

#define instance_ret(type,inst,err)   \
   type *inst = (type*)hf;            \
   if (!inst || inst->sign!=HUSER_SIGN) return err;

typedef Strings<u64t> TLinks;

struct hdata;

struct huser {
   u32t          sign;
   hdata          *fp;
   qshandle   thandle;  ///< handle of the caller thread
};

struct hdata {
   u32t          sign;
   TINIFile      *src;
   List<huser*> users;  ///< active users
   TStrings    topics;
   TStrings   schemes;
   TStrings     alias;
   int          index;
   u32t      minwidth;
   spstr        title;
};

Strings<hdata*>   HFL;

/* - all data is stored in a global memory (DLL owned, actually);
   - the help file remains loaded after the last user closed it;
   - all unused files are freed in help_release(), it also takes care
     of missing help_close() calls if the caller process is exited already;
   - this file is shared with PC build, so it written without using QSINIT
     features (so unhandy) */
qserr _std help_open(const char *name, qsh_file *hf) {
   if (!hf) return E_SYS_ZEROPTR;
   spstr fp;
   // compilation error
   if (sizeof(u64t)!=sizeof(qsh_link)) return E_SYS_SOFTFAULT;
   // use fullpath() to normalize incoming full path else _searchenv
   if (name[0] && name[1]==':') {
      if (!_fullpath(fp.LockPtr(_MAX_PATH+1), name, _MAX_PATH+1)) {
         fp.UnlockPtr(0);
         return E_SYS_NOPATH;
      }
   } else
      _searchenv(name, "HELPPATH", fp.LockPtr(_MAX_PATH+1));
   fp.UnlockPtr();
   // access to global data starts here ...
   HelpCritSec lk;
   /* a lot of memory will be lost if thread will be killed below, so
      lock it until the end of this function */
   SysLock  syslk;  
      
   int   idx = HFL.IndexOfICase(fp());
   hdata *fe;

   if (idx<0) {
      TINIFile *src = 0;
      qserr     err;
#ifdef __QSINIT__
      { // try to unzip it first
         TStrings Src;
         err = help_unpack(fp(),Src);
         if (!err) src = new TINIFile(&Src);
      }
#endif
      if (!src) src = new TINIFile(fp(),true);
      if (!src) return E_SYS_NOMEM;

      if (!src->SectionExists("main") || !src->SectionExists("@@default")) {
         err = src->GetFileText().Count()==0 ? E_SYS_NOFILE : E_SYS_BADFMT;
         delete src;
         return err;
      }
      // new help file
      fe = new hdata;
      fe->sign  = HDATA_SIGN;
      fe->src   = src;

      src->ReadSectionOrgNames(fe->topics);
      // remove all service section names (including duplicates)
      static const char *service[] = { "main", "alias", 0 };
      for (idx=0; service[idx]; idx++)
         while (1) {
            int pos = fe->topics.IndexOfICase(service[idx]);
            if (pos<0) break;
            fe->topics.Delete(pos);
         }
      // move scheme sections into the own list
      for (idx=0; idx<fe->topics.Count(); ) {
         if (fe->topics[idx][0]=='@' && fe->topics[idx][1]=='@') {
            spstr name(fe->topics[idx],2);
            // "default" is forced to be scheme 0
            if (stricmp(name(),"default")==0) fe->schemes.Insert(0,name);
               else fe->schemes.Add(name);
            fe->topics.Delete(idx);
            continue;
         }
         idx++;
      }
      // read aliases and find topic id for them
      src->ReadSection("alias", fe->alias);
      fe->alias.TrimEmptyLines();
      for (idx=0; idx<fe->alias.Count(); ) {
         spstr key = fe->alias.Name(idx),
               src = fe->alias.Value(idx).trim();
         int   pos = fe->topics.IndexOfICase(src());

         if (!src || !key || pos<0) fe->alias.Delete(idx); else {
            fe->alias[idx] = key;
            fe->alias.Objects(idx) = pos;
            idx++;
         }
      }

      fe->title       = src->ReadStr("main", "title").trim();
      fe->minwidth    = src->ReadInt("main", "minwidth");
      spstr indexname = src->ReadStr("main", "index").trim();
      fe->index       = fe->topics.IndexOf(indexname);
      // a new file added
      HFL.AddObject(fp, fe);
   } else
      fe = HFL.Objects(idx);

   huser *ref   = new huser;
   ref->sign    = HUSER_SIGN;
   ref->fp      = fe;
#ifdef __QSINIT__
   // use TID 1 to verify (below) that caller process terminated
   ref->thandle = mt_threadhandle(0,1);
   io_setstate(ref->thandle, IOFS_DETACHED, 1);
#endif
   fe->users.Add(ref);

   *hf = ref;
   return 0;
}

qserr _std help_close(qsh_file hf) {
   HelpCritSec lk;

   instance_ret(huser, ref, E_SYS_INVOBJECT);
   hdata *fe = ref->fp;
   if (fe->sign!=HDATA_SIGN) return E_SYS_INVOBJECT;
   // disable thread kill
   SysLock  syslk;  
   // remove from the help file owners
   int pos = fe->users.IndexOf(ref);
   if (pos<0) {
      log_printf("no help user %X in the list! (%s)\n", ref, fe->src->FileName()());
   } else {
      fe->users.Delete(pos);
   }
#ifdef __QSINIT__
   if (ref->thandle) mt_closehandle(ref->thandle);
#endif
   ref->sign = 0;
   delete ref;
   return 0;
}

void _std help_release(void) {
   HelpCritSec lk;

   u32t idx = 0, fcnt = 0;
   // loop all cached files
   while (idx<HFL.Count()) {
      hdata *fe = HFL.Objects(idx);
      u32t uidx = 0;
      // check for missing help_close() calls
      while (uidx<fe->users.Count()) {
         huser *ref = fe->users[uidx];
         if (ref->sign!=HUSER_SIGN) {
            log_printf("broken help user %X in the list! (%s)\n", ref,
               fe->src->FileName()());
            // it`s just unsafe to free it
            fe->users.Delete(uidx);
            continue;
         }
#ifdef __QSINIT__
         u32t flags;
         // handle is invalid or thread is "broken"? assume instance as closed
         if (io_getstate(ref->thandle, &flags) || (flags & IOFS_BROKEN)) {
            help_close(ref);
            continue;
         }
#endif
         uidx++;
      }
      if (fe->users.Count()==0) {
         HFL.Delete(idx);
         fe->sign = 0;
         delete fe->src;
         delete fe;
         fcnt++;
         continue;
      }
      idx++;
   }
   if (fcnt) log_printf("%u help file(s) unloaded\n", fcnt);
}

qsh_info* _std help_info(qsh_file hf) {
   HelpCritSec lk;
   instance_ret(huser, ref, 0);
   hdata *fe = ref->fp;
   if (fe->sign!=HDATA_SIGN) return 0;

   char     tmp[2];
   u32t    tlen = 1, slen = 1;
   // calc list sizes
   str_unroll(fe->topics.Str, 0, 0, &tmp, tlen, 0);
   str_unroll(fe->schemes.Str, 0, 0, &tmp, slen, 0);

   u32t   items = fe->topics.Count(),
        schemes = fe->schemes.Count(),
          asize = tlen + slen + sizeof(qsh_info) + items*sizeof(char*) +
                  schemes*sizeof(char*) + fe->title.length() + 1;
   qsh_info *rc = (qsh_info*)malloc_local(asize);

   if (!items) memset(rc, 0, asize); else {
      rc->nschemes = schemes;
      rc->item     = (char**)(rc + 1);
      rc->scheme   = rc->item + items;
      void *tp     = rc->scheme + schemes,
           *sp     = (char*)tp + tlen;
      rc->title    = (char*)sp + slen;
      strcpy(rc->title, fe->title());
      rc->index    = fe->index<0 ? 0 : fe->index + 1;
      rc->minwidth = fe->minwidth;

      if (!str_unroll(fe->topics.Str, 0, 0, tp, tlen, rc->item) ||
         !str_unroll(fe->schemes.Str, 0, 0, sp, slen, rc->scheme))
            log_printf("unroll error?\n");
   }
   rc->ntopics  = items;
   return rc;
}

#define GET_CODE      0x0001  ///< preformatted text
#define GET_INLINK    0x0002  ///< active <l> tag
#define GET_VLINE     0x0004  ///< vertical line in the indent
#define GET_AMASK     0x0030  ///< align mask
#define GET_ALEFT     0x0000
#define GET_ARIGHT    0x0010
#define GET_ACENTER   0x0020
#define GET_AJUSTIFY  0x0030

struct BufferState {
   u16t        *out;
   u32t        line;
   u32t       lines;
   u32t    scrwidth;
   qsh_topic   *res;

   u32t     lcommit;  ///< first unaligned link (in the current line/cell)
   int       lgroup,
              lback;
   spstr      clink;  ///< current active link
   TLinks     links;
};

struct TextState {
   TStrings     out,  ///< ready text, last line can be incomplete
                att;  ///< the same, but color attributes
   u32t       width,  ///< output width
              state,  ///< state flags
               scrx;  ///< left position on the screen
   u8t         lcol,  ///< current link color
               icol,  ///< current indent color
                ind,  ///< current indent value
              txcol;  ///< current default text color

   BufferState *pbs;
};

#define GC_INDENT   1
#define GC_NOLINK   2

inline u8t getcolbyte(TextState &s, u32t fl=0) {
   u8t col = fl & GC_INDENT ? s.icol :
           !(fl & GC_NOLINK) && (s.state&GET_INLINK) ? s.lcol : s.txcol;
   // deny zero value in any case
   return col ? col : DEF_BLACK;
}

/** fill/align string (and colors attr string) up to width characters.
    @return position of the source string in the result string */
static u32t str_align(spstr &str, spstr &attr, u32t width, u32t state, u8t color) {
   u32t    pos = 0;
   int    diff = width - str.length(),
         align = state & GET_AMASK;
   /// always guarantee "width" length
   if (diff>0) {
      // just fill end of line in preformatted mode
      if (state & GET_CODE) align = GET_ALEFT;
      switch (align) {
         case GET_ALEFT   :
            str += spstr(' ', diff);
            attr+= spstr((char)color, diff);
            break;
         case GET_AJUSTIFY: // fixme!

         case GET_ACENTER : {
            u32t dright = diff>>1;
            if (dright) {
               str += spstr(' ', dright);
               attr+= spstr((char)color, dright);
            }
            diff -= dright;
            pos   = diff;
            if (!diff) break;
         }
         case GET_ARIGHT  :
            str.insert(spstr(' ',diff),0);
            attr.insert(spstr(!attr?(char)color:attr[0],diff),0);
            pos   = diff;
            break;
      }
   }
   return pos;
}

static void complete_line(TextState &s) {
   if (s.out.Count()>0) {
      u32t pos = 0,
            ii = s.out.Max();

      if (s.out[ii].length() <= s.width - s.ind) {
         pos = str_align(s.out[ii], s.att[ii], s.width-s.ind, s.state,
                         getcolbyte(s,GC_NOLINK)) + s.ind;
         // add indent into the aligned string
         if (s.ind) {
            char icolor = (char)getcolbyte(s,GC_INDENT);
            s.out[ii].insert(spstr(' ',s.ind), 0);
            if (s.state & GET_VLINE) s.out[ii][0] = CHAR_VBAR;
            s.att[ii].insert(spstr(icolor,s.ind), 0);
         }
      }
#if 0
      if (pos)
         log_printf("link [pos=%u ind=%u n=%u/%i [%s]]\n", pos, s.ind,
            s.pbs->lcommit, s.pbs->links.Count(), s.out[ii]());
#endif
      // make sure we`re update only *current* cell
      for (ii=s.pbs->lcommit; ii<s.pbs->links.Count(); ii++) {
         qsh_link *lnk = (qsh_link*)&s.pbs->links.Objects(ii);

         if (lnk->xpos>=s.scrx && lnk->xpos<s.scrx+s.width) {
            lnk->xpos += pos;
            if (ii==s.pbs->lcommit) s.pbs->lcommit++;
         }
      }
   }
}

/* add a (long) line to the string list.
    @param s       text buffers
    @param wl      words/lines to add */
static void add_lines(TextState &s, TStrings &wl) {
   if (!wl.Count()) return;
   spstr   tln, aln;
   int     idx = s.out.Count()>0 && s.out[s.out.Max()].length()<s.width 
                 ? s.out.Max() : -1;
   u32t    len = idx>=0 ? s.out[idx].length() : 0,
          ilen = len,
           ind = s.ind;
   char  color = (char)getcolbyte(s);

   if (len) {
      tln = s.out[idx];
      aln = s.att[idx];
   }
   // drop invalid value
   if (ind>s.width-1) ind = 0;
   // split line to strings no longer than width
   for (u32t ii=0; ii<wl.Count(); ) {
      u32t    nlen = wl[ii].length();
      int  noncomp = 0;
      if (len + nlen <= s.width - ind) {
         tln += wl[ii];
         aln += spstr(color,nlen);
         len += nlen;
         if (ii++<wl.Max()) continue;
         noncomp = len < s.width - ind;
      }
      if (!len) {
         spstr cut = wl[ii].deltostr(0, s.width-ind);
         tln += cut;
         if (cut.length()) aln += spstr(color, cut.length());
      }
      // added text is a link?
      if ((s.state & GET_INLINK) && (len - ilen)) {
         union {
            qsh_link  lnk;
            u64t       lv;
         };
         lnk.xpos = s.scrx + ilen;
         lnk.ypos = s.pbs->line + (idx>=0 ? idx : s.out.Count());
         lnk.len  = len - ilen;
         lnk.back = s.pbs->lgroup == s.pbs->links.Count() && s.pbs->lback>=0
                    ? s.pbs->lback : 0xFFFF;
         s.pbs->links.AddObject(s.pbs->clink, lv);
      }    
      // replace the first (originally incomplete) line or add a new one
      if (idx<0) { s.out.Add(tln); s.att.Add(aln); } else {
         s.out[idx] = tln;
         s.att[idx] = aln;
         ilen = 0;
         idx  = -1;
      }
      if (!noncomp) complete_line(s);

      tln.clear(); aln.clear();
      len = 0;
   }
}

/** format text to strings.
    function stops on <.
    last line at exit is finished and aligned only if string length
    equal to width.

    @param s       text buffers
    @param src     source line
    @param notag   add string as is
    @return src position where it stops (points to 0 or to '<'). */
static const char *get_next(TextState &s, const char *src, int notag=0) {
   TStrings   wl;
   spstr    line;
   u32t      ind = s.ind, statesv;
   if (ind>s.width-1) ind = 0;
   // force prefomatted mode for <x20>
   if (notag) { statesv = s.state; s.state|=GET_CODE; }

   if (s.state & GET_CODE) {
      while (*src) {
         u8t cc = *src++;
         if (cc=='<' && !notag) { src--; break; }
         if (cc=='\t') line+="    "; else
            if (cc!='\n') line+=cc<' ' ? ' ' : (char)cc;
         if (line.length()>=s.width-ind || cc=='\n') {
            wl.Add(line);
            add_lines(s,wl);
            line.clear();
            wl.Clear();
            if (cc=='\n') complete_line(s);
         }
      }
   } else {
      while (*src) {
         u8t cc = *src++;
         if (cc=='<' && !notag) { src--; break; }
         if (cc>' ') line+=(char)cc; else {
            int spc = 0;
            /* any of <=' ' acts as a space, but leading space counted only
               for the space itself or \t */
            while (cc && cc<=' ') {
               if (cc==' ' || cc=='\t') spc++; else spc = 0;
               cc = *src++;
            }
            if (line.length()) { line+=" "; wl.Add(line); line.clear(); }
            // single/leading space (but not at the start of the line)
            if (!wl.Count() && spc) {
               int lidx = s.out.Max();
               if (lidx>=0 && s.out[lidx].length() && s.out[lidx].length()<s.width)
                  wl.Add(spstr(" "));
            }
            src--;
         }
      }
   }
   // flush it
   if (line.length()) wl.Add(line);
   if (wl.Count()) add_lines(s,wl);
   if (notag) s.state = statesv;
   return src;
}

/// return error code
qserr get_int(const char *&src, u32t width, int &res, int mode) {
   char *ep = 0;
   if (mode & rINT_HEX)
      res = strtoul(src, &ep, 16);
   else
      res = mode & rINT_UNS ? strtoul(src, &ep, 10):
                              strtol (src, &ep, 10);
   if (*ep=='%' && (mode & rINT_PER)) {
      int neg = 1;
      if (res < 0) { res = -res; neg = -1; }
      ep++;

      if (res > 100) res = 100;
      res = res * width / 100 * neg;
   }
   if (ep==src) return E_HLP_BADINT;
   src = ep;
   return 0;
}

/** read " = params >"
    @param res     result string
    @param midspc  allow spaced inside */
static int get_params(const char *&src, spstr &res, int midspc = 0) {
   while (*src==' ') src++;
   if (*src=='=') src++;
   while (*src==' ') src++;
   const char *ep = src;
   while (*ep && (midspc || *ep!=' ') && *ep!='>') ep++;
   if (ep==src) res.clear(); else {
      res = spstr(src, ep-src);
      // trim trailing spaces
      if (midspc) res.trimright();
      src = ep;
   }
   return 1;
}

static void add_br(TextState &s, u32t n_br=1) {
   complete_line(s);
   while (n_br--)
      if (n_br) {
         spstr lt(' ',s.width),
               la((char)getcolbyte(s),s.width);
         // fix indent color and add a vertical bar if enabled
         if (s.ind) {
            char icol = (char)getcolbyte(s,GC_INDENT);
            if (s.state & GET_VLINE) lt[0] = CHAR_VBAR;
            for (u32t ii=0; ii<s.ind; ii++) la[ii] = icol;
         }
         s.out.Add(lt);
         s.att.Add(la);
      } else {
         s.out.Add(spstr());
         s.att.Add(spstr());
      }
}

/** commit text to the output buffer.
    function commit all complete lines.
    @param s       text data
    @param xpos    rectange x position
    @return 0 if no memory in realloc (should never occur) */
static int commit_buf(TextState &s, u32t xpos) {
   BufferState &bs = *s.pbs;
   // realloc error has occured before
   if (!bs.out) return 0;
   if (!s.out.Count()) return 1;
   // expand buffer
   u32t ln = 0, ii;

   while (ln<s.out.Count() && s.out[ln].length()==s.width) {
      if (bs.line + 1 >= bs.lines) {
         u32t  nofs = bs.lines * bs.scrwidth;
         bs.lines  += LINE_ALLOC_STEP;
         // realloc result buffer
         qsh_topic *nres = (qsh_topic*)realloc(bs.res, bs.lines *
                            bs.scrwidth * 2 + sizeof(qsh_topic));
         if (!nres) { free(bs.res); bs.res = 0; bs.out = 0; return 0; }

         bs.res = nres;
         bs.out = (u16t*)(nres + 1);
      }
      u16t *lp = bs.out + bs.line*bs.scrwidth + xpos;
      // some kind of assert (lengths MUST be equal)
      if (s.out[ln].length()!=s.att[ln].length())
         s.att[ln] = spstr(DEF_BADATTR, s.out[ln].length());
      // draw a line
      for (ii=0; ii<s.width; ii++) {
         if (ii+xpos>=bs.scrwidth) break;
         // char can be signed in SysView PC build
         lp[ii] = (u16t)s.att[ln][ii]<<8 | (u8t)s.out[ln][ii];
      }
      bs.line++;
      ln++;
   }
   if (ln) {
#if 0
      static int counter = 0;
      s.out.SaveToFile(spstr().sprintf("dump%03u.txt",counter++)());
#endif
      s.out.Delete(0,ln);
      s.att.Delete(0,ln);
   }
   return 1;
}

static u32t getcolor(TStrings &styles, const char *name, u32t def = DEF_BADATTR) {
   if (styles.IndexOfName(name)>=0) return styles.DwordValue(name);
   return def;
}

static qserr parse_align(TextState &s, const char *&src) {
   u32t newal = 0;
   spstr   al;
   get_params(src, al, 1);
   al.lower();
   if (al=="j" || al=="justify") newal = GET_AJUSTIFY; else
      if (al=="c" || al=="center") newal = GET_ACENTER; else
         if (al=="r" || al=="right") newal = GET_ARIGHT; else
            if (al!="l" && al!="left") return E_HLP_BADALIGN;
   s.state = s.state & ~GET_AMASK | newal;
   return 0;
}

#define LCELLS   16

#define FreeCells()            \
   for (ii=0; ii<LCELLS; ii++) \
      if (ts[ii]) { delete ts[ii]; ts[ii] = 0; }

static qsh_topic* _std help_drawtext(qsh_file hf, u32t scheme_id,
                                     const char *src, const char *topicname,
                                     u32t width, qserr *perr)
{
   TStrings style;
   spstr    topic(topicname);
   u32t    coldef, collink, colsel;
   // objects & args should already be verified in caller
   if (perr) *perr = E_SYS_INVPARM;
   // lock help data only for a short time
   if (!hf) return 0; else {
      HelpCritSec lk;
      instance_ret(huser, ref, 0);
      hdata *fe = ref->fp;
      if (fe->sign!=HDATA_SIGN) return 0;
      // invalid value dropped to the default
      if (scheme_id>=fe->schemes.Count()) scheme_id = 0;
      spstr sn("@@");
      if (scheme_id<fe->schemes.Count()) sn+=fe->schemes[scheme_id];
      // read selected color scheme section
      fe->src->ReadSection(sn, style);
      colsel   = getcolor(style, "selected", DEF_LINKSELATTR);
      coldef   = getcolor(style, "default", DEF_ATTR);
      collink  = getcolor(style, "link", DEF_LINKATTR);
   }
   TStrings   marks;
   // output buffer
   BufferState   bs;
   bs.lines    = LINE_ALLOC_STEP;
   bs.line     = 0;
   bs.res      = (qsh_topic*)malloc_th(bs.lines*width*2 + sizeof(qsh_topic));
   bs.out      = (u16t*)(bs.res + 1);
   bs.scrwidth = width;
   bs.lgroup   = -1;
   bs.lback    = -1;
   bs.lcommit  = 0;
   if (!bs.res) {
      if (perr) *perr = E_SYS_NOMEM;
      return 0;
   }
   // text/table buffers
   TextState  ms,         // main text state
             *ts[LCELLS]; // cell state array
   u32t    cellw[LCELLS], ccell, ii, tabw,
          tsline,         // start line of the current table row
           ncell = 0;
   qserr     err = 0;
   memset(&ts, 0, sizeof(ts));
   ms.lcol     = collink;
   ms.txcol    = coldef;
   ms.width    = width;
   ms.pbs      = &bs;
   ms.ind      = 0;
   ms.scrx     = 0;
   ms.state    = 0;

   while (*src && !err) {
      TextState &cs = ncell ? *ts[ccell] : ms;
      src = get_next(cs, src);
      // "no memory" error should never occur ...
      if (!ncell) 
         if (!commit_buf(cs,0)) break;
      // only unfinished line can be in buf here ...
      if (*src && *src=='<') {
         while (*++src==' ');
         const char *epos = src;
         int        close = 0,
                 ctabline = 0,  // 1 - just line, -1 - close table
                  skipeol = 0;  // ignore end of line next to the end of this tag
         if (*src=='!') { close = 1; src++; }
         switch (*src++) {
            case 'a':
               if (close) err = E_HLP_TAGA; else
                  err = parse_align(cs, src);
               break;
            case 'b':
               if (*src++!='r' || close) err = E_HLP_TAGBR; else {
                  spstr bstr;
                  int    nbr;
                  get_params(src, bstr);
                  if (!bstr) nbr = 1; else {
                     // a bad way to use a string ...
                     const char *num = bstr();
                     err = get_int(num, 0, nbr, rINT_UNS);
                  }
                  if (!err)
                     if (!nbr) err = E_HLP_TAGBR; else add_br(cs, nbr);
               }
               break;
            case 'c':
               if (close || !ncell) err = E_HLP_TAGC; else {
                  cs.state &= ~GET_INLINK;
                  if (++ccell==ncell) ctabline = 1;
               }
               break;
            case 'f':
               if (Xor(close,(cs.state&GET_CODE))) err = E_HLP_TAGF; else {
                  if (close) cs.state &= ~GET_CODE; else {
                     cs.state |= GET_CODE;
                     // allow <f> to be at a separate line
                     skipeol   = 1;
                  }
                  add_br(cs);
               }
               break;
            case 'i': {
               spstr iparm, istr;
               int   indent, wcnt, ii;
               // w/o br!
               complete_line(cs);
               // !i acts as i0
               if (close) { indent = 0; wcnt = 0; } else {
                  get_params(src, iparm, 1);
                  istr = iparm.word(1);
                  wcnt = iparm.words();
                  // a bad way to use a string ...
                  const char *num = istr();
                  err  = get_int(num, cs.width, indent, rINT_UNS|rINT_PER);
               }
               // reset indent state to default
               cs.icol  = cs.txcol;
               cs.state&= ~GET_VLINE;
               if (!err) {
                  if (indent>=width || indent>255 || indent<0) indent = 0;
                  cs.ind = indent;
                  // check other parms
                  for (ii=1; ii<wcnt; ii++) {
                     spstr wp(iparm.word(ii+1));
                     if (wp=="v") {
                        if (indent>0) cs.state|=GET_VLINE;
                     } else
                     if (wp[0]=='s') {
                        wp.del(0).trim();
                        if (wp[0]=='=') wp.del(0);
                        cs.icol = getcolor(style, wp());
                     } else {
                        err = E_HLP_TAGIP;
                        break;
                     }
                  }
               }
               break;
            }
            case 'l':
               if (Xor(close,cs.state&GET_INLINK)) err = E_HLP_TAGL; else
               if (close) cs.state &= ~GET_INLINK; else
               if (!get_params(src, bs.clink) || !bs.clink) err = E_HLP_TAGLN; else {
                  //cs.lcol   = collink;
                  cs.state |= GET_INLINK;
                  bs.lback  = bs.lgroup;
                  bs.lgroup = bs.links.Count();
               }
               break;
            case 'm': {
               spstr mark;
               if (!get_params(src, mark) || !mark) err = E_HLP_TAGM; else
                  marks.AddObject(mark, bs.line + cs.out.Count());
               break;
            }
            case 'n':
               get_params(src, topic, 1);
               break;
            case 's': {
               u8t &dst = cs.state&GET_INLINK ? cs.lcol : cs.txcol;
               if (close) dst = cs.state&GET_INLINK ? collink : coldef; else {
                  spstr sc;
                  if (!get_params(src, sc) || !sc) err = E_HLP_TAGS; else
                     dst = getcolor(style, sc());
               }
               break;
            }
            case 't':
               if (Xor(close,ncell)) err = E_HLP_TAGT; else {
                  cs.state &= ~GET_INLINK;
                  if (close) ctabline = -1; else {
                     add_br(cs);
                     // commit and remember starting line
                     if (!commit_buf(cs,0)) break;
                     tsline = bs.line;

                     spstr   cw;
                     get_params(src, cw, 1);
                     u32t  cols = cw.words(","), ci;
                     // random # of cells with constant width
                     if (cw[0]=='w') {
                        if (cols>1) err = E_HLP_TAGTC; else {
                           // a bad way to use a string ...
                           const char *num = cw() + 1;
                           int         wdt;
                           err = get_int(num, ms.width, wdt, rINT_UNS|rINT_PER);
                           if (!err)
                              if (!wdt) err = E_HLP_TAGTW; else {
                                 cols = (ms.width - ms.ind) / wdt;
                                 if (cols<2) err = E_HLP_TAGTW; else {
                                    if (cols>LCELLS) cols = LCELLS;
                                    for (ci=0; ci<cols; ci++) cellw[ci] = wdt;
                                 }
                              }
                        }
                     } else { // normal table
                        if (cols<2 || cols>=LCELLS) { err = E_HLP_TAGTC; break; }

                        for (ci=0; ci<cols; ci++) {
                           spstr ccw = cw.word(ci+1,",").trim();
                           if (!ccw) cellw[ci] = 0; else {
                              // a bad way to use a string ...
                              const char *num = ccw();
                              int         wdt;
                              err = get_int(num, ms.width, wdt, rINT_UNS|rINT_PER);
                              if (err) break;
                              cellw[ci] = wdt;
                           }
                        }
                     }
                     if (err) break;

                     if (cols<2) err = E_HLP_TAGTC; else {
                        u32t sum = 0, size = 0, ci;

                        for (ci = 0; ci<cols; ci++)
                           if (!cellw[ci]) size++; else sum += cellw[ci];

                        if (size) {
                           size = sum >= ms.width-ms.ind ? 1 :
                                  (ms.width-sum-ms.ind) / size;
                           if (size<1) size = 1;
                           // set for all w/o size
                           for (ci = 0; ci<cols; ci++)
                              if (!cellw[ci]) cellw[ci] = size;
                        }
                        ncell = cols;
                        ccell = 0;

                        for (ci=0, sum=ms.ind; ci<ncell; ci++) {
                           if (!ts[ci]) ts[ci] = new TextState; else {
                              ts[ci]->out.Clear();
                              ts[ci]->att.Clear();
                           }
                           ts[ci]->state   = ms.state & GET_AMASK;
                           ts[ci]->lcol    = ms.lcol;
                           ts[ci]->txcol   = ms.txcol;
                           ts[ci]->ind     = 0;
                           ts[ci]->pbs     = &bs;
                           ts[ci]->width   = cellw[ci];
                           ts[ci]->scrx    = sum;
                           sum += cellw[ci];
                        }
                        tabw  = sum;
                     }
                  }
               }
               break;
            case 'x': {
               int cv;
               err = get_int(src, 0, cv, rINT_HEX);
               if (!err) {
                  char str[2];
                  str[0] = (u8t)cv;
                  str[1] = 0;
                  get_next(cs, str, 1);
               }
               break;
            }
            default:
               err = E_HLP_TAGUNKNOWN;
         }
         while (*src && *src!='>') src++;
         if (*src=='>') src++;
         if (skipeol)
            while (*src && (*src=='\r' || *src=='\n'))
               if (*src++=='\n') break;
         
         // commit table line
         if (ctabline) {
            u32t  maxline = 0;
            // first complete all lines and calc max line number
            for (ii=0; ii<ncell; ii++) {
               TextState &s = *ts[ii];
               complete_line(s);
               if (s.out.Count() > maxline) maxline = s.out.Count();
            }
            u32t xpos = ms.ind;
            // draw indent or right border if table width < screen width
            if (xpos || tabw<bs.scrwidth) {
               add_br(ms, maxline);
               complete_line(ms);
               if (!commit_buf(ms,0)) break;
               bs.line = tsline;
            }
            // add <br>s up to maxline and commit
            for (ii=0; ii<ncell; ii++) {
               TextState &s = *ts[ii];
               if (s.out.Count() < maxline) {
                  add_br(s, maxline - s.out.Count());
                  complete_line(s);
               }
#if 0
               static int row = 0;
               s.out.SaveToFile(spstr().sprintf("cell[%u,%u].txt",row,ii)());
               if (ii==ncell-1) row++;
#endif
               bs.line = tsline;
               if (!commit_buf(s,xpos)) break;
               xpos   += cellw[ii];
            }
            bs.line = tsline += maxline;
            // table end?
            if (ctabline<0) {
               FreeCells();
               ncell = 0;
            } else {
               // clear buffers for the next line 
               for (ii=0; ii<ncell; ii++) {
                  ts[ii]->out.Clear();
                  ts[ii]->att.Clear();
                  ts[ii]->state   = ms.state & GET_AMASK;
                  ts[ii]->lcol    = ms.lcol;
                  ts[ii]->txcol   = ms.txcol;
                  ts[ii]->ind     = 0;
               }
               ccell = 0;
            }
         }
         if (err) log_printf("wrong tag %-10s (%X)\n", epos, err);
      }
   }
   FreeCells();

   if (err) {
      if (bs.res) free(bs.res);
      if (perr) *perr = err;
      return 0;
   } else {
      add_br(ms);
      commit_buf(ms,0);
   }

   if (bs.links.Count() > 0) {
      qsh_link *l0 = (qsh_link*)&bs.links.Objects(0);
      // make circular link for the first entry
      if (bs.lgroup>0) l0->back = bs.lgroup;

      if (bs.lcommit<bs.links.Count())
         log_printf("failed to commit %i link(s)\n", bs.links.Count()-bs.lcommit);
   }

   char      tmp[2];
   u32t  marklen = 1,
         linklen = 1;
   // calc list sizes
   str_unroll(marks.Str, 0, 0, &tmp, marklen, 0);
   str_unroll(bs.links.Str, 0, 0, &tmp, linklen, 0);

   u32t addmem = bs.links.Count() * sizeof(qsh_link) +   // link refs
                 marks.Count() * sizeof(u16t) +          // mark line numbers
                 marklen +                               // mark strings
                 linklen +                               // link strings
                 marks.Count() * sizeof(char*) +         // marks string array
                 bs.links.Count() * sizeof(char*) +      // links string array
                 topic.length() + 1;                     // topic name

   qsh_topic *res = (qsh_topic*)realloc(bs.res, bs.lines *
                    bs.scrwidth * 2 + sizeof(qsh_topic) + addmem);
   if (!res) {
      free(bs.res);
      if (perr) *perr = E_SYS_NOMEM;
      return 0;
   }
   /* maybe a bit stupid but returning all data in a single heap block is very
      easy to handle on user side. Since this is an API - it SHOULD be EASY */
   res->text   = (u16t*)(res + 1);
   res->nlinks = bs.links.Count();
   res->nmarks = marks.Count();
   res->width  = width;
   res->lines  = bs.line;
   res->selcol = colsel;
   res->reserv = 0;

   u8t *lp     = (u8t*)(res->text + bs.lines * bs.scrwidth);
   // build marks ...
   if (res->nmarks) {
      res->mark  = (char**)lp;
      res->mlist = (u16t*)(res->mark + marks.Count());
      lp         = (u8t*)(res->mlist + marks.Count());
      // line numbers
      for (u32t mi=0; mi<res->nmarks; mi++) res->mlist[mi] = marks.Objects(mi);
      // both unrolls should return the same result ...
      str_unroll(marks.Str, 0, 0, lp, marklen, res->mark);
      lp += marklen;
   } else {
      res->mark  = 0;
      res->mlist = 0;
   }
   // build links ...
   if (res->nlinks) {
      res->link  = (char**)lp;
      res->llist = (qsh_link*)(res->link + bs.links.Count());
      lp         = (u8t*)(res->llist + bs.links.Count());
      // link refs
      memcpy(res->llist, bs.links.LList.Value(), bs.links.Count() * sizeof(qsh_link));
      // both unrolls should return the same result ...
      str_unroll(bs.links.Str, 0, 0, lp, linklen, res->link);
      lp += linklen;
   } else {
      res->link  = 0;
      res->llist = 0;
   }
   res->topic = (char*)lp;
   if (!topic) *lp = 0; else strcpy((char*)lp, topic());
#ifdef __QSINIT__
   mem_localblock(res);
#endif
   if (perr) *perr = 0;
   return res;
}

qsh_topic* _std help_draw(qsh_file hf, u32t scheme_id, u32t topic_id,
                          u32t width, qserr *err)
{
   if (!hf || !topic_id) {
      if (err) *err = E_SYS_INVPARM;
      return 0;
   }
   TStrings  lst;
   spstr   topic;
   { // lock it only to read the topic
      HelpCritSec lk;
      if (err) *err = E_SYS_INVOBJECT;

      instance_ret(huser, ref, 0);
      hdata *fe = ref->fp;
      if (fe->sign!=HDATA_SIGN) return 0;

      if (fe->minwidth && fe->minwidth > width) {
         if (err) *err = E_HLP_MINWIDTH;
         return 0;
      }
      if (--topic_id >= fe->topics.Count()) {
         if (err) *err = E_HLP_TOPICID;
         return 0;
      }
      topic = fe->topics[topic_id];
      fe->src->ReadSection(topic, lst);
      // trim spaces at the end of all lines
      for (u32t ii=0; ii<lst.Count(); ii++) lst[ii].trimright();
   }
   char *text = lst.GetText("\x0A");
#ifdef __QSINIT__
   mem_threadblock(text);
#endif
   lst.Clear();
   qsh_topic *rc = help_drawtext(hf, scheme_id, text, topic(), width, err);
   free(text);
   return rc;
}

u32t _std help_topicid(qsh_file hf, const char *topic) {
   if (hf) {
      HelpCritSec lk;
      instance_ret(huser, ref, 0);
      hdata *fe = ref->fp;
      if (fe->sign!=HDATA_SIGN) return 0;
      // quick way to ask for an index topic
      if (!topic)
         return fe->index<0 ? 0 : fe->index + 1;

      int index = fe->topics.IndexOfICase(topic);
      if (index<0) {
         index = fe->alias.IndexOfICase(topic);
         if (index>=0) index = fe->alias.Objects(index);
      }
      if (index>=0) return index + 1;
   }
   return 0;
}

s32t _std help_schemeid(qsh_file hf, const char *scheme) {
   if (hf && scheme) {
      HelpCritSec lk;
      instance_ret(huser, ref, 0);
      hdata *fe = ref->fp;
      if (fe->sign!=HDATA_SIGN) return -1;
      return fe->schemes.IndexOfICase(scheme);
   }
   return -1;
}

// search for a link in the current visible area
int _std help_findlink(qsh_topic *td, int prev, u32t dir, u32t ypos, u32t ysize) {
   if (!td || !td->nlinks) return -1;
   if (dir>HLINK_ANY && prev<0) dir = HLINK_ANY;
   if (dir==HLINK_ANY) prev = -1;
   // prev can be outside of th visible area!
   u32t ii;
   int  skip = 0, s_x1=-1, s_y1=-1, s_x2=-1, s_y2=-1,
       group = -1,
         res = -1,
        diff = 0;
   // get current link rect
   if (prev>=0)
      for (ii=prev; ii<td->nlinks; ii++) {
         qsh_link &lnk = td->llist[ii];
         // next group?
         if (ii!=prev && lnk.back!=0xFFFF) break;
#if 0
         log_printf("%i,%i - %i,%i <-%u,%u - %u\n", s_x1, s_y1, s_x2, s_y2,
            lnk.xpos, lnk.ypos, lnk.len);
#endif         
         if (s_y1<0 || s_y1>lnk.ypos) s_y1 = lnk.ypos;
         if (s_y2<0 || s_y2<lnk.ypos) s_y2 = lnk.ypos;
         // not that x1 can be > x2 in multiline link!
         if (s_x1<0 || s_x1>lnk.xpos) s_x1 = lnk.xpos;
         if (s_x2<0 || s_x2<lnk.xpos+lnk.len-1) s_x2 = lnk.xpos+lnk.len-1;
      }
//   log_printf("%i,%i - %i,%i\n", s_x1, s_y1, s_x2, s_y2);

   for (ii=0; ii<td->nlinks; ii++) {
      qsh_link &lnk = td->llist[ii];
      // start of the group
      if (!ii || lnk.back!=0xFFFF) {
         group = ii;
         skip  = ii==prev;
      }
      if (!skip) {
         u32t yy = lnk.ypos;
         if (yy<ypos || yy>=ypos+ysize) continue;

         if (dir==HLINK_ANY) {
            res = group;
            break;
         } else {
            int tdx1 = lnk.xpos + lnk.len - 1 - s_x1,
                tdx2 = lnk.xpos - s_x2,
                tdxv = abs(lnk.xpos - s_x1),
                tdy1 = yy - s_y1, tdy1a = abs(tdy1),
                tdy2 = yy - s_y2, tdy2a = abs(tdy2),
               ndiff = 0;
#if 0
            log_printf("%u..%u,%u dx=%i,%i, dy=%i,%i\n", lnk.xpos,
               lnk.xpos + lnk.len - 1, yy, tdx1, tdx2, tdy1, tdy2);
#endif
            switch (dir) {
               case HLINK_UP   :
                  if (tdy1<0) ndiff = -tdy1 + tdxv * OTHER_AXIS_W;
                  break;
               case HLINK_DOWN :
                  if (tdy2>0) ndiff = tdy2 + tdxv * OTHER_AXIS_W;
                  break;
               case HLINK_LEFT :
                  if (tdx1<=0) ndiff = -tdx1 + min(tdy1a,tdy2a) * OTHER_AXIS_W;
                  break;
               case HLINK_RIGHT:
                  if (tdx2>=0) ndiff = tdx2 + min(tdy1a,tdy2a) * OTHER_AXIS_W;
                  break;
            }
            if (ndiff && (res<0 || diff>ndiff)) {
               diff = ndiff;
               res  = group;
               skip = 1;
            }
         }
      }
   }
   return res;
}
