//
// QSINIT help module
// system functions
//
#include "stdlib.h"
#include "qsbase.h"
#include "errno.h"
#include "helploc.h"
#include "qs_rt.h"
#include "vioext.h"
extern "C" {
#include "../../init/zip/unzip.h"
// function is not published in vio.h
void _std vio_getmodefast(u32t *cols, u32t *lines);
}
#ifdef __WATCOMC__
#pragma disable_message (919);
#endif

#define TOP_B          1
#define SIDE_B         1
#define BOTTOM_B       2
#define FILL_C         (VIO_COLOR_BLUE<<4|VIO_COLOR_LWHITE)
#define MINSIZEX      24   // window cannot be smaller
#define MINSIZEY       8

static const char *Help_Topic  = "HELP";
static const char *Scheme_Name = "WINDOW";

struct histentry {
   u32t      hfindex;
   u32t      topicid;
   u32t         line;
   s32t      sellink;
};

static void draw_scrollbar(u32t x, u32 y, u32t size, u32t vpos, u32t vlen) {
   if (!size) return;
   if (vpos > size-1) vpos = size-1;
   if (!vlen) vlen = 1;
   if (vpos + vlen > size) vlen = size - vpos;
   // use vio_writebuf() to prevent random scrolling effects
   u16t *buf = (u16t*)malloc_th(size << 1);

   for (u32t py=0; py<size; py++)
      buf[py] = FILL_C<<8 | (py<vpos || py>=vpos+vlen ? 0xB0 : 0xB2);

   vio_writebuf(x, y, 1, size, buf, 2);
   free(buf);
}

static int draw_link(qsh_topic *td, int link, int sel,
                     u32t x, u32t y, u32t ypos, u32t ysize)
{
   u16t *lb = sel ? (u16t*)malloc_th(td->width*2) : 0;
   int lidx = link,
        vis = 0;

   while (lidx>=0 && lidx<td->nlinks) {
      qsh_link &lnk = td->llist[lidx];
      // next group?
      if (lidx!=link && lnk.back!=0xFFFF) break;
      lidx++;
      if (lnk.ypos<ypos || lnk.ypos>=ypos+ysize) continue;

      u16t *bptr = td->text + lnk.ypos*td->width + lnk.xpos;
      if (sel) {
         for (u32t ii=0; ii<min(lnk.len,td->width); ii++)
            lb[ii] = bptr[ii] & 0xFF | td->selcol<<8;
         bptr = lb;
      }
      vio_writebuf(x+2+lnk.xpos, y+1+lnk.ypos-ypos, lnk.len, 1, bptr, 0);
      vis++;
   }
   if (lb) free(lb);
   // is at least part visible?
   return vis;
}

static int help_findload(spstr &link, TPtrStrings &openlist, qserr *err=0) {
   if (link[0]=='?') {
      int   wp2 = link.wordpos(2,"/");
      spstr  ft;
      if (wp2<0) {
         ft = link;
         link.clear();
      } else {
         ft = link.deltostr(0,wp2);
         // delete '/'
         ft.dellast();
      }
      if (err) *err = 0;
      ft.del(0).upper();
      for (u32t ii=0; ii<openlist.Count(); ii++)
         if (openlist[ii]==ft) return ii;

      spstr    fn(ft);
      qsh_file hf;
      fn += ".QSH";

      qserr res = help_open(fn(), &hf);
      if (err) *err = res;

      if (!res) {
         log_it(3, "open %s = %X\n", ft(), hf);
         return openlist.AddObject(ft, hf);
      }
   } else
   if (err) *err = E_HLP_INVLINK;
   return -1;
}

qserr _std help_show(const void *name, const char *topicname, const char *mark,
                     int x, int y, u32t width, u32t height, u32t flags)
{
   u32t  mx, my;
   qsh_file  hf = 0;
   qserr     rc = 0;
   int    quiet = flags & HSFL_QUIET;
   if (!name) return E_SYS_ZEROPTR;

   vio_getmodefast(&mx, &my);

   if (!width || !height) {
      spstr hstr(getenv("HELPSIZE")), ws;
      if (hstr.length() && hstr.words(",")==2) {
         const char *sptr;
         int         resv;
         if (!width) {
            ws   = hstr.word(1,",").trim();
            sptr = ws();
            if (!get_int(sptr, mx, resv, rINT_PER|rINT_UNS))
               width = resv; 
         }
         if (!height) {
            ws   = hstr.word(2,",").trim();
            sptr = ws();
            if (!get_int(sptr, my, resv, rINT_PER|rINT_UNS)) height = resv;
         }
      }
   }
   // fix sizes
   if (width<MINSIZEX || width > mx-SIDE_B*2) width = mx-SIDE_B*2;
   if (height<MINSIZEY || height > my-TOP_B-BOTTOM_B) height = my-TOP_B-BOTTOM_B;

   if (x<0) x = mx - width >> 1; else
      if (x+width > mx-1) x = mx - 1 - width;
   if (y<0) y = my - height >> 1; else
      if (y+height > my-1) y = my - 1 - height;
   // file short name in links
   spstr        filetag;
   spstr       gotomark(mark);
   TPtrStrings openlist;
   u32t         topicid = 0;
   /* open a file
      HSFL_HANDLE file is not added to the list because of unknown name */
   if (flags & HSFL_HANDLE) hf = (qsh_file)name; else {
      if (*(const char*)name=='?') {
         spstr lnk((const char*)name);
         int idx = help_findload(lnk, openlist, &rc);

         if (idx>=0) {
            hf = openlist.Objects(idx);
            // it should be 0 in such call
            topicname = 0;
            if (lnk.length()) {
               topicid  = help_topicid(hf, lnk.word(1,"/")());
               gotomark = lnk.word(2,"/");
            }
         } else
         if (!rc) rc = E_SYS_SOFTFAULT;
      } else {
         _splitpath((const char*)name, 0, 0, filetag.LockPtr(QS_MAXPATH+1), 0);
         filetag.UnlockPtr();
         filetag.upper();
         rc = help_open((const char*)name, &hf);
         openlist.AddObject(filetag, hf);
      }
   }

   if (rc && quiet) return rc;


   spstr errpfx;

   if (rc) {
      errpfx  = "Error opening \"";
      errpfx += (const char*)name;
      errpfx += "\":^";
   } else {
      u32t  curtopic = 0,
             curfile = 0;

      if (!topicid) topicid = help_topicid(hf, topicname?topicname:0);
      // go to index if topic was not found
      if (!topicid && (flags & HSFL_DEFINDEX) && topicname)
         topicid = help_topicid(hf, 0);

      if (!topicid) rc = E_HLP_NOTOPIC; else {
         u32t  cx, cy;
         u16t   shape = vio_getshape(),
                color = vio_getcolor();
         vio_getpos(&cy, &cx);
         vio_setshape(VIO_SHAPE_NONE);

         void        *buf = malloc_th((width+1) * (height+1) * 2);
         qsh_topic    *td = 0;
         TPtrList history;

         int          upd = 0,
                   goback = 0,
                 gotolink = -1,
                 gotofile = -1,
                     link = -1;
         u32t        vhgt = height - 2,  // visible text height
                  gotopos = 0,
                      pos,
                  cvlines;
         char     *outprn = 0;
         int           cs = help_schemeid(hf, Scheme_Name);

         vio_readbuf(x, y, width+1, height+1, buf, 0);
         vio_drawshadow(x, y, width, height);

         // main help loop
         while (1) {
            if (topicid /*&& curtopic!=topicid*/ ||
                gotofile>=0 && curfile!=gotofile)
            {
               if (td) free(td);
#if 0
               log_printf("goto file %i topic %u line %u link %i\n", gotofile,
                  topicid, gotopos, gotolink);
#endif
               // switch to another file
               if (gotofile>=0 && curfile!=gotofile) {
                  hf = openlist.Objects(gotofile);
                  cs = help_schemeid(hf, Scheme_Name);
               }
               // parse a new topic!
               td = help_draw(hf, cs<0?0:cs, topicid, width-4, &rc);
               // update border in any case
               vio_drawborder(x, y, width, height, FILL_C);
               vio_fillrect(x+1, y+1, 1, height-2, ' ', FILL_C);
               vio_fillrect(x+width-2, y+1, 1, height-2, ' ', FILL_C);
               if (!td) break; else {
                  if (!goback && curtopic) {
                     histentry *he = new histentry;
                     he->hfindex = curfile;
                     he->topicid = curtopic;
                     he->line    = pos;
                     he->sellink = link;
                     history.Add(he);
                  }
                  if (gotofile>=0) curfile = gotofile;
                  curtopic = topicid;
                  // go back or goto mark?
                  if (!gotomark) {
                     link = gotolink<(int)td->nlinks ? gotolink : -1;
                     pos  = gotopos;
                  } else {
                     link = -1;
                     pos  =  0;
                     for (u32t mm=0; mm<td->nmarks; mm++) {
                        if (!stricmp(gotomark(),td->mark[mm])) {
                           pos = td->mlist[mm];
                           break;
                        }
                     }
                     //log_printf("mark \"%s\", line %u\n", gotomark(), pos);
                  }
                  // verify/fix external seek in this topic
                  if (pos>0)
                     if (pos>=td->lines) pos = 0; else
                        if (pos+vhgt > td->lines)
                           pos = td->lines<vhgt ? 0 : td->lines - vhgt;

                  upd  = 1;
                  if (td->topic) {
                     spstr ts(td->topic);
                     // cut too long topic name
                     if (ts.length() > width>>1) {
                        ts.del(width>>1, x7FFF);
                        ts += "...";
                     }
                     if (ts.length()) {
                        vio_setcolor(FILL_C);
                        vio_setpos(y, x+width-7-ts.length());
                        vio_charout('[');
                        vio_charout(' ');
                        vio_strout(ts());
                        vio_charout(' ');
                        vio_charout(']');
                     }
                  }
               }
            }
            // reset in any case
            topicid  = 0;
            goback   = 0;
            gotopos  = 0;
            gotolink = -1;
            gotofile = -1;
            gotomark.clear();

            if (upd && td) {
               u32t top = pos + vhgt, diff;
               cvlines  = top > td->lines ? td->lines - pos : vhgt,
               diff     = vhgt - cvlines;
               // fill empty bottom
               if (diff)
                  vio_fillrect(x+1, y+1+cvlines, width-2, diff, ' ', FILL_C);
               // draw text
               vio_writebuf(x+2, y+1, width-4, cvlines, td->text + pos*td->width, 0);
               // draw a scrollbar (use fixed point calculations)
               if (vhgt < td->lines) {
                  u32t  vpos = (pos*vhgt<<8)/td->lines,
                       vsize = (vhgt*vhgt<<8)/td->lines;
                  if ((vpos & 255) > 127) vpos += 256;
                  if ((vsize & 255) > 127) vsize += 256;
                  draw_scrollbar(x+width-1, y+1, vhgt, vpos>>8, vsize>>8);
               }
               // check that currect link is still on the screen
               if (link>=0)
                  if (!draw_link(td, link, 1, x, y, pos, vhgt)) link = -1;
               // else select the first found
               if (link<0) {
                  int nlink = help_findlink(td, -1, HLINK_ANY, pos, vhgt);
                  if (nlink>=0)
                     draw_link(td, link=nlink, 1, x, y, pos, vhgt);
               }
               upd = 0;
            }
            u16t key = key_read();
            u8t keyh = key>>8,
                keyl = key&0xFF;
            if (keyl==0x1B) break; // Esc
            if (keyh==0x44 && !(flags & HSFL_NOF10)) { // F10
               if (td) free(td);
               td = help_draw(hf, 0, curtopic, width - 4, &rc);
               if (td) outprn = vio_buftoansi(td->text+pos*td->width, td->width,
                  td->width, cvlines, 0);
               break;
            }

            if (link>=0) {
               if (keyl=='\r' || keyl==' ') {
                  spstr lnk(td->link[link]);
                  if (lnk[0]=='?') gotofile = help_findload(lnk, openlist);
                  // search for a topic
                  topicid = help_topicid(gotofile>=0 ?
                                         openlist.Objects(gotofile) : hf,
                                         !lnk ? 0 : lnk.word(1,"/")());
                  if (topicid) gotomark = lnk.word(2,"/");
                  continue;
               }
               u32t linkdst = 0;
               switch (keyh) {
                  case 0x48: linkdst = HLINK_UP; break;
                  case 0x4B: linkdst = HLINK_LEFT; break;
                  case 0x4D: linkdst = HLINK_RIGHT; break;
                  case 0x50: linkdst = HLINK_DOWN; break;
               }
               if (linkdst) {
                  int nlink = help_findlink(td, link, linkdst, pos, vhgt);
                  if (nlink>=0) {
                     draw_link(td,  link, 0, x, y, pos, vhgt);
                     draw_link(td, nlink, 1, x, y, pos, vhgt);
                     link = nlink;
                     continue;
                  }
               }
            }

            switch (keyh) {
               case 0x0F:
                  if (td->nlinks>0) {
                     u32t dst = 0;
                     if (keyl) { // Tab
                        if (link>=0) {
                           for (dst = link; dst<td->nlinks; dst++)
                              if (dst!=link && td->llist[dst].back!=0xFFFF)
                                 break;
                           if (dst==td->nlinks) dst = 0;
                        }
                        // can found self, actually 
                        if (dst==link) dst = 0xFFFF;
                     } else { // Shift-Tab
                        dst = td->llist[link<0?0:link].back;
                        // a single invisible link
                        if (dst==0xFFFF && link<0) dst = 0;
                     }
                     if (dst!=0xFFFF) {
                        int y1 = 0xFFFF, y2 = -1;
                        for (int lidx = dst; lidx<td->nlinks; lidx++) {
                           qsh_link &lnk = td->llist[lidx];
                           // next group
                           if (lidx!=dst && lnk.back!=0xFFFF) break;
                           if (lnk.ypos<y1) y1 = lnk.ypos;
                           if (lnk.ypos>y2) y2 = lnk.ypos;
                        }
                        u32t newy = pos;
                        if (y2 >= pos + vhgt) newy = y2 - vhgt + 1;
                        if (y1 <= pos) newy = y1 > 1 ? y1 - 1 : 0;
                        // update screen or links only
                        if (newy!=pos) { upd = 1; pos = newy; } else {
                           draw_link(td, link, 0, x, y, pos, vhgt);
                           draw_link(td,  dst, 1, x, y, pos, vhgt);
                        }
                        link = dst;
                     }
                  }
                  break;
               case 0x68: // Alt-F1
               case 0x0E: // Backspace - topic history
                  if (history.Count()>0) {
                     histentry *he = (histentry*)history[history.Max()];;
                     gotofile = he->hfindex;
                     topicid  = he->topicid;
                     gotolink = he->sellink;
                     gotopos  = he->line;
                     delete he;
                     history.Delete(history.Max());
                     goback   = 1;
                  }
                  break;
               case 0x3B: // F1 - help on help
                  topicid = help_topicid(hf, Help_Topic);
                  break;
               case 0x49: // PgUp
                  if (pos>vhgt) { upd = 1; pos-=vhgt; break; }
               case 0x47: // Home
                  if (pos) { upd = 1; pos = 0; }
                  break;
               case 0x48: // Up
                  if (pos) { upd = 1; pos--; }
                  break;
               case 0x51: // PgDn
                  if (pos + vhgt*2 < td->lines) { upd = 1; pos+=vhgt; break; }
               case 0x4F: // End
                  if (td->lines>vhgt) { upd = 1; pos = td->lines - vhgt; }
                  break;
               case 0x50: // Down
                  if (td->lines>vhgt && pos < td->lines-vhgt)
                     { upd = 1; pos++; }
                  break;
               case 0x54: // Shift-F1
                  topicid = help_topicid(hf, 0);
                  break;
            }
         }
         FreeListItems(history);
         if (td) free(td);
         // restore the original vio settings & screen
         vio_setpos(cy, cx);
         vio_writebuf(x, y, width+1, height+1, buf, 0);
         free(buf);
         if (outprn) { // F10 output
            puts(outprn);
            free(outprn);
         }
         vio_setshape(shape);
         vio_setcolor(color);
      }
      if (!(flags&HSFL_HANDLE)) help_close(hf);
   }
   // free all help files (handle in "name" parameter is not included here)
   for (u32t ii=0; ii<openlist.Count(); ii++) help_close(openlist.Objects(ii));

   if (rc && !quiet) {
      char *et = cmd_shellerrmsg(EMSG_QS, rc);
      errpfx += et;
      free(et);
      vio_msgbox("Help", errpfx(), MSG_OK|MSG_WIDE|MSG_BLUE, 0);
   }
   return rc;
}

u32t _std shl_help(const char *cmd, str_list *args) {
   int     tty = 0;
   TStrings al;
   str_getstrs(args,al);

   if (al.Count()) {
      if (al[0].upper()=="/RELEASE" || al[0]=="/REL") {
         help_release();
         return 0;
      }
      int idx = al.IndexOfICase("/c");
      if (idx>=0) { tty = 1; al.Delete(idx); }
      idx = al.IndexOf("/?");
      if (idx>=0) { tty = 1; al[idx] = "HELP"; }
   }
   al.TrimEmptyLines();
   if (tty) {
      spstr topic;
      if (!args->count || al.IndexOf("/?")>=0) topic = Help_Topic;
         else topic = al[0];
      cmd_shellhelp(topic(), CLR_HELP);
   } else
      help_show(SYS_HELP_FILE, al.Count()?al[0]():0, 0, -1, -1, 0, 0,
                HSFL_DEFINDEX);
   return 0;
}

char* _std help_drawansi(qsh_file hf, u32t scheme_id, u32t topic_id,
                         u32t xpos, u32t width, qserr *err)
{
   qsh_topic *res = help_draw(hf, scheme_id, topic_id, width, err);
   if (res) {
      char *rc = vio_buftoansi(res->text, res->width, res->width, res->lines, xpos);
      free(res);
      return rc;
   }
   return 0;
}

qserr help_unpack(const char *path, TStrings &text) {
   ZIP   hzip;
   u32t  size;

   text.Clear();
   qserr   rc = zip_openfile(&hzip, path, 0);
   if (rc) return rc;
   rc = zip_nextfile(&hzip, 0, &size);
   // limit help zip size
   if (!rc && size>_16MB) rc = E_SYS_NOMEM;
   if (!rc && size) {
      void *buf = 0;
      rc = zip_readfile(&hzip, &buf);
      if (!rc && buf) {
         text.SetText((char*)buf, size);
         hlp_memfree(buf);
      }
   }
   zip_close(&hzip);
   return rc;

}

static qserr _exicc hcc_open(EXI_DATA, const char *name, qsh_file *hf) {
   return help_open(name, hf);
}

static qserr _exicc hcc_close(EXI_DATA, qsh_file hf) {
   return help_close(hf);
}

static void _exicc hcc_release(EXI_DATA) { return help_release(); }

static qsh_info* _exicc hcc_info(EXI_DATA, qsh_file hf) {
   return help_info(hf);
}

static u32t _exicc hcc_topicid(EXI_DATA, qsh_file hf, const char *topic) {
   return help_topicid(hf, topic);
}

static s32t _exicc hcc_schemeid(EXI_DATA, qsh_file hf, const char *scheme) {
   return help_schemeid(hf, scheme);
}

static qsh_topic* _exicc hcc_draw(EXI_DATA, qsh_file hf, u32t scheme,
                                  u32t topic, u32t width, qserr *err)
{
   return help_draw(hf, scheme, topic, width, err);
}

static char* _exicc hcc_drawansi(EXI_DATA, qsh_file hf, u32t scheme, u32t topic,
                                 u32t xpos, u32t width, qserr *err)
{
   return help_drawansi(hf, scheme, topic, xpos, width, err);
}

static qserr _exicc hcc_show(EXI_DATA, const void *name, const char *topic,
   const char *mark, int x, int y, u32t width, u32t height, u32t flags)
{
   return help_show(name, topic, mark, x, y, width, height, flags);
}

static void *qs_hcc_list[] = { hcc_open, hcc_close, hcc_release, hcc_info,
                               hcc_topicid, hcc_schemeid, hcc_draw,
                               hcc_drawansi, hcc_show };

extern "C"
int register_class(void) {
   return exi_register("qs_helpctrl", qs_hcc_list,
                       sizeof(qs_hcc_list)/sizeof(void*), 0, 0, 0, 0, 0);
}
