//
// QSINIT
// JFS boot code (micro-FSD)
// ----------------------------------------
// * note that 512 bytes sector is forced and hardcoded
// * memory model is tiny (cs=ds=ss)
//
#include "sys/fs/jfs.h"
#include "filetab.h"
#include "jsegdef.h"

#define PRINT_FNAME         0   ///< debug output level

extern
struct filetable_s       ftab;
extern u8t          code_nsec;
static u16t              mtop,  ///< top of available memory in seg
                         mbuf,  ///< start of available memory in seg
                     mbufmark;  ///< start of current file extent list
extern u8t          mfsd_bits;
extern char          bfname[];

#define PGSHIFT         (12-9)  ///< page -> sector shift
static u8t            clshift;  ///< cluster -> sector shift

#define STACK_SIZE    (6*_1KB)
#define PHYS2FP(x)    ((void far*)MAKEFAR16((x)>>PARASHIFT,(x)&0xF))
#define FP2PHYS(x)    (((u32t)SELECTOROF(x)<<PARASHIFT) + OFFSETOF(x))
#define MAXNAMELEN        127   ///< limit for the file name length

/* memory layout:
   SELF_SEG_LEN = 60k
   seg:0                  micro / bpb
   seg:code_nsec*512      inode buffer                   512
   seg:(code_nsec+1)*512  root dir inode                 512
   seg:(code_nsec+2)*512  sector for unaligned read      512
   seg:(code_nsec+3)*512  jfs_xtreepage / jfs_dtpage      4k
   seg:mbuf               extent list for the inode file
                          extent list for the current file
   seg:SELF_SEG_LEN-6k    stack bottom  (mtop)
   seg:SELF_SEG_LEN       stack top
 */

static jfs_xtreepage     *xtp;
static jfs_xad           *xad;
static jfs_dinode        *ino,
                       *iroot;
static u32t           x_index,
                  x_lastindex;
static u32t             nclus;
static u8t              *usec;

typedef struct {
   u32t                 start;  ///< first cluster in chain
   u32t                   len;  ///< number of clusters
} cllist;

typedef struct {
   u32t                  size;  ///< file size
   u32t                   ncl;  ///< number of clusters in the file
   u16t                extnum;  ///< # of extents
   cllist              ext[1];
} file;

static file             *inof,  ///< inode file record
                        *curf = 0;  ///< current file

/// read up to 127 sectors (64k - 512)
u8t volread(u32t sector, u8t count, void far *buf);
#pragma aux volread      \
   value [al]            \
   parm nomemory [dx ax] [cx] [di si] \
   modify exact [ax cx dx es di si];

u16t get_ds(void);
#pragma aux get_ds = "mov ax,ds" value [ax] modify exact [ax];

void _stdcall _fmemcpy(void far*, const void far*, u16t);
void _stdcall panic_nofile(void);
void _stdcall panic(char *str);
#pragma aux panic parm [si] aborts;

/* fmt see in \system\src\init\misc\prn16.asm
   linked actually only if PRINT_FNAME above >0, in default mode linker
   optimization just removes it */
void __cdecl printf16(char far *fmt, ...);

#if PRINT_FNAME>1
#define log_it printf16
#define log_m  printf16
#elif PRINT_FNAME>0
#define log_it printf16
#define log_m(...)
#else
#define log_it(...)
#define log_m(...)
#endif

/// get first extent
static jfs_xad* first_extent(jfs_dinode *ino) {
   jfs_xtreepage *etp = (jfs_xtreepage *)&ino->xtroot;

   x_index = XTENTRYSTART;
   xad     = &etp->xad[XTENTRYSTART];

   if (etp->header.flag&BT_LEAF) x_lastindex = etp->header.nextindex; else {
      do {
         // broken allocation data
         if (xad->addr1&0xFF000000) return 0;
         if (!volread(xad->addr2<<clshift, 1<<PGSHIFT, xtp)) return 0;

         xad = &xtp->xad[XTENTRYSTART];
      } while ((xtp->header.flag&BT_LEAF)==0);
      x_lastindex = xtp->header.nextindex;
   }
   return xad;
}

static jfs_xad* next_extent(void) {
   if (++x_index < x_lastindex) return ++xad;
   if (xtp->header.next) {
      if (!volread((u32t)xtp->header.next<<clshift, 1<<PGSHIFT, xtp)) return 0;
      x_lastindex  = xtp->header.nextindex;
      x_index      = XTENTRYSTART;
      return xad = &xtp->xad[XTENTRYSTART];
   } else
      return 0;
}

/** opens a file.
    used by file open and for inode file itself */
static file* open_file(jfs_dinode *ino) {
   jfs_xad   *xad;
   file      *loc = (file*)mbuf;
   u16t   locnext = 0,
          locsize = sizeof(file) - sizeof(cllist);
   u32t      nblk = ino->di_nblocks,
              ofs = 0,
              fsz = (u32t)ino->di_size,
         fsize_cl = fsz+(1<<9+clshift)-1>>(9+clshift);

   if (((u32t*)&ino->di_size)[1]) {
      log_it("size >4Gb (%lx%lx)\n", ino->di_size);
      return 0;
   }
   /* allocated size<file size? (this point should deny sparse files?)
      Native JFS code may store nblk > inode`s file size ... */
   if (fsize_cl>nblk) {
      log_it("size mismatch (%ld vs %ld)\n", fsize_cl, nblk);
      return 0;
   }
   /* i have several partitions with wrong nblk for INODE FILE itself (+1 more),
      but real number of allocated blocks is equal to the file size */
   if (nblk>fsize_cl) nblk = fsize_cl;

   xad = first_extent(ino);
   if (!xad) {
      log_it("bad first extent\n");
      return 0;
   }

   while (xad && nblk) {
      u32t pos = xad->addr2,
           len = xad_len(xad);
      // check for 64-bit cluster number, offset match, cluster number & length
      if ((xad->addr1&0xFF000000) || xad->off1 || xad->off2!=ofs || len>nblk ||
         pos+len>nclus)
      {
         log_it("bad cluser number\n");
         return 0;
      }
      // too fragmented - no memory for the extent list
      if (mbuf + locsize + sizeof(cllist) > mtop) {
         log_it("too fragmented (%x,%d)\n", mbuf, locsize);
         return 0;
      }
      loc->ext[locnext].start = pos;
      loc->ext[locnext].len   = len;
      locnext++;

      locsize += sizeof(cllist);
      nblk    -= len;
      ofs     += len;
      if (nblk)
         if ((xad=next_extent())==0) {
            log_it("bad next extent\n");
            return 0;
         }
   }
   // commit allocation
   mbuf += locsize;
   // complete file record
   loc->extnum = locnext;
   loc->size   = fsz;
   loc->ncl    = fsize_cl;
   log_m("open %x extnum %d sz %ld(%ld)\n", loc, locnext, fsz, fsize_cl);
   return loc;
}

/** calculate number of sectors to read.
   @param [out] first   number of bytes at the end of start sector
   @param [out] last    number of bytes at the start of last sector
   @return full # of *complete* sectors */
static u16t calcsector(u32t pos, u32t size, u16t *first, u16t *last) {
   u16t  una = pos & 512 - 1;
   *last = 0;
   if (una) {
      una = 512 - una;
      if (una>size) { *first = size; return 0; }
      *first = una;
      pos   += una;
      size  -= una;
   } else
      *first = 0;
   una = size & 512 - 1;
   if (una) {
      *last  = una;
      size  -= una;
   }
   return size >> 9;
}

typedef struct {
   u16t     index;     ///< index of chain in list
   u16t      srem;     ///< sectors, remain in the cluster
   u32t       ofs;     ///< cluster position in chain
} rwcall_state;

/** find location of the file sector in the chain list.
   @param fd            file data
   @param filesec       file sector to search for
   @param [out] rws     state record for rwcall()
   @return success flag */
static int listpos(file *fd, u32t filesec, rwcall_state *rws) {
   u16t   idx = 0;
   u32t  stcl = 0;

   while (idx<fd->extnum) {
      u32t len = fd->ext[idx].len;

      if (stcl<=filesec>>clshift && stcl+len>filesec>>clshift) {
         rws->index = idx;
         rws->ofs   = (filesec>>clshift) - stcl;
         rws->srem  = (1<<clshift) - (filesec & (1<<clshift) - 1);
         return 1;
      }
      stcl += len;
      idx++;
   }
   return 0;
}

/** read sectors & update position in the cluster.
   @param fd            file rec
   @param [in,out] rws  state of the operation
   @param sectors       # of sectors to read
   @param buf           buffer for data
   @param [out] act     # of actually updated sectors
   @return success flag */
static int rwcall(file *fd, rwcall_state *rws, u16t sectors, void far *buf,
                  u16t *act)
{
   u16t     idx = rws->index, until_eoc;
   u32t    clen = fd->ext[idx].len, sector,
           // use physaddr to simplify calc
           bpos = FP2PHYS(buf);
   // both should be internal error
   if (rws->index>=fd->extnum) return 0;
   if (rws->ofs>=clen) return 0;
   // sectors until the end of current chain
   until_eoc = (clen - rws->ofs - 1 << clshift) + rws->srem;
   sector    = (fd->ext[idx].start + rws->ofs + 1 << clshift) - rws->srem;
   *act      = 0;

   while (sectors) {
      u16t readsz, rem, csize;
      // rsw may point to the end of chain
      if (!until_eoc) {
         // internal error: size mismatch should be filtered in caller
         if (++idx==fd->extnum) return 0;
         sector    = fd->ext[idx].start << clshift;
         until_eoc = fd->ext[idx].len << clshift;
      }
      readsz = until_eoc;
      if (readsz>sectors) readsz = sectors;
      // segment limit
      if (readsz>127) readsz = 127;
      // read/write it
      if (!volread(sector, readsz, PHYS2FP(bpos))) break;

      sectors -= readsz;
      bpos    += readsz<<9;
      *act    += readsz;
      /* update position. Allow "end of chain" pos here to avoid "end of file"
         special case after update */
      rem      = until_eoc - readsz,
      csize    = 1<<clshift;
      clen     = fd->ext[idx].len;
      rws->index = idx;

      if (rem) {
         rws->ofs  = clen - 1 - (rem-1 >> clshift);
         rws->srem = (rem-1 & csize-1) + 1;
      } else {
         rws->ofs  = clen - 1;
         rws->srem = 0;
      }
      until_eoc -= readsz;
   }
   return !sectors;
}

/** read a file.
   @param src        file rec
   @param dst        target buffer
   @param pos        position in file
   @param size       # of bytes to read
   @param [out] act  number of bytes was actually read
   @return success flag (act can be < size!) */
static int read_file(file *src, void far *dst, u32t pos, u32t size, u32t *act) {
   rwcall_state rws;
   int          res = 0;
   u32t          bp = FP2PHYS(dst);

   *act = 0;
   if (!size) return 1;
   if (pos>src->size) return 0;
   // just a check
   if (size>_512KB) return 0;

   if (pos+size>src->size) size = src->size - pos;
   log_m("read %x dst %lx pos %lx sz %ld\n", src, dst, pos, size);

   if (size) {
      u16t  una1, una2, rwres,
            nsec = calcsector(pos, size, &una1, &una2);

      if (!listpos(src, pos>>9, &rws)) return 0;
      // printf16("index %d srem %d ofs %d\n", rws.index, rws.srem, rws.ofs);
      while (una1 || una2 || nsec) {
         void far *bpf = PHYS2FP(bp);
         u16t   opsize;
         // read/update first/last misaligned sector
         if (una1 || !nsec && una2) {
            opsize = una1?una1:una2;
            if (!rwcall(src, &rws, 1, usec, &rwres)) break;

            if (una1) _fmemcpy(bpf, usec+((u16t)pos & 512-1), una1);
               else _fmemcpy(bpf, usec, una2);
            *act  += opsize;
            if (una1) una1 = 0; else una2 = 0;
         } else
         if (nsec) {
            if (!rwcall(src, &rws, nsec, bpf, &rwres)) break;
            opsize = rwres<<9;
            *act  += opsize;
            nsec   = 0;
         }
         size -= opsize;
         bp   += opsize;
         pos  += opsize;
      }
   }
   // no error if some data is ready ...
   return *act?1:0;
}

/** read inode
   @return success flag */
static int read_inode(u32t inum, jfs_dinode *di) {
   jfs_pxd pxd;
   // offset of IAG
   u32t    ofs = ((inum >> L2INOSPERIAG) << L2INOSPERIAG) + PSIZE, pos, act;
   // extent # in inoext
   u16t     xd = (inum & INOSPERIAG - 1) >> L2INOSPEREXT;
   // read extent address
   if (!read_file(inof, &pxd, ofs + FIELDOFFSET(jfs_iag,inoext) +
      xd*sizeof(jfs_pxd), sizeof(pxd), &act)) return 0;
   // who knows?
   if (act!=sizeof(pxd)) return 0;

   pos = pxd.addr2;
   log_m("inode %ld iag ofs %lx ext %x clus %lx\n", inum, ofs, xd, pos);
   // check high part and volume size match
   if ((pxd.addr1&0xFF000000)==0 && pos<nclus)
      return volread((pos<<clshift) + (inum & INOSPEREXT - 1), 1, di);
   return 0;
}

/// return # of non-ascii characters (chr>=0x80)
static int w2char(char *dst, const void far *sname, u16t len) {
   u16t far *src = (u16t far*)sname;
   int  err = 0;
   while (len--) {
      u16t chr = *src++;
      *dst++ = chr>=0x80?'?':(chr>0x60&&chr<=0x7A?chr-0x20:chr);
      err   += chr>=0x80;
   }
   return err;
}

/** loop over root directory and search for a name
    @param   name         entry name to search (non-zero term)
    @param   namelen      length of the "name" string
    @param   [out] inode  inode number (if found)
    @return success flag */
static int read_dir(const char far *name, u16t namelen, u32t *inode) {
   // reuse xtp buffer here
   jfs_dtpage    *pb = (jfs_dtpage*)xtp;
   jfs_dtroot   *dtr = &iroot->dtroot;
   u32t    lastindex,
               index = 0;
   int           emb = 0;  // embedded data

   if (!namelen) {
      const char far *np = name;
      while (*np++) namelen++;
   }
   // there is no reason for a larger size (QSINIT also uses 128 bytes limit)
   if (!namelen || namelen>=MAXNAMELEN) return 0;

   // initialization
   do {
      if (dtr->header.flag&BT_LEAF) {
         lastindex = dtr->header.nextindex;
         emb = 1;
      } else {
         jfs_pxd      *xd;
         jfs_idtentry *ie;
         int  idx = dtr->header.stbl[0];
         // check it
         if (idx<0) return 0;

         ie = (jfs_idtentry*)pb->slot;
         xd = &((jfs_idtentry*)dtr->slot)[idx].xd;
         while (1) {
            if (xd->addr1&0xFF000000) return 0;
            if (!volread(xd->addr2<<clshift, 1<<PGSHIFT, pb)) return 0;

            if (pb->header.flag & BT_LEAF) break;
            // check for negative value at least
            idx = pb->header.stblindex;
            if (idx<0) return 0;
            idx = ((s8t far*)&ie[idx])[0];
            if (idx<0) return 0;
            xd = &ie[idx].xd;
         }
         lastindex = pb->header.nextindex;
      }
   } while (0);

   *inode = 0;
   // entry read loop
   while (1) {
      jfs_ldtentry *de = 0;
      if (!emb) {          
         // read next page
         if (index>=lastindex && (u32t)pb->header.next) {
            if (!volread((u32t)pb->header.next<<clshift, 1<<PGSHIFT, pb))
               return 0;
            lastindex = pb->header.nextindex;
            index     = 0;
         }
         if (index<lastindex) {
            jfs_ldtentry *le = (jfs_ldtentry*)pb->slot;
            s8t       *table;
            int          idx = pb->header.stblindex;
            if (idx<0) return 0;
            table = (s8t*)&le[idx];
            idx   = table[index++];
            if (idx<0) return 0;
            de    = le + idx;
         }
      } else
      if (index<lastindex) {
         int idx = dtr->header.stbl[index++];
         if (idx<0) return 0;
         de = (jfs_ldtentry*)&dtr->slot[idx];
      }
      if (!de) break;

      if (namelen==de->namlen) {
         // 6k stack is enough for 128 bytes buffer
         char entry[MAXNAMELEN+1];
         u16t  nlen = de->namlen;
         int    err = 0;

         entry[nlen] = 0;

         if (de->next<0) err = w2char(entry, &de->name, nlen); else {
            s8t  next;
            char  *wp = entry;
            err  += w2char(wp, &de->name, DTLHDRDATALEN);
            wp   += DTLHDRDATALEN;
            nlen -= DTLHDRDATALEN;
            next  = de->next;

            while (1) {
               jfs_dtslot *ds = (emb?dtr->slot:pb->slot) + next;
               next = ds->next;
               err += w2char(wp, &ds->name, next<0?nlen:DTSLOTDATALEN);

               if (next<0) break;
               wp   += DTSLOTDATALEN;
               nlen -= DTSLOTDATALEN;
            }
         }
         // name has no national characters
         if (!err) {
            u16t idx;
            // log_misc("cmpname [%s] ws [%s]\n", name, entry);
            for (idx=0; idx<namelen; idx++)
               if (name[idx]!=entry[idx]) break;
            // got it!
            if (idx==namelen) { *inode = de->inumber; break; }
         }
      }
   }
   log_it("read_dir(%s) = %lx\n", name, *inode);
   return *inode?1:0;
}

void _stdcall mfsd_close(void) {
   // drop the extent list
   mbuf = mbufmark;
   curf = 0;
}

u16t _stdcall mfsd_open(char far *name, u32t far *fsize) {
   u32t inum;

   if (curf) mfsd_close();
   if (!read_dir(name, 0, &inum)) return 1;
   if (!read_inode(inum, ino)) return 1;

   curf = open_file(ino);
   log_it("%s = %x (%ld)\n", name, curf, curf?curf->size:0);
   if (!curf) return 1;

   *fsize = curf->size;
   return 0;
}

u32t _stdcall mfsd_read(u32t ofs, char far *dst, u32t len) {
   u32t act;
   int  res;
   if (!curf) return 0;
   res = read_file(curf, dst, ofs, len, &act);
   log_it("read %ld at %lx to %lx = %d (%ld)\n", len, ofs, dst, res, act);
   return res?act:0;
}

void _stdcall mfsd_init() {
   jfs_superblock *sb;
   // memory allocation
   ino   = (jfs_dinode*)(code_nsec << 9);
   iroot = ino + 1;
   usec  = (u8t*)iroot + 512;
   xtp   = (jfs_xtreepage*)(usec + 512);
   // memory between mbuf..mtop is available
   mbuf  = (u16t)xtp + sizeof(jfs_xtreepage);
   mtop  = (SELF_SEG_LEN<<10) - STACK_SIZE;
   // use the same buffer
   sb    = (jfs_superblock*)xtp;
   log_m("ino %x iroot %x usec %x xtp %x mbuf %x..%x\n", ino, iroot, usec, xtp, mbuf, mtop);
   // panic occurs if error
   volread(JFS_SUPER1_OFS >> 9, 1, sb);
   // check for signature and zero in the high dword of the volume size
   if (sb->sign==JFS_SUPER_SIGN && sb->version==JFS_VERSION &&
      ((u32t*)&sb->size)[1]==0)
   {
      clshift = sb->l2bsize - 9;
      if (clshift<=3) {
         nclus   = (u32t)sb->size>>sb->l2bfactor;
         // read fileset inode
         volread(JFS_AITBL_OFF + FILESYSTEM_I*DISIZE >> 9, 1, ino);

         inof = open_file(ino);
         if (inof && read_inode(ROOT_I, iroot)) {
            //printf16("root mode %lx\n", iroot->di_mode);
            // verify inode type
            if ((iroot->di_mode&IFMT)==IFDIR) {
               u32t fsz;
               // memory below is used by inode file
               mbufmark = mbuf;

               if (!mfsd_open("OS2BOOT", &fsz))
                  if (fsz <= BOOT_SEG - MFSD_SEG << PARASHIFT)
                     if (mfsd_read(0, (void far*)MAKEFAR16(MFSD_SEG,0), fsz)==fsz) {
                        /* mini-FSD is actually here.
                           QSINIT is able to work without it, at least
                           until system boot itself */
                        mfsd_bits|=BF_MINIFSD;
                        ftab.ft_mfsdlen = fsz;
                     }
               // no OS2LDR?
               if (mfsd_open(bfname, &fsz)) panic_nofile();
               // file should fits between 1000:0 and our own segment
               if ((u32t)(get_ds() - BOOT_SEG)<<PARASHIFT >= fsz &&
                  mfsd_read(0, (void far*)MAKEFAR16(BOOT_SEG,0), fsz)==fsz)
               {
                  ftab.ft_loaderlen = fsz;
               } else
                  panic("Valid loader file not found.");
               // ok
               return;
            }
         }
      }
   }
   panic("Filesystem is not recognized.");
}
