//
// QSINIT
// JFS formatting code
// -----------------------------
// number of sectors/clusters in volume fixed to be 32-bit (QSINIT limit)
//
#include "mkjfs.h"

extern char    jfs_bsect[],    ///< JFS boot sector + micro-FSD code
           jfs_bsect_end[];

/** format JFS log.
   @param di            original volume information
   @param clshift       bit shift of cluster size
   @param supflags      fs superblock flags
   @param logstart      start cluster of inline log
   @param loglen        log length in clusters
   @return error code or 0 */
static qserr create_log(disk_volume_data *di, u8t clshift, u32t supflags,
                        u32t logstart, u32t loglen)
{
   u8t     secshift = bsf32(di->SectorSize),
                c2s = clshift - secshift;
   logpage_t  *logp = malloc_th(4 * LOGPSIZE);       // 16k
   logsuper_t *lsup = malloc_th(sizeof(logsuper_t)); // 4k
   qserr         rc = 0;
   u32t   secinpage = LOGPSIZE >> secshift, ii, idx,
               wpos,                                 // no 64-bit size volumes!
             npages = ((u64t)loglen << clshift) >> L2LOGPSIZE;

   // init log superblock: log page 1
   memset(lsup, 0, sizeof(logsuper_t));
   lsup->magic   = LOGMAGIC;
   lsup->version = LOGVERSION;
   lsup->state   = LOGREDONE;
   lsup->flag    = supflags;
   lsup->size    = npages;
   lsup->bsize   = 1 << clshift;
   lsup->l2bsize = clshift; 
   lsup->end     = 2*LOGPSIZE + LOGPHDRSIZE + LOGRDSIZE;

   // log superblock location 
   wpos = (logstart<<c2s) + secinpage;
   rc   = j_io(di, 0, wpos, secinpage, lsup);
   free(lsup); lsup = 0;

   if (!rc) {
      /* note from the original JFS format:
      
         init device pages 2 to npages-1 as log data pages ...
         log page sequence number (lpsn) initialization:
         the N (= npages-2) data pages of the log is maintained as a circular
         file for the log records; lpsn grows by 1 monotonically as each log
         page is written to the circular file of the log;
         We have to simulate the log wrap situation at the log format time.
         The 1st log page written will have the highest lpsn. Then the
         succeeding log pages will have ascending order of the lspn
         starting from 0...(N-2) */
      for (ii=0; ii<4; ii++) {
         lrd_t *lrdp = (lrd_t*)&logp[ii].data;
         lrdp->logtid      = 0;
         lrdp->backchain   = 0;
         lrdp->type        = LOG_SYNCPT;
         lrdp->length      = 0;
         lrdp->syncpt.sync = 0;
         logp[ii].h.eor    = LOGPHDRSIZE;
         logp[ii].t.eor    = LOGPHDRSIZE;
      }
      /* first 2 log pages: lpsn = N-1, 0
         Since the log is always an even number of meg, if we write 2 pages
         before entering the loop, we are assured that the log will end after
         a 4 page buffer */
      logp[0].h.eor  = logp[0].t.eor  = LOGPHDRSIZE + LOGRDSIZE;
      logp[0].h.page = logp[0].t.page = npages - 3;
      logp[1].h.page = logp[1].t.page = 0;
      
      wpos += secinpage;
      rc    = j_io(di, 0, wpos, secinpage*2, logp);
      if (!rc) {
         // write 4 pages at a time
         logp[0].h.eor = logp[0].t.eor = LOGPHDRSIZE;
         // initialize succeeding log pages: lpsn = 1, 2, ..., (N-2)
         wpos += secinpage*2;

         for (ii=1; ii<npages-4 && !rc; ii+=4) {
            for (idx=0; idx<4; idx++)
               logp[idx].h.page = logp[idx].t.page = ii + idx;
            rc    = j_io(di, 0, wpos, secinpage*4, logp);
            wpos += secinpage*4;
         }
      }
   }   
   free(logp);

   return rc;
}

/** initialize inode fields for an inode.
   @param ni            pointer to inode to be initialized
   @param fileset       fileset number
   @param inode         inode number
   @param clsize        # of allocated clusters
   @param size          size in bytes
   @param first_block   offset of first block of inode's extent
   @param mode          mode for inode
   @param type          type of inode to be initialized (INO_TYPE_*)
   @param inoext_addr   address of inode extent containing this inode
   @param clshift       bit shift of cluster size
   @param stamp         stamp used to identify inode as belonging to fileset */
void j_fill_inode(jfs_dinode *ni, u32t fileset, u32t inode, u32t clsize,
                  u64t size, u64t first_block, u32t mode, u32t type,
                  u64t inoext_addr, u8t clshift, u32t stamp)
{
   ni->di_inostamp  = stamp;
   ni->di_fileset   = fileset;
   ni->di_number    = inode;
   ni->di_gen       = 1;
   pxd_addr_set(&ni->di_ixpxd, inoext_addr);
   pxd_len_set (&ni->di_ixpxd, INODE_EXTENT_SIZE>>clshift);
   ni->di_mode      = mode;
   ni->di_nblocks   = clsize;
   ni->di_size      = size;
   ni->di_nlink     = 1;

   if (type==INO_TYPE_INLINE) {
      ni->special.dxd.flag = DXD_INLINE;
      dxd_len_set (&ni->special.dxd,
                   sizeof(jfs_dinode) - offsetof(jfs_dinode, special.rdev));
      dxd_addr_set(&ni->special.dxd, 0);
   } else {
      jfs_xtreepage *xtp = (jfs_xtreepage*)&ni->xtroot;

      xtp->header.flag      = DXD_INDEX | BT_ROOT | BT_LEAF;
      xtp->header.next      = 0;
      xtp->header.prev      = 0;
      xtp->header.maxentry  = XTROOTMAXSLOT;
      xtp->xad[XTENTRYSTART].flag  = 0;
      xtp->xad[XTENTRYSTART].rsvrd = 0;

      if (type==INO_TYPE_NODATA) {
         xtp->header.nextindex = XTENTRYSTART;
      } else {
         xtp->header.nextindex = XTENTRYSTART+1;
         xad_ofs_set (&xtp->xad[XTENTRYSTART], 0);
         xad_len_set (&xtp->xad[XTENTRYSTART], clsize);
         xad_addr_set(&xtp->xad[XTENTRYSTART], first_block);
      }
   }
   ni->di_atime.tv_sec = ni->di_ctime.tv_sec =
   ni->di_mtime.tv_sec = ni->di_otime.tv_sec = time(0);
}

/// read/write volume inode.
qserr j_rwinode(disk_volume_data *di, jfs_dinode *ino, u32t inum, int write,
                u32t secvtab, u8t clshift)
{
   u32t     npass = write && secvtab ? 2 : 1, inodeofs, ii;
   qserr       rc = 0;
   // volume inode table is just one inode extent
   if (inum >= INOSPEREXT) return E_SYS_SOFTFAULT;

   // second pass for the secondary table
   for (ii=0; ii<npass; ii++) {
      if (ii) {
         inodeofs = (inum << L2DISIZE) + (secvtab<<clshift);
         pxd_addr_set(&ino->di_ixpxd, secvtab);
      } else {
         inodeofs = (inum << L2DISIZE) + JFS_AITBL_OFF;
         pxd_addr_set(&ino->di_ixpxd, JFS_AITBL_OFF>>clshift);
      }
      // bad case - sector size > DISIZE (512)
      if (DISIZE < di->SectorSize) {
         jfs_dinode *sb = malloc_th(di->SectorSize);
         // read full sector
         rc = j_io(di, JIO_BO|JIO_READ, inodeofs, 1, sb);
         if (!rc) {
            u32t sbidx = (inodeofs & di->SectorSize-1) >> L2DISIZE;
            if (write) {
               memcpy(sb + sbidx, ino, sizeof(jfs_dinode));
               // write full sector back
               rc = j_io(di, JIO_BO, inodeofs, 1, sb);
            } else
               memcpy(ino, sb + sbidx, sizeof(jfs_dinode));
         }
         free(sb);
      } else 
      if (DISIZE==di->SectorSize)
         rc = j_io(di, JIO_BO|(write?0:JIO_READ), inodeofs, 1, ino);
      else
         rc = E_SYS_SOFTFAULT;
   }
   return rc;
}


/// initialize one dmap page.
static void idmap_page(dmap_t *page, u32t nblocks) {
   u32t ii, nwords, bit;
   page->nblocks = page->nfree = nblocks;
   /* Partial dmap page?
      If there are not enough blocks to cover an entire dmap page the ones
      which represent blocks which don`t exist will be marked as allocated.

      nwords will indicate the first word beyond the end of existing blocks
      bit will indicate if this block does not fall on a 32-bit boundary */
   nwords = nblocks >> L2DBWORD;
   bit    = nblocks & DBWORD - 1;
   // need to mark a partial word allocated
   if (bit) {
      page->wmap[nwords] = page->pmap[nwords] = FFFF >> bit;
      nwords++;
   }
   // set the rest of the words in the page to 0xFF.
   for (ii = nwords; ii<LPERDMAP; ii++)
      page->pmap[ii] = page->wmap[ii] = FFFF;
}

/** initialize first extent of inode allocation map.
   Inode size forced to INODE_EXTENT_SIZE and inode map has 2 pages.
   Map pages are marked in bitmap.

   @param di            volume information
   @param clshift       bit shift of cluster size
   @param pmap          map state/data buffer
   @param loc           start cluster of first extent of inode table
   @param maploc        start cluster of first extent of inode map
   @param initnum       number of initialized fileset inodes
   @param agsize        allocation group size in clusters
   @param type          table type (fileset or volume)
   @return error code or 0 */
static qserr init_inode_map(disk_volume_data *di, u8t clshift, mapdata *pmap,
                            u32t loc, u32t maploc, u32t initnum, u32t agsize,
                            u32t type)
{
   dinomap_t    *cp;
   jfs_iag     *iag;
   jfs_pxd  *inoext;
   u32t       *wmap, ii, bit, agend, agnum;
   qserr         rc;
   // control & initial pages of inode allocation map
   void       *imap = calloc_th(1, PSIZE*2);
   if (!imap) return E_SYS_NOMEM;

   // initialize control page
   cp = imap;
   cp->freeiag     = -1;
   cp->nextiag     = 1;
   cp->numinos     = INOSPEREXT;
   cp->numfree     = INOSPEREXT - initnum;
   cp->nbperiext   = INODE_EXTENT_SIZE >> clshift;
   cp->l2nbperiext = bsf32(cp->nbperiext);
   // initialize space for first extent of fileset inode map
   iag = (jfs_iag*)((u8t*)imap + PSIZE); 
   // initialize control section of first IAG of inode allocation map
   iag->iagnum      = 0;
   iag->agstart     = loc / agsize * agsize;
   iag->inofreefwd  = -1;
   iag->inofreeback = -1;
   iag->extfreefwd  = -1;
   iag->extfreeback = -1;
   iag->iagfree     = -1;
   iag->nfreeinos   = INOSPEREXT - initnum;
   iag->nfreeexts   = EXTSPERIAG - 1;

   // initialize working and persistent maps of first IAG
   if (type==AGGREGATE_I) iag->wmap[0] = iag->pmap[0] = 0xF8008000;
      else iag->wmap[0] = iag->pmap[0] = 0xF0000000;
   // initialize first inode extent of IAG
   pxd_len_set(&iag->inoext[0], INODE_EXTENT_SIZE>>clshift);
   pxd_addr_set(&iag->inoext[0], loc);
   /* initialize extent summary map from the inode extents of the iag,
      initialize inode summary map from working/permanent map of the iag */
   for (inoext=&iag->inoext[0], wmap=&iag->wmap[0], ii=0; ii<SMAPSZ; ii++) {
      u32t *extsmap = iag->extsmap + ii,
           *inosmap = iag->inosmap + ii;
      for (bit=0; bit<32; bit++, inoext++, wmap++) {
         *extsmap <<= 1;
         *inosmap <<= 1;
         if (pxd_addr(inoext)) {
            // extent allocated
            *extsmap |= 1;
            // all allocated
            if (*wmap==FFFF) *inosmap |= 1;
         } else
            *inosmap |= 1;
      }
   }
   /* initialize the AG free list ...
      figure out which AG the inode extent belongs in and initialize AG free
      list entry to point to it, all other will point to an empty entry */
   for (agend = agsize, agnum = 0; agnum<MAXAG; agnum++) {
      if (loc < agend) break;
      agend += agsize;
   }
   /* broke out of for loop by hitting the maximum number of aggregates,
      either the agsize is wrong, or the "loc" value is wrong */
   if (loc >= agend) rc = E_SYS_SOFTFAULT; else {
      for (ii=0; ii<MAXAG; ii++) {
         iagctl_t *pe = cp->agctl + ii;
         if (ii==agnum) {
            // initial inode
            pe->inofree = iag->iagnum;
            pe->extfree = iag->iagnum;
            pe->numinos = INOSPEREXT;
            pe->numfree = INOSPEREXT - initnum;
         } else {
            // no inode extents for this AG
            pe->inofree = -1;
            pe->extfree = -1;
            pe->numinos = 0;
            pe->numfree = 0;
         }
      }
      // write control page and first extent of inode map to disk
      rc = j_io(di, JIO_BO|JIO_BS, (u64t)maploc<<clshift, PSIZE*2, imap);
   }
   free(imap);
   if (rc) return rc;
   // mark 2 pages in block allocation map
   rc = j_mark(pmap, maploc, MARK_ALLOC | PSIZE*2>>clshift);
   return rc;
}


/** initialize volume inodes.
   @param di            volume information
   @param clshift       bit shift of cluster size
   @param pmap          map state/data buffer
   @param loc           start cluster of first extent of inode table
   @param vino          array of initial volume inodes
   @param initnum       number initialized of volume inodes
   @param loc           start cluster of first extent of inode table
   @param maploc        start cluster of first extent of inode map
   @param stamp         inode stamp value to be used
   @return error code or 0 */
static qserr init_vol_inodes(disk_volume_data *di, u8t clshift, mapdata *pmap,
                             jfs_dinode *vino, u32t initnum,
                             u32t loc, u32t maploc, u32t stamp)
{
   qserr     rc = 0;
   // allocate space for first inode extent, and clear
   void    *buf = calloc_th(INODE_EXTENT_SIZE, 1);
   if (!buf) return E_SYS_NOMEM;

   // initialize inodes: 0, 1, 3, and 4; inode 2 has already been initialized
   vino[AGGR_RESERVED_I].di_nlink = 1;
   // initialize inode 1: "self" inode
   j_fill_inode(vino + AGGREGATE_I, AGGREGATE_I, AGGREGATE_I, PSIZE*2 >> clshift,
                PSIZE*2, maploc, IFJOURNAL|IFREG, INO_TYPE_MAXEXTENT, loc,
                clshift, stamp);
   // equal to inode->_data2._imap._gengen = 1
   vino[AGGREGATE_I].extdata[4] = 1;
   // initialize inode 3: inline log inode
   j_fill_inode(vino + LOG_I, AGGREGATE_I, LOG_I, 0, 0, 0,
                IFJOURNAL|IFREG, INO_TYPE_NODATA, loc, clshift, stamp);
   // initialize inode 4: bad block inode
   j_fill_inode(vino + BADBLOCK_I, AGGREGATE_I, BADBLOCK_I, 0, 0, 0,
                IFJOURNAL|IFREG|ISPARSE, INO_TYPE_NODATA, loc, clshift, stamp);
   // copy initialized inodes to buffer
   memcpy(buf, vino, initnum * sizeof(jfs_dinode));

   // write inode extent to disk
   rc = j_io(di, JIO_BO|JIO_BS, (u64t)loc<<clshift, INODE_EXTENT_SIZE, buf);
   free(buf);
   if (rc) return rc;
   // mark as allocated in block allocation map
   rc = j_mark(pmap, loc, MARK_ALLOC | INODE_EXTENT_SIZE>>clshift);
   return rc;
}


/** initialize fileset inodes and write to disk.
   @param di            volume information
   @param clshift       bit shift of cluster size
   @param pmap          map state/data buffer
   @param loc           start cluster of fileset
   @param maploc        start cluster of fileset inode map
   @param stamp         inode stamp value to be used
   @return error code or 0 */
static qserr init_fset_inodes(disk_volume_data *di, u8t clshift, mapdata *pmap,
                              u32t loc, u32t maploc, u32t stamp)
{
   void        *buf;
   jfs_dinode   *bp;
   jfs_dtroot *root;
   u32t          ii;
   qserr         rc = 0;

   /* The first fileset inode extent is the first thing written for the
      fileset, so its location is the start of the fileset */
   buf = calloc_th(INODE_EXTENT_SIZE, 1);
   if (!buf) return E_SYS_NOMEM;
   bp  = (jfs_dinode*)buf;
   // inode 0 - reserved inode
   j_fill_inode(bp, FILESYSTEM_I, FILESET_RSVD_I, 0, 0, 0, IFJOURNAL|IFREG,
                INO_TYPE_NODATA, loc, clshift, stamp);
   bp++;
   // inode 1 - 2nd half of fileset superblock information (not used)
   j_fill_inode(bp, FILESYSTEM_I, FILESET_EXT_I, 0, 0, 0, IFJOURNAL|IFREG,
                INO_TYPE_NODATA, loc, clshift, stamp);
   bp++;
   // inode 2 - root directory
   j_fill_inode(bp, FILESYSTEM_I, ROOT_I, 0,
                sizeof(jfs_dinode) - offsetof(jfs_dinode, special.rdev), 0,
                IFJOURNAL|IFDIR|IREAD|IWRITE|IEXEC|IDIRECTORY, INO_TYPE_NODATA,
                loc, clshift, stamp);
   // set number of links for root to account for '.' entry
   bp->di_nlink = 2;
   // initialize the directory B+-tree header for the root inode.
   root = &bp->dtroot;
   root->header.flag      = DXD_INDEX | BT_ROOT | BT_LEAF;
   root->header.nextindex = 0;
   // determine how many slots will fit
   root->header.freecnt   = DTROOTMAXSLOT - 1;
   root->header.freelist  = DTENTRYSTART;
   root->header.idotdot   = 2;

   for (ii = DTENTRYSTART; ii<DTROOTMAXSLOT; ii++) {
      root->slot[ii].cnt  = 1;
      root->slot[ii].next = ii+1;
   }
   // last entry should end the free list
   root->slot[--ii].next = -1;
   bp++;

   // inode 3 - ACL file
   j_fill_inode(bp, FILESYSTEM_I, ACL_I, 0, 0, 0, IFJOURNAL|IFREG,
               INO_TYPE_NODATA, loc, clshift, stamp);

   // write inode extent to disk
   rc = j_io(di, JIO_BO|JIO_BS, (u64t)loc<<clshift, INODE_EXTENT_SIZE, buf);
   free(buf);
   if (rc) return rc;

   // mark as allocated in block allocation map
   rc = j_mark(pmap, loc, MARK_ALLOC | INODE_EXTENT_SIZE>>clshift);
   return rc;
}

qserr _std jfs_format(u8t vol, u32t flags, u32t unitsize, read_callback cbprint) {
   disk_volume_data    di;
   disk_geo_data      geo;
   u32t            sizecl, logsize, logstart, ii,
                   hidden = 0;
   char            drivel = 0;
   u8t            clshift,
                 secshift;
   long            volidx;
   u32t            sflags = JFS_OS2 | JFS_GROUPCOMMIT | JFS_INLINELOG;
   u64t             sizeb,
                 logsizeb,
                   badcnt = 0;
   qserr               rc;
   mapdata          *pmap = 0;
   jfs_dinode       *vino = 0;
   jfs_superblock     *sb = 0;
   struct Boot_Record *br = 0;

   if (flags & DFMT_ONEFAT) return E_SYS_INVPARM;
   if (flags & DFMT_SPARSE) sflags |= JFS_SPARSE;

   hlp_volinfo(vol, &di);
   if (!di.TotalSectors) return E_DSK_NOTMOUNTED;
   if (di.InfoFlags&VIF_VFS) return E_DSK_NOTPHYS;
   secshift = bsf32(di.SectorSize);
   if (secshift<9 || secshift>12) return E_DSK_SSIZE;
   // verify min JFS size
   if ((u64t)di.TotalSectors<<secshift < MINJFS) return E_DSK_VSMALL;

   if (!unitsize) unitsize = 4096;
   clshift = bsr32(unitsize);
   if (clshift != bsf32(unitsize)) return E_DSK_CLSIZE;
   // the only supported clusters are 512/1k/2k/4k
   if (clshift<9 || clshift>12) return E_DSK_CLSIZE;
   // just expand cluster to sector without an error ...
   if (clshift<secshift) {
      clshift  = secshift;
      unitsize = 1<<clshift;
   }
   // get geometry for bpb
   if (dsk_getptgeo(di.Disk,&geo))
      if (!hlp_disksize(di.Disk,0,&geo)) {
         geo.Heads       = 255;
         geo.SectOnTrack =  63;
      }

   volidx = vol_index(vol,0);
   // allow floppies and big floppies (for test reason)
   if (volidx<0 && di.StartSector) return E_PTE_RESCAN;
   if (volidx>=0) {
      u32t ptflags = 0;
      if (dsk_isgpt(di.Disk,volidx)==1) {
         // GPT partition
         flags  |= DFMT_NOPTYPE;
         ptflags = DPTF_PRIMARY;
         hidden  = di.StartSector>=_4GBLL ? FFFF : di.StartSector;
      } else {
         // MBR or hybrid partition
         dsk_ptquery(di.Disk, volidx, 0, 0, 0, &ptflags, &hidden);
         if (hidden==FFFF) return E_PTE_EXTERR;
      }
      // query drive letter for the primary partition (this can be GPT too)
      if (ptflags & DPTF_PRIMARY) {
         drivel = lvm_ismounted(di.Disk, volidx);
         if (drivel<'C' || drivel>'Z') drivel = 0;
      }
   }
   // unmount will clear all cache, all r/w ops below are use QDSK_DIRECT flag
   if (io_unmount(vol, flags&DFMT_FORCE?IOUM_FORCE:0)) return E_DSK_UMOUNTERR;

   sizeb    = (u64t)di.TotalSectors << secshift;
   /* there is no logsize option in our format, so use default JFS logic:
      1/256 of the volume size */
   logsizeb = sizeb >> 8;
   if (logsizeb > _32MB) logsizeb = _32MB;
   // round log size up to Mb
   logsizeb = logsizeb+_1MB-1 & ~(_1MB-1);

   sizecl   = sizeb >> clshift;
   // convert logsize into clusters
   logsize  = logsizeb >> clshift;
   logstart = sizecl - logsize;
   sizecl  -= logsize;

   log_printf("Formatting vol %c, disk %02X, index %d to JFS (%u)\n",
      vol+'A', di.Disk, volidx, unitsize);
   log_it(2, "log pos %010LX(%X) size %ukb, vol size = %u clusters\n",
      di.StartSector + (logstart << clshift-secshift), logstart,
         (u32t)logsizeb>>10, sizecl);

   do {
      jfs_xad  inodemap;
      u32t      numbits, npages, fchklen, fchkpos, fchklog, agsize,
                secvmap, secvtab, fsetpos, fsetmap;
      u32t       vstamp = time(0);
      // create journal (log)
      rc = create_log(&di, clshift, sflags, logstart, logsize);
      if (rc) break;
      // can`t be overlow here since size was cut above for a log ...
      numbits = sizecl + BPERDMAP - 1 >> L2BPERDMAP << L2BPERDMAP;
      /* Alloc fsck working space and mark those blocks.
         The fsck working space is always at the very end of the volume.
         Need enough pages to cover:
         - 1 bit per block in volume rounded up to BPERDMAP boundary
         - 1 extra 4k page to handle control page and intermediate level pages
         - 50 extra 4k pages for the chkdsk service log */
      npages  = (numbits + PSIZE*8 - 1) / (PSIZE*8) + 1 + 50;
      
      fchklen = npages << L2PSIZE >> clshift;
      fchkpos = sizecl - fchklen;
      fchklog = 50 << L2PSIZE >> clshift;
      // cut chkdsk working space
      sizecl -= fchklen;
      vino    = calloc_th(INIT_NUM_AGGR_INODES, sizeof(jfs_dinode));
      pmap    = calloc_th(sizeof(mapdata), 1);
      // initialize bitmaps & volume inodes
      rc = j_calcmap(vino, sizecl, clshift, &agsize, pmap, vstamp);
      if (rc) break;
      // zero area before superblock
      rc = dsk_emptysector(di.Disk|QDSK_DIRECT, di.StartSector + 1,
                           JFS_SUPER1_OFS - di.SectorSize >> secshift);
      if (rc) break;
      // and mark it as used
      rc = j_mark(pmap, 0, MARK_ALLOC | JFS_SUPER1_OFS>>clshift);
      if (rc) break;
      // zero old superblocks to be safe
      rc = dsk_emptysector(di.Disk|QDSK_DIRECT, JFS_SUPER1_OFS>>secshift,
                           PSIZE>>secshift);
      if (rc) break;
      rc = dsk_emptysector(di.Disk|QDSK_DIRECT, JFS_SUPER2_OFS>>secshift,
                           PSIZE>>secshift);
      if (rc) break;
      // mark superblocks
      rc = j_mark(pmap, JFS_SUPER1_OFS>>clshift, MARK_ALLOC|PSIZE>>clshift);
      if (rc) break;
      rc = j_mark(pmap, JFS_SUPER2_OFS>>clshift, MARK_ALLOC|PSIZE>>clshift);
      if (rc) break;
#if MKJFS_DBLEVEL>0
      log_it(3, "volume inodes ...\n");
#endif
      // initialize first extent of volume inode allocation map
      rc = init_inode_map(&di, clshift, pmap,
                          JFS_AITBL_OFF >> clshift, JFS_AIMAP_OFF >> clshift,
                          INIT_NUM_AGGR_INODES + 1, agsize, AGGREGATE_I);
      if (rc) break;
      // initialize first inode extent of volume inode table
      rc = init_vol_inodes(&di, clshift, pmap, vino, INIT_NUM_AGGR_INODES,
                           JFS_AITBL_OFF >> clshift, JFS_AIMAP_OFF >> clshift,
                           vstamp);
      if (rc) break;
      /* initialize the secondary volume inode table and map

         We can use the same aggr_inodes we already initialized except for the
         volume self inode. This will be updated by init_vol_inodes() to
         point to the secondary table.
       
         The location will follow the block map. Since the block map might
         be sparse we need to use the number of blocks instead of length of
         the extents.  This works since the extents are in the inode for mkfs */
      inodemap = ((jfs_xtreepage*)&vino[BMAP_I].xtroot)->xad[XTENTRYSTART];
      secvmap  = xad_addr(&inodemap) + vino[BMAP_I].di_nblocks;
      secvtab  = secvmap + (PSIZE*2 >> clshift);
#if MKJFS_DBLEVEL>0
      log_it(3, "sec volume inode tab %u map %u\n", secvmap, secvtab);
#endif
      rc = init_inode_map(&di, clshift, pmap, secvtab, secvmap,
                          INIT_NUM_AGGR_INODES + 1, agsize, AGGREGATE_I);
      if (rc) break;
      // modify the volume inodes ixpxd fields
      pxd_addr_set(&vino[BMAP_I].di_ixpxd, secvtab);
      rc = init_vol_inodes(&di, clshift, pmap, vino, INIT_NUM_AGGR_INODES,
                           secvtab, secvmap, vstamp);
      if (rc) break;
      // mark blocks for the block map
      rc = j_mark(pmap, JFS_BMAP_OFF>>clshift, MARK_ALLOC|vino[BMAP_I].di_nblocks);
      if (rc) break;
#if MKJFS_DBLEVEL>0
      log_it(3, "creating fileset ...\n");
#endif
      /* create a fileset.
         Since the secondary volume inode table is the last thing written for
         the volume, the fileset will start here.
         The first extent of the fileset inode allocation map follows the first
         extent of the first extent of the fileset inodes at the beginning of
         the fileset */
      fsetpos  = secvtab + (INODE_EXTENT_SIZE >> clshift);
      fsetmap  = fsetpos + (INODE_EXTENT_SIZE >> clshift);
      // create volume inode for fileset
      if (!rc) {
         jfs_dinode  fseti;
         memset(&fseti, 0, sizeof(fseti));
         j_fill_inode(&fseti, AGGREGATE_I, FILESYSTEM_I, PSIZE*2 >> clshift,
                      PSIZE*2, fsetmap, IFJOURNAL|IFREG, INO_TYPE_MAXEXTENT,
                      JFS_AITBL_OFF>>clshift, clshift, vstamp);
         // equal to inode->_data2._imap._gengen = 1
         fseti.extdata[4] = 1;
         // write to both primary and secondary inode tables
         rc = j_rwinode(&di, &fseti, FILESYSTEM_I, 1, secvtab, clshift);
         if (rc) break;
      }
      // create fileset inode table - first extent
      rc = init_fset_inodes(&di, clshift, pmap, fsetpos, fsetmap, vstamp);
      if (rc) break;
      // create fileset inode allocation map - first extent
      rc = init_inode_map(&di, clshift, pmap, fsetpos, fsetmap, ACL_I + 1,
                          agsize, FILESYSTEM_I);
      if (rc) break;
      // read/erase data area
      if ((flags&DFMT_WIPE) || (flags&DFMT_QUICK)==0) {
         dd_list bclist = NEW(dd_list);
         void    *iobuf = malloc_th(_32KB);
#if MKJFS_DBLEVEL>0
         log_it(3, "surface scan (%u..%u) ...\n", pmap->lastalloc+1, sizecl-1);
#endif
         rc = vol_scansurface(di.Disk, flags, di.StartSector, 0, sizecl,
                              pmap->lastalloc+1, 1<<clshift-secshift,
                              bclist, iobuf, cbprint,0,0);
#if MKJFS_DBLEVEL>1
         for (ii=1000; ii<sizecl; ii+=1000)
            if (ii>pmap->lastalloc) bclist->add(ii);
#endif
         badcnt = bclist->count();
         if (!rc)
            rc = j_write_bb(&di, pmap, clshift, secvtab, vino + BADBLOCK_I, bclist);
         DELETE(bclist);
         free(iobuf);
         if (rc) break;
      }
#if MKJFS_DBLEVEL>0
      log_it(3, "writing bitmap ...\n");
#endif
      // block allocation map should be complete, write to disk
      rc = j_write_bm(&di, pmap, sizecl);
      if (rc) break;
      sb = calloc_th(PSIZE, 1);

      sb->sign        = JFS_SUPER_SIGN;
      sb->version     = JFS_VERSION;
      sb->size        = (u64t)sizecl << clshift - secshift;
      sb->bsize       = unitsize;
      sb->l2bsize     = clshift;
      sb->l2bfactor   = clshift - secshift;
      sb->pbsize      = di.SectorSize;
      sb->l2pbsize    = secshift;
      sb->agsize      = agsize;

      sb->flag        = sflags;
      sb->state       = FM_CLEAN;
      sb->fsckloglen  = fchklog;
      sb->time.tv_sec = vstamp;
      
      pxd_addr_set(&sb->ait2, secvtab);
      pxd_len_set (&sb->ait2, INODE_EXTENT_SIZE >> clshift);
      pxd_addr_set(&sb->aim2, secvmap);
      pxd_len_set (&sb->aim2, PSIZE*2 >> clshift);

      pxd_addr_set(&sb->fsckpxd, fchkpos);
      pxd_len_set (&sb->fsckpxd, fchklen);
      pxd_addr_set(&sb->logpxd, logstart);
      pxd_len_set (&sb->logpxd, logsize);
      
      // write both superblocks
      rc = j_io(&di, JIO_BO|JIO_BS, JFS_SUPER2_OFS, PSIZE, sb);
      if (rc) break;
      rc = j_io(&di, JIO_BO|JIO_BS, JFS_SUPER1_OFS, PSIZE, sb);
      if (rc) break;

      br = calloc_th(di.SectorSize, 1);
      br->BR_BPB.BPB_BytePerSect = di.SectorSize;
      br->BR_BPB.BPB_ResSectors  = 1;
      if (di.TotalSectors < 0x10000) br->BR_BPB.BPB_TotalSec = di.TotalSectors;
         else br->BR_BPB.BPB_TotalSecBig = di.TotalSectors;
      br->BR_BPB.BPB_MediaByte   = di.StartSector?0xF8:0xF0;
      br->BR_BPB.BPB_SecPerTrack = geo.SectOnTrack;
      br->BR_BPB.BPB_Heads       = geo.Heads;
      br->BR_BPB.BPB_HiddenSec   = hidden;
      // use current time as VSN
      br->BR_EBPB.EBPB_VolSerial = tm_getdate();
      if ((di.Disk&QDSK_FLOPPY)==0) br->BR_EBPB.EBPB_PhysDisk = 0x80;
      br->BR_EBPB.EBPB_Sign      = 0x29;
      br->BR_EBPB.EBPB_Dirty     = drivel?0x80+(drivel-'C'):0;
      memcpy(br->BR_EBPB.EBPB_FSType, "JFS     ", 8);
      memcpy(br, jfs_bsect, 11);
      memcpy(br+1, (struct Boot_Record *)&jfs_bsect+1, 512-sizeof(struct Boot_Record));

      rc = j_io(&di, 0, 0, 1, br);
      if (rc) break;
      /* write the rest of boot code only if sector size is 512 bytes.
         First sector will print "Boot code is broken." in any hypothetical
         case of 4k physical sector. */
      if (di.SectorSize==512) {
         /* code is in a separate object of this LX module, so unaligned end
            of it will be filled with zeroes by LX loader */
         u32t nsec = Round512((u32t)&jfs_bsect_end - (u32t)&jfs_bsect) >> 9;
         if (nsec>1) {
            rc = j_io(&di, 0, 1, nsec-1, &jfs_bsect[512]);
            if (rc) break;
         }
      }
      // on success - update partition type & mount volume back in format_done()
      if (!rc) {
         int ptbyte = volidx>=0 && (flags&DFMT_NOPTYPE)==0 ? PTE_07_MSFS : -1;
         rc = format_done(vol, di, ptbyte, volidx, badcnt);
      }
   } while (0);
   // fix write error code
   if (rc==E_DSK_ERRWRITE) rc = E_DSK_BADSYSAREA;
   // free map pages
   if (pmap) {
      if (pmap->maps) {
         for (ii=0; ii<pmap->nmaps; ii++)
            if (pmap->maps[ii] && pmap->maps[ii]!=pmap->emptypage)
               free(pmap->maps[ii]);
         free(pmap->maps);
      }
      if (pmap->emptypage) free(pmap->emptypage);
      free(pmap);
   }
   if (vino) free(vino);
   if (sb) free(sb);
   if (br) free(br);

   return rc;
}
