#include "mkjfs.h"

/* BMAPSZTOLEV() macro and some checks below are for 64-bit volume size,
   leave is as it is and disable "always true/false" message in OW 1.9 */
#pragma disable_message(124);

/* used to determine the maximum free string in a character of a wmap word.
   the actual bit values of the character serve as the index into this array
   and the value of the array at that index is the max free string */
static s8t budtab[256] = {
   3,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
   2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
   2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0,
   2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0,2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0,
   2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0,
   2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0,2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0,
   2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0,
   2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,0,2,1,1,1,1,0,0,0,1,0,0,0,1,0,0,-1 };

qserr j_io(disk_volume_data *di, u8t flags, u64t ofs, u64t size, void *buf) {
   u8t secshift = bsf32(di->SectorSize);
   qserr     rc = 0;
   if (!buf) return E_SYS_ZEROPTR;
   // convert bytes to sectors
   if (flags & JIO_BO)
      if (ofs & di->SectorSize-1) return E_SYS_ALIGN;
         else ofs>>=secshift;
   if (flags & JIO_BS)
      if (size & di->SectorSize-1) return E_SYS_ALIGN;
         else size>>=secshift;
   // make sure to be safe
   if (ofs + size > di->TotalSectors) {
      log_it(2, "Volume range violation (%LX+%LX > %X)!\n", ofs, size,
         di->TotalSectors);
      return E_SYS_SOFTFAULT;
   }
   if ((flags & JIO_READ ? hlp_diskread : hlp_diskwrite)(di->Disk|QDSK_DIRECT,
      di->StartSector + ofs, (u32t)size, buf) != (u32t)size)
         rc = flags & JIO_READ ? E_DSK_ERRREAD : E_DSK_ERRWRITE;
   return rc;
}

static void idmap_page(dmap_t *page, u32t size);

qserr j_mark(mapdata *pmap, u32t clnum, u32t flags) {
   u32t nclus = flags & MARK_NMASK;
   // default zero means one cluster
   if (!nclus) nclus++;
#if MKJFS_DBLEVEL>0
   log_it(3, "markit(%c:%u,%u)\n", (flags&~MARK_NMASK)==MARK_BAD? 'B' :
      flags&MARK_ALLOC ? 'A' : 'F', clnum, nclus);
#endif
   while (nclus--) {
      u32t page, rem, word, bit, agno;
      dmap_t     *pp;
      // keep track of the last allocated block for the block map inode
      if (clnum > pmap->lastalloc && !(flags & (MARK_BAD & ~MARK_ALLOC)))
         pmap->lastalloc = clnum;

      // calculate page number in map, and word and bit number in word.
      page = clnum >> L2BPERDMAP;
      rem  = clnum - (page << L2BPERDMAP);
      word = rem >> L2DBWORD;
      bit  = rem - (word << L2DBWORD);

      if (page > pmap->nmaps) return E_SYS_SOFTFAULT;
      // allocate dmap page
      if (!pmap->maps[page]) {
         u64t left = pmap->cpage.mapsize - (page << L2BPERDMAP);

         pmap->maps[page] = calloc_th(sizeof(dmap_t),1);
         if (!pmap->maps[page]) return E_SYS_NOMEM;

         idmap_page(pmap->maps[page], min(BPERDMAP,left));
      }

      pp   = pmap->maps[page];
      agno = clnum >> pmap->cpage.agl2size;
      bit  = 0x80000000 >> bit;

      if (flags & MARK_ALLOC) {
         pp->pmap[word] |= bit;
         pp->wmap[word] |= bit;
         pp->nfree--;
         pmap->cpage.nfree--;
         pmap->cpage.agfree[agno]--;
      } else {
         pp->pmap[word] &= ~bit;
         pp->wmap[word] &= ~bit;
         pp->nfree++;
         pmap->cpage.nfree++;
         pmap->cpage.agfree[agno]++;
      }
      clnum++;
   }
   return 0;
}

/// determine shift of allocation group size based on volume size
static u32t getagshift(u64t size) {
   if (size<BPERDMAP*MAXAG) return L2BPERDMAP; else {
      u32t l2sz = bsr64(size);
      if ((u64t)1<<l2sz < size) l2sz++;
      return l2sz - L2MAXAG;
   }
}

qserr j_calcmap(jfs_dinode *aggi, u32t sizecl, u8t clshift, u32t *agsize,
                mapdata *pmap, time_t stamp)
{
   // number dmap pages required + one extra for extendfs()
   u32t   npages = (sizecl + BPERDMAP - 1 >> L2BPERDMAP) + 1, ii, nb, l2nl,
          ndmaps = npages, nl0pages;
   // this is impossible in our case
   if (ndmaps<<L2BPERDMAP > MAXMAPSIZE) return E_DSK_VLARGE;
   // # of L0 pages to cover dmap pages
   nl0pages = ndmaps + LPERCTL - 1 >> L2LPERCTL;
   npages += nl0pages;

   if (nl0pages>1)
      // one L2 and actual L1 pages to cover L0 pages
      npages += 1 + (nl0pages + LPERCTL - 1 >> L2LPERCTL);
   else
      // logical L2 and one logical L1 pages
      npages += 2;
   // global control page of map
   npages++;
   // init the block allocation map inode
   j_fill_inode(aggi+BMAP_I, AGGREGATE_I, BMAP_I, npages << L2PSIZE-clshift,
                (u64t)npages << L2PSIZE, JFS_BMAP_OFF>>clshift,
                IFJOURNAL|IFREG, INO_TYPE_MAXEXTENT, JFS_AITBL_OFF>>clshift,
                clshift, stamp);
   // alloc/init dmap page pointer array
   pmap->maps   = calloc_th(ndmaps, sizeof(dmap_t*));
   pmap->nmaps  = ndmaps;
   if (!pmap->maps) return E_SYS_NOMEM;

   // control page - initialize base information
   pmap->cpage.l2nbperpage = L2PSIZE - clshift;
   pmap->cpage.mapsize  = sizecl;
   pmap->cpage.nfree    = sizecl;
   pmap->cpage.maxlevel = BMAPSZTOLEV(sizecl);
   // pmap->cpage.maxfreebud is computed at finalization

   // initialize allocation group information
   pmap->cpage.agl2size = getagshift(sizecl);
   pmap->cpage.agsize   = (u64t)1 << pmap->cpage.agl2size;
   *agsize              = pmap->cpage.agsize;
   pmap->cpage.numag    = sizecl / pmap->cpage.agsize;
   pmap->cpage.numag   += (sizecl % pmap->cpage.agsize) ? 1 : 0;

   log_it(2, "ag shift %u, map size %u, l0=%u, agsize=%u\n", pmap->cpage.agl2size,
      ndmaps, nl0pages, *agsize);

   for (ii=0, nb=sizecl; ii<pmap->cpage.numag; ii++, nb-=*agsize)
      pmap->cpage.agfree[ii] = min(nb,*agsize);

   pmap->cpage.aglevel  = BMAPSZTOLEV(pmap->cpage.agsize);
   l2nl = pmap->cpage.agl2size - (L2BPERDMAP + pmap->cpage.aglevel * L2LPERCTL);
   pmap->cpage.agheigth = l2nl >> 1;
   pmap->cpage.agwidth  = 1 << l2nl - (pmap->cpage.agheigth << 1);

   for (ii=5-pmap->cpage.agheigth, pmap->cpage.agstart=0, nb=1; ii>0; ii--) {
      pmap->cpage.agstart += nb;
      nb <<= 2;
   }
   // pmap->cpage.maxag is computed at finalization
   pmap->cpage.agpref   = 0;
   // reset last_allocated to ignore the fsck working space
   pmap->lastalloc      = 0;

   return 0;
}

/** determines the maximum string of free blocks.
   Returns maximum string of free blocks within a word of the wmap or pmap
   consistent with binary buddy.
   @param cp            pointer to wmap or pmap word.
   @return maximum string of free blocks within word */
static s8t maxbuddy(u8t *cp) {
   s8t max12, max34;
   // check if the wmap or pmap word is all free
   if (*(u32t*)cp==0) return BUDMIN;
   // check if the wmap or pmap word is half free
   if (*(u16t*)cp==0 || *((u16t*)cp+1)==0) return BUDMIN-1;
   // determine the size via table using quarters of the wmap or pmap word
   max12 = max(budtab[cp[0]], budtab[cp[1]]);
   max34 = max(budtab[cp[2]], budtab[cp[3]]);
   return max(max12,max34);
}

/** calculate the tree of a dmap or dmapctl.
   This routine first works against the leaves of the tree to calculate
   the maximum free string for leaf buddys. Once this is accomplished the
   values of the leaf nodes are bubbled up the tree.
   @param tp            pointer to the top of the tree
   @param l2leaves      number of leaf nodes, p2
   @param l2min         # of clusters actually covered by a leaf of the tree, p2
 */
static s8t adjtree(s8t *tp, u32t l2leaves, u32t l2min) {
   u32t  nleaves, nlevel, l2max, ii, l2free, leaf, nextp, leaf_index, nextb, bsize;
   s8t       *cp = tp, *cp0, *cp1;
   /* determine the number of leaves of the tree and the index of the first leaf.
      Funny note from IBM coder:
      I don`t know why the leaf_index calculation works, but it does */
   nleaves    = 1 << l2leaves;
   leaf_index = (nleaves - 1) / 3;
   // determine the maximum free string possible for the leaves
   l2max      = l2min + l2leaves;
   /* try to combine buddies starting with a buddy size of 1 (i.e. two leaves).
      At a buddy size of 1 two buddy leaves can be combined if both buddies
      have a maximum free of l2min; the combination will result in the
      left-most buddy leaf having a maximum free of l2min+1. After processing
      all buddies for a certain size, process buddies at the next higher buddy
      size (i.e. current size * 2) and the next maximum free (current free+1).
      This continues until the maximum possible buddy combination yields
      maximum free */
   for (l2free = l2min, bsize = 1; l2free < l2max; l2free++, bsize = nextb) {
      nextb = bsize << 1;
      for (cp0 = cp+leaf_index, ii = 0; ii < nleaves; ii+=nextb, cp0+=nextb)
         if (*cp0==l2free && cp0[bsize]==l2free) {
            *cp0 = l2free + 1;
            cp0[bsize] = -1;
         }
   }
   /* With the leaves reflecting maximum free values bubble this information up
      the tree. Starting at the leaf node level, the four nodes described by
      the higher level parent node are compared for a maximum free and this
      maximum becomes the value of the parent node.  All lower level nodes are
      processed in this fashion then we move up to the next level (parent
      becomes a lower level node) and continue the process for that level. */
   for (leaf=leaf_index, nlevel=nleaves>>2; nlevel>0; nlevel>>=2, leaf=nextp) {
      nextp = leaf - 1 >> 2;
      /* Process all lower level nodes at this level setting the value of the
         parent node as the maximum of the four lower level nodes */
      for (cp0=cp+leaf, cp1=cp+nextp, ii=0; ii<nlevel; ii++, cp0+=4, cp1++)
         *cp1 = TREEMAX(cp0);
   }
   return *cp;
}

/** Initialize one dmap page.
   Blocks which don't actually exist in the volume will be marked as
   allocated in the dmap page. The total number of blocks will only account
   for the actually existing blocks.

   @param cp            pointer to wmap or pmap word.
   @param page          pointer to map page
   @param size          number of blocks in this page */
static void idmap_page(dmap_t *page, u32t size) {
   u32t     ii,
        nwords = size >> L2DBWORD,
           bit = size%DBWORD;
   page->nblocks = page->nfree = size;
   // partial word allocated
   if (bit) {
      page->wmap[nwords] = page->pmap[nwords] = FFFF >> bit;
      nwords++;
   }
   // set the rest of the words in the page to 0xFF
   for (ii=nwords; ii<LPERDMAP; ii++) {
      page->pmap[ii] = FFFF;
      page->wmap[ii] = FFFF;
   }
}

/** Fill in rest of dmap fields from wmap/pmap already initialized.
   @param dp            dmap to complete
   @param start         starting block number for this dmap
   @param [out] tmax    filled in with max free for this dmap */
static void complete_dmap(dmap_t *dp, u32t start, s8t *tmax) {
   dmaptree_t *tp;
   s8t        *cp;
   u32t        ii;

   dp->start    = start;
   tp           = &dp->tree;
   tp->height   = 4;
   tp->leafidx  = LEAFIND;
   tp->nleafs   = LPERDMAP;
   tp->l2nleafs = L2LPERDMAP;
   tp->budmin   = BUDMIN;
   // pick up the pointer to the first leaf of the dmap tree.
   cp = tp->stree + tp->leafidx;
   /* set the initial state for the leaves of the dmap tree,
      they will reflect the current allocation state of the wmap words */
   for (ii=0; ii<LPERDMAP; ii++) cp[ii] = maxbuddy((u8t*)&dp->wmap[ii]);
   // with the leaves of the dmap initialized initialize the dmap's tree
   *tmax = adjtree(tp->stree, L2LPERDMAP, BUDMIN);
}

/** initialize a dmap for the specified block range
   @param di            volume information
   @param pmap          map state/data buffer
   @param start         starting block number for this dmap
   @param size          number of blocks in this dmap
   @param loc           volume byte offset of where this dmap should live
   @return error code or 0 */
static qserr initdmap(disk_volume_data *di, mapdata *pmap, u32t start,
                      u32t size, s8t *treemax, u64t loc)
{
   qserr rc = 0;
   // determine if the dmap already exists
   if (pmap->maps[pmap->curdmap]==0) {
      if (size==BPERDMAP) {
         // alloc/init a template empty full page buffer
         if (!pmap->emptypage) {
            pmap->emptypage = calloc_th(sizeof(dmap_t),1);
            if (!pmap->emptypage) return E_SYS_NOMEM;
            idmap_page(pmap->emptypage, size);
            complete_dmap(pmap->emptypage, start, treemax);
         } else {
            // customize/reuse the template empty page
            pmap->emptypage->start = start;
            *treemax = pmap->emptypage->tree.stree[0];
         }
         pmap->maps[pmap->curdmap] = pmap->emptypage;
      } else {
         // alloc/init a special dmap page with the correct size
         pmap->maps[pmap->curdmap] = calloc_th(sizeof(dmap_t),1);
         if (!pmap->maps[pmap->curdmap]) return E_SYS_NOMEM;

         idmap_page(pmap->maps[pmap->curdmap], size);
         complete_dmap(pmap->maps[pmap->curdmap], start, treemax);
      }
   } else
      // fill in rest of fields of special existing dmap page
      complete_dmap(pmap->maps[pmap->curdmap], start, treemax);
   // write the dmap page and free it if special buffer
   rc = j_io(di, JIO_BO|JIO_BS, loc, PSIZE, pmap->maps[pmap->curdmap]);
   if (rc) return rc;

   if (pmap->maps[pmap->curdmap] != pmap->emptypage) {
      free(pmap->maps[pmap->curdmap]);
      pmap->maps[pmap->curdmap] = 0;
   }
   pmap->curdmap++;
   return rc;
}

/** initialize a dmapctl for the specified block range.
   Initialize a dmapctl for the range of (start..start+size-1) and
   level and initialize all dmapctls and dmaps under this dmapctl.

   This routine is called recursively. On first invocation it is called for
   the top level dmapctl of the tree. For each leaf of the dmapctl, the lower
   level dmap (level == 0) or dmapctl (level > 0) is created for the block
   range covered by the leaf and the leaf is set to the maximum free string
   found in the lower level object. If the lower level object is a dmap,
   initdmap is called to handle it's initialization. Otherwise, initctl is
   called recursively to initialize the lower level dmapctl with the level
   specified as the current level - 1; once all leaves have been initialized
   adjtree is called to combine all appropriate buddies and update the higher
   level of the tree to reflect the result of the buddy combination.
   The maximum free string of the dmapctl (i.e. the root value of the tree)
   is returned in treemax.

   @param di            volume information
   @param pmap          map state/data buffer
   @param start         starting disk block number to be covered by this dmapctl
   @param size          number of blocks covered by this dmapctl
   @param level         the level of this dmapctl
   @param [out] treemax maximum free string found in this dmapctl
   @param [in,out] loc  volume byte offset of where this page should live
   @return error code or 0 */
static qserr initctl(disk_volume_data *di, mapdata *pmap, u32t start, u32t size,
                     u32t level, s8t *treemax, u64t *loc)
{
   qserr       rc = 0;
   u32t        ii, l2cblks, nchild;
   s8t        *cp, max;
   u64t     cblks, nb, nextpage;
   // alloc/init current level dmapctl page buffer
   dmapctl_t *dcp = calloc_th(sizeof(dmapctl_t),1);
   if (!dcp) return E_SYS_NOMEM;

   dcp->height   = 5;
   dcp->leafidx  = CTLLEAFIND;
   dcp->nleafs   = LPERCTL;
   dcp->l2nleafs = L2LPERCTL;
   dcp->budmin   = L2BPERDMAP + level * L2LPERCTL;
   // pick up the pointer to the first leaf of the dmapctl tree */
   cp = dcp->stree + dcp->leafidx;

   /* determine how many lower level dmapctls or dmaps will be described by
      this dmapctl based upon the number of blocks covered by this dmapctl */
   l2cblks  = L2BPERDMAP + level * L2LPERCTL;
   cblks    = 1 << l2cblks;
   nchild   = size >> l2cblks;
   nchild   = (size & (cblks-1)) ? nchild+1 : nchild;
   nextpage = *loc + PSIZE;

   for (ii=0; ii<nchild; ii++, size-=nb, start+=nb) {
      // determine how many blocks the lower level dmapctl or dmap will cover
      nb = min(cblks, size);
      /* if this is a level 0 dmapctl, initialize the dmap for the block range
         (i.e. start thru start+nb-1).  Otherwise, initialize the lower level
         dmapctl for this block range.  in either case, the pointer to the
         leaf covering this block range is passed down and will be set to the
         length of the maximum free string of blocks found at the lower level */
      if (level==0) {
         rc += initdmap(di, pmap, start, nb, cp+ii, nextpage);
         nextpage += PSIZE;
      } else
         rc += initctl(di, pmap, start, nb, level-1, cp+ii, &nextpage);
   }
   /* Initialize the leaves for this dmapctl that were not covered by the
      specified input block range (i.e. the leaves have no low level dmapctl
      or dmap */
   for (; ii<LPERCTL; ii++) cp[ii] = NOFREE;
   // with the leaves initialized, adjust the tree for this dmapctl.
   max = adjtree(dcp->stree, L2LPERCTL, l2cblks);
   // write and release the dmapctl page
   rc  = j_io(di, JIO_BO|JIO_BS, *loc, PSIZE, dcp);
   free(dcp);
   // set the treemax return value with the maximum free described by this dmapctl.
   *treemax = max;
   *loc     = nextpage;
   return rc;
}

qserr j_write_bm(disk_volume_data *di, mapdata *pmap, u32t nclus) {
   u8t   secshift = bsf32(di->SectorSize);
   u32t     level;
   u64t     start;
   qserr       rc;
   /* At this point all of the dmaps have been initialized except for their
      trees.  Now we need to build the other levels of the map and adjust the
      tree for each of the dmaps */
   pmap->curdmap     = 0;
   pmap->cpage.maxag = pmap->lastalloc / pmap->cpage.agsize;

   /* the bmap control page is created. Next, the number dmapctl level
      required to described the volume size is determined. initctl()
      is then called to initialize the appropriate dmapctl levels and
      corresponding dmaps. */

   /* get the level for the actual top dmapctl for the volume and
      its physical address */
   level = BMAPSZTOLEV(nclus);
   start = JFS_BMAP_OFF + PSIZE + PSIZE * (2 - level);

   /* initialize only the dmapctls and the dmaps they describe
      that covers the actual volume size */
   rc = initctl(di, pmap, 0, nclus, level, &pmap->cpage.maxfreebud, &start);
   if (rc) return rc;

   // write the control page
   rc = j_io(di, JIO_BO|JIO_BS, JFS_BMAP_OFF, PSIZE, &pmap->cpage);

   return rc;
}

/** allocate the specified number of blocks
   @param       pmap    map state/data buffer
   @param       xlen    number of clusters to allocate (power of 2)
   @param [out] xaddr   starting clusters of allocated chain
   @return error code or 0 */
static qserr bmalloc(mapdata *pmap, u32t xlen, u32t *xaddr) {
   u32t       page, word, bitno,
              l2nb = bsr32(xlen);
   dmap_t      *pg;
   /* searching for a contiguous extent of xlen blocks. Since we may have bad
      blocks intermixed we can`t just take blocks starting at the last block
      allocated. However, lastalloc won`t be updated with bad blocks, so it`s
      a hint to start looking. */
   for (page = pmap->lastalloc >> L2BPERDMAP,
        word = (pmap->lastalloc & BPERDMAP - 1) >> L2DBWORD;
        page < pmap->nmaps; page++, word = 0)
   {
      // if page is not allocated yet then all blocks in it are free
      if (!pmap->maps[page]) {
         *xaddr = page << L2BPERDMAP;
         return j_mark(pmap, *xaddr, MARK_ALLOC|xlen);
      }
      /* check for free blocks starting with "word" to the end of this page,
         then go to the next page if failed here */
      pg = pmap->maps[page];
     
      for (; word<LPERDMAP; word++) {
         // determine if the leaf describes sufficient free space
         s8t leafw = maxbuddy((u8t*)&pg->wmap[word]);
         if (leafw < l2nb) continue;
         // this word has sufficient free space, find it and allocate it
         *xaddr = (page<<L2BPERDMAP) + (word<<L2DBWORD);
         
         if (leafw<BUDMIN) {
            u32t mask = FFFF << DBWORD-xlen, bitno,
                 cmap = ~(pg->wmap[word]);
            // scan the word for xlen free bits
            for (bitno=0; mask; bitno++, mask>>=1)
               if ((mask&cmap)==mask) break;
            *xaddr += bitno;
         }
         // allocate the blocks
         return j_mark(pmap, *xaddr, MARK_ALLOC|xlen);
      }
   }
   return E_SYS_SOFTFAULT;
}

/** split full root of xtree.
   @param pmap          map state/data buffer
   @param ino           inode of xtree
   @param offset        offset of extent to add (in clusters)
   @param blk           starting cluster of extent to add
   @param nblocks       number of clusters for extent to add
   @param xroot         root of xtree
   @return error code or 0 */
static qserr xtSplitRoot(mapdata *pmap, jfs_dinode *ino, u32t offset,
                         u32t blk, u32t nblocks, xtlist *xroot)
{
   jfs_xtreepage *root, *npage;
   u32t           next, xlen, xaddr;
   jfs_xad        *xad;
   qserr            rc;
   xtlist         *buf;
   // allocate and initialize buffer for new page to accomodate the split
   buf = malloc_th(sizeof(xtlist));
   if (!buf) return E_SYS_NOMEM;

   buf->prev = xroot;
   if (!xroot->next) pmap->bbpages = buf;
      else xroot->next->prev = buf;

   buf->dyn    = 1;
   buf->next   = xroot->next;
   xroot->next = buf;
   buf->page   = calloc_th(PSIZE,1);
   npage       = buf->page;
   if (!npage) return E_SYS_NOMEM;

   // allocate disk blocks for new page
   xlen = 1 << pmap->cpage.l2nbperpage;
   rc   = bmalloc(pmap, xlen, &xaddr);
   if (rc) return rc;
#if MKJFS_DBLEVEL>0
   log_it(3, "split root to %u (%u) ...\n", xaddr, xlen);
#endif
   root = xroot->page;
   // initialize new page
   npage->header.flag = root->header.flag & BT_LEAF ? BT_LEAF : BT_INTERNAL;
   pxd_len_set (&(npage->header.self), xlen);
   pxd_addr_set(&(npage->header.self), xaddr);

   npage->header.nextindex = XTENTRYSTART;
   npage->header.maxentry  = PSIZE >> L2XTSLOTSIZE;
   // initialize sibling pointers
   npage->header.next = 0;
   npage->header.prev = 0;
   // copy the in-line root page into new right page extent
   next = root->header.maxentry;
   memcpy(&npage->xad[XTENTRYSTART], &root->xad[XTENTRYSTART],
          (next - XTENTRYSTART) << L2XTSLOTSIZE);
   // insert the new entry into the new right/child page
   xad = &npage->xad[next];
   xad_ofs_set (xad, offset);
   xad_len_set (xad, nblocks);
   xad_addr_set(xad, blk);
   // update page header
   npage->header.nextindex = next + 1;
   // init root with the single entry for the new right page
   xad = &root->xad[XTENTRYSTART];
   xad_ofs_set (xad, 0);
   xad_len_set (xad, xlen);
   xad_addr_set(xad, xaddr);
   // update page header of root
   root->header.flag &= ~BT_LEAF;
   root->header.flag |= BT_INTERNAL;
   root->header.nextindex = XTENTRYSTART + 1;
   // update nblocks for inode to account for new page
   ino->di_nblocks += xlen;

   return 0;
}

static qserr xtAppend(disk_volume_data *di, mapdata *pmap, u8t clshift,
                      jfs_dinode *ino, u32t offset, u32t blk, u32t nblocks,
                      xtlist *xpage);

/** split non-root page of xtree.
   @param di            volume information
   @param pmap          map state/data buffer
   @param clshift       bit shift of cluster size
   @param ino           inode of xtree splitting
   @param offset        offset of extent to add (in clusters)
   @param blk           starting cluster of extent to add
   @param nblocks       number of clusters of new extent to add
   @param xpage         page to split
   @return error code or 0 */
static qserr xtSplitPage(disk_volume_data *di, mapdata *pmap, u8t clshift,
                         jfs_dinode *ino, u32t offset, u32t blk, u32t nblocks,
                         xtlist *xpage)
{
   jfs_xtreepage   *pg;
   qserr            rc = 0;
   u32t          xaddr;  // new right page block number
   jfs_xad        *xad;
   u32t           ppos;
   u32t           xlen = 1 << pmap->cpage.l2nbperpage;
   // allocate disk space for the new xtree page
   rc = bmalloc(pmap, xlen, &xaddr);
   if (rc) return rc;

   /* modify xpage's next entry to point to the new disk space,
      write the xpage to disk since we won`t be needing it anymore */
   pg   = xpage->page;
   pg->header.next = xaddr;
   ppos = pxd_addr(&(pg->header.self));
   rc   = j_io(di, JIO_BS|JIO_BO, (u64t)ppos<<clshift, PSIZE, pg);
   if (rc) return rc;
#if MKJFS_DBLEVEL>0
   log_it(3, "split page to %u ...\n", ppos);
#endif
   // since done with the xpage, we can re-use this buffer
   pxd_len_set (&(pg->header.self), xlen);
   pxd_addr_set(&(pg->header.self), xaddr);
   pg->header.flag = pg->header.flag & BT_TYPE;
   // initialize sibling pointers
   pg->header.next = 0;
   pg->header.prev = ppos;
   // insert entry at the first entry of the new right page
   xad = &pg->xad[XTENTRYSTART];
   xad_ofs_set (xad, offset);
   xad_len_set (xad, nblocks);
   xad_addr_set(xad, blk);

   pg->header.nextindex = XTENTRYSTART + 1;
   // now append new page to parent page
   rc = xtAppend(di, pmap, clshift, ino, offset, xaddr, xlen, xpage->prev);
   // update inode to account for new page
   ino->di_nblocks += xlen;

   return rc;
}

/** append an extent to the specified file.
   xpage points to its parent in the xtree and its rightmost child (if it
   has one). It also points to the buffer for the page.

   @param di            volume information
   @param pmap          map state/data buffer
   @param clshift       bit shift of cluster size
   @param ino           inode to add extent to
   @param offset        offset of extent to add (in clusters)
   @param blk           start cluster of extent to add
   @param nblocks       number of clusters in extent to add
   @param xpage         xtree page to add extent to
   @return error code or 0 */
static qserr xtAppend(disk_volume_data *di, mapdata *pmap, u8t clshift,
                      jfs_dinode *ino, u32t offset, u32t blk, u32t nblocks,
                      xtlist *xpage)
{
   jfs_xtreepage *cpage = xpage->page;
   int    index = cpage->header.nextindex;
   qserr     rc = 0;

   // insert entry for new extent
   if (index == cpage->header.maxentry) {
     // no room in this page
     if (cpage->header.flag & BT_ROOT)
       // this is the root of the xtree
       rc = xtSplitRoot(pmap, ino, offset, blk, nblocks, xpage);
     else
       /* non-root page: add new page at this level, xtSplitPage() calls
          xtAppend again to propogate up the new page entry */
       rc = xtSplitPage(di, pmap, clshift, ino, offset, blk, nblocks, xpage);
   } else {
      // there is room to add the entry to this page
      jfs_xad *xad = &cpage->xad[index];
      xad_ofs_set (xad, offset);
      xad_len_set (xad, nblocks);
      xad_addr_set(xad, blk);
      // next available entry index
      cpage->header.nextindex++;
   }
   return rc;
}

qserr j_write_bb(disk_volume_data *di, mapdata *pmap, u8t clshift,
                 u32t secvtab, jfs_dinode *bb_inode, dd_list bclist)
{
   int   commit = 0;
   xtlist   *pg;
   u32t     idx, blsize;
   qserr     rc = 0;
   // initialize list of xtree append buffers
   pmap->bbpages = malloc_th(sizeof(xtlist));
   pmap->bbpages->prev = 0;
   pmap->bbpages->next = 0;
   pmap->bbpages->page = (jfs_xtreepage*)&bb_inode->xtroot;
   pmap->bbpages->dyn  = 0;

   blsize = bclist->count();
   // first mark all bad clusters as used and only then allocate bad block inode
   for (idx = 0; idx<blsize; idx++)
      if ((rc = j_mark(pmap, bclist->value(idx), MARK_BAD))) break;
   if (!rc)
      for (idx = 0; idx<blsize; idx++) {
         // add an extent for this bad block to the inode`s tree
         rc = xtAppend(di, pmap, clshift, bb_inode, bb_inode->di_size >> clshift,
                       bclist->value(idx), 1, pmap->bbpages);
         if (!rc) {
            bb_inode->di_size += 1 << clshift;
            bb_inode->di_nblocks++;
         } else
            break;
         // flag to update below
         commit = 1;
      }
   // need to update BADBLOCK_I inode
   if (!rc && commit) {
      // flush pages to disk
      xtlist *curpage = pmap->bbpages, *ppage;

      while (!(curpage->page->header.flag & BT_ROOT)) {
         u32t blk = pxd_addr(&(curpage->page->header.self));

         rc = j_io(di, JIO_BS|JIO_BO, (u64t)blk<<clshift, PSIZE, curpage->page);
         if (rc) break;
         curpage = curpage->prev;
      }
      // write the inode itself
      if (!rc) rc = j_rwinode(di, bb_inode, BADBLOCK_I, 1, secvtab, clshift);
   }
   // free buffers
   pg = pmap->bbpages;
   while (pg) {
      xtlist *ppg = pg->prev;
      if (pg->page && pg->dyn) free(pg->page);
      free(pg);
      pg = ppg;
   }
   pmap->bbpages = 0;
   return rc;
}
