//
// QSINIT "start" module
// external shell commands
//
#include "qsbase.h"
#include "stdlib.h"
#include "qs_rt.h"
#include "errno.h"
#include "vioext.h"
#include "syslocal.h"
#include "time.h"
#include "qsint.h"
#include "qsdm.h"
#include "stdarg.h"
#include "qcl/qshlp.h"
// delete of a pointer to void message
#pragma disable_message (919);

#define SPACES_IN_WIDE_MODE       4      ///< spaces between names in wide "dir" mode
#define PRNSEQ_NP        (PRNSEQ_INITNOPAUSE^PRNSEQ_INIT)

void cmd_shellerr(u32t errtype, int errorcode, const char *prefix) {
   if (!prefix) prefix = "\r";

   char *msg = cmd_shellerrmsg(errtype, errorcode);
   if (!msg)
      msg = sprintf_dyn(errtype?"Error code %d.":"Error code %X.", errorcode);
   // enable ANSI while help message printing...
   u32t state = vio_setansi(1);
   printf("%s%s\n", prefix, msg);
   free(msg);
   if (!state) vio_setansi(0);
}

char* _std cmd_shellerrmsg(u32t errtype, int errorcode) {
   char msgn[16];
   snprintf(msgn, 16, errtype?"_C_%05d":"_E_%06X", errorcode);
   return cmd_shellgetmsg(msgn);
}

static int pause_check(int init_counter) {
   u32t  pp_lines,
        pp_lstart = mt_tlsget(QTLS_SHLINE),
         pp_flags = mt_tlsget(QTLS_SHFLAGS);
   vio_getmodefast(0, &pp_lines);

   if (init_counter>=0 && (pp_flags&PRNSEQ_NP)==0 && 
      (int)(vio_ttylines(-1)-pp_lstart)>=(int)pp_lines-1) 
   {
      u16t col = vio_getcolor();
      vio_setcolor(VIO_COLOR_GRAY);
      vio_strout("Press any key to continue...");
      vio_setcolor(col);

      u16t key = key_read();
      // exit by ESC
      if ((key&0xFF)==27) { vio_charout('\n'); return 1; }
      // text with original color attribute
      vio_strout("\r                            \r");
      mt_tlsset(QTLS_SHLINE, vio_ttylines(-1));
   }
   return 0;
}

/** print command`s output.
    @param line          string to print, without \\n
    @param init_counter  <0 - no pause, 1 - init counter, 3 - copy to log
                         all strings until next init, 0 - print with pause,
                         bit 2 - force no pause until next init
    @param color         line color, 0 - default
    @return 1 if ESC was pressed in pause. */
static u32t pause_println(const char *line, int init_counter=0, u8t color=0) {
   // begin of sequence
   if (init_counter>0) {
      mt_tlsset(QTLS_SHLINE, vio_ttylines(-1));
      mt_tlsset(QTLS_SHFLAGS, init_counter);
   }
   if (!line) return 0;
   u32t  pp_flags = mt_tlsget(QTLS_SHFLAGS);
   if (pp_flags&PRNSEQ_LOGCOPY) log_it(2,"echo: %s\n", line);

   FILE     *stdo = get_stdout();
   int    p_check = init_counter>=0 && (pp_flags&PRNSEQ_NP)==0 && isatty(fileno(stdo));
   if (p_check) {
      const char *lp = line;
      while (*lp) {
         char *eolp = strchr(lp, '\n');
         int    len = eolp?eolp-lp+1:strlen(lp);

         if (pause_check(init_counter)) return 1;

         if (color) vio_setcolor(color);
         fwrite(lp, 1, len, stdo);
         lp += len;
      }
   } else {
      if (color) vio_setcolor(color);
      fputs(line, stdo);
   }
   if (color) vio_setcolor(VIO_COLOR_RESET);
   fputc('\n', stdo);
   return 0;
}

int __cdecl cmd_printf(const char *fmt, ...) {
   if (!fmt) return 0;
   va_list  argptr;
   va_start(argptr, fmt);
   char *outstr = vsprintf_dyn(fmt, argptr);
   va_end(argptr);
   int    prnlen = strlen(outstr);
   FILE    *stdo = get_stdout();
   u32t pp_flags = mt_tlsget(QTLS_SHFLAGS);
   fputs(outstr, stdo);
   if (pp_flags&PRNSEQ_LOGCOPY) log_it(2,"%s", outstr);
   free(outstr);
   if ((pp_flags&PRNSEQ_NP)==0 && isatty(fileno(stdo))) pause_check(0);
   return prnlen;
}

u32t __cdecl cmd_printseq(const char *fmt, int flags, u8t color, ...) {
   if (!fmt) return pause_println(0, flags, color);
   va_list argptr;
   va_start(argptr, color);
   char *outstr = vsprintf_dyn(fmt, argptr);
   va_end(argptr);
   int   prnlen = strlen(outstr);
   u32t     res = pause_println(outstr, flags, color);
   free(outstr);
   return res;
}

static void qsfwerr2str(spstr &msg, const spstrw &fname, qserr err) {
   char *emsg = cmd_shellerrmsg(EMSG_QS, err);
   if (!emsg) emsg = sprintf_dyn("Error %05X", err);
   msg.sprintf("%S : %s", fname(), emsg);
   free(emsg);
}

int ask_yn(int allow_all, int allow_skip) {
   u16t chv = 0;
   // reset pause counter
   mt_tlsset(QTLS_SHLINE, vio_ttylines(-1));
   while (1) {
      chv = key_read();
      char cch = toupper(chv&0xFF);
      if (allow_skip && cch=='S') return 3;
      if (allow_all && cch=='A') return 2;
      if (cch=='Y') return 1;
      if (cch=='N') return 0;
      if (cch==27)  return -1;
   }
}

int ask_overwrite(const spstr &fname) {
   ansi_push_set color(1);
   printf(ANSI_RESET "\rOverwrite %.44s ("
      ANSI_WHITE "y" ANSI_RESET "/"
      ANSI_WHITE "n" ANSI_RESET "/"
      ANSI_WHITE "a" ANSI_RESET "ll/"
      ANSI_WHITE "s" ANSI_RESET "kip all/"
      ANSI_WHITE "e" ANSI_RESET "sc)?", fname());
   return ask_yn(1,1);
}

int ask_overwritew(const spstrw &fname) {
   ansi_push_set color(1);
   printf(ANSI_RESET "\rOverwrite %.44S ("
      ANSI_WHITE "y" ANSI_RESET "/"
      ANSI_WHITE "n" ANSI_RESET "/"
      ANSI_WHITE "a" ANSI_RESET "ll/"
      ANSI_WHITE "s" ANSI_RESET "kip all/"
      ANSI_WHITE "e" ANSI_RESET "sc)?", fname());
   return ask_yn(1,1);
}

static void process_args_common(TPtrStrings &al, u32t flags, char* args,
   short *values, va_list argptr)
{
   char  arg[144], *ap = args;
   int   noslash = flags&SPA_NOSLASH ? 1 : 0;
   // a bit of safeness
   if (strlen(args)>=144) { log_it(3, "too long args!"); return; }

   while (*ap) {
      char *ep = strchr(ap,'|');
      int  *vp = va_arg(argptr,int*);
      if (noslash) arg[0] = '/';
      if (ep) {
         strncpy(arg+noslash, ap, ep-ap);
         arg[ep-ap+noslash] = 0;
         ap = ep+1;
      } else {
         strcpy(arg+noslash, ap);
         ap += strlen(ap);
      }
      int idx = al.IndexOfName(arg);
      if (idx>=0) { *vp = *values; al.Delete(idx); }
      // accept '-key' as well as '/key'
      if (arg[0]=='/' && !(flags&SPA_NODASH)) {
         arg[0] = '-';
         idx = al.IndexOfName(arg);
         if (idx>=0) { *vp = *values; al.Delete(idx); }
      }
      values++;
   }
}

static void args2list(str_list *lst, TPtrStrings &al) {
   str_getstrs(lst,al);
}

void process_args(TPtrStrings &al, u32t flags, char* args, short *values, ...) {
   va_list   argptr;

   va_start(argptr, values);
   process_args_common(al, flags, args, values, argptr);
   va_end(argptr);
}

str_list* str_parseargs(str_list *lst, u32t firstarg, u32t flags, char* args,
   short *values, ...)
{
   va_list   argptr;
   TPtrStrings   al;

   str_getstrs2(lst,al,firstarg);
   va_start(argptr, values);
   process_args_common(al, flags, args, values, argptr);
   va_end(argptr);
   return flags&SPA_RETLIST ? str_getlist_local(al.Str) : 0;
}

#define AllowSet " ^\r\n"
#define CarrySet "-"

static u32t cmd_printtext(TStrings &res, int flags, u8t color) {
   if (flags!=0 && flags!=-1 && (flags&PRNSEQ_INIT)) {
      pause_println(0, flags);
      flags = flags&PRNSEQ_NP?-1:0;
   }
   spstr out = res.GetTextToStr("\n");
   if (out.lastchar()=='\n') out.dellast();
   if (pause_println(out(), flags, color)) return 1;
   return 0;
}

u32t _std cmd_printtext(const char *text, int flags, u8t color) {
   if (!text) return 0;
   TStrings  res;
   u32t   dX, ln;
   int        np = flags==-1,
        helpmode = !np && (flags&PRNTXT_HELPMODE);
   /* "flags" is a bit crazy arg: -1 here is a special value, but bit flags are
      used too. Dumb way, but pause_println() used too often to change it */
   if (helpmode) flags &= ~PRNTXT_HELPMODE;

   vio_getmode(&dX,0);

   splittext(text, dX, res, helpmode?SplitText_HelpMode:0);
   return cmd_printtext(res, flags, color);
}

qserr _std cmd_printfile(qshandle handle, long len, int flags, u8t color) {
   if (!len) return E_SYS_INVPARM;
   // query file first
   u64t  fsz;
   qserr res = io_size(handle, &fsz);
   if (res) return res;
   u64t fpos = io_pos(handle);
   if (fpos==FFFF64) return io_lasterror(handle);
   // the same note as in cmd_printtext() above
   int split = flags!=-1 && (flags&PRNFIL_SPLITTEXT);
   if (split) flags &= ~PRNFIL_SPLITTEXT;
   // validate data length
   if (len<0) len = fsz-fpos>x7FFF ? x7FFF : fsz - fpos; else
      if (fsz-fpos < len) len = fsz-fpos;
   if (!len) return E_SYS_EOF;

   char *txbuf = (char*)hlp_memalloc(len+1, QSMA_RETERR|QSMA_NOCLEAR|QSMA_LOCAL);
   if (!txbuf) return E_SYS_NOMEM;

   if (io_read(handle, txbuf, len) < len) res = E_DSK_ERRREAD; else {
      if (txbuf[len-1]=='\n') len--;
      if (len && txbuf[len-1]=='\r') len--;
      if (len) {
         txbuf[len] = 0;
         if (split) {
            if (cmd_printtext(txbuf, flags, color)) res = E_SYS_UBREAK;
         } else {
            if (flags!=0 && flags!=-1 && (flags&PRNSEQ_INIT)) {
               pause_println(0, flags);
               flags = flags&PRNSEQ_NP?-1:0;
            }
            if (pause_println(txbuf, flags, color)) res = E_SYS_UBREAK;
         }
      }
   }
   hlp_memfree(txbuf);
   return res;
}


static void _std freadfull_callback(u32t percent, u32t ready) {
   char *cp = (char*)mt_tlsget(QTLS_SFINT1);
   if (cp) {
      u32t cols = 80,
            len = strlen(cp);
      vio_getmode(&cols, 0);
      // we need to cut long file name (once per load)
      if (len>cols-24) {
         char *dp = cp+len-cols+24+3;
         cp[0]='.'; cp[1]='.'; cp[2]='.';
         memmove(cp+3, dp, strlen(dp)+1);
         // update it
         len = strlen(cp);
         cp[len+1] = 0;
      }
      if (percent) {
         printf("\rcaching %s ... %02u%% ", cp, percent);
         // signal about "\n" in the caller
         cp[len+1] = 1;
      } else
      if ((ready & _1MB-1)==0) {
         printf("\rcaching %s ... %u Mb ", cp, ready>>20);
         cp[len+1] = 1;
      }
   }
}


void* hlp_freadfull_progress(const char *name, u32t *bufsize) {
   u32t     nlen = strlen(name);
   if (!nlen) {
      if (bufsize) *bufsize = 0;
      return 0;
   }
   char *srcname = (char*)malloc_th(nlen + 2);
   strcpy(srcname, name);
   // zero flag
   srcname[nlen+1] = 0;
   mt_tlsset(QTLS_SFINT1, (u32t)srcname);
   void *ptr = hlp_freadfull(name, bufsize, freadfull_callback);
   // printf was used in callback at least once
   if (srcname[nlen+1]) printf("\n");
   free(srcname);
   return ptr;
}

/// convert source ascii or utf-8 string to utf-16
qserr get_w_arg(const spstr &src, spstrw &dst, int utf8) {
   qserr qserr = 0;
   c16t  *dstw = utf8 ? 0 : hlp_atow(src(), 0, 0, &qserr);
   dst = dstw;
   if (dstw) free(dstw);
   if (utf8)
      if (!dst.setutf8(src)) qserr = E_SYS_UTFSURR;
   return qserr;
}

u32t _std shl_copy(const char *cmd, str_list *args) {
   char  *errstr = 0;
   int     quiet = 0, rc = -1;
   u32t  errtype = EMSG_CLIB;

   if (args->count>0) {
      int frombp = 0, cpattr = 0, beep = 0, ovwmode = 0, pause = 0, utf8 = 0;

      TPtrStrings al;
      str_getstrs(args,al);
      // is help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "boot|q|a|beep|o|p|c|u";
      static short argval[] = {   1,1,1,   1,1,1,2,1};
      process_args(al, SPA_NOSLASH, argstr, argval, &frombp, &quiet, &cpattr,
                   &beep, &ovwmode, &pause, &ovwmode, &utf8);

      al.TrimEmptyLines();
      if (al.Count()>=2) {
         pause_println(0, pause?PRNSEQ_INIT:PRNSEQ_INITNOPAUSE);
         // dst is ascii in the current codepage or utf8 string
         spstrw    dst;
         qserr    qerr = get_w_arg(al[al.Max()], dst, utf8);
         int  dst_mask = dst.cpos('?')>=0 || dst.cpos('*')>=0;
         al.Delete(al.Max());

         if (qerr) {
            errtype = EMSG_QS;
            rc      = qerr;
         } else
         if (al[al.Max()].lastchar()!='+') {
            // must catch all combination of "+", quotes and spaces (i hope ;)
            spstr merged = al.GetTextToStr("\x01");
            int   is_sum = al.ParseCmdLine(merged.replace("\x01",""),'+') > 1;
            al.TrimEmptyLines();
            if (al.Count()>0) {
               u8t     *is_boot = (u8t*)malloc_th(al.Count()),
                       *is_mask = (u8t*)malloc_th(al.Count());
               u32t     bt_type = hlp_boottype(), opena,
                          bt_vi = frombp? hlp_volinfo(DISK_BOOT, 0) : FST_NOTMOUNTED,
                         sfattr = _A_ARCH;
               timeq_t   sftime = 0,
                         sctime = 0;
               u64t     sumsize = 0;
               spstr    bootdir;

               if (frombp && hlp_hosttype()==QSHT_EFI) {
                  char *eroot = (char*)sto_data(STOKEY_EFIROOT);
                  bootdir = sprintf_dyn("A:%s\\", eroot ? eroot : "\\EFI\\BOOT");
               }
               // zero source pointers/handles (before any possible error)
               for (idx=0; idx<al.Count(); idx++) al.Objects(idx) = 0;
               rc = 0;
               // check that mask is not used for a real boot source
               for (idx=0; idx<al.Count(); idx++) {
                  is_boot[idx] = 0;
                  // source is a wildcard
                  is_mask[idx] = al[idx].cpos('?')>=0 || al[idx].cpos('*')>=0;
                  // normal files can be inserted with full path in /boot mode
                  if (frombp)
                     if (al[idx][1]!=':') {
                        is_boot[idx] = 1;
                        // switch it to normal file i/o if any FS is available ...
                        if (bt_type!=QSBT_PXE && bt_vi>=FST_FAT12 && bt_vi<=FST_OTHER) {
                           int forceroot = al[idx][0]=='/' || al[idx][0]=='\\';
                           // "forceroot" actually should be only for EFI
                           if (!bootdir || forceroot) {
                              if (forceroot) al[idx].del(0);
                              al[idx].insert("A:\\",0);
                           } else
                              al[idx].insert(bootdir,0);
                           is_boot[idx] = 0;
                        }
                     }
                  // drop wildcards for a real boot source
                  if (is_boot[idx] && is_mask[idx]) {
                     errtype = EMSG_QS;
                     rc      = E_SYS_INVNAME;
                     errstr  = sprintf_dyn("%s : ", al[idx]());
                     break;
                  }
               }
               // actual list
               TStringsW actl;
               // check - who is here, get size and wildcard lists
               if (!rc)
               for (idx=0; idx<al.Count(); idx++) {
                  // boot source? wildcards was denied above ...
                  if (is_boot[idx]) {
                     u32t  len;
                     void *ptr = quiet?hlp_freadfull(al[idx](), &len, 0):
                                       hlp_freadfull_progress(al[idx](), &len);
                     if (ptr) {
                        // make sure of size because it used below
                        if (hlp_memgetsize(ptr)!=len)
                           ptr = hlp_memrealloc(ptr, len);
                        if (ptr) {
                           actl.AddObject(al[idx], idx);
                           if (is_sum) sumsize += len;
                           al.Objects(idx) = ptr;
                        }
                     }
                  } else {
                     spstrw fnc;
                     /* utf-16 note: any error printing???
                        this code was intially written on Jan,1 so I`m extremely
                        surprised that it works at all ;) */
                     if (get_w_arg(al[idx], fnc, utf8)) continue;
                     if (!fullpathw(fnc)) continue;

                     if (is_mask[idx]) {
                        spstrw  wdir, wname;
                        splitfnamew(fnc, wdir, wname);
                        // read tree
                        dirw_t *info = 0;
                        u32t    cnt = _dos_readtreew(wdir(),wname(),&info,0,0,0,0);

                        if (cnt) {
                           for (dirw_t *pd=info; pd; pd=pd->d_next)
                              if (!(pd->d_attr & (_A_SUBDIR|_A_VOLID))) {
                                 actl.AddObject(spstrw(pd->d_openpath) +
                                    spstrw(pd->d_name), idx);
                                 // continue to calc the result file size ...
                                 if (is_sum) sumsize += pd->d_size;
                              }
                           _dos_freetree(info);
                        }
                     } else {
                        io_handle_info fi;
                        // check file type
                        if (!io_pathinfow(fnc(),&fi))
                           if (!(fi.attrs & (_A_SUBDIR|_A_VOLID))) {
                              actl.AddObject(fnc, idx);
                              if (is_sum) sumsize += fi.size;
                           }
                     }
                  }
               }
               // was found at least someone?
               if (!rc && actl.Count()==0) rc = ENOENT;
               // 800 bytes !!! (well, cmd.exe stack is 32k)
               wchar_t  fnbuf[QS_MAXPATH + QS_MAXPATH/2];
               spstrw  dstdir, dstname;
               splitfnamew(dst, dstdir, dstname);
               if (dstdir.length() && dstdir.lastchar()!='\\') dstdir+='\\';

               /* fix a case of "copy *.txt + *.doc *.inf" command:
                  get first name as a target */
               if (!rc && dst_mask && is_sum) {
                  spstrw    sdir, sname, mdir, mname, al0;
                  qserr  nameerr = get_w_arg(al[actl.Objects(0)], al0, utf8);
                  if (!nameerr) {
                     splitfnamew(actl[0], sdir, sname);
                     splitfnamew(al0, mdir, mname);
                  }
                  if (nameerr || !_replacepatternw(mname(), dstname(), sname(), fnbuf)) {
                     errtype  = EMSG_QS;
                     rc       = nameerr ? nameerr : E_SYS_INVNAME;
                     errstr   = sprintf_dyn("%S : ", dst());
                  } else {
                     dst_mask = 0;
                     dstname  = fnbuf;
                  }
               }
               /* these indexes do not match the actl array, so delete them to
                  avoid using below */
               free(is_boot); is_boot = 0;
               free(is_mask); is_mask = 0;

               if (!rc) {
                  u32t memsize = _128KB, maxblock;
                  hlp_memavail(&maxblock, 0, QSMA_LOCAL);
                  // 1-4Mb buffer
                  if (maxblock>_64MB) memsize<<=5; else
                     if (maxblock>_32MB) memsize<<=3;
                  void *buf = hlp_memalloc(memsize,QSMA_RETERR|QSMA_LOCAL);
                  if (!buf) rc = ENOMEM;

                  io_handle dstf = 0,
                              ff = 0;
                  spstrw     dfp = dstdir + dstname;
                  u32t   succcnt = 0;
                  qserr  lasterr = 0;
                  // target is directory name
                  if (!dst_mask)
                     if (hlp_isdirw(dfp())) {
                        dstdir = dfp;
                        if (dstdir.lastchar()!='\\') dstdir+='\\';
                        dstname.clear();
                     }

                  for (idx=0; idx<actl.Count() && !rc; idx++) {
                     spstrw  rdir, rname;
                     splitfnamew(actl[idx], rdir, rname);
                     if (rdir.length() && rdir.lastchar()!='\\') rdir+='\\';

                     if (dst_mask) {
                        spstrw mdir, mname, fnx;
                        qserr  ferr = get_w_arg(al[actl.Objects(idx)], fnx, utf8);
                        if (!ferr) splitfnamew(fnx, mdir, mname);

                        if (!ferr && _replacepatternw(mname(), dstname(), rname(), fnbuf)) {
                           dfp = dstdir + spstrw(fnbuf);
                        } else {
                           // replacepattern failed? just ignore this file
                           continue;
                        }
                     } else
                     if (!dstname) dfp = dstdir + rname;

                     void *bootf = al.Objects(actl.Objects(idx));
                     // a lost file after "continue;" below ...
                     if (ff) { io_close(ff); ff = 0; }
                     // not a boot source?
                     if (!bootf) {
                        // open source file
                        rc = io_openw(actl[idx](), IOFM_OPEN_EXISTING|IOFM_READ|
                                IOFM_SHARE_READ|IOFM_SHARE_DEL|IOFM_SHARE_REN,
                                   &ff, &opena);
                        // get time for the target
                        if (cpattr && (dst_mask || !is_sum && !dstname || !sftime)) {
                           io_handle_info fi;

                           if (!io_info(ff, 0, &fi)) {
                              sftime = io_iototimeq(&fi.wtime);
                              sctime = io_iototimeq(&fi.ctime);
                              /* creation time zeroed by OS/2 on FAT and often
                                 can be equal, flag it as missing too in the
                                 second case */
                              if (sftime==sctime) sctime = 0;
                              sfattr = fi.attrs;
                           }
                        }
                        // no source?
                        if (rc) { lasterr = rc; rc = 0; continue; }
                     } else
                        sftime = 0;

                     // create destination
                     if (!dstf) {
                        io_handle_info di;
                        int          doit = 1;
                        u32t      flagset = IOFM_CREATE_ALWAYS|IOFM_WRITE|
                                            IOFM_SHARE_READ|IOFM_CLOSE_DEL;

                        if (io_pathinfow(dfp(), &di)==0) {
                           if (di.attrs & IOFA_DIR) doit = 0; else
                           if (di.attrs&IOFA_DEVICE) {
                              // IOFM_CLOSE_DEL is not acceptable for device
                              flagset &= ~IOFM_CLOSE_DEL;
                              flagset |= IOFM_SHARE_WRITE;
                           } else {
                              doit = ovwmode;

                              if (doit==2) {
                                 doit = ask_overwritew(dfp);
                                 cmd_printf("\n");
                                 if (doit<0) {
                                    rc = EINTR;
                                    break; 
                                 } else
                                 if (doit==2) ovwmode = 1; else  // all
                                    if (doit==3) { ovwmode = 0; doit = 0; } // skip
                              }
                           }
                           if (!doit) lasterr = E_SYS_EXIST;
                        }
                        // skip this target
                        if (!doit) continue;
                        // close on del on destination! will be reset on success
                        rc = io_openw(dfp(), flagset, &dstf, &opena);
                        // allow to see E_SYS_FSLIMIT just on open
                        if (!rc && is_sum) rc = io_setsize(dstf, sumsize);
                        if (rc)
                           // single destination file error
                           if (!dst_mask) {
                              errtype = EMSG_QS;
                              errstr  = sprintf_dyn("%S : ", dfp());
                              break;
                           } else {
                              lasterr = rc; rc = 0;
                              continue;
                           }
                     }
                     // we have both source and destination, can about it ...
                     if (!quiet) cmd_printf("%S ", rname());

                     if (bootf) {
                        u32t len = hlp_memgetsize(bootf);
                        if (len)
                           if (io_write(dstf, bootf, len) != len) {
                              errtype = EMSG_QS;
                              rc = io_lasterror(dstf);
                           }
                     } else {
                        u64t  len, pos = 0;

                        rc = io_size(ff, &len);
                        if (rc) errtype = EMSG_QS;
                           else
                        while (len) {
                           u32t copysize = len<memsize ? len : memsize,
                                  actual = io_read(ff, buf, copysize);

                           if (actual!=copysize) { rc = EIO; break; }
                           actual = io_write(dstf, buf, copysize);

                           if (actual!=copysize) {
                              errtype = EMSG_QS;
                              rc = io_lasterror(dstf);
                           }
                           len-=copysize; pos+=copysize;
                           // allow Ctrl-Esc during very long file copying
                           if (!mt_active()) key_pressed();
                        }
                        io_close(ff); ff = 0;
                     }
                     /* - wildcard destination
                        - directory destination
                        - error
                        - last entry */
                     if (dst_mask || !is_sum && !dstname || rc || idx+1 >= actl.Count()) {
                        if (!rc) {
                           // remove "delete on close" flag
                           io_setstate(dstf, IOFS_DELONCLOSE, 0);
                           rc = io_close(dstf);
                           succcnt++;
                           // creation time can be 0
                           if (!rc && cpattr && sftime) {
                              io_handle_info  hi;
                              hi.attrs = sfattr;
                              io_timeqtoio(&hi.wtime, sftime);
                              io_timeqtoio(&hi.ctime, sctime?sctime:sftime);
                              io_setinfow(dfp(), &hi, IOSI_ATTRS|IOSI_CTIME|IOSI_WTIME);
                           }
                        } else
                           io_close(dstf);
                        dstf = 0;
                     }
                     if (!quiet) cmd_printf("\n");
                  }
                  if (buf) hlp_memfree(buf);
                  if (dstf) io_close(dstf);
                  if (ff) io_close(ff);
                  // no one? use any error from file open or default one
                  if (!rc && !succcnt) {
                     errtype = EMSG_QS;
                     rc      = lasterr ? lasterr : E_SYS_NOFILE;
                  }
               }
               for (idx=0; idx<al.Count(); idx++)
                  if (al.Objects(idx)) hlp_memfree(al.Objects(idx));
            }
         }
         if (rc>0) log_it(3, "copy err %X, dst \"%S\"\n", rc, dst());
         if (beep) {
            vio_beep(rc>0?200:700,40);
            while (vio_beepactive()) usleep(5000);
            vio_beep(rc>0?100:600,40);
         }
      }
   }
   if (rc<0) rc = EINVAL;
   if (!quiet && rc && rc!=EINTR) cmd_shellerr(errtype, rc, errstr);
   if (errstr) free(errstr);
   // convert error code to common errno
   if (rc && errtype==EMSG_QS) rc = qserr2errno(rc);
   return rc;
}

u32t _std shl_type(const char *cmd, str_list *args) {
   int rc=-1;
   if (args->count>0) {
      int frombp=0, nopause=0, igneof=0;

      TPtrStrings al;
      str_getstrs(args,al);
      // is help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "boot|np|ie";
      static short argval[] = {   1, 1, 1};
      process_args(al, SPA_NOSLASH, argstr, argval, &frombp, &nopause, &igneof);

      al.TrimEmptyLines();
      if (al.Count()>=1) {
         u32t bt_vi = frombp? hlp_volinfo(DISK_BOOT, 0) : FST_NOTMOUNTED;
         // init "paused" output
         pause_println(0, nopause?PRNSEQ_INITNOPAUSE:PRNSEQ_INIT);
         // print files
         for (idx=0; idx<al.Count(); idx++) {
            spstr name(al[idx]);
            u32t  is_boot = 0, f_size = 0;
            char *f_data = 0;

            if (frombp)
               if (name[1]!=':' && name.cpos('\\')<0 && name.cpos('/')<0)
                  is_boot = 1;
            if (is_boot)
               f_data = (char*)hlp_freadfull(name(), &f_size, 0);
            // non-boot source
            if (!is_boot) f_data = (char*)freadfull(name(), &f_size);
            if (!f_data) { rc=ENOENT; break; }
            // end at \x1A by default
            if (f_size && !igneof) {
               char *eofpos = (char*)memchr(f_data, 0x1A, f_size);
               if (eofpos) {
                  *eofpos = 0;
                  f_size  = eofpos - f_data;
               }
            }
            if (al.Count()>1) {
               name.insert("** ",0);
               name+=" **";
               pause_println(name(), 0, 0x0A);
            }
            TPtrStrings ft;
            ft.SetText(f_data, f_size);
            hlp_memfree(f_data);
            for (u32t ii=0;ii<ft.Count();ii++)
               if (pause_println(ft[ii](),0)) return EINTR;
         }
         if (rc<0) rc=0;
      }
   }
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return rc;
}

// -------------------------------------------------------------------------
//  dir

struct dir_cbinfo {
   int    dir_npmode;
   u64t    sizetotal; // size in kb
   u32t     file_cnt;
   u32t      dir_cnt;
   int     check_esc;
   int      altnames;
   int      widemode;
   int      shortfmt;
   int       utfmode;
   u32t     lcounter; // counter for esc check
   int        ubreak; // user breaks this
};

static void readtree_prn(dir_cbinfo *cbi, const char *str) {
   int rc = pause_println(str, cbi->dir_npmode)?-1:1;
   // check for ESC every 1024 files in non-pause mode
   if (cbi->dir_npmode<0 && cbi->check_esc && (++cbi->lcounter&1023)==0) {
      u16t key = key_wait(0);
      if ((key&0xFF)==27) rc = -1;
   }
   // flag user break
   if (rc<0) cbi->ubreak = 1;
}

static void readtree_prnw(dir_cbinfo *cbi, const wchar_t *str) {
   char *out = cbi->utfmode==2 ? hlp_wto8(str,0,0,0) : hlp_wtoa(str,0,0,0,'?');
   if (out) {
      readtree_prn(cbi, out);
      free(out);
   }
}

static void dir_total(dir_cbinfo *cbi, disk_volume_data *vi, int eol) {
   spstr sum;
   if (cbi->file_cnt + cbi->dir_cnt) {
      u32t fcnt = cbi->file_cnt,
           dcnt = cbi->dir_cnt;
      sum.sprintf(dcnt>1 ? "\n%s in %u file%s and %u directories" :
         "\n%s in %u file%s", dsk_formatsize(1024,cbi->sizetotal,10,0),
            fcnt, fcnt!=1?"s":"", dcnt);
      pause_println(sum(), cbi->dir_npmode);
   }
   sum.sprintf("%s free%s", dsk_formatsize(vi->SectorSize, (u64t)vi->ClSize*
               vi->ClAvail, 10, 0), eol?"\n":"");
   pause_println(sum(), cbi->dir_npmode);

   cbi->sizetotal = 0;
   cbi->file_cnt  = 0;
   cbi->dir_cnt   = 0;
}

// file entry data for directory enumeration
typedef struct {
   io_time    time;
   u64t       size;
   u32t      attrs;
} dir_en_rec;

static void dir_std(dir_cbinfo *cbi, TPtrStringsW &lst, const spstrw &dir) {
   for (u32t ii=0; ii<lst.Count(); ii++) {
      dir_en_rec *de = (dir_en_rec*)lst.Objects(ii);
      char       buf[80];
      // format time
      int        pos = sprintf(buf,"%02d.%02d.%04d  %02d:%02d ", de->time.day,
                          de->time.mon+1, de->time.year+1781, de->time.hour,
                              de->time.min);
      if (pos>0)
         if (de->attrs & IOFA_DIR) strcpy(buf+pos, "   <DIR>     ");
            else sprintf(buf+pos, "%11Lu  ", de->size);

      spstrw pp(buf);
      if (cbi->altnames) {
         wchar_t *altn = 0;
         if (de->attrs & IOFA_ALTNAME) {
            spstrw nm(dir);
            if (nm.lastchar()!='\\') nm+='\\';
            nm += lst[ii];
            altn = io_altnamew(nm());
         }
         pp += spstrw().sprintf((wchar_t*)L"%-14S", altn?altn:(wchar_t*)L"");
         if (altn) free(altn);
      }
      pp += lst[ii];

      readtree_prnw(cbi, pp());
      if (cbi->ubreak) break;
   }
}

static void dir_short(dir_cbinfo *cbi, TPtrStringsW &lst, spstrw dir) {
   // dir /b /s relative path ...
   if (dir.length() && dir.lastchar()!='\\') dir+='\\';

   for (u32t ii=0; ii<lst.Count() && !cbi->ubreak; ii++)
      if (lst[ii]!="." && lst[ii]!="..")
         if (!dir) readtree_prnw(cbi, lst[ii]()); else
            readtree_prnw(cbi, (dir+lst[ii])());
}

static void dir_wide(dir_cbinfo *cbi, TPtrStringsW &lst) {
   u32t ii, maxlen = 0, nperline = 1, cols;

   for (ii=0; ii<lst.Count(); ii++) {
      u32t ln = lst[ii].length();
      // +2 for []
      if (((dir_en_rec*)lst.Objects(ii))->attrs & IOFA_DIR) ln+=2;
      if (ln>maxlen) maxlen = ln;
   }
   vio_getmode(&cols, 0);
   if (maxlen) nperline = cols/(maxlen+SPACES_IN_WIDE_MODE);
   if (nperline==0) nperline++;

   spstrw  outline;
   // just avoid a lot of spstr mallocs for each name ...
   wchar_t  *cname = new wchar_t[maxlen+SPACES_IN_WIDE_MODE+1], *cp;

   for (ii=0; ii<lst.Count(); ii++) {
      int  dir = ((dir_en_rec*)lst.Objects(ii))->attrs & IOFA_DIR,
           eol = ii%nperline==nperline-1 || ii==lst.Max();
      cp = cname;

      if (dir) *cp++='[';
      memcpy(cp, lst[ii](), lst[ii].length()<<1);  cp+=lst[ii].length();
      if (dir) *cp++=']';
      if (!eol)
         while (cp-cname < maxlen+SPACES_IN_WIDE_MODE) *cp++=' ';
      *cp = 0;
      outline += cname;

      if (eol) {
         readtree_prnw(cbi, outline());
         outline.del(0,FFFF);
         if (cbi->ubreak) break;
      }
   }
   delete cname;
}

/** read directory
    mask can be 0 for *.*
    dstack is filled with all found directories if not 0 */
static qserr dir_enum(dir_cbinfo *cbi, const spstrw &path, TPtrStringsW &list,
                      TPtrStringsW *dstack, const wchar_t *mask, int crtime)
{
   dir_handle dh;
   qserr     res = io_diropenw(path(), &dh);

   if (!res) {
      // dir /w|b /x - return short names just in the list
      if (cbi->altnames && (cbi->widemode||cbi->shortfmt))
         io_setstate(dh,IOFS_ALTNAME,1);

      while (1) {
         io_direntry_infow de;
         res = io_dirnextw(dh, &de);
         if (res==E_SYS_NOFILE) { res = 0; break; }
         if (res) break;

         int dir = de.attrs & IOFA_DIR ? 1 : 0;
         if (dir && !wcscmp(de.name,(wchar_t*)L".")) continue;
         if (dir && wcscmp(de.name,(wchar_t*)L"..")) dir++;

         if (!mask || _matchpatternw(mask, de.name)) {
            // use thread alloc
            dir_en_rec *drec = (dir_en_rec*)malloc_th(sizeof(dir_en_rec));
            drec->attrs = de.attrs;
            drec->size  = dir ? 0 : de.size;
            drec->time  = crtime ? de.ctime : de.wtime;
            list.AddObject(de.name, drec);

            if (!dir) {
               cbi->file_cnt++;
               cbi->sizetotal += de.size>>10;
               if (de.size & _1KB-1) cbi->sizetotal++;
            } else
            if (dir>1) cbi->dir_cnt++;
         }
         // push a full path dir to the stack
         if (dstack && dir>1) {
            spstrw nm(de.dir);
            if (nm.lastchar()!='\\') nm+='\\';
            nm += de.name;
            dstack->Add(nm);
         }
      }
      io_dirclose(dh);
   }
   return res;
}

/// sort directory list
static void dir_sort(TPtrStringsW &lst, int sortmode, int dirtop) {
   if (lst.Count()>1 && (sortmode || dirtop)) {
      u32t iCnt = lst.Count(),
           iOfs = iCnt/2;
      int   inv = sortmode<0;
      if (inv) sortmode = -sortmode;
      while (iOfs) {
         u32t iLimit=iCnt-iOfs, iSwitch;
         do {
            iSwitch = 0;
            for (u32t iRow=0; iRow<iLimit; iRow++) {
               dir_en_rec *l1 = (dir_en_rec*)lst.Objects(iRow),
                          *l2 = (dir_en_rec*)lst.Objects(iRow+iOfs);
               int         w1 = 5,
                           w2 = 5;
               s64t      diff = 0;
               if (dirtop) {
                  if (l1->attrs & IOFA_DIR) w1 = 0;
                  if (l2->attrs & IOFA_DIR) w2 = 0;
               }
               switch (sortmode) {
                  case 1: diff = _wcsicmp(lst[iRow](),lst[iRow+iOfs]());
                          break;
                  case 2: diff = l1->size - l2->size;
                          break;
                  case 3: diff = _wcsicmp(getextw(lst[iRow])(),
                                          getextw(lst[iRow+iOfs])());
                          break;
                  case 4: diff = io_iototimeq(&l1->time) - io_iototimeq(&l2->time);
                          break;
               }
               if (diff<0) w2++; else
                  if (diff>0) w1++;

               if (w1!=w2)
                  if (!inv && w1>w2 || inv && w1<w2) {
                     lst.Exchange(iRow,iRow+iOfs);
                     iSwitch = iRow;
                  }
            }
         } while (iSwitch);
         iOfs /= 2;
      }
   }
}

u32t _std shl_dir(const char *cmd, str_list *args) {
   int rc=-1, shortfmt=0, nopause=0, subdirs=0, crtime=0, wide=0, altnames=0,
       sort=0, dirtop=0, utf8=0;

   TPtrStrings al;
   str_getstrs(args,al);
   // is help?
   int idx = al.IndexOf("/?");
   if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
   // process args
   static char *argstr   = "s|np|b|tc|w|x|on|o-n|os|o-s|oe|o-e|od|o-d|g|u|u+";
   static short argval[] = {1, 1,1, 1,1,1, 1, -1, 2, -2, 3, -3, 4, -4,1,1, 2};
   process_args(al, SPA_NOSLASH, argstr, argval, &subdirs, &nopause, &shortfmt,
                &crtime, &wide, &altnames, &sort, &sort, &sort, &sort, &sort,
                &sort, &sort, &sort, &dirtop, &utf8, &utf8);

   al.TrimEmptyLines();
   if (!al.Count()) al<<spstr(".");
   // disable UTF-8 output in wide mode ...
   if (wide && utf8==2) utf8=1;
   // init "paused" output
   pause_println(0,1);
   dir_cbinfo cbi;
   cbi.dir_npmode = nopause?-1:0;
   cbi.sizetotal  = 0;
   cbi.file_cnt   = 0;
   cbi.dir_cnt    = 0;
   cbi.check_esc  = nopause && isatty(fileno(get_stdout()));
   cbi.lcounter   = 0;
   cbi.ubreak     = 0;
   cbi.altnames   = altnames;
   cbi.shortfmt   = shortfmt;
   cbi.widemode   = wide;
   cbi.utfmode    = utf8;
   // current volume data
   disk_volume_data  vi;
   char      lvol = 0;
   int     vol1st = 1;

   for (idx=0; idx<al.Count(); idx++) {
      // source is ascii in the current codepage or utf8 string
      spstrw    pp, drive, name;
      qserr  error = get_w_arg(al[idx], pp, utf8);
      // name conversion error?
      if (!error) {
         // fix and split path
         splitfnamew(pp.replacechar('/','\\'), drive, name);

         // w/o mask?
         if (name.cpos('*')<0 && name.cpos('?')<0) {
            io_handle_info fi;
            error = io_pathinfow(pp(), &fi);
            if (!error && (fi.attrs & IOFA_DIR)) { drive = pp; name.clear(); }
         }
      }
      if (!error)
         if (!drive) getcurdirw(drive); else
            error = qs_fullpathw(drive);

      // check volume
      if (drive[0] != lvol) {
         // total for previous volume (if header was printed)
         if (lvol && !shortfmt && vi.FsVer && !vol1st) dir_total(&cbi, &vi, 1);
         // another volume, stats cleared
         lvol  = drive[0];
         // ignore error here since fullpath() must return it earlier
         qserr vierr = io_volinfo(lvol-'A', &vi);
         if (!error) error = vierr;
         if (!error && !vi.FsVer) {
            spstr str;
            str.sprintf("File system in drive %c is not recognized", drive[0]);
            readtree_prn(&cbi, str());
            continue;
         }
         vol1st = 1;
      }
      if (error) {
         cmd_shellerr(EMSG_QS, error, 0);
         rc = qserr2errno(error);
         continue;
      }
      TPtrStringsW dstack;
      dstack.Add(drive);
      // loop over dir tree ...
      while (dstack.Count() && !cbi.ubreak) {
         TPtrStringsW lst;
         error = dir_enum(&cbi, dstack[0], lst, subdirs?&dstack:0,
                          name.length()?name():0, crtime);

         if (lst.Count() || !name) {
            if (!shortfmt) {
               spstrw  head;

               if (vol1st) {
                  int lbon = vi.Label[0];
                  head.sprintf((wchar_t*)
                               L"  Directory of \n"
                               L"  Volume in drive %c %ss %s\n"
                               L"  Volume Serial Number is %04X-%04X\n",
                               lvol, lbon?"i":"ha", lbon?vi.Label:"no label",
                               vi.SerialNum>>16, vi.SerialNum&0xFFFF);
               } else
                  head = (wchar_t*)L"\n  Directory of ";
               // insert after "..ry of " (since direct %S will drop all ch>=128)
               head.insert(dstack[0],head.cpos('f')+2);

               readtree_prnw(&cbi, head());
            }
            vol1st = 0;
            
            if (!error && !cbi.ubreak) {
               if (sort || dirtop) dir_sort(lst, sort, dirtop);
            
               if (wide) dir_wide(&cbi, lst); else
                  if (!shortfmt) dir_std(&cbi, lst, dstack[0]); else
                     dir_short(&cbi, lst, spstrw(dstack[0]).del(0,drive.length()));
            }
            // free dir_en_rec(s)
            FreeItems(lst);
         }
         if (error && !cbi.ubreak) cmd_shellerr(EMSG_QS, error, 0);
         dstack.Delete(0);
      }

      if (cbi.ubreak) {
         printf("User break\n");
         return EINTR;
      } else
      if (vol1st && !error && vi.FsVer) cmd_shellerr(EMSG_CLIB, rc = ENOENT, 0);
   }
   if (!shortfmt && lvol && vi.FsVer && !vol1st) dir_total(&cbi, &vi, 0);

   if (rc<0) rc=0;
   return rc;
}

// -------------------------------------------------------------------------

u32t _std shl_restart(const char *cmd, str_list *args) {
   int    rc = -1;
   char *ini =  0;
   if (args->count>0) {
      TPtrStrings al;
      str_getstrs(args,al);
      // help?
      if (al.IndexOf("/?")>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      u32t bootaddr=0, bpbvol=0;
      int    nosend=0, boot=0, efi=0, msg=0, os=0, bmode=0;

      al.TrimEmptyLines();
      // options must be at the start of the line
      while (al.Count()>=1 && (al[0][0]=='/' || al[0][0]=='-')) {
         al[0].del(0,1).upper();
         if (al[0]=="NOSEND") nosend = 1; else
            if (al[0]=="MSG") msg = 1; else
               if (al[0]=="OS") os = QRST_OSMODE; else
                  if (al[0]=="PAE") os = QRST_PAE; else
                     if (al[0]=="RAW") { boot = 1; efi = 0; bootaddr = 0; } else
                     if (al[0](0,4)=="RAW:") {
                        u32t seg = strtoul(al[0].word(2,":")(),0,16),
                             ofs = strtoul(al[0].word(3,":")(),0,16);
                        if (seg>0xFFFF || ofs>0xFFFF) { rc = EINVAL; break; }
                        bootaddr = seg<<16|ofs;
                        boot = 1;
                        efi  = 0;
                     } else
                     if (al[0]=="BM+") bmode = QRST_BMNTRW; else
                     if (al[0]=="BM-") bmode = QRST_BMNTNO; else
                     if (al[0]=="BMRO") bmode = QRST_BMNTRO; else
                     if (al[0]=="EFI") { boot = 0; efi = 1; } else
                     if (al[0](0,4)=="BPB:") {
                        char vol = al[0][4];
                        if (vol>='C' && vol<='Z' && (al[0][5]==':' || !al[0][5]))
                           bpbvol = (u32t)(vol-'A') << QRST_BPBSHIFT;
                        else {
                           rc = EINVAL;
                           break;
                        }
                     } else
                     if (al[0](0,4)=="INI:") {
                        if (hlp_hosttype()==QSHT_EFI)
                           rc = EEFIHOST;
                        else {
                           spstr  fn(al[0],4);
                           u32t   inilen = 0;
                           void *inidata = 0; 
                           fn.trim();
                           if (fn[1]!=':')
                              inidata = (char*)hlp_freadfull(fn(), &inilen, 0);
                           if (!inidata)
                              inidata = (char*)freadfull(fn(), &inilen);
                           if (inidata) {
                              ini = (char*)malloc_th(inilen+1);
                              memcpy(ini, inidata, inilen);
                              ini[inilen] = 0;
                              hlp_memfree(inidata);
                           } else {
                              rc = ENOENT;
                              break;
                           }
                        }
                     } else
                        break;
         al.Delete(0);
      }
      // /raw should not be combined with INI or BM ...
      if (boot && (ini || bmode>0 || bpbvol)) rc = EINVAL;

      if (rc<0 && al.Count()>=1) {
         qserr res = 0;
         if (hlp_hosttype()==QSHT_EFI) {
            if (boot || ini || bmode>0 || bpbvol) res = E_SYS_EFIHOST; else {
               spstr rm = al[0], args;

               al.Delete(0);
               if (al.Count()) args = al.MergeCmdLine();
               // convert a:\ to \ to launch it from a file instead of memory
               if (toupper(rm[0])=='A' && rm[1]==':' && rm[2]=='\\') rm.del(0,2);

               res = exit_efiboot(rm(), !args?0:args());
            }
         } else 
         if (efi) res = E_SYS_BIOSHOST; else {
            spstr  ldr(al[0]);
            if (boot) res = exit_rawcode(ldr(),bootaddr); else {
               char  *env = 0;
               al.Delete(0);
               // we have keys!
               if (al.Count()) {
                  str_list *lst = str_getlist(al.Str);
                  env = env_create(lst, 0);
                  free(lst);
               }
               res = exit_restart((char*)ldr(), env, ini, os|bmode|bpbvol|
                                  (nosend?QRST_NOHIST|QRST_NOCP:0));
               if (env) free(env);
            }
         }
         // we have QS error OR clib error
         if (res) {
            if (msg) {
               char *msg = cmd_shellerrmsg(EMSG_QS, res);
               vio_msgbox("RESTART", msg, MSG_OK|MSG_RED, 0);
               free(msg);
            } else 
               cmd_shellerr(EMSG_QS, res, 0);
            rc = 0;
         }
      }
   }
   if (ini) free(ini);
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return rc;
}

int _std cmd_shellhelp(const char *cmd, u8t color) {
   static qs_helpctrl help = 0;
   if (!help) help = NEW_G(qs_helpctrl);
   if (!help) return 0;
   qsh_file hf;
   qserr    rc = help->open(SYS_HELP_FILE, &hf);
   if (rc) {
      printf("File " SYS_HELP_FILE " not found!\n");
      return 0;
   }
   u32t topicid = help->topicid(hf, cmd), scrw;
   vio_getmodefast(&scrw, 0);
   if (topicid) {
      char *msg = help->drawansi(hf, color, topicid, 0, scrw-1, 0);
      u32t  len = msg ? strlen(msg) : 0;
      if (!len) topicid = 0; else {
         int redirect = !isatty(fileno(get_stdout())),
                 ansi = -1;
         u16t   svcol = 0;
         // save ansi state and vio color if output device is not redirected
         if (!redirect) {
            ansi  = vio_setansi(1);
            svcol = vio_getcolor();
         }
         if (msg[len-1]=='\n') msg[len-1] = 0;
         pause_println(msg, PRNSEQ_INIT);
         // restore user settings ...
         if (!redirect) {
            vio_setcolor(svcol);
            vio_setansi(ansi);
         }
      }
      free(msg);
   }
   if (!topicid) printf("There is no help entry for \"%s\"!\n", cmd);
   help->close(hf);

   return topicid?1:0;
}

u32t _std shl_mkdir(const char *cmd, str_list *args) {
   int  utf8 = 0, quiet = 0;
   u32t   rc = E_SYS_INVPARM;
   TPtrStrings al;
   if (args->count>0) {
      str_getstrs(args,al);
      // is it help?
      if (al.IndexOf("/?")>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "/q|/u";
      static short argval[] = { 1, 1};
      process_args(al, 0, argstr, argval, &quiet, &utf8);
   }
   al.TrimEmptyLines();
   if (al.Count()>=1) {
      spstrw path;
      rc = ascii2utf(al[0], path, utf8);
      if (!rc) rc = io_mkdirw(path());
      if (rc>0) log_it(3, "mkdir err %X, dir \"%S\"\n", rc, path());
   }
   if (rc && !quiet) cmd_shellerr(EMSG_QS,rc,0);
   return qserr2errno(rc);
}

/// internal chdir for CD / PUSHD / POPD commands
static qserr chdir_int(const spstrw &path) {
   spstrw pp(path);
   qserr err = qs_fullpathw(pp);
   if (!err) err = io_chdirw(pp());
   //if (err) log_it(2,"chdir_int(%S)=%X\n", pp(), err);

   if (!err && pp[1]==':') {
      char disk = toupper(pp[0]);
      if (disk>='0' && disk<='9') disk+='A'-'0';
      io_setdisk(disk - 'A');
   }
   return err;
}

u32t _std shl_chdir(const char *cmd, str_list *args) {
   int rc=-1, utf8=0, quiet=0;
   TPtrStrings al;
   if (args->count>0) {
      str_getstrs(args,al);
      // is it help?
      if (al.IndexOf("/?")>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "/q|/u";
      static short argval[] = { 1, 1};
      process_args(al, 0, argstr, argval, &quiet, &utf8);
   }
   // process
   al.TrimEmptyLines();
   qserr   err = 0;
   spstrw  dir;
   // chdir a:
   if (al.Count()>=1 && al[0].length()==2 && al[0][1]==':' && isalnum(al[0][0])) {
      int drive = toupper(al[0][0]);
      drive    -= drive<='9'?'0':'A';
      c16t  *dd = _wgetdcwd(drive,0,0);

      if (!dd) rc = get_errno(); else {
         dir = dd;
         free(dd);
      }
      al.Clear();
   }

   if (rc>0) ; else {
      rc = 0;
      if (al.Count()>=1) {
         err = ascii2utf(al[0], dir, utf8);
         if (!err) err = chdir_int(dir);
      } else
      if (!dir && !getcurdirw(dir)) rc = get_errno(); else {
         spstr dira;
         // show ? instead of error for ascii target
         if (utf8) err = utf2ascii(dir, dira, 1); else {
            char  *cp = dira.LockPtr(dir.length()+1);
            cp = hlp_wtoa(dir(), cp, dir.length()+1, &err, '?');
            dira.UnlockPtr();
         }
         if (!err) printf("%s\n", dira());
      }
   }
   if (err) {
      if (!quiet) cmd_shellerr(EMSG_QS,err,0);
      rc = qserr2errno(err);
   } else {
      if (rc<0) rc = EINVAL;
      if (rc && !quiet) cmd_shellerr(EMSG_CLIB,rc,0);
   }
   return rc;
}

void dir_to_list(const spstrw &Dir, dirw_t *info, d exclude_attr, d only_attrs,
                 TStringsW &rc, TStringsW *dirs)
{
   TPtrStringsW stack;
   rc.Clear();
   if (!info) return;
   dirw_t *di = info;

   spstrw dir(Dir);
   if (dir.length()) {
      if (dir.lastchar()=='\\') dir.dellast();
      if (dir.lastchar()!='/') dir+='/';
   }

   do {
      while (di) {
         spstrw cname(dir);
         cname+=di->d_name;
         // add a file
         if ((di->d_attr&exclude_attr)==0 &&
            (!only_attrs || (di->d_attr&only_attrs)))
               rc.AddObject(cname,di->d_attr);
         // add dir to stack
         if ((di->d_attr&_A_SUBDIR) && wcscmp(di->d_name,(wchar_t*)L"..")) {
            stack.AddObject(cname, di->d_subdir);
            if (dirs) dirs->AddObject(cname, di->d_attr);
         }
         di = di->d_next;
      }
      while (stack.Count()) {
         di  = (dirw_t*)stack.Objects(0);
         dir = stack[0];
         dir+= "/";
         stack.Delete(0);
         if (di) break;
      }
   } while (di&&di->d_name[0]||stack.Count());
   rc.Sort();
}


u32t _std shl_del(const char *cmd, str_list *args) {
   int rc=-1;
   if (args->count>0) {
      int askall=0, force=0, subdir=0, quiet=0, nopause=0, mask=0, nquiet=0,
          edirs=0, breakable=1, fileonly=-1, ignore=0, utf8=0;

      TPtrStrings al;
      str_getstrs(args,al);
      // is help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "p|f|s|q|e|qn|np|nb|ad|af|i|u";
      static short argval[] = {1,1,1,1,1, 1, 1, 0, 0, 1,1,1};
      process_args(al, SPA_NOSLASH, argstr, argval,
                   &askall, &force, &subdir, &quiet, &edirs, &nquiet,
                   &nopause, &breakable, &fileonly, &fileonly, &ignore, &utf8);

      al.TrimEmptyLines();
      if (al.Count()>=1) {
         u32t fcount = 0;
         // init "paused" output
         pause_println(0,1);
         // delete files
         for (idx=0; idx<al.Count(); idx++) {
            dirw_t   fi;
            spstrw  dir, name, pp;
            // errors should be rare here ...
            qserr err = ascii2utf(al[idx], pp, utf8);
            
            if (err) rc = qserr2errno(err); else
            if (!fullpathw(pp)) rc = get_errno(); else {
               splitfnamew(pp, dir, name);
               mask = name.cpos('*')>=0 || name.cpos('?')>=0;

               if (!mask) {
                  if (_dos_statw(pp(), &fi)) rc = ENOENT; else
                  if (fi.d_attr&_A_SUBDIR) {
                     if (fileonly>0) rc = EISDIR; else {
                        dir   = pp;
                        name  = (wchar_t*)L"*.*";
                        mask  = true;
                     }
                  } else
                  if (fileonly==0) rc = ENOTDIR;
               }
            }
            if (rc>0) {
               if (ignore) {
                  spstr np;
                  np.sprintf("%S: ", pp());
                  cmd_shellerr(EMSG_CLIB, rc, np());
                  rc = 0;
                  continue;
               } else break;
            }
            if (dir.length() && dir.lastchar()!='\\') dir+='\\';
            TStringsW  dellist, dirlist;
            spstrw dname = dir + name;

            if (mask || subdir) {
               // read tree
               dirw_t *info = 0;
               u32t cnt = _dos_readtreew(dir(), name(), &info, subdir, 0, 0, 0);
               if (info) {
                  if (cnt) dir_to_list(dir, info, (force?0:_A_RDONLY)|
                     _A_SUBDIR|_A_VOLID, 0, dellist, edirs?&dirlist:0);
                  _dos_freetree(info);

                  u32t todel = dellist.Count()+dirlist.Count();

                  if (todel && !quiet && (mask || todel>1)) {
                      printf("Delete %S (%d files) (y/n/esc)?", dname(), todel);
                      int yn = ask_yn();
                      printf("\n");
                      if (yn<0) return EINTR;
                      if (!yn) { dellist.Clear(); dirlist.Clear(); }
                  }
               }
            } else
               dellist.AddObject(dname, fi.d_attr);

            if (dellist.Count() + dirlist.Count()) {
               int ii;
               spstr msg;  // put it here to prevent malloc on every file

               for (ii=0; ii<dellist.Count(); ii++) {
                  int doit = 1;
                  spstrw dfname(dellist[ii]);
                  dfname.replacechar('/','\\');
                  
                  if (askall) {
                     printf("Delete %S (y/n/esc)?", dfname());
                     doit = ask_yn();
                     printf("\n");
                     if (doit<0) return EINTR;
                  } else
                  if (breakable && (ii&7)==0) // check for ESC key pressed
                     if (key_pressed())
                        if ((key_read()&0xFF)==27) {
                           printf("Break this DEL command (y/n)?");
                           int yn = ask_yn();
                           printf("\n");
                           if (yn==1) return EINTR;
                        }
                  if (doit) {
                     qserr res = 0;
                     /* unlike previous variant - do not ignore get/set attr
                        errors */
                     if ((dellist.Objects(ii)&_A_RDONLY)!=0)
                        if (force) {
                           io_handle_info fi;
                           res = io_pathinfow(dfname(), &fi);
                           if (!res) {
                              fi.attrs &= ~IOFA_READONLY;
                              res = io_setinfow(dfname(), &fi, IOSI_ATTRS);
                           }
                        }
                     if (!res)
                        res = dellist.Objects(ii)&_A_SUBDIR? io_rmdirw(dfname()) :
                              io_removew(dfname());
                     if (res==0) {
                        msg.sprintf("Deleted file - \"%S\"", dfname());
                        fcount++;
                     } else
                        qsfwerr2str(msg, dfname, res);
                     if (!nquiet)
                        if (pause_println(msg(),nopause?-1:0)) return EINTR;
                  }
               }
               for (ii=dirlist.Max(); ii>=0; ii--) {
                  qserr err = io_rmdirw(dirlist[ii]());
                  if (err && !nquiet) {
                     qsfwerr2str(msg, dirlist[ii], err);
                     if (pause_println(msg(),nopause?-1:0)) return EINTR;
                  }
               }
            }
         }
         if (rc<0) rc=0;
         if (!rc && !nquiet) printf("%d file(s) deleted.\n",fcount);
      }
   }
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return rc;
}

u32t _std shl_beep(const char *cmd, str_list *args) {
   int  wait=0, before=0, idx;
   TPtrStrings al;
   str_getstrs(args,al);
   // process
   al.TrimEmptyLines();
   if (al.Count()>=1) {
      // is help?
      idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "w|b";
      static short argval[] = {1,1};
      process_args(al, SPA_NOSLASH, argstr, argval, &wait, &before);
   }
   pause_println(0,1);
   if (before)
      while (vio_beepactive()) usleep(10*1000);

   if (!al.Count()) vio_beep(500,500);
      else
   for (idx=0; idx<al.Count(); idx++) {
      u32t freq = al[idx].word_Dword(1,","),
            dur = al[idx].word_Dword(2,",");
      if (!freq||!dur) continue;
      vio_beep(freq,dur);

      if (idx<al.Max())
         while (vio_beepactive()) usleep(10*1000);
   }
   if (wait)
      while (vio_beepactive()) usleep(10*1000);
   return EZERO;
}

u32t _std shl_rmdir(const char *cmd, str_list *args) {
   TPtrStrings al;
   int   subdir=0, quiet=0, utf8=0;
   str_getstrs(args,al);
   // is help?
   int idx = al.IndexOf("/?");
   if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
   // process args
   static char *argstr   = "q|s|u";
   static short argval[] = {1,1,1};
   process_args(al, SPA_NOSLASH, argstr, argval, &quiet, &subdir, &utf8);
   // process command
   al.TrimEmptyLines();
   qserr err = E_SYS_INVPARM;
   if (al.Count()>=1)
   do {
      dirw_t fi;
      spstrw rp;
      err = get_w_arg(al[0], rp, utf8);
      if (err) break;
      err = qs_fullpathw(rp);
      if (err) break;
      err = _dos_statw(rp(), &fi);
      if (err) break;
      if ((fi.d_attr&_A_SUBDIR)==0) { err=E_SYS_NOTDIR; break; }
      // call del tree with UTF-8 path but ignore all errors except user ESC
      if (subdir) {
         spstr dcs;
         dcs.sprintf("/s /qn /np /e /u /f %s \"%s\"", quiet?"/q":"", rp.getutf8()());
         if (cmd_shellcall(shl_del, dcs(), 0)==EINTR) { err=0; break; }
      }
      // try it anyway, just to get the error code ...
      err = io_rmdirw(rp());
   } while (false);

   if (err && !quiet) cmd_shellerr(EMSG_QS,err,0);
   return err?qserr2errno(err):0;
}

u32t _std shl_loadmod(const char *cmd, str_list *args) {
   int rc=-1, quiet=0, unload=0, list=0, nopause = 0;
   if (args->count>0) {
      TPtrStrings     al;
      ansi_push_set ansi(1);

      str_getstrs(args,al);
      // is help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "q|u|list|l|np";
      static short argval[] = {1,1, 1,  1, 1};
      process_args(al, SPA_NOSLASH, argstr, argval, &quiet, &unload, &list,
                   &list, &nopause);
      // process command
      al.TrimEmptyLines();
      // wrong args?
      if (unload && list || !Xor(list,al.Count())) rc = EINVAL; else {
         // init line counter
         pause_println(0, nopause?PRNSEQ_INITNOPAUSE:PRNSEQ_INIT);

         if (list) {
            module_information *mi = 0;
            u32t     cmi = mod_enum(&mi), ii;
            int  ioredir = !isatty(fileno(get_stdout()));

            cmd_printf(" N  module name               spec   pid   parent  code sz   data sz\n"
                       "--- --------------------- --- ------ ------ ------ --------- ---------\n");

            for (ii=0; ii<cmi; ii++) {
               int islib = mi[ii].flags&MOD_LIBRARY,
                   issys = mi[ii].flags&MOD_SYSTEM;
               char pstr[48], cstr[24], dstr[24];

               if (mi[ii].flags & MOD_EXECPROC) {
                  u32t ppid = 0,
                        pid = mod_getmodpid(mi[ii].handle, &ppid);
                  snprintf(pstr, 48, ppid?"%6u %6u ":"%6u        ", pid, ppid);
               } else 
                  strcpy(pstr, "              ");

               u32t dsz = 0, csz = 0, obj, osz, ofl;
               for (obj=0; obj<mi[ii].objects; obj++)
                  if (!mod_objectinfo(mi[ii].handle, obj, 0, &osz, &ofl, 0))
                     if (ofl&4) csz+=osz; else // OBJEXEC
                        if ((ofl&0x88)==0) dsz+=osz; // all except OBJRSRC & OBJINVALID
               if (csz<1024) strcpy(cstr, csz?"    <1 kb ":"          "); else
                  snprintf(cstr, 24, " %8s ", dsk_formatsize(1,csz,0,0));
               if (dsz<1024) strcpy(dstr, dsz?"    <1 kb ":"          "); else
                  snprintf(dstr, 24, " %8s ", dsk_formatsize(1,dsz,0,0));

               if (ioredir)
                  cmd_printf("%3u  %-20s %s %s%s%s%s\n", ii+1, mi[ii].name,
                     islib?"lib":"EXE", issys?"system":"      ", pstr, cstr, dstr);
               else
                  cmd_printf("%3u  " ANSI_WHITE "%-20s" ANSI_RESET " %s %s%s%s%s\n",
                     ii+1, mi[ii].name, islib?ANSI_YELLOW "lib" ANSI_RESET:
                        ANSI_LGREEN "EXE" ANSI_RESET, issys?
                           ANSI_LRED "system" ANSI_RESET:"      ", pstr, cstr, dstr);
            }
            rc = 0;
         } else {
            for (idx=0; idx<al.Count(); idx++) {
               spstr mdn(al[idx]);
               if (unload) {
                  // test of module unload
                  u32t mod = mod_query(mdn(), MODQ_NOINCR), rc;
                  if (!mod) {
                     if (!quiet) cmd_printf("There is no active module \"%s\"\n", mdn());
                     rc = ENOENT;
                  } else {
                     u32t res = mod_free(mod);
                     if (res && !quiet) {
                        char *errmsg = cmd_shellerrmsg(EMSG_QS, res);
                        if (errmsg) {
                           cmd_printf("%s (\"%s\")\n", errmsg, mdn());
                           free(errmsg);
                        } else
                           cmd_printf("Unload error %X for module \"%s\"\n", rc, mdn());
                     }
                     if (res) rc = EBUSY;
                  }
               } else {
                  u32t error = 0;
                  module *md = load_module(mdn, &error);
                  if (md) rc = 0; else {
                     rc = ENOENT;
                     if (!quiet) {
                        cmd_printf("Error loading module \"%s\".\n",mdn());
                        char *msg = cmd_shellerrmsg(EMSG_QS, error);
                        if (msg) {
                           cmd_printseq("(%s)",0,VIO_COLOR_GRAY,msg);
                           free(msg);
                        }
                     }
                  }
               }
            }
            if (rc<0) rc = 0;
         }
      }
   }
   if (rc<0) rc = EINVAL;
   if (rc && rc!=ENOENT && rc!=EBUSY) cmd_shellerr(EMSG_CLIB,rc,0);
   return rc;
}

u32t _std shl_power(const char *cmd, str_list *args) {
   int quiet=0, rc=-1;
   if (args->count>0) {
      TPtrStrings al;
      str_getstrs(args,al);
      // is help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      idx = al.IndexOfName("/q");
      if (idx>=0) { quiet=1; al.Delete(idx); }

      if (al.Count()>0) {
         if (al[0].upper()=="ON") {
            printf("This PC is already powered on.\n");
            rc = 0;
         } else
         if (al[0]=="OFF" || al[0]=="S" || al[0]=="SUSPEND") {
            int susp = al[0][0]=='S'?1:0;

            if (hlp_hosttype()==QSHT_EFI) {
               // suspend is not supported on EFI host
               if (susp) {
                  rc = EEFIHOST;
                  if (quiet) return rc;
               }
            } else {
               u32t ver = hlp_querybios(QBIO_APM);
               int apmerr = ver&0x10000;
               if (apmerr) {
                  log_it(2,"APM error %02X\n", ver&0xFF);
                  if (!quiet) printf("There is no APM on this PC.\n");
                  return ENODEV;
               }
               log_it(2,"APM version %d.%d\n", ver>>8&0xFF,ver&0xFF);
            }
            if (rc<=0) {
               int yn = 1;
               if (!susp && !quiet) {
                  printf("Power off (y/n)?");
                  yn = ask_yn();
               }
               if (yn==1) exit_poweroff(susp);
               rc = 0;
            }
         }
      }
   }
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}

// abort entering on -1
static int _std label_cb(u16t key, key_strcbinfo *inputdata) {
   char cch = toupper(key&0xFF);
   if (cch==27) return -1;
   return 0;
}

u32t _std shl_label(const char *cmd, str_list *args) {
   int rc=-1;
   TPtrStrings al;
   str_getstrs(args,al);
   // is help?
   int idx = al.IndexOf("/?");
   if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }

   if (al.Count()>0) {
      u8t drive;
      int   pos = 0;
      if (al[0][1]==':') {
         char dch = toupper(al[0][0]);
         // correct A:..Z:
         if (dch>='0'&&dch<='9') dch+='A'-'0';
         drive = dch-'A';
         pos++;
      } else
         drive = io_curdisk();

      rc = io_vollabel(drive,al.Count()>pos?al[pos]():0);
      if (rc) { cmd_shellerr(EMSG_QS, rc, 0); rc = EACCES; }
   } else {
      disk_volume_data vinf;
      u8t drive = io_curdisk();
      hlp_volinfo(drive,&vinf);
      int  lbon = vinf.Label[0];

      printf(lbon?"Volume in drive %c is %s\n":"Volume in drive %c has no label\n",
         drive+'A', vinf.Label);
      printf("Volume Serial Number is %04X-%04X\n",
         vinf.SerialNum>>16,vinf.SerialNum&0xFFFF);
      printf("Volume label (11 characters, ENTER - none, ESC)? ");
      char* label = key_getstr(label_cb);
      printf("\n");
      if (label) {
         rc = io_vollabel(drive,*label?label:0);
         if (rc) { cmd_shellerr(EMSG_QS, rc, 0); rc = EACCES; }
         free(label);
      } else {
         rc = 0;
      }
   }
   if (rc<0) cmd_shellerr(EMSG_CLIB, rc = EINVAL, 0); else rc = qserr2errno(rc);
   return rc;
}

u32t _std shl_trace(const char *cmd, str_list *args) {
   int quiet=0, rc=-1, nopause=0;
   if (args->count>0) {
      TPtrStrings al;
      str_getstrs(args,al);
      // is help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      static char *argstr   = "q|np";
      static short argval[] = {1, 1};
      process_args(al, SPA_NOSLASH, argstr, argval, &quiet, &nopause);

      if (al.Count()>0) {
         al[0].upper();
         if (al[0]=="ON" || al[0]=="OFF") {
            int on = al[0][1]=='N';
            // don`t allow empty module name in ON
            if (!on || al.Count()>1) {
               rc = 0;
               if (al.Count()<=2) trace_onoff(al.Count()>1?al[1]():0,0,quiet,on);
                  else
               for (int ii=2; ii<al.Count(); ii++) {
                  rc = 0;
                  if (!trace_onoff(al[1](),al[ii](),quiet,on)) break;
               }
            }
         } else
         if (al[0]=="LIST") {
            trace_list(al.Count()>1?al[1]():0,al.Count()>2?al[2]():0,!nopause);
            rc = 0;
         } else
         if (al[0]=="PIDLIST") {
            if (al.Count()==1) {
               trace_pid_list(!nopause);
               rc = 0;
            }
         } else {
            int lnot = 0, lnew = 0;
            if (al[0][0]=='!') { lnot=1; al[0].del(0); }
            if (al[0].lastchar()=='+') { lnew=1; al[0].dellast(); }

            if (al[0]=="PID") {
               TList lpid;
               for (int ii=1; ii<al.Count(); ii++) {
                  u32t pv = al[ii].Dword();
                  if (!pv || !mod_checkpid(pv)) { rc=ESRCH; break; }
                  lpid.Add(pv);
               }
               if (rc<0) {
                  rc = 0;
                  trace_pid(lpid, lnot, lnew);
               }
            }
         }
      }
   }
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}

/** common RENAME/MOVE processing.
    @param  al       argument list
    @param  is_ren   action type (ren/move)
    @param  quiet    be quiet flag
    @param  utf8     UFT-8 names
    @return error code or 0 */
static qserr processRenMove(TPtrStrings &al, int is_ren, int quiet, int utf8) {
   spstrw target, tgt_path, tgt_name;
   qserr      err = get_w_arg(al[al.Max()], target, utf8),
          copyerr = 0;
   int   tgt_mask = target.cpos('?')>=0 || target.cpos('*')>=0;
   u32t   counter = 0;
   al.Delete(al.Max());

   if (!err) {
      splitfnamew(target, tgt_path, tgt_name);
      err = qs_fullpathw(tgt_path);
      if (!err) {
         if (tgt_path.lastchar()!='\\') tgt_path+='\\';
         // wildcards
         if (tgt_path.cpos('?')>=0||tgt_path.cpos('*')>=0) err = E_SYS_INVNAME;
      }
      if (!err) {
         // target is directory
         if (!is_ren && tgt_name.length())
             if (hlp_isdirw(target())) {
                tgt_path += tgt_name;
                tgt_path += '\\';
                tgt_name.clear();
             }
         if (is_ren && al.Count()>1 || !al.Count()) err = E_SYS_INVPARM;
      }
   }

   if (!err)
   while (al.Count()) {
      spstrw al0w, dir, name;
      err = get_w_arg(al[0], al0w, utf8);
      if (err) break;

      splitfnamew(al0w, dir, name);
      err = qs_fullpathw(dir);
      if (err) break;
      if (dir.lastchar()!='\\') dir+='\\';

      if (is_ren && _wcsicmp(dir(),tgt_path())) {
         // rename to another directory
         if (!quiet) printf("Rename to another directory is not possible.\n");
         return E_SYS_DISKMISMATCH;
      } else
      if (dir[0]!=tgt_path[0]) {
         // move to another drive
         if (!quiet) printf("Rename/move to another drive is not possible.\n");
         return E_SYS_DISKMISMATCH;
      } else
      if (dir.length()>=NAME_MAX-1 || dir.length()+tgt_name.length()>=NAME_MAX) {
         err = E_SYS_LONGNAME;
         break;
      } else {
         TStringsW slst, dlst;
         // alloc a lot, just to be safe
         c16t      *buf = (c16t*)malloc_th(NAME_MAX*4*2);
         int matchcount = 0, idx;
         wcscpy(buf, dir());

         if (name.cpos('?')>=0 || name.cpos('*')>=0) {
            dirw_t *dt = _wopendir(dir()), *drec = dt;
            if (dt) {
               while (drec) {
                  drec = _wreaddir(dt);
                  if (drec)
                     if (_replacepatternw(name(), !tgt_name?name():tgt_name(),
                                          drec->d_name, buf))
                     {
                        slst.Add(dir + spstrw(drec->d_name));
                        dlst.Add(tgt_path + spstrw(buf));
                        matchcount++;
                     }
               }
               _wclosedir(dt);
            }
         } else {
            if (!tgt_mask) {
               slst.Add(dir+name);
               dlst.Add(tgt_path+tgt_name);
               matchcount++;
            } else
            if (_replacepatternw(name(), tgt_name(), name(), buf)) {
               slst.Add(dir + name);
               dlst.Add(tgt_path + spstrw(buf));
               matchcount++;
            }
         }
         free(buf);

         if (!matchcount) {
            if (!quiet)
               printf("There is no replacement for specified criteria.\n");
            return E_SYS_NOTFOUND;
         } else
         for (idx=0; idx<slst.Count(); idx++) {
            if (dlst[idx].length() > NAME_MAX) copyerr = E_SYS_LONGNAME;
               else copyerr = io_movew(slst[idx](), dlst[idx]());
            if (!copyerr) counter++;
            if (copyerr && !quiet) {
               spstr msg;
               qsfwerr2str(msg, slst[idx], copyerr);
               printf("%s\n", msg());
            }
         }
      }
      al.Delete(0);
   }
   if (err && !quiet) cmd_shellerr(EMSG_QS,err,0);
   if (counter && !quiet && !is_ren) printf("%d file(s) moved.\n",counter);
   return err ? err : copyerr;
}

static u32t renmove_common(int ren, const char *cmd, str_list *args) {
   int quiet=0, rc=-1, utf8=0;
   if (args->count>0) {
      TPtrStrings al;
      str_getstrs(args,al);
      // is it help?
      if (al.IndexOf("/?")>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      static char *argstr   = "q|u";
      static short argval[] = {1,1};
      process_args(al, SPA_NOSLASH, argstr, argval, &quiet, &utf8);

      if (ren && al.Count()==2 || al.Count()>=2)
         rc = qserr2errno(processRenMove(al, ren, quiet, utf8));
   }
   if (rc<0) {
      rc = EINVAL;
      if (!quiet) cmd_shellerr(EMSG_CLIB,rc,0);
   }
   return rc;
}

u32t _std shl_ren(const char *cmd, str_list *args) {
   return renmove_common(1, cmd, args);
}

u32t _std shl_move(const char *cmd, str_list *args) {
   return renmove_common(0, cmd, args);
}

u32t _std shl_date(const char *cmd, str_list *args) {
   int setd=0, quiet=0, rc=-1, idx;
   TPtrStrings al;
   struct tm tme;
   time_t    tmv;

   str_getstrs(args,al);
   // process
   al.TrimEmptyLines();
   if (al.Count()>=1) {
      // is help?
      idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "q|s";
      static short argval[] = {1,1};
      process_args(al, SPA_NOSLASH, argstr, argval, &quiet, &setd);
   }
   if (al.Count()==0 || al.Count()==1 && !setd) {
      if (al.Count()==0 && !quiet) {
         char str[32];
         time(&tmv);
         localtime_r(&tmv,&tme);
         strftime(str, sizeof(str), "%A, %d.%m.%Y", &tme);
         printf("The current date is %s\n", str);
      }

      if (setd) {
         if (al.Count()!=0) rc = EINVAL; else {
            printf("Enter new date in DD.MM.YYYY format: ");
            char* tms = key_getstr(label_cb);
            printf("\n");
            if (tms) {
               al.Add(tms); al[0].trim();
               // empty string?
               if (!al[0]) rc = 0;
               free(tms);
            } else rc = 0;
         }
      }
      if (rc<0 && al.Count()==1) {
         u32t dostime = 0;
         time(&tmv);
         localtime_r(&tmv,&tme);

         if (al[0].words(".")>=2) {
            int svyear = tme.tm_year+1900;
            tme.tm_mday = al[0].word_Int(1,".");
            tme.tm_mon  = al[0].word_Int(2,".")-1;
            tme.tm_year = al[0].word_Int(3,".");
            if (!tme.tm_year) tme.tm_year = svyear;

            if (tme.tm_mday>=1 && tme.tm_mday<32 && tme.tm_mon<12 &&
               tme.tm_mon>=0 && tme.tm_year>=1980 && tme.tm_year<=2099)
            {
               tme.tm_year-=1900;
               tmv = mktime(&tme);
               dostime = timetodostime(tmv);
               tm_setdate(dostime);
               if (!setd && !quiet) printf("The new date is: %02d.%02d.%04d\n",
                  tme.tm_mday, tme.tm_mon+1, tme.tm_year+1900);
            }
         }
         if (!dostime) printf("Date format is invalid!\n");
      }
      if (rc<0) rc = 0;
   }

   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}

u32t _std shl_time(const char *cmd, str_list *args) {
   int setd=0, quiet=0, rc=-1, idx;
   TPtrStrings al;
   struct tm tme;
   time_t    tmv;

   str_getstrs(args,al);
   // process
   al.TrimEmptyLines();
   if (al.Count()>=1) {
      // is help?
      idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "q|s";
      static short argval[] = {1,1};
      process_args(al, SPA_NOSLASH, argstr, argval, &quiet, &setd);
   }
   if (al.Count()==0 || al.Count()==1 && !setd) {
      if (al.Count()==0 && !quiet) {
         time(&tmv);
         localtime_r(&tmv,&tme);
         printf("The current time is: %02d:%02d:%02d\n", tme.tm_hour, tme.tm_min,
            tme.tm_sec);
      }

      if (setd) {
         if (al.Count()!=0) rc = EINVAL; else {
            printf("Enter new time in hh:mm:sec format (24 hour time): ");
            char* tms = key_getstr(label_cb);
            printf("\n");
            if (tms) {
               al.Add(tms); al[0].trim();
               // empty string?
               if (!al[0]) rc = 0;
               free(tms);
            } else rc = 0;
         }
      }
      if (rc<0 && al.Count()==1) {
         u32t dostime = 0;
         time(&tmv);
         localtime_r(&tmv,&tme);

         if (al[0].words(":")>=2) {
            tme.tm_hour = al[0].word_Int(1,":");
            tme.tm_min  = al[0].word_Int(2,":");
            tme.tm_sec  = al[0].word_Int(3,":");

            if (tme.tm_hour<24 && tme.tm_min<60 && tme.tm_sec<60 &&
               tme.tm_hour>=0 && tme.tm_min>=0 && tme.tm_sec>=0)
            {
               tmv = mktime(&tme);
               dostime = timetodostime(tmv);
               tm_setdate(dostime);
               if (!setd && !quiet) printf("The new time is: %02d:%02d:%02d\n",
                  tme.tm_hour, tme.tm_min, tme.tm_sec);
            }
         }
         if (!dostime) printf("Time format is invalid!\n");
      }
      if (rc<0) rc = 0;
   }

   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}

// "GRAY".."WHITE" are reused in mbox.cpp
const char *msgbox_opts[] = { "GRAY", "CYAN", "GREEN", "LGREEN",
   "BLUE", "LBLUE", "RED", "LRED", "WHITE", "CENTER", "LEFT", "RIGHT",
   "JUSTIFY", "WIDE", "OK", "OKCANCEL", "YESNO", "YNCANCEL",
   "DEF1", "DEF2", "DEF3", "POPUP", "ENVCOL", 0 };

u32t _std shl_msgbox(const char *cmd, str_list *args) {
   int rc=-1, idx, flags=0, ii;
   TPtrStrings al;

   str_getstrs(args,al);
   // process
   al.TrimEmptyLines();
   if (al.Count()>=1) {
      // is help?
      idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }

      if (al.Count()>=2) {
         static u16t values[] = { MSG_GRAY, MSG_CYAN, MSG_GREEN, MSG_LIGHTGREEN,
             MSG_BLUE, MSG_LIGHTBLUE, MSG_RED, MSG_LIGHTRED, MSG_WHITE,
             MSG_CENTER, MSG_LEFT, MSG_RIGHT, MSG_JUSTIFY, MSG_WIDE, MSG_OK,
             MSG_OKCANCEL, MSG_YESNO, MSG_YESNOCANCEL, MSG_DEF1, MSG_DEF2,
             MSG_DEF3, MSG_POPUP, 0x8000 };

         for (idx=2; idx<al.Count(); idx++) {
            for (ii=0; msgbox_opts[ii]; ii++)
               if (stricmp(al[idx](),msgbox_opts[ii])==0) { flags|=values[ii]; break; }
            if (!msgbox_opts[ii]) { rc = EINVAL; break; }
         }
         if (rc<0) {
            // MSG_ENVCOLOR does not fit in u16t
            if (flags&0x8000) flags=flags&~0x8000|MSG_ENVCOLOR;

            rc = vio_msgbox(al[0](), al[1](), flags, 0);
            set_errorlevel(rc);
            return rc;
         }
      }
   }
   // return 0 (cancel), but display EINVAL error message
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}

// query/load handler for the "mode" device
cmd_eproc _std cmd_modeget(const char *dev) {
   cmd_eproc proc = cmd_modermv(dev,0);
   // no proc for this device type - trying to load it
   if (!proc) {
      spstr ename, mdname;
      ename.sprintf("MODE_%s_HANDLER",dev);
      ename = getenv(ename());
      ename.trim();
      if (!ename) {
         mdname = ecmd_readstr("MODE", dev);
         if (mdname.trim().length())
            ename = ecmd_readstr("MODULES", mdname());
      }
      // load module from env. var OR extcmd.ini list
      if (ename.trim().length()) {
         u32t error = 0;
         if (!load_module(ename, &error)) {
            log_it(2,"unable to load MODE %s handler, error %u\n", dev, error);
         } else
            proc = cmd_modermv(dev,0);
      }
   }
   return proc;
}

u32t _std shl_mode(const char *cmd, str_list *args) {
   int rc=-1, idx, flags=0, ii;
   TPtrStrings al;

   str_getstrs(args,al);
   // process
   al.TrimEmptyLines();
   if (al.Count()>=1) {
      // is help?
      idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }

      // mode x,y and mode co80 support
      if (al.Count() && isdigit(al[0][0]) || al.Count()==1) {
         spstr  ln = al.GetTextToStr(" ");
         u32t cols = 0,
             lines = 0;

         if (ln.words(",")==2) {
            cols  = ln.word_Dword(1,","),
            lines = ln.word_Dword(2,",");
         } else
         if (ln.upper().trim()=="CO80") { cols = 80; lines = 25; } else
         // "unsupported feature" message
         if (ln=="BW80" || ln=="BW40" || ln=="CO80") rc = ENOLCK;

         if (cols && lines) {
            cmd_eproc proc = cmd_modeget("CON");
            if (!proc) {
               rc = vio_setmodeex(cols,lines) ? 0 : ENODEV;
            } else {
               char ln[32];
               snprintf(ln, sizeof(ln), "CON cols=%u lines=%u", cols, lines);

               return cmd_shellcall(proc, ln, 0);
            }
         }
      } else 
      /*if (al.Count()>=2)*/ {
         cmd_eproc proc = cmd_modeget(al[0]());
         if (!proc) rc = ENODEV;
            else
         return cmd_shellcall(proc, 0, args);
      }
   }
   // return 0 (cancel), but display EINVAL error message
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}

u32t _std shl_mode_sys(const char *cmd, str_list *args) {
   int rc = -1;
   TPtrStrings al;
   str_getstrs(args,al);
   al.TrimEmptyLines();
   if (al.Count()>=2) {
      if (al[1].upper()=="STATUS") {
         u32t      rate = 0,
                   host = hlp_hosttype(),
                    cmv = hlp_cmgetstate(),
                    pgm = sys_pagemode();
         u16t      port = hlp_seroutinfo(&rate);
         int        vma = pgm & QPGM_VMA,
                  pgbit = bsf32(pgm & ~QPGM_VMA) + 1;
         static const char *pgmstr[4] = {"Flat", "32-bit page", "PAE page",
                                         "64-bit page"};

         printf("%s host.\n", host==QSHT_EFI?"EFI":"BIOS");
         cmd_printseq("%s mode%s.", -1, VIO_COLOR_LWHITE, pgbit > 3 ?
                      "Unknown" : pgmstr[pgbit], vma?" (with VMA)":"");
         // should not use get_mtlib() here, because it forces MTLIB loading 
         if (mt_active())
            cmd_printseq("MT mode is active.\n", -1, VIO_COLOR_LGREEN);
         if (!port) {
            printf("There is no debug COM port in use now.\n");
         } else {
            printf("Debug COM port address: %04hX, baudrate: %d%s.\n", port, rate,
               mod_secondary->dbport_lock?" (output disabled)":"");
         }
         if (cmv && cmv<CPUCLK_MAXFREQ) {
            cmv = 10000/CPUCLK_MAXFREQ*cmv;
            printf("CPU at %u.%2.2u%% speed.\n", cmv/100, cmv%100);
         }
         rc = 0;
      } else
      if (al[1]=="LIST") {
         if (al.Count()==3 && al[2].upper()=="FS") fs_list_int(printf);
            else printf("Unsupported for this device.\n");
         rc = 0;
      } else {
         u32t  port = al.DwordValue("DBPORT"),
               baud = al.DwordValue("BAUD"),
                cmv = al.DwordValue("CM");
         // call this first, else DBCARD will use wrong BAUD
         if (port || baud) {
            if (!hlp_seroutset(port, baud)) printf("Invalid port or baud rate value.\n");
            rc = 0;
         }
         if (cmv) {
             if (!hlp_cmgetstate()) printf("There is no clock modulation in CPU.\n"); else
             if (cmv==0 || cmv>CPUCLK_MAXFREQ) printf("Invalid CM value.\n"); else
             if (hlp_cmsetstate(cmv)) printf("CM setup error.\n");
             rc = 0;
         }
         if (al.IndexOfName("HEAPFLAGS")>=0) {
            spstr mstr = al.Value("HEAPFLAGS");
            if (!mstr) {
               u32t flags = mem_getopts();
               mstr.clear();
               if (flags&QSMEMMGR_PARANOIDALCHK) mstr+="paranoidal, ";
               if (flags&QSMEMMGR_ZEROMEM) mstr+="zero memory, ";
               if (flags&QSMEMMGR_HALTONNOMEM) mstr+="halt on lack of memory";
               if (mstr.length()) {
                  mstr.insert("(",0);
                  if (mstr.lastchar()==' ') mstr.del(mstr.length()-2,2);
                  mstr+=")";
               }
               printf("Current heap flags: %02X %s\n", flags, mstr());
            } else {
               u32t flags = mstr.Dword();
               if ((flags&~(QSMEMMGR_PARANOIDALCHK|QSMEMMGR_ZEROMEM|QSMEMMGR_HALTONNOMEM))) {
                  printf("Invalid heap flags value.\n");
               } else
                  mem_setopts(flags);
            }
            rc = 0;
         }
         if (al.IndexOfName("CPU")>=0) {
            u32t  len = sys_queryinfo(QSQI_CPUSTRING,0), data[4],
                  avf = sys_isavail(FFFF);
            char *cpu = 0;
            
            if (len>1) {
               cpu = (char*)malloc_th(len);
               sys_queryinfo(QSQI_CPUSTRING,cpu);
            }
            if (!cpu) cpu = (char*)sprintf_dyn("CPU is %s", avf&SFEA_INTEL?
               "Intel":avf&SFEA_AMD?"AMD":avf&SFEA_MSR?"rare":avf&SFEA_NOFPU?
                  "80486sx":"80486dx");
            // function will return 0 on 486dx :)
            if (hlp_getcpuid(1,data)) {
               char sb[40];
               sprintf(sb, " (%u.%u.%u", data[0]>>8&0xF, data[0]>>4&0xF, data[0]&0xF);
               cpu = strcat_dyn(cpu, sb);
               len = avf&(SFEA_PAE|SFEA_PAT|SFEA_X64|SFEA_LAPIC);
               if (avf&len) {
                  sprintf(sb, " %s", avf&SFEA_X64?"x64":(avf&SFEA_PAE?"PAE":""));
                  if (avf&SFEA_PAT) strcat(sb, " PAT");
                  if (avf&SFEA_LAPIC) strcat(sb, " LAPIC");
                  cpu = strcat_dyn(cpu, sb);
               }
               cpu = strcat_dyn(cpu, ")");
            }
            printf("%s\n", cpu);
            free(cpu);

            if ((avf&SFEA_NOFPU)==0) {
               len = fpu_statesize();
               if (len) printf("FPU state    : %u bytes\n", len);
            }
            len = sys_acpiroot();
            if (len) printf("ACPI root    : %08X\n", len);

            static const char *vn[] = { "BIOS Vendor", "BIOS Version", "BIOS Date",
               "Sys Vendor", "Sys Product", "Sys Version", "Sys Serial", "Sys UUID",
               "MB Vendor", "MB Product", "MB Version", "MB Serial" };
            for (u32t ii=SMI_BIOS_Vendor; ii<=SMI_MB_Serial; ii++) {
               char *info = sys_dmiinfo(ii);
               if (info) printf("%-13s: %s\n", vn[ii], info);
               if (info) free(info);
            }
            rc = 0;
         }
         if (al.IndexOfName("NORESET")>=0) {
            spstr  mstr = al.Value("NORESET");
            // empty value acts as 1 here
            u32t svmode = !mstr ? 1 : mstr.Dword();

            sto_savedword(STOKEY_CMMODE, svmode);
            rc = 0;
         }
         if (al.IndexOfName("DBCARD")>=0) {
            rc = shl_dbcard(al.Value("DBCARD"), printf);
            if (rc!=EINVAL) rc = 0;
         }
         if (al.IndexOfICase("PAE")>=0) {
            int err = pag_enable(PGRN_SHELL);
            rc = 0;
            if (err==E_SYS_INITED) printf("QSINIT already in paging mode.\n"); else
            if (err==E_SYS_EFIHOST) printf("QSINIT is EFI hosted and paging mode cannot be changed.\n"); else
            if (err==E_SYS_UNSUPPORTED) printf("There is no PAE support in CPU unit.\n");
               else rc = err;
         }
         if (al.IndexOfICase("MT")>=0) {
            qs_mtlib mtlib = get_mtlib();
            rc = 0;
            if (!mtlib) printf("MTLIB module is missing.\n"); else {
               u32t err = mtlib->initialize();
               if (err==E_SYS_DONE) printf("QSINIT already in MT mode.\n"); else
                  if (err) cmd_shellerr(EMSG_QS, err, 0);
            }
         }
      }
   }
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}

u32t _std shl_autostub(const char *cmd, str_list *args) {
   int rc=-1, idx, flags=0, ii;
   TPtrStrings al;

   str_getstrs(args,al);
   al.TrimEmptyLines();
   // print help without loading module
   if (al.Count()>=1) {
      idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
   }
   // try to load module, assigned to this command
   spstr mdname = ecmd_readstr("commands", cmd);
   if (mdname.trim().length()) {
      spstr mdfname = ecmd_readstr("MODULES", mdname());

      if (mod_query(mdfname(),MODQ_NOINCR)==0) {
         u32t error = 0;
         module *md = load_module(mdfname, &error);
         if (md) {
            /* module must replace our function in ext.cmd shell list,
               query current command processor - if it was changed - call it,
               else fall to "not loaded" message */
            cmd_eproc func = cmd_shellrmv(cmd,0);

            if (func!=shl_autostub) return cmd_shellcall(func, 0, args);
         } else
            log_printf("Error %08X on loading \"%s\" handler module \"%s\"\n",
               error, cmd, mdfname());
      }
   }
   printf("Command is not loaded.\n");
   return 0;
}

#if 0
u32t _std shl_log(const char *cmd, str_list *args) {
   int rc=-1, nopause=0, usedate=0, usetime=1, usecolor=1, level=-1, useflags=0;
   if (args->count>0) {
      TPtrStrings al;
      str_getstrs(args,al);
      // is help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      static char *argstr   = "f|d|np|nt|nc";
      static short argval[] = {1,1, 1, 0, 0};
      process_args(al, SPA_NOSLASH, argstr, argval,
                   &useflags, &usedate, &nopause, &usetime, &usecolor);

      idx = al.IndexOfName("/l");
      if (idx>=0) {
         spstr lv(al.Value(idx));
         level = !lv?3:lv.Int();
         al.Delete(idx);
      }

      if (al.Count()>0 && level>=-1 && level<=3) {
         al[0].upper();
         // init "paused" output
         pause_println(0,1);

         if (al[0]=="SAVE" || al[0]=="TYPE") {
            int is_save = al[0][0]=='S';

            if (!is_save || al.Count()==2) {
               u32t flags = LOGTF_DOSTEXT;

               if (usedate) flags|=LOGTF_DATE;
               if (usetime) flags|=LOGTF_TIME;
               if (useflags) flags|=LOGTF_FLAGS;
               if (level>=0) flags|=LOGTF_LEVEL|level; else
               if (usecolor) flags|=LOGTF_LEVEL|LOG_GARBAGE;

               char *cp = log_gettext(flags);
               TPtrStrings ft;
               ft.SetText(cp);
               free(cp);

               if (is_save) rc = ft.SaveToFile(al[1]()) ? 0 : EACCES; else {
                  u32t ii;
                  for (ii=0; ii<ft.Count(); ii++) {
                     spstr pln(ft[ii]);
                     u8t   color = 0;

                     if (usecolor) {
                        int cp = pln.cpos('[');
                        if (cp>=0 && cp<19) {
                           int lp = cp;
                           while (!isdigit(pln[lp]) && pln[lp]) lp++;
                           switch (pln[lp]) {
                              case '0': color = VIO_COLOR_YELLOW; break;
                              case '1': color = VIO_COLOR_LWHITE; break;
                              case '2': color = VIO_COLOR_WHITE; break;
                              case '3': color = VIO_COLOR_CYAN; break;
                           }
                           if (level<0) {
                              int ep = pln.cpos(']', cp);
                              if (ep>0 && ep-cp<=4) pln.del(cp,ep-cp+2);
                           }
                        }
                     }
                     if (pause_println(pln(), nopause?-1:0, color)) return EINTR;
                  }
                  rc = 0;
               }
            }
         } else
         if (al[0]=="CLEAR") {
            log_clear();
            rc = 0;
         }
      }
   }
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}
#endif

u32t _std shl_attrib(const char *cmd, str_list *args) {
   TPtrStrings al;
   int subdir=0, nopause=0, mask=0, idirs=0, a_R=0, a_A=0, a_S=0, a_H=0,
       quiet=0, idx, utf8=0;
   qserr err=0;

   if (args->count>0) {
      str_getstrs(args,al);
      // is help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "/s|/d|/np|+r|-r|+s|-s|+h|-h|+a|-a|/q|/u";
      static short argval[] = { 1, 1,  1, 1,-1, 1,-1, 1,-1, 1,-1, 1, 1};

      process_args(al, SPA_NODASH, argstr, argval,
                   &subdir, &idirs, &nopause, &a_R, &a_R, &a_S, &a_S,
                   &a_H, &a_H, &a_A, &a_A, &quiet, &utf8);
      al.TrimEmptyLines();
   }
   if (!al.Count()) al.Add("*");
   // init "paused" output
   pause_println(0,1);
   // process multiple paths
   for (idx=0; idx<al.Count(); idx++) {
      // fix and split path
      dirw_t  fi;
      spstrw  pp, drive, name;
      // errors should be rare here ...
      err = ascii2utf(al[idx], pp, utf8);

      if (!err) {
         splitfnamew(pp.replacechar('/','\\'), drive, name);
         mask = name.cpos('*')>=0 || name.cpos('?')>=0;
         
         if (!mask) {
            if (_dos_statw(pp(),&fi)) {
               if (!subdir) err = E_SYS_NOPATH;
            } else
            if (fi.d_attr&_A_SUBDIR) {
               drive = pp;
               name  = (wchar_t*)L"*.*";
               mask  = true;
            }
         }
      }
      if (!err)
         if (!drive) getcurdirw(drive); else
            err = qs_fullpathw(drive);
      if (!err) {
         TStringsW  flist;
         spstrw     dname(drive);
         if (dname.length() && dname.lastchar()!='\\') dname+='\\';
         dname+=name;

         if (mask || subdir) {
            // read tree
            dirw_t *info = 0;
            u32t     cnt = _dos_readtreew(drive(),name(),&info,subdir,0,0,0);
            if (info) {
               if (cnt) dir_to_list(drive, info, (idirs?0:_A_SUBDIR)|
                  _A_VOLID, 0, flist, 0);
               _dos_freetree(info);
            }
         } else
            flist.AddObject(dname, fi.d_attr);

         if (flist.Count()) {
            int ii;
            u32t andmask = FFFF, ormask = 0;
            if (a_R||a_A||a_S||a_H) {
               if (a_R) {
                  andmask&=~_A_RDONLY; if (a_R>0) ormask|=_A_RDONLY;
               }
               if (a_H) {
                  andmask&=~_A_HIDDEN; if (a_H>0) ormask|=_A_HIDDEN;
               }
               if (a_A) {
                  andmask&=~_A_ARCH;   if (a_A>0) ormask|=_A_ARCH;
               }
               if (a_S) {
                  andmask&=~_A_SYSTEM; if (a_S>0) ormask|=_A_SYSTEM;
               }
            }

            spstr msg;
            for (ii=0; ii<flist.Count(); ii++) {
               io_handle_info fi;
               flist[ii].replacechar('/','\\');

               qserr aerr = io_pathinfow(flist[ii](), &fi);
               if (!aerr)
                  if (andmask!=FFFF) {
                     fi.attrs = fi.attrs&andmask|ormask;
                     aerr     = io_setinfow(flist[ii](), &fi, IOSI_ATTRS);
                     msg.clear(); 
                  } else {
                     msg.sprintf("%c%c%c%c%c  %S", fi.attrs&_A_ARCH?'A':' ',
                        fi.attrs&_A_RDONLY?'R':' ', fi.attrs&_A_HIDDEN?'H':' ',
                           fi.attrs&_A_SYSTEM?'S':' ', fi.attrs&_A_SUBDIR?'D':' ',
                              flist[ii]());
                  }
               if (aerr && !quiet) qsfwerr2str(msg, flist[ii], aerr);

               if (msg.length())
                  if (pause_println(msg(),nopause?-1:0)) return EINTR;
            }
         }
      }
      if (err)
         if (!quiet) cmd_shellerr(EMSG_QS, err, 0);
   }
   // return last found error
   return qserr2errno(err); 
}

// free pushd stack
void pushd_free(void* lptr) {
   if (lptr) delete (TStringsW*)lptr;
}

u32t _std shl_pushd(const char *cmd, str_list *args) {
   int quiet=0, rc=-1, utf8=0;
   TPtrStrings al;

   if (args->count>0) {
      str_getstrs(args,al);
      // is help?
      if (al.IndexOf("/?")>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "/q|/u";
      static short argval[] = { 1, 1};
      process_args(al, 0, argstr, argval, &quiet, &utf8);
   }
   if (al.Count()<=1) {
      process_context *pq = mod_context();
      if (!pq) rc = EFAULT; else {
         spstrw cdir;
         if (!getcurdirw(cdir)) rc = get_errno(); else {
            mt_swlock();
            TStringsW *pdlist = (TStringsW*)pq->rtbuf[RTBUF_PUSHDST];
            if (!pdlist) {
               pdlist = new TStringsW;
               pq->rtbuf[RTBUF_PUSHDST] = (u32t)pdlist;
            }
            pdlist->Add(cdir);
            mt_swunlock();
         }
         if (al.Count()==1) {
            qserr err = ascii2utf(al[0], cdir, utf8);
            if (!err) err = chdir_int(cdir);
            rc = qserr2errno(err);
         }
         if (rc<0) rc = 0;
      }
   }
   if (quiet) return rc;
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}

u32t _std shl_popd(const char *cmd, str_list *args) {
   int quiet=0, rc=-1;
   TPtrStrings al;

   if (args->count>0) {
      str_getstrs(args,al);
      // is help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      idx = al.IndexOfName("/q");
      if (idx>=0) { quiet=1; al.Delete(idx); }
   }
   // get PUSHD stack from process context data
   process_context *pq = mod_context();
   if (!pq) rc = EFAULT; else {
      mt_swlock();
      TStringsW *pdlist = (TStringsW*)pq->rtbuf[RTBUF_PUSHDST];
      if (pdlist)
         if (pdlist->Count()) {
            rc = qserr2errno(chdir_int((*pdlist)[pdlist->Max()]));
            pdlist->Delete(pdlist->Max());
         }
      if (rc<0) rc = 0;
      mt_swunlock();
   }
   if (quiet) return rc;
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}

u32t _std shl_ansi(const char *cmd, str_list *args) {
   int quiet=0, rc=-1;
   TPtrStrings al;

   if (args->count>0) {
      str_getstrs(args,al);
      // is help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
   }
   if (al.Count()>0) {
      if (al[0].upper()=="ON") { vio_setansi(1); rc = 0; } else
      if (al[0]=="OFF") { vio_setansi(0); rc = 0; }
   } else {
      printf("ANSI is %s.\n", vio_setansi(-1)?"on":"off");
      rc = 0;
   }
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}

u32t _std shl_reboot(const char *cmd, str_list *args) {
   int quiet=0, rc=-1, warm=1;
   TPtrStrings al;

   if (args->count>0) {
      str_getstrs(args,al);
      // is help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "/q|WARM|COLD";
      static short argval[] = { 1,   1,   0};
      process_args(al, 0, argstr, argval, &quiet, &warm, &warm);
      al.TrimEmptyLines();
   }
   if (al.Count()==0) {
      int yn = 1;
      if (!quiet) {
         printf("Reboot (y/n)?");
         yn = ask_yn();
      }
      if (yn==1) exit_reboot(warm);
      rc = 0;
   }
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}

u32t _std shl_delay(const char *cmd, str_list *args) {
   int rc = -1;
   if (args->count==1) {
      // is it help?
      if (strcmp(args->item[0],"/?")==0) {
         cmd_shellhelp(cmd,CLR_HELP); 
         return 0; 
      } else {
         u32t value = str2ulong(args->item[0]);
         if (value && value<FFFF/1000) { usleep(value*1000); rc = 0; }
      }
   } else
   if (args->count==0) {
      usleep(CLOCKS_PER_SEC);
      rc = 0;
   }
   if (rc<0) rc = EINVAL;
   if (rc) cmd_shellerr(EMSG_CLIB,rc,0);
   return 0;
}

u32t _std shl_ver(const char *cmd, str_list *args) {
   int quiet=0;
   if (args->count>0) {
      TPtrStrings al;
      str_getstrs(args,al);
      // is it help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      idx = al.IndexOfName("/q");
      if (idx>=0) { quiet=1; al.Delete(idx); }
   }
   u32t  vlen = sys_queryinfo(QSQI_VERSTR, 0);
   char *vstr = (char*)malloc_th(vlen+1), *cpos;

   sys_queryinfo(QSQI_VERSTR,vstr);
   // cut it to "QSinit x.yy, rev zzzz"
   cpos = strrchr(vstr,',');
   if (cpos) *cpos = 0;

   if (quiet) {
      cpos = strrchr(vstr,' ');
      if (cpos) setenv("QSBUILD",cpos+1,1);
      cpos = strrchr(vstr,',');
      if (cpos) *cpos = 0;
      cpos = strchr(vstr,' ');
      if (cpos) setenv("QSVER",cpos+1,1);
   } else
      printf("%s\n", vstr);
   free(vstr);

   return 0;
}

u32t _std shl_touch(const char *cmd, str_list *args) {
   int rc=-1, quiet=0;

   if (args->count>0) {
      int nopause=0, create=1, ta=0, tc=0, tm=0, utf8=0;

      TPtrStrings al;
      str_getstrs(args,al);
      // is help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "/c|/q|/u|/np|/tm|/tc|/ta";
      static short argval[] = { 0, 1, 1,  1, IOSI_WTIME, IOSI_CTIME, IOSI_ATIME};

      process_args(al, 0, argstr, argval, &create, &quiet, &utf8, &nopause,
                   &tm, &tc, &ta);
      al.TrimEmptyLines();
      // default is all times
      if (ta+tc+tm==0) { ta = IOSI_ATIME; tc = IOSI_CTIME; tm = IOSI_WTIME; }

      int        srcmode = -1;
      io_time        tmv;
      io_handle_info ref;

      // get time and show an error about only one of /r /u /d was expected
      while (rc==-1 && al.Count()>0 && (al[0][0]=='-'||al[0][0]=='/')) {
         int nm = -1;
         switch (toupper(al[0][1])) {
            case 'R': nm = 0; break;
            case 'X': nm = 1; break;
            case 'D': nm = 2; break;
         }
         if (nm<0) break;

         if (srcmode>=0) {
            if (!quiet) printf("More than one time source key!\n");
            rc = EINVAL;
            break;
         }
         al.Delete(0);
         switch (srcmode = nm) {
            case 0: // /r reffile
               if (al.Count()<2) rc = -2; else {
                  spstrw fn;
                  qserr res = ascii2utf(al[0], fn, utf8);
                  if (!res) res = io_pathinfow(fn(), &ref);
                  if (res) {
                     rc = qserr2errno(res);
                     if (!quiet) printf("%s \"%s\"\n", res==E_SYS_NOFILE ||
                        res==E_SYS_NOPATH ? "File not found:" : "Access error to", al[0]());
                  }
                  al.Delete(0);
               }
               break;
            case 1: // /u unixtime
               if (al.Count()<2) rc = -2; else {
                  char *eptr;
                  u32t tm = strtoul(al[0](), &eptr, 0);
                  // verify conversion
                  if (*eptr) {
                     if (!quiet) printf("Invalid time value (%s)\n", al[0]());
                     rc = ENOENT;
                  } else
                     io_timetoio (&tmv, tm);
                  al.Delete(0);
               }
               break;
            case 2: // /d [dd.mm.yyyy] [hh:mm[:ss]]
               // touch /d file (just use current time)
               if (al.Count()<2) srcmode = -1; else {
                  io_timeqtoio(&tmv, sys_timeq());
                  if (al[0].countchar('.')>1) {
                     u32t day, month, year;

                     int sr = sscanf(al[0](), "%u.%u.%u", &day, &month, &year);
                     if (sr<2 || !day || !month || day>31 || month>12 ||
                        sr>2 && year<1781)
                     {
                        if (!quiet) printf("Invalid date string (%s)!\n", al[0]());
                        rc = EINVAL;
                        break;
                     }
                     if (sr>2) tmv.year = year-1781;
                     tmv.day = day;
                     tmv.mon = month-1;
                     al.Delete(0);
                  }
                  if (al.Count()>1 && al[0].countchar(':')>0) {
                     u32t hour, min, sec;

                     int sr = sscanf(al[0](), "%u:%u:%u", &hour, &min, &sec);
                     if (sr==2) sec = 0;
                     if (sr>=2) {
                        if (hour>23 || min>59 || sec>59) {
                           if (!quiet) printf("Invalid time string (%s)!\n", al[0]());
                           rc = EINVAL;
                           break;
                        }
                        tmv.sec  = sec;
                        tmv.min  = min;
                        tmv.hour = hour;
                        al.Delete(0);
                     }
                  }
               }
               break;
         }
      }
      if (srcmode<0) {
         io_timeqtoio(&tmv, sys_timeq());
         srcmode = 2;
      }
      // no error code set
      if (rc==-1 && al.Count()>0) {
         // init "paused" output
         pause_println(0,1);
         // process multiple paths
         for (idx=0; idx<al.Count(); idx++) {
            spstrw pp, drive, name;
            qserr err = ascii2utf(al[idx], pp, utf8);

            if (err && !quiet) {
               al[idx] += ": ";
               cmd_shellerr(EMSG_QS, err, al[idx]());
               continue;
            }
            // fix and split path
            splitfnamew(pp.replacechar('/','\\'), drive, name);
         
            int   mask = name.cpos('*')>=0 || name.cpos('?')>=0;
         
            if (!drive) getcurdirw(drive); else fullpathw(drive);

            TStringsW  flist;
            spstrw     dname(drive);
            if (dname.length() && dname.lastchar()!='\\') dname+='\\';
            dname+=name;
         
            if (mask) {
               // read tree
               dirw_t *info = 0;
               u32t     cnt = _dos_readtreew(drive(),name(),&info,0,0,0,0);
               if (info) {
                  if (cnt) dir_to_list(drive,info,_A_VOLID|_A_SUBDIR,0,flist,0);
                  _dos_freetree(info);
               }
            } else
               flist.AddObject(dname,0);
         
            for (u32t ii=0; ii<flist.Count(); ii++) {
               io_handle_info info;
               flist[ii].replacechar('/','\\');

               err = io_pathinfow(flist[ii](), &info);
               // create missing file (or directory)
               if ((err==E_SYS_NOPATH || err==E_SYS_NOFILE) && create)
                  if (!srcmode && (ref.attrs & IOFA_DIR))
                     err = io_mkdirw(flist[ii]());
                  else {
                     io_handle fh;
                     err = io_openw(flist[ii](),IOFM_WRITE|IOFM_CREATE_ALWAYS,&fh,0);
                     io_close(fh);
                  }
               if (!err) {
                  if (tc>0) info.ctime = srcmode>0 ? tmv : ref.ctime;
                  if (ta>0) info.atime = srcmode>0 ? tmv : ref.atime;
                  if (tm>0) info.wtime = srcmode>0 ? tmv : ref.wtime;
                  err = io_setinfow(flist[ii](), &info, ta+tc+tm);
               }
               if (err && !quiet) {
                  char *emsg = cmd_shellerrmsg(EMSG_QS, err);
                  spstr  msg;
                  msg.sprintf("Unable to set time for \"%S\". Error %X: %s",
                     flist[ii](), err, emsg?emsg:"");
                  if (emsg) free(emsg);
                  if (pause_println(msg(),nopause?-1:0)) return EINTR;
               }
            }
         }
         if (rc<0) rc = 0;
      }
   }
   if (rc<0) {
      rc = EINVAL;
      if (!quiet) cmd_shellerr(EMSG_CLIB, rc, 0);
   }
   return rc;
}

u32t _std shl_subst(const char *cmd, str_list *args) {
   qserr res = E_SYS_INVPARM;

   if (args->count>0) {
      int del=0, utf8=0;
      TPtrStrings al;
      str_getstrs(args,al);
      // is help?
      int idx = al.IndexOf("/?");
      if (idx>=0) { cmd_shellhelp(cmd,CLR_HELP); return 0; }
      // process args
      static char *argstr   = "/d|/u";
      static short argval[] = { 1, 1};

      process_args(al, 0, argstr, argval, &del, &utf8);
      al.TrimEmptyLines();

      if (al.Count() && al.Count()<=2 && al[0].length()==2 && al[0][1]==':') {
         char vol = toupper(al[0][0]);

         if (vol<'C' || vol>'Z') res = E_DSK_BADVOLNAME; else {
            u8t drive = vol - 'A';
            if (!del && al.Count()==2) {
               if (utf8) {
                  spstrw path;
                  if (!path.setutf8(al[1])) res = E_SYS_UTFSURR; else
                     res = io_substw(&drive, path());
               } else
                  res = io_subst(&drive, al[1]());
            } else 
            if (del && al.Count()==1) {
               c16t *root = io_substroot(drive);
               if (!root) res = E_DSK_FIXED; else {
                  free(root);
                  res = io_unmount(drive, IOUM_FORCE|IOUM_NOASK);
               }
            }
         }
      }
   } else {
      for (u32t vol='C'; vol<='Z'; vol++) {
         c16t *root = io_substroot(vol-'A');
         if (root) {
            printf("%c:\\ => %S\n", vol, root);
            free(root);
         }
      }
      res = 0;
   }
   if (res) cmd_shellerr(EMSG_QS, res, 0);

   return qserr2errno(res);
}


#include "zz.cpp"
#include "shpci.cpp"

extern "C"
void setup_shell(void) {
   // install shell commands
   cmd_shelladd("COPY"   , shl_copy   );
   cmd_shelladd("TYPE"   , shl_type   );
   cmd_shelladd("DIR"    , shl_dir    );
   cmd_shelladd("RESTART", shl_restart);
   cmd_shelladd("MKDIR"  , shl_mkdir  );
   cmd_shelladd("MD"     , shl_mkdir  );
   cmd_shelladd("CHDIR"  , shl_chdir  );
   cmd_shelladd("CD"     , shl_chdir  );
   cmd_shelladd("DEL"    , shl_del    );
   cmd_shelladd("ERASE"  , shl_del    );
   cmd_shelladd("BEEP"   , shl_beep   );
   cmd_shelladd("RMDIR"  , shl_rmdir  );
   cmd_shelladd("RD"     , shl_rmdir  );
   cmd_shelladd("UNZIP"  , shl_unzip  );
   cmd_shelladd("LM"     , shl_loadmod);
   cmd_shelladd("LOADMOD", shl_loadmod);
   cmd_shelladd("POWER"  , shl_power  );
   cmd_shelladd("LABEL"  , shl_label  );
   cmd_shelladd("TRACE"  , shl_trace  );
   cmd_shelladd("REN"    , shl_ren    );
   cmd_shelladd("DATE"   , shl_date   );
   cmd_shelladd("TIME"   , shl_time   );
   cmd_shelladd("MSGBOX" , shl_msgbox );
   cmd_shelladd("MODE"   , shl_mode   );
   cmd_shelladd("ATTRIB" , shl_attrib );
   cmd_shelladd("PUSHD"  , shl_pushd  );
   cmd_shelladd("POPD"   , shl_popd   );
   cmd_shelladd("ANSI"   , shl_ansi   );
   cmd_shelladd("MOVE"   , shl_move   );
   cmd_shelladd("REBOOT" , shl_reboot );
   cmd_shelladd("DELAY"  , shl_delay  );
   cmd_shelladd("SM"     , shl_sm     );
   cmd_shelladd("VER"    , shl_ver    );
   cmd_shelladd("TOUCH"  , shl_touch  );
   cmd_shelladd("SUBST"  , shl_subst  );

   // install MODE SYS handler
   cmd_modeadd("SYS", shl_mode_sys);
   /* install stubs for pre-defined external commands */
   TStrings autoload;
   ecmd_commands(autoload);
   for (int ii=0; ii<autoload.Count(); ii++) cmd_shelladd(autoload[ii](), shl_autostub);
}
