/*
   QSINIT: set /a - arithmetic expressions

   Expression parser operates with signed 64-bit integers.

   Hex values are acceptable (with 0x prefix), but octals are not
   supported, i.e. 0777=777, not 511.
   Use @int(var,0) for any variable that can be an octal string.

   The string contants are not allowed, but single characters can
   be used:
       set /a is_a = str[0]=='A' || str[0]=='a'

   Functions can only get SET variable as an argument, string expressions
   are not supported, result of any expression is always an integer value.

   Function list:
     @len[w](var)        get variable length
     @int(var,base)      convert variable to integer (c language strtoll)
     @pos[w](var,char)   first position of a character in variable (-1 if absent)
     @rpos[w](var,char)  last position of a character in variable
     @ccnt[w](var,char)  number of specified character in variable
     @random(range)      random in 0..range-1 (32-bit unsigned)
     @get[w](var,index)  return a character in variable (0 if out of range
                         or if empty, negative index value counts from the end
                         of string)
     @set[w](var,index,char)
                         replace a character in variable (negative index
                         value counts from the end of string, function returns
                         previous character or 0 if index is out of range).
     @eq[i[w]](var,var)  compare two variables as strings (return 1/0),
                         @eqi - case insensitive, @eqiw - utf-8 case insensitive
     @mem(addr,len)      read 1-8 bytes of physical memory. Note that address
                         above 4Gb will turn PAE mode on.
     @memavail(type)     return size of available memory in kb, type value
                         defines what to return: 0 - max free block size
                         for the current process, 1 - total free memory for
                         the current process, 2 - size of QSINIT memory space,
                         3 - total physical memory size, 4 - physical memory
                         below 4Gb, 5 - size of physical memory above 4Gb.

   Functions with w suffix in the name assume string as utf-8 and work
   in unicode (i.e. char value in @posw() can be 0..65535, @getw() returns
   unicode character value, etc)

   Note that [] operator does not support negative values unlike @get[w]()

   This expression parser only accepts environment variable names that do not
   start with a number and can only contain A-Z, a-z, 0-9 and # $ . @ _ ` { }

   set /ax   - hex result string (w/o 0x)
   set /ao   - octal result string

   Priority of operators:
   () @fn() a[b]       - grouping, function, indexing in a variable
   a++ a--             - postfix operators
   ! ~ +a -a ++a --a   - unary operators
   * / %               - arithmetic operators
   + -                 - arithmetic operators
   << >>               - logical shift
   < > <= >=           - relational
   == !=               - equality
   &                   - bitwise AND
   ^                   - bitwise XOR
   |                   - bitwise OR
   &&                  - logical AND
   ^^                  - logical XOR
   ||                  - logical OR
   a ? b : c           - conditional
   = *= /= %= += -=    - assignment
     &= ^= |= <<= >>=
   a,b                 - expression separator
*/

#include "qstypes.h"
#include "qserr.h"
#include "sp_str.h"
#include "sp_strw.h"
#include <stdlib.h>

#ifndef __QSINIT__
#define random RandInt
#else
#include "qsbase.h"
#include "qsshell.h"
#include "syslocal.h"
#endif

typedef enum { tErr, tEnd,
   tParL, tParR, tBracL, tBracR, tComma, tCond, tDDot, tInt, tVar,

   tAnd, tNot, tLNot, tOr, tXor, tShl, tShr,
   tAdd, tSub, tMul, tDiv, tMod, tEQ, tLT, tGT, tLE, tGE, tNE,
   tLOr, tLXor, tLAnd, tInc, tDec,
   // must be grouped here 
   tSet, tSetMul, tSetDiv, tSetMod, tSetAdd, tSetSub, tSetAnd, tSetXor,
   tSetOr, tSetShl, tSetShr,

   // functions must be at the end (odd values assumed as utf8 functions)
   tFnLen = 64, tFnLenw, tFnInt, tFnIntw, tFnPos, tFnPosw, tFnRPos, tFnRPosw,
   tFnCCnt, tFnCCntw, tFnGet, tFnGetw, tFnSet, tFnSetw, tFnEQi, tFnEQiw,
   tFnEQ = 100, tFnRnd = 102, tFnMem = 104, tFnMemAv = 106
} Token;

struct TokRef {
   const char *str;
   Token     value;
};

static TokRef tklist[] = {
   {"@LEN"     , tFnLen   },
   {"@LENW"    , tFnLenw  },
   {"@INT"     , tFnInt   },
   {"@POS"     , tFnPos   },
   {"@POSW"    , tFnPosw  },
   {"@RPOS"    , tFnRPos  },
   {"@RPOSW"   , tFnRPosw },
   {"@CCNT"    , tFnCCnt  },
   {"@CCNTW"   , tFnCCntw },
   {"@RANDOM  ", tFnRnd   },
   {"@GET"     , tFnGet   },
   {"@GETW"    , tFnGetw  },
   {"@SET"     , tFnSet   },
   {"@SETW"    , tFnSetw  },
   {"@EQ"      , tFnEQ    },
   {"@EQI"     , tFnEQi   },
   {"@EQIW"    , tFnEQiw  },
   {"@MEM"     , tFnMem   },
   {"@MEMAVAIL", tFnMemAv },
   {0, tErr}
};

class SetCalc {
   Token  tknext();          // <0 - token, >0 - qserr
   void     init();
   void   tkback();
   spstr    cstr,            // processing string
            lvar;            // variable name if tVar
   Token  nexttk, prevtk;
   s64t     lval;
   qserr    lerr;
   s64t     geti(const spstr &varname);
   void     gets(const spstr &varname, spstr &value);
   void     seti(const spstr &varname, s64t value, int radix=0);

   qserr  elvl00(s64t &val); // () @fn() a[b]
   qserr  elvl01(s64t &val); // a++ a--
   qserr  elvl02(s64t &val); // ! ~ +a -a ++a --a
   qserr  elvl03(s64t &val); // * / %
   qserr  elvl04(s64t &val); // + -
   qserr  elvl05(s64t &val); // << >>
   qserr  elvl06(s64t &val); // < > <= >=
   qserr  elvl07(s64t &val); // == !=    
   qserr  elvl08(s64t &val); // &
   qserr  elvl09(s64t &val); // ^
   qserr  elvl10(s64t &val); // |
   qserr  elvl11(s64t &val); // &&
   qserr  elvl12(s64t &val); // ^^
   qserr  elvl13(s64t &val); // ||
   qserr  elvl14(s64t &val); // a ? b : c

   qserr  eassgn(s64t &val); // = *= /= %= += -= &= ^= |= <<= >>=
   qserr  ecomma(s64t &val); // a,b
public:
   SetCalc(const char *str) : cstr(str) { init(); }
   SetCalc(const spstr &str) : cstr(str) { init(); }

   s64t   result;
   int    aradix;            // radix for all assignment ops
   qserr    calc(spstr *errpos);
   // if /a () processing
   qserr   calcp(spstr *errpos);
};

void SetCalc::init() {
   nexttk  = tEnd;
   prevtk  = tEnd;
   result  = 0;
   lerr    = 0;
   aradix  = 10;
   cstr.replacechar('\t',' ').replacechar(0x1A,' ').replacechar('\r',' ').
        replacechar('\n',' ').replacechar(0x0C,' ').trim();
}

s64t SetCalc::geti(const spstr &varname) {
#ifdef __QSINIT__
   char *ev = env_getvar(varname(), 1);
   spstr vv(ev);
   free(ev);
#else
   spstr vv(getenv(varname()));
#endif
   if (!vv) return 0;
   return vv.Int64();
}

void SetCalc::gets(const spstr &varname, spstr &value) {
   char *val = getenv(varname());
   //printf("[%s]=\"%s\"\n", varname(), val?val:"NULL");
   if (val) value = val; else value.clear();
}

void SetCalc::seti(const spstr &varname, s64t value, int radix) {
   char buf[48];
   if (!radix) radix = 10;
   lltoa(value, buf, radix);
   setenv(varname(), buf, 1);
}

void SetCalc::tkback() {
    nexttk = prevtk;
}

Token SetCalc::tknext() {
   // expect that lval/lvar are untouched
   if (nexttk!=tEnd) {
      nexttk = tEnd;
      return prevtk;
   }
   if (!cstr) return prevtk = tEnd;
   // at least one char is present, so cstr[1] will be the next one or zero
   char ch1st = toupper(cstr[0]),
          ch2 = cstr[1];
   int    len = 1;
   Token   tk = tErr;
   lval = 0;
   switch (ch1st) {
      case '(': tk = tParL; break;
      case ')': tk = tParR; break;
      case '[': tk = tBracL; break;
      case ']': tk = tBracR; break;
      case '!': if (ch2=='=') { tk = tNE; len = 2; } else tk = tLNot; break;
      case '~': tk = tNot; break;
      case '-': if (ch2=='-') { tk = tDec; len = 2; } else
                   if (ch2=='=') { tk = tSetSub; len = 2; } else tk = tSub;
                break;
      case '+': if (ch2=='+') { tk = tInc; len = 2; } else
                   if (ch2=='=') { tk = tSetAdd; len = 2; } else tk = tAdd;
                break;
      case '*': if (ch2=='=') { tk = tSetMul; len = 2; } else tk = tMul; break;
      case '/': if (ch2=='=') { tk = tSetDiv; len = 2; } else tk = tDiv; break;
      case '%': if (ch2=='=') { tk = tSetMod; len = 2; } else tk = tMod; break;
      case '<': if (ch2=='=') { tk = tLE; len = 2; } else
                   if (ch2=='>') { tk = tNE; len = 2; } else  // allow <> too
                      if (ch2=='<') {
                         if (cstr[2]=='=') { tk = tSetShl; len = 3; } else
                            { tk = tShl; len = 2; }
                      } else tk = tLT;
                break;
      case '>': if (ch2=='=') { tk = tGE; len = 2; } else
                   if (ch2=='>') {
                      if (cstr[2]=='=') { tk = tSetShr; len = 3; } else
                         { tk = tShr; len = 2; }
                   } else tk = tGT;
                break;
      case '&': if (ch2=='=') { tk = tSetAnd; len = 2; } else
                   if (ch2=='&') { tk = tLAnd; len = 2; } else tk = tAnd;
                break;
      case '|': if (ch2=='=') { tk = tSetOr; len = 2; } else
                   if (ch2=='|') { tk = tLOr; len = 2; } else tk = tOr;
                break;
      case '^': if (ch2=='=') { tk = tSetXor; len = 2; } else
                   if (ch2=='^') { tk = tLXor; len = 2; } else tk = tXor;
                break;
      case '?': tk = tCond; break;
      case ':': tk = tDDot; break;
      case ',': tk = tComma; break;
      case '=': if (ch2=='=') { tk = tEQ; len = 2; } else tk = tSet; break;
      case '\'': // character constant
         if (cstr.length()<3) { lerr = E_EXP_CHARCONST; return tErr; }
         if (ch2=='\\') {
            switch (cstr[2]) {
               case '\\': lval = '\\'; break;
               case 'a': lval =  7; break;
               case 'b': lval =  8; break;
               case 't': lval =  9; break;
               case 'n': lval = 10; break;
               case 'v': lval = 11; break;
               case 'f': lval = 12; break;
               case 'r': lval = 13; break;
               default : lval = cstr[2]; break;
            }
            if (cstr[2]!='\'') len++;
         } else
            lval = ch2;
         len+=2;
         if (cstr.length()<len || cstr[len-1]!='\'')
            { lerr = E_EXP_CHARCONST; return tErr; }
         tk = tInt;
         lvar.clear();
         break;
   }
   if (tk==tErr) {
      len--;
      // integer constant (w/o octal!)
      if (ch1st>='0' && ch1st<='9') {
         int  base = ch2=='x'||ch2=='X' ? 16 : 10;
         char  *ep;
         lval = strtoull(cstr(), &ep, base);
         len  = ep - cstr();
         tk   = tInt;
         lvar.clear();
      } else
      while (1) {
         char ch = cstr[len];
         // deny " anywhere ...
         if (ch && (ch>=0x7F || ch<' ' || ch=='\"'))
            { lerr = E_EXP_INVCHAR; return tErr; }
         // only allowed in variable name
         if (ch && (isalnum(ch) || strchr("#$.@_`{}",ch))) len++; else {
            if (len>0) {
               spstr vname(cstr,0,len);
               vname.upper();
               if (vname[0]=='@') {
                  TokRef *fn = tklist;
                  while (fn->str)
                     if (vname==fn->str) {
                        tk = fn->value;
                        break;
                     } else
                        fn++;
               }
               if (tk==tErr) {
                  lvar = vname;
                  lval = geti(lvar);
                  tk   = tVar;
               }
            }
            break;
         }
      }
   }
   if (tk==tErr) { lerr = E_EXP_SYNTAX; return tErr; }

   if (len) {
      // also trim left spaces in this .del()
      if (len<cstr.length())
         while (cstr[len]==' ') len++;
      cstr.del(0,len);
   }
   prevtk = tk;
   return tk;
}

qserr SetCalc::elvl00(s64t &val) {
   Token  tk = tknext();
   qserr err;

   if (tk==tErr) return lerr;
   // function call
   if (tk>=tFnLen) {
      Token Fn = tk;
      tk = tknext();
      if (tk==tErr) return lerr;
      if (tk!=tParL) return E_EXP_LPAR;

      if (Fn==tFnRnd) {
         err = eassgn(val);
         if (err) return err;
         if (val<0) val = 0; else
            val = random(val>FFFF ? FFFF : (u32t)val);
      } else
      if (Fn==tFnMemAv) {
         u32t  m1, m2;
         err = eassgn(val);
         if (err) return err;
#ifdef __QSINIT__
         if (val==0 || val==1) { // max/total free memory
            m1 = hlp_memavail(&m2, 0, QSMA_LOCAL);
            val = (val ? m1 : m2) >> 10;
         } else
         if (val==2) {           // QS memmgr total
            hlp_memavail(0, &m1, QSMA_PHYS);
            val = m1>>10;
         } else
         if (val>=3 && val<=5) { // PC physical memory
            m1 = sys_ramtotal(&m2);
            if (val==3) val = int12mem() + (m1<<2) + (m2<<2); else
               if (val==4) val = int12mem() + (m1<<2); else val = m2<<2;
         } else
#endif
           val = 0;
      } else
      if (Fn==tFnMem) {
         err = eassgn(val);
         if (err) return err;
         if (tknext()!=tComma) return E_EXP_COMMA; else {
            s64t clen;
            err = eassgn(clen);
            if (err) return err;
            if (clen>0) {
               u32t len = clen>8?8:clen;
               u64t pos = (u64t)val;
               val = 0;
#ifdef __QSINIT__
               // both functions return 0 if exception occured
               len = pos>=sys_endofram() ? sys_memhicopy((u32t)&val, pos, len)
                   : hlp_memcpy(&val, (void*)pos, len, MEMCPY_PG0);
               // return FF FF in all requested bytes
               if (!len) val = clen<8 ? -1LL & (1<<(u8t)clen*8) - 1 : -1LL;
#endif
            } else
               val = 0;
         }
      } else {
         tk = tknext();
         if (tk==tErr) return lerr;
         if (tk!=tVar) return E_EXP_NEEDVAR;
         spstr   v1, vname(lvar);
         spstrw v1w;
         gets(lvar,v1);
         // utf-8 function?
         if (Fn & 1) v1w.setutf8(v1);

         if (Fn==tFnLen) val = v1.length(); else
         if (Fn==tFnLenw) val = v1w.length(); else
         if (tknext()!=tComma) return E_EXP_COMMA; else
         if (Fn==tFnEQ || Fn==tFnEQi || Fn==tFnEQiw) {
            tk = tknext();
            if (tk==tErr) return lerr;
            if (tk!=tVar) return E_EXP_NEEDVAR;
            spstr v2;
            gets(lvar,v2);
         
            if (Fn==tFnEQ) val = v1==v2 ? 1 : 0; else
            if (v1.length()!=v2.length()) val = 0; else
            if (Fn==tFnEQi) val = stricmp(v1(),v2()) ? 0 : 1; else {
               spstrw v2w;
               v2w.setutf8(v2);
               val = _wcsicmp(v1w(),v2w()) ? 0 : 1;
            }
         } else {
            err = eassgn(val);
            if (err) return err;
            switch (Fn) {
               case tFnInt :
                  if (val && (val<2 || val>36)) val = 0; else
                     val = strtoull(v1(), 0, (int)val);
                  break;
               case tFnPos  :
               case tFnRPos :
                  if (val<0 || val>=0x100) val = -1; else
                     val = Fn==tFnPos ? v1.cpos((char)val) : v1.crpos((char)val);
                  break;
               case tFnPosw :
               case tFnRPosw:
                  if (val<0 || val>=0x10000) val = -1; else
                     val = Fn==tFnPosw ? v1w.cpos((wchar_t)val) :
                                         v1w.crpos((wchar_t)val);
                  break;
               case tFnGet :
               case tFnGetw: 
               case tFnSet :
               case tFnSetw: {
                  int  lv = Fn&1 ? v1w.length() : v1.length();
                  s64t sv;

                  if (Fn==tFnSet || Fn==tFnSetw) {
                     if (tknext()!=tComma) return E_EXP_COMMA;
                     err = eassgn(sv);
                     if (err) return err;
                  }
                  if (val<0 && -val>lv || val>=0 && val>=lv) val = 0; else {
                     lv  = val<0?lv+val:val;
                     val = Fn&1 ? v1w[lv] : v1[lv];

                     if (Fn==tFnSetw) {
                        v1w[lv] = (wchar_t)sv;
                        setenv(vname(), v1w.getutf8()(), 1);
                     } else
                     if (Fn==tFnSet) {
                        v1[lv] = (char)sv;
                        setenv(vname(), v1(), 1);
                     }
                  }
                  break;
               }
               case tFnCCnt :
                  val = val<0 || val>=0x100 ? 0 : v1.countchar((char)val);
                  break;
               case tFnCCntw:
                  val = val<0 || val>=0x10000 ? 0 : v1w.countchar((wchar_t)val);
                  break;
               default: // unknown function ...
                  return E_EXP_SYNTAX;
            }
         }
      }
      lvar.clear();
      tk = tknext();
      if (tk==tErr) return lerr;
      if (tk!=tParR) return E_EXP_RPAR;
      return 0;
   } else
   if (tk==tParL) {
      err = ecomma(val);
      if (err) return err;
      tk = tknext();
      if (tk==tErr) return lerr;
      if (tk!=tParR) return E_EXP_RPAR;
   } else
   if (tk==tVar) {
      spstr vname(lvar);
      val = lval;
      tk  = tknext();
      if (tk==tErr) return lerr;
      // var[x]
      if (tk==tBracL) {
         err = ecomma(val);
         if (err) return err;
         if (val<0) val = 0; else {
            spstr vv;
            gets(vname,vv);
            val = val>=vv.length() ? 0 : vv[(u32t)val];
         }
         tk = tknext();
         if (tk==tErr) return lerr;
         if (tk!=tBracR) return E_EXP_RBRAC;
      } else
         tkback();
   } else
   if (tk==tInt) val = lval;
      else return E_EXP_SYNTAX;

   return 0;
}

/// var++ var--
qserr SetCalc::elvl01(s64t &val) {
   qserr rc = elvl00(val);
   if (rc) return rc;
   Token tk = tknext();

   while (tk==tInc || tk==tDec) {
      if (!lvar) return E_EXP_LVALUE;
      val = geti(lvar);
      seti(lvar, val+(tk==tInc?1:-1), 0);
      tk = tknext();
      lvar.clear(); // drop l-value
   }
   if (tk==tErr) return lerr;
   tkback();
   return 0;
}

/// !a ~a -a +a ++var --var
qserr SetCalc::elvl02(s64t &val) {
   Token   tk = tknext(), opstack[16];
   u32t oppos = 0;

   while (tk==tAdd || tk==tSub || tk==tNot || tk==tLNot || tk==tInc || tk==tDec) {
      if (oppos<16) opstack[oppos++] = tk; else return E_EXP_PREFIX;
      tk = tknext();
   }
   if (tk==tErr) return lerr;
   tkback();

   qserr   rc = elvl01(val);
   if (rc) return rc;

   while (oppos--) {
      tk = opstack[oppos];
      if (tk==tInc || tk==tDec) {
         if (!lvar) return E_EXP_LVALUE;
         val = geti(lvar) + (tk==tInc?1:-1);
         seti(lvar, val, 0);
      } else
      switch (tk) {
         case tSub : val=-val; break;
         case tNot : val=~val; break;
         case tLNot: val=val?0:1; break;
      }
      lvar.clear(); // drop l-value
   }

   return 0;
}

/// * / %
qserr SetCalc::elvl03(s64t &val) {
   qserr rc = elvl02(val);
   if (rc) return rc;
   Token tk = tknext();

   while (tk==tMul || tk==tDiv || tk==tMod) {
      s64t v2;
      rc = elvl02(v2);
      if (rc) return rc;
      switch (tk) {
         case tMul : val = val*v2; break;
         case tDiv : val = val/v2; break;
         case tMod : val = val%v2; break;
      }
      tk = tknext();
      lvar.clear();
   }
   if (tk==tErr) return lerr;
   tkback();
   return 0;
}

/// + -
qserr SetCalc::elvl04(s64t &val) {
   qserr rc = elvl03(val);
   if (rc) return rc;
   Token tk = tknext();

   while (tk==tSub || tk==tAdd) {
      s64t v2;
      rc = elvl03(v2);
      if (rc) return rc;
      if (tk==tSub) val-=v2; else val+=v2;
      tk = tknext();
      lvar.clear();
   }
   if (tk==tErr) return lerr;
   tkback();
   return 0;
}

/// << >>
qserr SetCalc::elvl05(s64t &val) {
   qserr rc = elvl04(val);
   if (rc) return rc;
   Token tk = tknext();

   while (tk==tShr || tk==tShl) {
      s64t v2;
      rc = elvl04(v2);
      if (rc) return rc;
      /* negative shift value is undefined case: 4>>-1 prints 8 in icc and 0
         in watcom */
      if (v2<0) { v2=-v2; tk = tk==tShl?tShr:tShl; }
      if (v2) {
         u8t shift = (u8t)v2;
         if (tk==tShl) val<<=shift; else val>>=shift;
      }
      tk = tknext();
      lvar.clear();
   }
   if (tk==tErr) return lerr;
   tkback();
   return 0;
}

/// < > <= >=
qserr SetCalc::elvl06(s64t &val) {
   qserr rc = elvl05(val);
   if (rc) return rc;
   Token tk = tknext();

   while (tk==tLT || tk==tGT || tk==tLE || tk==tGE) {
      s64t v2;
      rc = elvl05(v2);
      if (rc) return rc;
      switch (tk) {
         case tLT: val = val<v2 ?1:0; break;
         case tLE: val = val<=v2?1:0; break;
         case tGT: val = val>v2 ?1:0; break;
         case tGE: val = val>=v2?1:0; break;
      }
      tk = tknext();
      lvar.clear();
   }
   if (tk==tErr) return lerr;
   tkback();
   return 0;
}

/// == !=
qserr SetCalc::elvl07(s64t &val) {
   qserr rc = elvl06(val);
   if (rc) return rc;
   Token tk = tknext();

   while (tk==tEQ || tk==tNE) {
      s64t v2;
      rc = elvl06(v2);
      if (rc) return rc;
      val = (tk==tEQ?val==v2:val!=v2)?1:0;
      tk = tknext();
      lvar.clear();
   }
   if (tk==tErr) return lerr;
   tkback();
   return 0;
}

#define OPFN(name,call,opname,op) \
qserr SetCalc::name(s64t &val) {  \
   qserr rc = call(val);          \
   if (rc) return rc;             \
   Token tk = tknext();           \
                                  \
   while (tk==opname) {           \
      s64t v2;                    \
      rc = call(v2);              \
      if (rc) return rc;          \
      val = op;                   \
      tk = tknext();              \
      lvar.clear();               \
   }                              \
   if (tk==tErr) return lerr;     \
   tkback();                      \
   return 0;                      \
}

OPFN(elvl08,elvl07,tAnd ,val&v2)
OPFN(elvl09,elvl08,tXor ,val^v2)
OPFN(elvl10,elvl09,tOr  ,val|v2)
OPFN(elvl11,elvl10,tLAnd,val&&v2?1:0)
OPFN(elvl12,elvl11,tLXor,(val?1:0)^(v2?1:0))
OPFN(elvl13,elvl12,tLOr ,val||v2?1:0)

/// a ? b : c
qserr SetCalc::elvl14(s64t &val) {
   qserr rc = elvl13(val);
   if (rc) return rc;
   Token tk = tknext();
   if (tk==tErr) return lerr;
   if (tk==tCond) {
      s64t v1, v2;
      rc = elvl13(v1);
      if (rc) return rc;
      if (tknext()!=tDDot) return E_EXP_DDOT;
      rc = elvl14(v2);
      if (rc) return rc;
      val = val ? v1 : v2;
      lvar.clear();
   } else
      tkback();
   return 0;
}

/// = *= /= %= += -= &= ^= |= <<= >>=
qserr SetCalc::eassgn(s64t &val) {
   qserr  rc = elvl14(val);
   if (rc) return rc;
   spstr dst(lvar);
   Token  tk = tknext();
   if (tk==tErr) return lerr;
   if (tk>=tSet && tk<=tSetShr) {
      if (!dst) return E_EXP_LVALUE;
      rc = eassgn(val);
      if (rc) return rc;

      if (tk!=tSet) {
         s64t dv = geti(dst);
         switch (tk) {
            case tSetMul: val = dv*val; break;
            case tSetDiv: val = dv/val; break;
            case tSetMod: val = dv%val; break;
            case tSetAdd: val = dv+val; break;
            case tSetSub: val = dv-val; break;
            case tSetAnd: val = dv&val; break;
            case tSetXor: val = dv^val; break;
            case tSetOr : val = dv|val; break;
            case tSetShl:
            case tSetShr:
               if (val<0) { val=-val; tk = tk==tSetShl?tSetShr:tSetShl; }
               if (val) {
                  u8t shift = (u8t)val;
                  if (tk==tSetShl) val=dv<<shift; else val=dv>>shift;
               }
               break;
         }
      }
      seti(dst, val, aradix);
   } else
      tkback();
   return 0;
}

/// a,b
qserr SetCalc::ecomma(s64t &val) {
   qserr  rc = eassgn(val);
   if (rc) return rc;
   Token  tk = tknext();
   while (tk==tComma) {
      rc = eassgn(val);
      if (rc) return rc;
      tk = tknext();
   }
   if (tk==tErr) return lerr;
   tkback();
   return 0;
}

qserr SetCalc::calc(spstr *errpos) {
   lerr = ecomma(result);
   if (!lerr && tknext()!=tEnd) lerr = E_EXP_SYNTAX;
   if (lerr && errpos) *errpos = cstr;
   return lerr;
}

qserr SetCalc::calcp(spstr *errpos) {
   if (tknext()!=tParL) lerr = E_EXP_LPAR; else {
      tkback();
      lerr = elvl00(result);
   }
   // always return pos where we stopped
   if (errpos) *errpos = cstr;
   return lerr;
}

#ifdef __QSINIT__
extern "C"
qserr _std START_EXPORT(env_calcexpr)(const char *str, u32t radix,
                                      char **errpos, s64t *result)
{
//   hlp_dumpregs();
   SetCalc expr(str);
   spstr  errst;
   int   parens = radix & EXPR_PARENS ?  1 : 0;
   radix &= ~EXPR_PARENS;

   if (radix>=2 && radix<=36) expr.aradix = radix;
   qserr    res = parens ? expr.calcp(&errst) : expr.calc(&errst);

   if ((res || parens) && errpos) *errpos = strdup(errst()); else *errpos = 0;
   if (result) *result = res?-1:expr.result;

   return res;
}

#else
// just a test (wcl386 -DSRT setexpr.cpp sp_defs.cpp sp_str.cpp sp_strw.cpp)
void main(int argc, char *argv[]) {
   if (argc!=2)
      printf("use: setexpr \"expr\"\n");
   else {
      SetCalc expr(argv[1]);
      spstr  errst;
      int    ptest = argv[1][0]=='(';
      qserr    res = ptest ? expr.calcp(&errst) : expr.calc(&errst);
      
      if (res) printf("Error %05X: %-20s\n", res, errst()); else
         if (ptest) printf("%Li (%s)\n", expr.result, errst()); else
            printf("%Li\n", expr.result);
   }
}
#endif
