#include "pci.h"
#include "acpi.h"
#include "qslog.h"
#include "syslocal.h"
#include "qsstor.h"
#include "qsint.h"
#include "qsbase.h"

static u32t  num_buses = 0;
static u8t  *funcmask  = 0;  ///<256 * 32 array of bits for function presence
static 
struct ACPI_MCFG *mcfg = 0;
static u32t     mcflen = 0;  ///< # of entries in MCFG table

typedef u8t    pciebus[32][8][4096];
static pciebus* busmap[256]; ///< mapped address or 0

static u8t *read_pcie_space(u8t bus, u8t slot, u8t func) {
   u64t mem = hlp_pcigetmem(0, bus, slot, func);
   if (mem) {
      u32t *mm = malloc_th(4096),
           ven = hlp_pciread(bus, slot, func, 0, 4);
      int   ok = 0;
      if (hlp_hosttype()==QSHT_EFI && mem>=_4GBLL)
         ok = sys_memhicopy((u32t)mm, mem, 4096);
      else {
         void *mp = pag_physmap(mem, 4096, 0);
         if (mp) {
            ok = hlp_memcpy(mm, mp, 4096, MEMCPY_DI);
            pag_physunmap(mp);
         }
      }
      if (ok)
         if (mm[0]!=ven || mm[64]==FFFF) ok = 0;
      if (ok) return (u8t*)mm;
      free(mm);
   }
   return 0;
}

static void* map_pcie_space(u16t seg, u8t bus, u8t slot, u8t func, u32t len) {
   u64t mem = hlp_pcigetmem(seg, bus, slot, func);
   if (mem)
      if (hlp_hosttype()==QSHT_EFI && mem>=_4GBLL) return 0; else {
         u32t ven = seg ? 0 : hlp_pciread(bus, slot, func, 0, 4);
         u32t *mp = (u32t*)pag_physmap(mem, len?len:4096, 0);
         if (mp) {
            if (ven && mp[0]!=ven /*|| mp[64]==FFFF*/) {
               pag_physunmap(mp);
               return 0;
            }
            return mp;
         }
      }
   return 0;
}

static void pci_devdump(printf_function pfn, u8t bus, u8t slot, u8t func) {
   u32t    buf[4], *mm;
   int      ii;
   if (pfn==0) pfn = log_printf;
   pfn("pci %02X:%02X.%d config space:\n", bus, slot, func);
   for (ii=0; ii<64; ii++) {
      buf[ii&3] = hlp_pciread(bus, slot, func, ii<<2, 4);
      if ((ii&3)==3) pfn("%03X: %16b\n", (ii&~3)<<2, &buf);
   }
   mm = (u32t*)read_pcie_space(bus, slot, func);
   if (mm) {
      for (ii=0; ii<240; ii++)
         pfn("%03X: %16b\n", ii+16<<4, mm+(ii+16<<2));
      free(mm);
   }
}

// do not use max values here for debug reason
void _std log_pcidump(printf_function pfn) {
   u32t bus, slot, func;
   log_it(3, "log_pcidump() full scan\n");
   if (pfn==0) pfn = log_printf;
   for (bus=0; bus<256; bus++)
      for (slot=0; slot<32; slot++)
         for (func=0; func<8; func++) {
            if (hlp_pciread(bus,slot,func,PCI_CLASS_REV,4) == 0xFFFFFFFF)
               continue;
            pci_devdump(pfn, bus, slot, func);

            if (func == 0)
               if ((hlp_pciread(bus,slot,func,PCI_HEADER_TYPE,1)&0x80) == 0)
                  break;
         }
}

static void init_enum(void) {
   static int processing = 0;
   mt_swlock();
   // simultaneous init call from different threads?
   if (processing) {
      mt_swunlock();
      while (processing) usleep(32000);
      return;
   } else {
      u32t bus, slot, func;
      u32t scan_all = sto_dword(STOKEY_PCISCAN),
            max_bus = 256, rc = 0;
      processing = 1;
      mt_swunlock();
      // init counters
      if (!funcmask) funcmask = (u8t*)malloc(256*32);
      mem_zero(funcmask);
      // ask PCI BIOS
      if (!scan_all && hlp_hosttype()!=QSHT_EFI) {
         max_bus = (hlp_querybios(QBIO_PCI)>>8&0xFF) + 1;
         log_it(2, "PCI BIOS max bus : %u\n", max_bus - 1);
      }

      for (bus=0; bus<max_bus; bus++)
         for (slot=0; slot<32; slot++)
            for (func=0; func<8; func++) {
               if (hlp_pciread(bus,slot,func,PCI_CLASS_REV,4) == 0xFFFFFFFF)
                  continue;
               // update max values
               funcmask[bus<<5|slot]|=1<<func;
               if (rc<=bus) rc = bus+1;

               if (func == 0)
                  if ((hlp_pciread(bus,slot,func,PCI_HEADER_TYPE,1)&0x80) == 0)
                     break;
            }
      log_it(2, "number of buses: %d\n", rc);
      num_buses = rc;
      processing = 0;
   }
}

/** enum PCI devices.
    @param  offsets  array of offsets in configuration data
    @param  sizes    array of sizes in configuration data
    @param  values   array of values to compare with configuration data
    @param  count    number of entries in offsets/sizes/values arrays, can be 0
    @param  index    zero-based index of device with required parameters
    @param  dev      device data (output, optional)
    @param  usenext  flag 1 to start enumeration from dev->bus,slot,func
    @param  nodupes  detect devices, mapped to all 8 functions of selected slot
    @return success flag (1/0) */
static int def_enum(u8t *offsets, u8t *sizes, u32t *values, int count,
                    u16t index, pci_location *dev, int usenext, int nodupes)
{
   u32t bus, slot, func,
        prevbus, prevslot;
   if (usenext && !dev || count<0) return 0;
   if (!num_buses) init_enum();
   // increment location
   if (usenext) {
      prevbus  = dev->bus;
      prevslot = dev->slot;
      if (++dev->func>=8) {
         dev->func = 0;
         if (++dev->slot>=32) {
            dev->slot = 0;
            if (++dev->bus>=num_buses) return 0;
         }
      }
      usenext = 1;
   }
   // searching
   for (bus=usenext==1?dev->bus:0; bus<num_buses; bus++) {
      for (slot=usenext==1?dev->slot:0; slot<32; slot++,usenext|=2)
         // at least one bit present
         if (funcmask[bus<<5|slot])
         for (func=usenext==1?dev->func:0; func<8; func++) {
            // used on duplicate check below
            if (usenext==1) usenext++;
            // is entry present?
            if ((funcmask[bus<<5|slot]&1<<func)!=0) {
               u32t clrev = hlp_pciread(bus,slot,func,PCI_CLASS_REV,4);
               if (clrev == 0xFFFFFFFF) continue; else {
                  int ii, success = !count;
                  for (ii=0; ii<count; ii++)
                     if (hlp_pciread(bus,slot,func,offsets[ii],sizes[ii])!=values[ii])
                        break; else
                     if (ii==count-1) success = 1;

                  if (success) {
                     u32t vendor = hlp_pciread(bus,slot,func,PCI_VENDOR_ID,4);
                     // additional check
                     if ((vendor&0xFFFF)==0 || (vendor&0xFFFF)==0xFFFF) continue;
                     // check for eight equal functions in one slot
                     if (usenext && nodupes && dev && func>0 && (funcmask[bus<<5|slot]&1)!=0)
                        if (prevbus==bus && prevslot==slot && dev->vendorid==(vendor&0xFFFF)
                           && dev->deviceid == vendor>>16)
                        {
                           u32t func0[64], funcx[64];
                           int kk;
                           for (kk=0; kk<64; kk++) {
                              funcx[kk] = hlp_pciread(bus, slot, func, kk<<2, 4);
                              func0[kk] = hlp_pciread(bus, slot, 0   , kk<<2, 4);
                           }
                           if (!memcmp(funcx, func0, 256)) continue;
                        }

                     if (index) index--; else {
                        if (dev) {
                           dev->bus       = bus;
                           dev->slot      = slot;
                           dev->func      = func;
                           dev->classcode = clrev>>16;
                           dev->progface  = clrev>>8;
                           dev->vendorid  = vendor;
                           dev->deviceid  = vendor>>16;
                           dev->header    = hlp_pciread(bus,slot,func,PCI_HEADER_TYPE,1);
                        }
                        return 1;
                     }
                  }
                  if (func == 0)
                     if ((hlp_pciread(bus,slot,func,PCI_HEADER_TYPE,1)&0x80) == 0)
                        break;
               }
            }
         }
   }
   return 0;
}

int _std START_EXPORT(hlp_pcifind)(u16t vendor_id, u16t device_id, u16t index,
                                   pci_location *dev, int nodupes)
{
   static u8t offsets[] = {PCI_VENDOR_ID, PCI_DEVICE_ID},
                sizes[] = {2, 2};
   u32t       values[2];
   values[0] = vendor_id;
   values[1] = device_id;
   return def_enum(offsets, sizes, values, 2, index, dev, 0, nodupes);
}

int _std hlp_pciclass(u32t classcode, u16t index, pci_location *dev, int nodupes) {
   static u8t offset = PCI_DEV_CLASS, size = 2;
   return def_enum(&offset, &size, &classcode, 1, index, dev, 0, nodupes);
}

/** return offset of extended PCI capability.
    "start" position here is not included in search and must be 0 or point
    to a capability header */
static u16t find_ext_cap(void *mem, u16t start, u16t cap) {
   u32t header;
   u16t    pos;
   // minimum 8 bytes per capability
   int   lcnt = 4096 - 256 >> 3;

   if (start && (start<256 || start>4096-8)) return 0;
   pos = start ? start : 256;
   // no more than max loops, for a case of incorrect data
   while (lcnt-->0) {
      u32t header = ((u32t*)mem)[pos>>2];
      if (!header) break;

      if ((header&0xFFFF)==cap && pos!=start) return pos;

      pos = header>>20 & 0xFFC;
      if (pos<256) break;
   }
   return 0;
}

static u32t _std hlp_pcigetbase_int(pci_location *dev, u64t *bases,
   u64t *sizes, u8t *indexmap, u32t *sizelist, u16t *capofs)
{
   if (!dev||!bases||!sizes) return 0;
   if (!num_buses) init_enum();
   if (!hlp_pciexist(dev)) return 0;

   if (dev->bus>=num_buses) return 0; else {
      u32t vendor = hlp_pciread(dev->bus,dev->slot,dev->func,PCI_VENDOR_ID,4),
            clrev = hlp_pciread(dev->bus,dev->slot,dev->func,PCI_CLASS_REV,4),
           maxcnt = 0,
          rccount = 0;
      int   svint;
      u8t     pos, imap[6];

      if (clrev == 0xFFFFFFFF) return 0;
      dev->vendorid  = vendor;
      dev->deviceid  = vendor>>16;
      if (dev->vendorid==0 || dev->vendorid==0xFFFF) return 0;
      dev->classcode = clrev>>16;
      dev->progface  = clrev>>8;
      dev->header    = hlp_pciread(dev->bus,dev->slot,dev->func,PCI_HEADER_TYPE,1);

      switch (dev->header&0x7F) {
         case PCI_HEADER_NORMAL : maxcnt = 6; break;
         case PCI_HEADER_BRIDGE : maxcnt = 2; break;
         case PCI_HEADER_CARDBUS: maxcnt = 1; break;
         default: return 0;
      }
      if (sizelist && !indexmap) indexmap = imap;

      /* address read is not an atomic op, but just disable ints here, instead
         of MT lock */
      svint = sys_intstate(0);

      for (pos=PCI_BASE_ADDR0; maxcnt>0; maxcnt--,pos+=4) {
         u32t addr = hlp_pciread(dev->bus, dev->slot, dev->func, pos, 4),
              size = 0;
         if (!addr) continue;

         hlp_pciwrite(dev->bus, dev->slot, dev->func, pos, 4, FFFF);
         size = hlp_pciread(dev->bus, dev->slot, dev->func, pos, 4);
         hlp_pciwrite(dev->bus, dev->slot, dev->func, pos, 4, addr);

         if (indexmap) indexmap[rccount] = pos - PCI_BASE_ADDR0 >> 2;

         if ((addr&PCI_ADDR_SPACE)==0) {
            u64t sizeh = 0,
                 addrh = 0;

            if ((addr&PCI_ADDR_MEMTYPE_MASK)==PCI_ADDR_MEMTYPE_64) {
               pos+=4; maxcnt--;
               addrh = hlp_pciread(dev->bus, dev->slot, dev->func, pos, 4);
               hlp_pciwrite(dev->bus, dev->slot, dev->func, pos, 4, FFFF);
               sizeh = hlp_pciread(dev->bus, dev->slot, dev->func, pos, 4);
               hlp_pciwrite(dev->bus, dev->slot, dev->func, pos, 4, addrh);
            } else
               sizeh = FFFF;
            sizeh = sizeh<<32 | size & PCI_ADDR_MEM_MASK;
            addrh = addrh<<32 | addr;

            bases[rccount] = addrh;
            /* sometimes high bits of the size can be erroneously cleared, so
               use the lowest one as linux does */
            sizes[rccount] = sizeh & ~(sizeh-1);
         } else {
            size = ~(size & PCI_ADDR_IO_MASK) + 1 & 0xFFFF;
            if (!size) continue;
            bases[rccount] = addr;
            sizes[rccount] = size;
         }
         rccount++;
      }
      sys_intstate(svint);

      if (sizelist && rccount) {
         u32t *mm = (u32t*)read_pcie_space(dev->bus, dev->slot, dev->func);
         // default to zero
         memset(sizelist, 0, sizeof(u32t)*rccount);
         if (capofs) memset(capofs, 0, sizeof(u16t)*rccount);

         if (mm) {
            u16t rcofs = find_ext_cap(mm, 0, PCIE_CAP_ID_REBAR);
            if (rcofs) {
               u32t   ii, mi,
                   nbars = (mm[rcofs + PCI_RBAR_CTRL >> 2] &
                            PCI_RBAR_CTRL_CNT_MASK) >> PCI_RBAR_CTRL_CNT_SHIFT;
                   
               for (ii=0; ii<nbars; ii++, rcofs+=8) {
                  u32t ctrl = mm[rcofs + PCI_RBAR_CTRL >> 2],
                        cap = mm[rcofs + PCI_RBAR_CAP >> 2];
                  u8t   idx = ctrl & PCI_RBAR_CTRL_BAR_IDX;

                  cap = (cap & PCI_RBAR_CAP_SIZES) >> PCI_RBAR_CAP_SHIFT;
                  if (!cap) continue;
                  /* fix from linux code: "Sapphire RX 5600 XT Pulse has an
                     invalid cap dword for BAR 0" */
                  if (dev->vendorid==0x1002 && dev->deviceid==0x731F &&
                     idx==0 && cap == 0x700) cap = 0x3F00;

                  for (mi=0; mi<rccount; mi++)
                     if (indexmap[mi]==idx) {
                        sizelist[mi] = cap;
                        if (capofs) capofs[mi] = rcofs + PCI_RBAR_CAP;
                     }
               }
            }
            free(mm);
         }
      }
      return rccount;
   }
}

u32t _std hlp_pcigetbaseex(pci_location *dev, u64t *bases, u64t *sizes,
                           u8t *indexmap, u32t *sizelist)
{
   return hlp_pcigetbase_int(dev, bases, sizes, indexmap, sizelist, 0);
}

u32t _std hlp_pcigetbase(pci_location *dev, u64t *bases, u64t *sizes) {
   return hlp_pcigetbase_int(dev, bases, sizes, 0, 0, 0);
}

qserr _std hlp_pcisetbase(pci_location *dev, u8t index, u64t base, u64t size) {
   int sbit = size ? bsf64(size) : -1;
   if (!base) return E_SYS_INVPARM;
   if (sbit>=0) {
      // deny size invalid size cases
      if (sbit!=bsr64(size) || sbit<20 || sbit>=40 || (base&PCI_ADDR_SPACE_IO))
         return E_SYS_INVPARM;
      sbit -= 20;
   }   
   if (!dev) return E_SYS_ZEROPTR;
   if (!num_buses) init_enum();

   if (!hlp_pciexist(dev)) return E_SYS_NOTFOUND; else {
      u64t   bases[6], sizes[6], addr;
      u16t  capofs[6], devcmd;
      u32t   setsz[6], count, ii,
             *emap = 0;
      int    svint,
            devoff = 0;
      u8t     imap[6],
               pos;
      count = hlp_pcigetbase_int(dev, bases, sizes, imap, sbit>=0?setsz:0,
                                 sbit>=0?capofs:0);
      // invalid index
      if (index >= count) return E_SYS_NOTFOUND;
      // verify type match!
      if ((bases[index] ^ base) & PCI_ADDR_SPACE) return E_SYS_INVPARM;
      if (base & PCI_ADDR_SPACE_IO) addr = base & PCI_ADDR_IO_MASK; else {
         if ((bases[index] ^ base) & PCI_ADDR_MEMTYPE_MASK)
            return E_SYS_INVPARM;
         if (sbit>=0 && (setsz[index] & 1<<sbit)==0) return E_SYS_UNSUPPORTED;
         addr = base & ~(u64t)0xF;
      }
      
      // and alignment to BAR size
      if (addr & (size ? size-1 : sizes[index]-1)) return E_SYS_ALIGN;

      if (sbit>=0)
         if ((emap = (u32t*)map_pcie_space(0, dev->bus, dev->slot, dev->func, 0))==0) {
            log_it(2, "map failed for %u.%u.%u\n", dev->bus, dev->slot, dev->func);
            return E_SYS_UNSUPPORTED;
         }
      // disable bus access only for non-atomic ops
      devoff = sbit>=0 || (base & 7) == PCI_ADDR_MEMTYPE_64;
      // calc BAR pos
      pos    = imap[index]*4 + PCI_BASE_ADDR0;
      // disable ints
      svint  = sys_intstate(0);

      if (devoff) {
         devcmd = hlp_pciread(dev->bus, dev->slot, dev->func, PCI_COMMAND, 2);
         hlp_pciwrite(dev->bus, dev->slot, dev->func, PCI_COMMAND, 2,
            devcmd & ~PCI_CMD_MEMORY);
      }
      if (sbit>=0) {
         u32t *ctrl = emap + (capofs[index]>>2) + 1,
                 cv = *ctrl;
         cv = cv &~ PCI_RBAR_CTRL_BAR_SIZE | sbit<<PCI_RBAR_CTRL_BAR_SHIFT;
         *ctrl = cv;
      }      
      hlp_pciwrite(dev->bus, dev->slot, dev->func, pos, 4, (u32t)base);
      if ((base & 7) == PCI_ADDR_MEMTYPE_64)
         hlp_pciwrite(dev->bus, dev->slot, dev->func, pos+4, 4, base>>32);

      // restore bus access
      if (devoff && (devcmd & PCI_CMD_MEMORY))
         hlp_pciwrite(dev->bus, dev->slot, dev->func, PCI_COMMAND, 2, devcmd);
         
      sys_intstate(svint);
      if (emap) pag_physunmap(emap);
      return 0;
   }
}

int _std START_EXPORT(hlp_pcigetnext)(pci_location *dev, int init, int nodupes) {
   if (!dev) return 0;
   return def_enum(0, 0, 0, 0, 0, dev, !init, nodupes);
}

int  _std hlp_pciexist(pci_location *dev) {
   if (!dev) return 0;
   if (!num_buses) init_enum();
   if (dev->bus>=num_buses || dev->slot>=32 || dev->func>=8) return 0;

   if ((funcmask[dev->bus<<5|dev->slot]&1<<dev->func)==0) return 0; else {
       u32t clrev = hlp_pciread(dev->bus,dev->slot,dev->func,PCI_CLASS_REV,4),
           vendor;
       if (clrev == 0xFFFFFFFF) return 0;
       vendor = hlp_pciread(dev->bus,dev->slot,dev->func,PCI_VENDOR_ID,4);
       // additional check
       if ((vendor&0xFFFF)==0 || (vendor&0xFFFF)==0xFFFF) return 0;
       dev->classcode = clrev>>16;
       dev->progface  = clrev>>8;
       dev->vendorid  = vendor;
       dev->deviceid  = vendor>>16;
       dev->header    = hlp_pciread(dev->bus,dev->slot,dev->func,PCI_HEADER_TYPE,1);
       return 1;
   }
}

u32t _std hlp_pcifindaddr(u64t addr, pci_location *dev) {
   int init = 0;
   if (!dev) return 0;
   if (addr<(u64t)sys_endofram()) return 0;

   while (def_enum(0,0,0,0,0,dev,init,1)) {
      u64t base[6], size[6];
      u32t cnt = hlp_pcigetbase(dev, base, size), ii;

      for (ii=0; ii<cnt; ii++)
         if ((base[ii]&PCI_ADDR_SPACE)==PCI_ADDR_SPACE_MEMORY) {
            u64t start = base[ii] & (((u64t)FFFF<<32)+PCI_ADDR_MEM_MASK);
            if (addr>=start && addr<start+size[ii]) return ii+1;
         }
      init = 1;
   }
   return 0;
}

u32t  _std hlp_pciatoloc(const char *str, pci_location *dev) {
   if (!dev) return PCILOC_ERROR;
   memset(dev, 0, sizeof(pci_location));
   if (!str || !strlen(str)) return PCILOC_ERROR;

   if (strchr(str,':') || strchr(str,'.')) {
      char *cp = strchr(str,':');
      if (cp) {
         dev->vendorid = strtoul(str,0,16);
         dev->deviceid = strtoul(cp+1,0,16);
         return PCILOC_VENDOR;
      } else {
         cp = strchr(str,'.');
         dev->bus  = str2long(str);
         dev->slot = str2long(cp+1);
         cp =  strchr(cp+1,'.');
         dev->func = cp?str2long(cp+1):0;
         return PCILOC_BUSSLOT;
      }
   }
   return PCILOC_ERROR;
}

u32t _std hlp_pcimaxbus(void) {
   if (!num_buses) init_enum();
   return num_buses ? num_buses-1 : num_buses;
}

static void init_mcfg(void) {
   mt_swlock();
   if (!mcfg) {
      struct ACPI_SDT_Header *mhdr = sys_getacpitable("MCFG");
      if (mhdr && mhdr->ACPTH_Length >= sizeof(struct ACPI_SDT_Header) +
         sizeof(struct ACPI_MCFG))
      {
         u32t ii;

         mcfg   = (struct ACPI_MCFG*)(mhdr + 1);
         mcflen = (mhdr->ACPTH_Length - sizeof(struct ACPI_SDT_Header) - 8) /
                   sizeof(struct MCFG_Entry);
         log_printf("mcfg: %X,%u\n", mcfg, mcflen);

         /* walk over MCFG, map all areas and save pointer for each existing
            bus */
         memset(busmap, 0, sizeof(busmap));

         for (ii=0; ii<mcflen; ii++) {
            struct MCFG_Entry *me = mcfg->MCFG_List + ii;
         
            if (me->MCE_SegGroup==0 && me->MCE_EndBus>=me->MCE_StartBus) {
               u32t size, bs;
               u8t  *map;

               size = (u32t)me->MCE_EndBus - me->MCE_StartBus + 1;
               // bus eats 1Mb (32*8*4096)
               size *= _1MB;
               // map it forever ...
               map  = (u8t*)map_pcie_space(0, me->MCE_StartBus, 0, 0, size);
               if (map) {
                  for (bs=me->MCE_StartBus; bs<=me->MCE_EndBus; bs++)
                     if (!busmap[bs])
                        busmap[bs] = (pciebus*)(map + (bs - me->MCE_StartBus << 20));
                  log_printf("bus %u..%u mapped at %08X\n", me->MCE_StartBus,
                     me->MCE_EndBus, map);
               }
            }
         }
      }
   }
   mt_swunlock();
}

u64t _std hlp_pcigetmem(u16t seg, u16t bus, u8t slot, u8t func) {
   u32t ii;
   if (func>7 || slot>31 || bus>255) return 0;

   if (!mcfg) init_mcfg();
   if (!mcfg) return 0;

   for (ii=0; ii<mcflen; ii++) {
      struct MCFG_Entry *me = mcfg->MCFG_List + ii;

      if (me->MCE_SegGroup==seg && me->MCE_StartBus<=bus && bus<=me->MCE_EndBus)
         return me->MCE_Base + (bus - me->MCE_StartBus << 20) + (slot << 15) +
             (func << 12);
   }
   return 0;
}

u32t _std hlp_pcieread(u16t seg, u8t bus, u8t slot, u8t func, u16t ofs, u8t size) {
   static u32t errv[4] = { 0xFF, 0xFFFF, 0xFFFFFF, FFFF};
   u32t res, ii;
   /* allow 3 despite of doc but only for PCI-E
      and read even on a border, out-of-range bytes will be FF */
   if (!size || size>4) size = 4;
   // default error value
   res = errv[size-1];

   if (!mcfg) init_mcfg();
   if (!mcfg) {
      if (ofs<256 && !seg && size!=3)
         res = hlp_pciread(bus, slot, func, ofs, size);
   } else
   if (ofs<4096)
      if (seg>0) {
         // this is just impossible so use a slow way
         u8t *mem = (u8t*)map_pcie_space(seg, bus, slot, func, 0);
         if (mem) {
            int sv = sys_intstate(0);
            for (ii=0; ii<size && ofs<4096; ii++, ofs++)
               BYTEOF(res,ii) = mem[ofs];
            sys_intstate(sv);
            pag_physunmap(mem);
         }
      } else {
         if (busmap[bus]) {
            int sv = sys_intstate(0);
            for (ii=0; ii<size && ofs<4096; ii++, ofs++)
               BYTEOF(res,ii) = (*busmap[bus])[slot][func][ofs];
            sys_intstate(sv);
         }
      }
   return res;
}

qserr _std hlp_pciewrite(u16t seg, u8t bus, u8t slot, u8t func, u16t ofs,
                         u8t size, u32t value)
{
   if (!size || size==3 || size>4) return E_SYS_INVPARM;

   if (!mcfg) init_mcfg();
   if (!mcfg) {
      if (!seg) {
         if (ofs<256) hlp_pciwrite(bus, slot, func, ofs, size, value);
         return 0;
      }
   } else
   if (ofs+size>4096) return E_SYS_INVPARM; else {
      void *unmap = 0,
            *wptr = 0;
      int      sv;
      if (seg>0) {
         // this is just impossible so use a slow way
         unmap = (u8t*)map_pcie_space(seg, bus, slot, func, 0);
         if (!unmap) return E_SYS_NOTFOUND;
         wptr  = (u8t*)unmap + ofs;
      } else {
         if (!busmap[bus]) return E_SYS_NOTFOUND;
         wptr  = (*busmap[bus])[slot][func]+ofs;
      }
      // write must be atomic
      sv = sys_intstate(0);
      switch (size) {
         case 1: *(u8t*) wptr = value; break;
         case 2: *(u16t*)wptr = value; break;
         case 4: *(u32t*)wptr = value; break;
      }
      sys_intstate(sv);

      if (unmap) pag_physunmap(unmap);
      return 0;
   }
   return E_SYS_NOTFOUND;
}
