//
// QSINIT "start" module
// directory change notification
// -------------------------------------
// included in sysio.c

/** check that path has open parent dirs and/or notify about directory changes.
   MUST be called in MT lock.

   Function has 2 special bits: IONC_NOSELF - skip self in calculation (for
   directories) and IONC_REVERSE - fullpath is parent, not a child, it calcs
   number of children in SFT in this case (both files and directories).

   @param  fullpath   full file or directory path
   @param  flags      flags (IONC_*), zeroed nofitication bits can be used to
                      enum open parent directories
   @return number of found SFT entries or 0. */
u32t sft_callnotify(const c16t *fullpath, u32t flags) {
   u32t   fecnt = 0,
        reverse = flags & IONC_REVERSE;
   //if (!reverse) log_printf("sft_callnotify(%S,%X)\n", fullpath, flags);

   // no open dirs? just skip it!
   if (nopendirs || reverse) {
      u32t   ii,
          fplen = wcslen(fullpath),
         noself = flags & IONC_NOSELF,
           bits = flags & ~(IONC_TREE|IONC_NOSELF|IONC_REVERSE),
         atypes = (reverse?IOHT_FILE:0) | IOHT_DIR;
      // a bit of order ;)
      if (bits & ~NOTIFY_VALID) {
         log_it(0, "wrong sft_callnotify(%X)\n", flags);
         return 0;
      }
      // skip entry 0
      for (ii=1; ii<sftable_alloc; ii++)
         if (sftable[ii]) {
            sft_entry *fe = sftable[ii];

            if (fe->sign!=SFTe_SIGN) sft_logerr(ii); else
               // reverse case enum files as well
               if ((fe->type & atypes) && fe->name_space==NAMESPC_FILEIO) {
                  u32t       namelen = wcslen(fe->fpath),
                                plen = reverse ? fplen : namelen,
                                clen = reverse ? namelen : fplen;
                  const c16t *parent = reverse ? fullpath : fe->fpath,
                              *child = reverse ? fe->fpath : fullpath;
                  // fix root dir case
                  if (parent[plen-1]=='\\') plen--;
                  /*if (!reverse) 
                     log_printf("(%S:%u,%S:%u)\n", child, clen, parent, plen);*/
                  // directory path here can be "b:\" or "b:\msg"
                  if (clen>=plen)
                     if (_wcsnicmp(child,parent,plen)==0 &&
                         (!child[plen] || child[plen]=='\\'))
                     {
                        int  tree = 0;
                        // found self
                        if (plen==clen || plen==clen-1 && child[plen]=='\\') {
                           if (noself) continue;
                        } else {
                           c16t *nxd = wcschr(child+plen+1,'\\');
                           // if has \\ and this not '\\\0'
                           tree = nxd && nxd[1];
                        }
                        fecnt++;
                        // reverse is only for enumeration ...
                        if (!reverse) {
                           u32t  idx = fe->ioh_first;

                           while (idx) {
                              io_handle_data *hd = ftable[idx];
                              if (hd->sign!=IOHd_SIGN) {
                                 ioh_logerr(fe, hd, idx);
                                 break;
                              } else {
                                 if (tree) hd->dir.actlogtree|=bits;
                                    else hd->dir.actlog|=bits;
                                 // update signaled state ...
                                 if (hd->dir.sigbits && hd->dir.sigbits!=FFFF
                                    && (hd->dir.sigbits & (hd->dir.actlog |
                                       (hd->dir.sigbits & IONC_TREE ?
                                          hd->dir.actlogtree:0))))
                                 {
                                    if (hd->dir.sigbits&IONC_TREE)
                                       hd->dir.actlogtree &= ~hd->dir.sigbits;
                                    hd->dir.actlog &= ~hd->dir.sigbits;
                                    // mark it as "signaled"
                                    hd->dir.sigbits = FFFF;
                                 }
                              }
                              idx = hd->next;
                           }
                        }
                     }
               }
         }
   }
   return fecnt;
}

/** drop notification cache in SFT file entries.
    Must be called in MT lock (sft access) */
void sft_notifydrop(void) {
   u32t ii;
   // skip entry 0
   for (ii=1; ii<sftable_alloc; ii++)
      if (sftable[ii]) {
         sft_entry *fe = sftable[ii];
         if (fe->sign!=SFTe_SIGN) sft_logerr(ii); else
            if (fe->type==IOHT_FILE && fe->name_space==NAMESPC_FILEIO)
               fe->file.notify = FFFF;
      }
}

/** action on a file entry.
    Must be called in MT lock (sft access) */
void sft_notifyfile(sft_entry *fe, u32t bits) {
   // this cache is dropped for all files with every io_diropen() call
   if (fe->type==IOHT_FILE)
      if (fe->file.notify) fe->file.notify = sft_callnotify(fe->fpath, bits);
}

qserr _std io_dirmodify(dir_handle dh, u32t *flags) {
   io_handle_data *fh;
   sft_entry      *fe;
   qserr          err;
   u32t            fl;
   if (!flags) return E_SYS_ZEROPTR;
   if (*flags & ~NOTIFY_VALID) return E_SYS_INVPARM;
   // this call LOCKs us if err==0
   err = io_check_handle(dh, IOHT_DIR, &fh, &fe);
   fl  = *flags;
   *flags = 0;
   if (err) return err;

   if (fe->broken) err = E_SYS_BROKENFILE; else {
      if ((fl&IONC_TREE) && (fh->dir.actlogtree&fl)) {
         *flags |= fh->dir.actlogtree & fl;
         fh->dir.actlogtree &= ~fl;
      }
      if (fh->dir.actlog & fl) {
         *flags |= fh->dir.actlog & fl;
         fh->dir.actlog &= ~fl;
      }
   }
   mt_swunlock();
   return err;
}

qserr _std io_dirnotify(dir_handle dh, u32t flags) {
   io_handle_data *fh;
   sft_entry      *fe;
   qserr          err;
   if (flags & ~NOTIFY_VALID) return E_SYS_INVPARM;
   // this call LOCKs us if err==0
   err = io_check_handle(dh, IOHT_DIR, &fh, &fe);
   if (err) return err;

   if (fe->broken) err = E_SYS_BROKENFILE; else
      fh->dir.sigbits = flags;

   mt_swunlock();
   return err;
}
