//
// QSINIT "start" module
// "ramfs" implementation for disk B:
// -------------------------------------
// * case insensitive
// * sorted directory list, unicode
// * expanded space after setsize is *always* zeroed
// * all files longer than 56k are actually sparse files
//
#include "qstypes.h"
#include "qserr.h"
#include "qsint.h"
#include "qslog.h"
#include "qsxcpt.h"
#include "stdlib.h"
#include "qcl/sys/qsvolapi.h"
#include "sysio.h"
#include "classes.hpp"

#define RAMFS_SIGN      0x444D4152   ///< RAMD string
#define RFSDIR_SIGN     0x45444D52   ///< RMDE string
#define RFSFH_SIGN      0x48464D52   ///< RMFH string
#define RFSFD_SIGN      0x44464D52   ///< RMFD string
#define RFS_DBLEVEL     0

#define FSIZE_MAX       x7FFF ///< file size limit
#define RANGELEN_BITS   16    ///< # of length bits in the range list entry
#define RANGELEN_SHIFT (64 - RANGELEN_BITS)
#define SLICE_SIZE      21    ///< size of an allocation slice (2Mb)
/** min file size for sparse mode.
    limit is equal to small block size in the memory manager (there is no
    reason to use malloc() instead of mem_alloc() if it will be called by
    malloc() itself */
#define FSIZE_SPARSE   ((56*1024)-16-1)

#define RSIZE(q)  ((u32t)((q) >> RANGELEN_SHIFT))
#define RPTR(q)   ((q) & (1LL << RANGELEN_SHIFT) - 1)

struct file_data {
   u32t           sign;     ///< RFSFD_SIGN
   u32t           size;

   u32t         in_use;     ///< # of users, including all sub-trees

   u32t          alloc;     ///< size of data. Can be < than size
   u8t           *data;     ///< file data
   TUQList     *ranges;

//   u32t          flags;     ///< FDF_*
   u16t          attrs;     ///< file attributes
   u16t        namelen;

   file_data   *parent;
   TPtrList       *dir;     ///< list of file_data for the directory (or 0)

   u32t          nhash;     ///< uppercased name hash
   c16t          *name,     ///< file name
                *uname;     ///< uppercased name (in the same buf with name)
   io_time       ctime;
   io_time       atime;
   io_time       wtime;

   TPtrList      users;     ///< file_handle + dir_enum_data pointers
};

struct volume_data {
   u32t           sign;     ///< RAMFS_SIGN
   int             vol;     ///< -1 or 0..9
   qs_sysvolume   self;     ///< ptr to self instance
   file_data     *root;
};

struct dir_enum_data {
   u32t           sign;
   file_data       *dd;
   u32t            pos;     ///< index in dir list
};

struct file_handle {
   u32t           sign;
   file_data       *fd;
   c16t             fp[QS_MAXPATH+1];
};

#define volinst_ret(inst,err)              \
   volume_data *inst = (volume_data*)data; \
   if (inst->sign!=RAMFS_SIGN) return err;

#define fdinst_ret(inst,err)               \
   file_data *inst = (file_data*)data;     \
   if (inst->sign!=RFSFD_SIGN) return err;

/** should be set by someone external to allow high memory alloc:
    now always enabled in PAE and UEFI (except safe mode) */
u32t           ramfs64 = 0;

#if RFS_DBLEVEL > 0
static void ranges_dump(file_data *fd) {
   log_it(3, "file \"%S\" fd:%08X data:%X size:%07X alloc:%u ranges:%X\n",
      fd->name, fd, fd->data, fd->size, fd->alloc, fd->ranges);
   u64t pos = 0;

   for (u32t ii = 0; ii<fd->ranges->Count(); ii++) {
      u32t rsize = RSIZE((*fd->ranges)[ii]);
      u64t   ptr = RPTR((*fd->ranges)[ii]);

      log_it(3, "%4u. %08LX %5u %12LX\n", ii+1, pos<<16, rsize, ptr);
      pos += rsize;
   }
}
#else
#define ranges_dump(fd)
#endif

/** calculate name hash for UPPERcased name.
    @param   name         name to calc
    @param   len          # of chars from name to use (or 0)
    @return hash value */
static u32t calc_hash(const c16t *name, u32t len = 0) {
   u32t res = 0;
   if (!len) len = x7FFF;
   while (*name && len--) {
      res += towupper(*name++);
      res = (res>>2) + (res<<30);
   }
   return res;
}

/// the same as calc_hash() but for *uppercased* unicode string
static u32t calc_hash_wu(const c16t *name, u32t len = 0) {
   u32t res = 0;
   if (!len) len = x7FFF;
   while (*name && len--) {
      res += *name++;
      res = (res>>2) + (res<<30);
   }
   return res;
}

static void usage_modify(file_data *fd, int incr) {
   while (fd) {
      if (incr) fd->in_use++; else fd->in_use--;
      fd = fd->parent;
   }
}

static int unlink_user(file_data *fd, void *user) {
   int idx = fd->users.IndexOf(user);
   if (idx<0) return 0;
   fd->users.Delete(idx);
   usage_modify(fd,0);
   return 1;
}

/// inc/dec directory enumeration indices after adding/removing file in it
static void fix_user_index(file_data *dir, u32t index, int incr) {
   for (u32t ii=0; ii<dir->users.Count(); ii++) {
      dir_enum_data *de = (dir_enum_data*)dir->users[ii];
      // both file_handle & dir_enum_data are listed here
      if (de->sign==RFSDIR_SIGN)
         if (de->pos>=index)
            if (incr) de->pos++; else de->pos--;
   }
}

/// link file to a directory
static qserr file_link(file_data *dir, file_data *fd) {
   // skip .. for subdirectories
   u32t pos = dir->parent==0?0:1;
   // dumb enumeration, still should be fast because this is RAM fs
   while (pos<dir->dir->Count()) {
      file_data *de = (file_data*)(*dir->dir)[pos];
      if (de->uname[0] < fd->uname[0]) { pos++; continue; }
      if (de->uname[0] > fd->uname[0]) break;

      int rc = wcscmp(de->uname, fd->uname);
      if (rc<0) { pos++; continue; }
      if (rc==0) return E_SYS_EXIST;
      break;
   }
   dir->dir->Insert(pos, fd);
   fix_user_index(dir, pos, 1);
   fd->parent = dir;
   return 0;
}

/** unlink file from the parent directory.
    file data is untouched after this call */
static qserr file_unlink(file_data *fd) {
   file_data *pd = fd->parent;
   if (!pd) return E_SYS_SOFTFAULT;
   int idx = pd->dir->IndexOf(fd);
   if (idx<0) return E_SYS_SOFTFAULT;
   pd->dir->Delete(idx);
   fix_user_index(pd, idx, 0);
   return 0;
}

static u64t alloc64(u32t n64k) {
   /* allocation should be zero-filled, also try to bind it to the max
      addresses to preserve 4Gb border for the ramdisk */
   return hlp_memalloc64((u64t)n64k<<16, "b:", QSMA_HIGH|QSMA_MAXADDR, 0, 0);
}

static void free64(u64t ptr) {
   qserr res = hlp_memfreephys(ptr);
   // should never occurs
   if (res) log_it(1, "free64(%LX)=%X!\n", ptr, res);
}

static u64t avail64(void) {
   u64t res;
   if (hlp_memavail64(0, &res, 0, QSMA_HIGH, 0)) return 0;
   return res;
}

// release all file data
static void file_nodata(file_data *fd) {
   // a large file?
   if (fd->ranges) {
      for (int ii=0; ii<fd->ranges->Count(); ii++) {
         u64t ptr = RPTR((*fd->ranges)[ii]);
         if (ptr)
            if (ptr<_4GBLL) hlp_memfree((void*)(u32t)ptr); else free64(ptr);
      }
      delete fd->ranges;
      fd->ranges = 0;
   }
   if (fd->data) { free(fd->data); fd->data = 0; fd->alloc = 0; }
}

static void file_delete(file_data *fd) {
   if (fd->sign!=RFSFD_SIGN) return;
#if 0
   // unschedule DCE callback first!
   if (fd->flags&FDF_DCE) {
      sys_dcedel(fd);
      fd->flags &= ~FDF_DCE;
   }
#endif
   if (fd->dir) {
      for (int ii=fd->dir->Max(); ii>=0; ii--) {
         file_delete((file_data*)((*fd->dir)[ii]));
         fd->dir->Delete(ii);
      }
      delete fd->dir;
      fd->dir = 0;
   }
   if (fd->users.Count()) {
      for (u32t ii=0; ii<fd->users.Count(); ii++) {
         u32t *ptr = (u32t*)fd->users[ii];
         // clean up signature to be safe
         if (*ptr==RFSDIR_SIGN || *ptr==RFSFH_SIGN) { *ptr = 0; free(ptr); } else
            log_it(3, "unlink(%S,%X) user data: %X = %X\n", fd->name, fd, ptr, *ptr);
      }
      fd->users.Clear();
   }
   if (fd->name) { free(fd->name); fd->name = 0; fd->namelen = 0; }
   file_nodata(fd);
   fd->parent = 0;
   fd->uname  = 0;
   fd->size   = 0;
   fd->sign   = 0;
   delete fd;
}

/** walk over tree to the selected path.
    @param   vd             volume data
    @param   path           full path (accepts '\' only)
    @param   [out] fd       file_data object
    @param   [out] failpos  if ptr is non-zero, then last reached fd value
                            will be returned (directory) and failpos set to
                            offset in the path buffer, where non-existing
                            path begin.
    @return error code or 0 */
static qserr follow_path(volume_data *vd, const c16t *path, file_data **fd,
                         u32t *failpos = 0)
{
   *fd = 0;
   // path MUST be full
   if (path[1]!=':' || path[0]-'A'!=vd->vol) return E_SYS_SOFTFAULT;

   const c16t *cp = path+2;
   // is this a root?
   while (*cp=='\\') cp++;
   if (*cp==0) { *fd = vd->root; return 0; }

   file_data *cdd = vd->root;
   qserr      res = 0;
   // directory read
   do {
      const c16t *ce = cp;
      while (*ce && *ce!='\\') ce++;

      u32t nmlen = ce-cp, *npos,
           nhash = calc_hash(cp, nmlen);
      // loop over directory
      res = E_SYS_NOPATH;
      for (u32t idx = 0; idx<cdd->dir->Count(); idx++) {
         file_data *de = (file_data*)(*cdd->dir)[idx];

         if (nmlen!=de->namelen) continue;
         if (nhash==de->nhash) {
            c16t       *fnp = de->uname;
            const c16t *cpp = cp;
            while (cpp!=ce)
               if (towupper(*cpp++) != *fnp++) break;
            // name match!
            if (cpp==ce) {
               cp = ce;
               while (*cp=='\\') cp++;
               // reach the end?
               if (!*cp) { *fd = de; res = 0; } else
                  if (de->attrs&IOFA_DIR) { cdd = de; res = 0; }
               break;
            }
         }
      }
   } while (!res && *cp);
   // only a part of path exists (for mkdir below)
   if (res==E_SYS_NOPATH && failpos) {
      *failpos = cp - path;
      *fd      = cdd;
   }
#if RFS_DBLEVEL>1
   log_it(3, "follow_path(%S, %X) = %X\n", path, *fd, res);
#endif
   return res;
}

/* function should save original name until complete (re)naming */
static qserr fd_setname(file_data *fd, const c16t *name) {
   u32t    len = wcslen(name), ii;
   c16t    *rc = (c16t*)malloc(len+1<<2),
          *urc = rc + (len+1);
   if (!rc) return E_SYS_NOMEM;
   mem_shareblock(rc);
   rc [len] = 0;
   urc[len] = 0;

   memcpy(rc, name, len+1<<1);
   for (ii=0; ii<len; ii++) urc[ii] = towupper(rc[ii]);

   if (fd->name) free(fd->name);
   fd->name    = rc;
   fd->uname   = urc;
   fd->namelen = len;
   fd->nhash   = calc_hash_wu(urc);
   return 0;
}

static qserr file_alloc(const c16t *name, int dir, file_data **pfd) {
   file_data *fd = new file_data;
   time_t    now = time(0);
   qserr     err = 0;
   fd->sign      = RFSFD_SIGN;
   fd->size      = 0;
   fd->in_use    = 0;
//   fd->flags     = 0;
   fd->alloc     = 0;
   fd->data      = 0;
   fd->attrs     = 0;
   fd->name      = 0;
   fd->uname     = 0;
   fd->parent    = 0;
   fd->dir       = 0;
   fd->nhash     = 0;
   fd->namelen   = 0;
   fd->ranges    = 0;
   *pfd = 0;
   // actually, should never fails now ...
   if (name) err = fd_setname(fd, name);
   if (!err) {
      io_timetoio (&fd->ctime, now);
      fd->atime = fd->ctime;
      fd->wtime = fd->ctime;

      if (dir) {
         file_data *ddot;
         err = file_alloc((const c16t*)L"..", 0, &ddot);
         if (!err) {
            fd->dir      = new TPtrList;
            fd->attrs    = IOFA_DIR;
            ddot->parent = fd;
            ddot->attrs  = IOFA_DIR;
            fd->dir->Add(ddot);
         }
      }
   }
   if (err) file_delete(fd); else *pfd = fd;
   return err;
}

static qserr _exicc ramfs_initialize(EXI_DATA, u8t vol, u32t flags,
                                     void *bootsec, const char *opts)
{
   volinst_ret(vd, E_SYS_INVOBJECT);

   vd->vol = vol;
   file_alloc((const c16t*)L".", 0, &vd->root);

   vd->root->dir   = new TPtrList;
   vd->root->attrs = IOFA_DIR;
   vol_data  *vdta = _extvol + vol;
   vdta->flags    |= VDTA_VFS;
   strcpy(vdta->fsname, "RAMFS");

   return 0;
}

/// unmount volume
static qserr _exicc ramfs_finalize(EXI_DATA) {
   volinst_ret(vd, E_SYS_INVOBJECT);

   if (vd->vol<0) return E_DSK_NOTMOUNTED; else {
      file_delete(vd->root);
      vd->root =  0;
      vd->vol  = -1;
      return 0;
   }
}

static int _exicc ramfs_drive(EXI_DATA) {
   volinst_ret(vd, -1);
   return vd->vol;
}

static qserr _exicc ramfs_open(EXI_DATA, const c16t *name, u32t mode,
                               io_handle_int *pfh, u32t *action)
{
   volinst_ret(vd, E_SYS_INVOBJECT);
   if (!pfh || !name || !action) return E_SYS_ZEROPTR;
   *pfh = 0;
   if (mode>SFOM_CREATE_ALWAYS) return E_SYS_INVPARM;
   if (wcslen(name)>QS_MAXPATH) return E_SYS_LONGNAME;
   if (wcschr(name,'?') || wcschr(name,'*')) return E_SYS_INVNAME;

   file_data *fd, *dd;
   u32t     fpos;
   c16t   *cpath = _wcsdup(name);
   c16t   *cname = 0;  // name to create
   qserr     res = follow_path(vd, cpath, &fd, &fpos);
   int        ro = !res && (fd->attrs&IOFA_READONLY);

   if (ro && mode==SFOM_CREATE_ALWAYS) res = E_SYS_ACCESS; else
      if (!res && (fd->attrs&IOFA_DIR)) res = E_SYS_ISDIR; else
         if (res==E_SYS_NOPATH) {
            cname = cpath+fpos;
            res   = wcschr(cname,'\\') ? E_SYS_NOPATH : 0;
         }
   if (!res)
      if (cname && mode==SFOM_OPEN_EXISTING) res = E_SYS_NOFILE; else
         if (!cname && mode==SFOM_CREATE_NEW) res = E_SYS_EXIST;
   if (!res) {
      if (cname) {
         dd  = fd;
         res = file_alloc(cname, 0, &fd);
         if (!res)
            if ((res = file_link(dd, fd)) != 0) file_delete(fd);
      } else
         dd  = fd->parent;
   }
   free(cpath);
   if (res) return res;

   // trunc file and release all memory
   if (mode==SFOM_CREATE_ALWAYS && !cname) {
      fd->size = 0;
      *action  = IOFN_TRUNCATED;
      file_nodata(fd);
   } else
      *action  = cname ? IOFN_CREATED : IOFN_EXISTED;

   file_handle *fh = (file_handle*)malloc(sizeof(file_handle));
   mem_shareblock(fh);
   wcscpy(fh->fp, name);
   fh->sign = RFSFH_SIGN;
   fh->fd   = fd;
   *pfh     = (io_handle_int)fh;
   // file & all of parents are in use
   usage_modify(fd, 1);
   fd->users.Add(fh);
#if RFS_DBLEVEL>0
   log_it(3, "open(%S) fd:%X, in_use:%u, sz:%u, alloc:%u, data:%08X\n",
      name, fd, fd->in_use, fd->size, fd->alloc, fd->data);
#endif
   return ro?E_SYS_READONLY:0;
}

/// expand range list to the selected number of blocks
static qserr ranges_expand(file_data *fd, u32t n64k, int sparse, int insert=-1) {
   if (!n64k) return 0;

   if (!sparse) {
      // shift block size step by step from 2Mb to 64k
      for (u32t step = SLICE_SIZE; step>=16 && n64k; step--) {
         u32t wb = n64k >> step - 16;
         if (!wb) continue;
         // try alloc64 first, switch to smaller size if failed
         u64t  ptr = 0;
         u32t acnt = wb << step - 16;
         if (ramfs64) ptr = alloc64(acnt);
         if (!ptr) ptr = (u32t)hlp_memallocsig(acnt<<16, "b:",
                                               QSMA_RETERR|QSMA_PHYS);
         if (!ptr)
            if (step==16) return E_SYS_NOMEM; else continue;
         // add/insert entry
         if (insert>=0) 
            fd->ranges->Insert(insert++, (u64t)acnt<<RANGELEN_SHIFT | ptr);
               else fd->ranges->Add((u64t)acnt<<RANGELEN_SHIFT | ptr);
         // remains
         n64k -= acnt;
      }
   } else
   if (insert>=0) fd->ranges->Insert(insert, (u64t)n64k<<RANGELEN_SHIFT); else {
      // add to the previous block if it`s sparse too
      if (fd->ranges->Count()>0 && !RPTR((*fd->ranges)[fd->ranges->Max()]))
         (*fd->ranges)[fd->ranges->Max()] += (u64t)n64k<<RANGELEN_SHIFT; else
            fd->ranges->Add((u64t)n64k<<RANGELEN_SHIFT);
   }
   ranges_dump(fd);
   return 0;
}

/// switch file data to range list
static qserr ranges_on(file_data *fd) {
   if (fd->ranges) return E_SYS_SOFTFAULT;
#if RFS_DBLEVEL>0
   log_it(3, "ranges on(%S, sz:%u, a:%u)\n", fd->name, fd->size, fd->alloc);
#endif
   fd->ranges = new TUQList;

   if (fd->size) {
      u32t ablk = Round64k(fd->alloc) >> 16,
           sblk = Round64k(fd->size) >> 16;
      // actual data size here should be <= FSIZE_SPARSE (56k)
      if (ablk && sblk) {
         void *mem = hlp_memallocsig(ablk<<16, "b:",
                                     QSMA_RETERR|QSMA_NOCLEAR|QSMA_PHYS);
         if (!mem) return E_SYS_NOMEM;

         u32t ncpy = fd->size<fd->alloc ? fd->size : fd->alloc;
         memcpy(mem, fd->data, ncpy);
         // clear remains ...
         if (ncpy & 0xFFFF)
            memset((u8t*)mem + ncpy, 0, (ablk<<16) - ncpy);
         
         fd->ranges->Add((u64t)ablk<<RANGELEN_SHIFT | (u32t)mem);
      }
      // add unallocated space ...
      if (ablk<sblk) {
         qserr res = ranges_expand(fd, sblk-ablk, 1);
         if (res) return res;
      }
      ranges_dump(fd);
   }
   if (fd->data) {
      free(fd->data);
      fd->alloc = 0;
      fd->data  = 0;
   }
   return 0;
}

/// shrink ranges memory list to the actual file size
static qserr ranges_shrink(file_data *fd) {
   if (!fd->ranges) return 0;

   u32t  b64k = Round64k(fd->size) >> 16,
         bact = 0;
   int delpos = -1;

   for (u32t ii = 0; ii<fd->ranges->Count(); ii++) {
      u32t rsize = RSIZE((*fd->ranges)[ii]);
      u64t   ptr = RPTR((*fd->ranges)[ii]);

      if (bact >= b64k) {
         // we need to delete this
         if (ptr)
             if (ptr<_4GBLL) hlp_memfree((void*)(u32t)ptr); else free64(ptr);
         if (delpos<0) delpos = ii;
      } else
      if (bact + rsize > b64k) {
         int ok = 1;
         // we need to cut this
         if (ptr) {
            u32t nsize = b64k - bact << 16;

            if (ptr<_4GBLL) {
               ptr = (u32t)hlp_memrealloc((void*)(u32t)ptr, nsize);
               // should never occurs, but who knows ...
               if (!ptr) {
                  log_it(3, "block cut failed (%08LX->%ukb)\n",
                     RPTR((*fd->ranges)[ii]), nsize>>10);
                  ok  = 0;
               }
            } else {
               // move the last block to low memory (since absent realloc64)
               void *mem = hlp_memallocsig(nsize, "b:",
                                           QSMA_RETERR|QSMA_NOCLEAR|QSMA_PHYS);
               
               if (!mem) ok = 0; else
                   if (!sys_memhicopy((u32t)mem, ptr, nsize)) {
                      hlp_memfree(mem);
                      ok  = 0;
                   } else {
                      free64(ptr);
                      ptr = (u32t)mem;
                   }
            }
         }
         if (ok)
            (*fd->ranges)[ii] = (u64t)(b64k - bact) << RANGELEN_SHIFT | ptr;
      }
      /* zero the end of the last block here, this is a bit simpler than do
         it in setsize (and preserves logic that all preallocated memory is
         zero-filled, even with a small overhead) */
      if ((fd->size & 0xFFFF) && ptr && bact<b64k && bact + rsize >= b64k) {
         u32t rem = fd->size & 0xFFFF,
              ofs = (b64k - 1 - bact << 16) + rem;

         if (ptr<_4GBLL)
            memset((u8t*)(u32t)ptr + ofs, 0, _64KB - rem);
         else
         if (!sys_memhiset(ptr + ofs, 0, _64KB - rem, 1))
            log_it(3, "memsethi failed (%LX,%u)\n", ptr + ofs, _64KB - rem);
      }

      bact += rsize;
   }
   // memory released above ...
   if (delpos>=0) fd->ranges->Truncate(delpos);
   ranges_dump(fd);
   return 0;
}

/// write ranges
static qserr rwr(file_data *fd, int idx, u32t ipos, void*&buffer, u32t *size,
                 u64t &pos)
{
   qserr  res = 0;
   // a new range!
   if (idx<0 || idx>=fd->ranges->Count()) {
      u32t nalloc = Round64k(*size) >> 16;
      idx = fd->ranges->Count();
      res = ranges_expand(fd, nalloc, 0);
   }
   if (res) return res;

   u32t bs64k, rofs, wsize;
   u64t   ptr = 0;
   // second pass must recalc all since the size of a new range can be anything 
   while (!res && !ptr) {
      bs64k = RSIZE((*fd->ranges)[idx]),
      rofs  = pos - (ipos<<16),
      wsize = (bs64k<<16) - rofs;
      ptr   = RPTR((*fd->ranges)[idx]);

      if (wsize > *size) wsize = *size;
      // range is missing, allocate only needed # of 64k blocks
      if (!ptr) {
         u32t r1n = rofs >> 16,
              rwn = Round64k((rofs & 0xFFFF) + wsize) >> 16;
         int  r2n = bs64k - rwn - r1n;
         // verify it!
         if (r2n<0) {
            res = E_SYS_SOFTFAULT;
            log_it(3, "error in ranges, pos %LX (%u..%u..%i) bs:%u rofs:%X wsz:%X\n",
               pos, r1n, rwn, r2n, bs64k, rofs, wsize);
            ranges_dump(fd);
         } else {
            fd->ranges->Delete(idx);
            if (r2n) res = ranges_expand(fd, r2n, 1, idx);
            if (!res) res = ranges_expand(fd, rwn, 0, idx);
            if (!res && r1n) res = ranges_expand(fd, r1n, 1, idx);
            // first entry is an empty area of r1n blocks
            if (r1n) { idx++; ipos += r1n; }
         }
      }
   }

   if (!res) {
      /* log_it(3, "wr(%08LX, %6b, %u), pos %LX, rofs %X\n", ptr + rofs, buffer,
         wsize, pos, rofs); */
      if (ptr<_4GBLL) memcpy((u8t*)(u32t)ptr + rofs, buffer, wsize); else
         if (!sys_memhicopy(ptr + rofs, (u32t)buffer, wsize))
            res = E_DSK_ERRWRITE;
      buffer = (u8t*)buffer + wsize;
      pos   += wsize;
      *size -= wsize;
   }
   return res;
}

/// read ranges
static qserr rrd(file_data *fd, u32t idx, u32t ipos, void*&buffer, u32t *size,
                 u64t &pos)
{
   u32t rsize = RSIZE((*fd->ranges)[idx]),
         rofs = pos - (ipos<<16);
   u64t   ptr = RPTR((*fd->ranges)[idx]);
   qserr   rc = 0;

   rsize = (rsize<<16) - rofs;
   if (rsize > *size) rsize = *size;

   if (!ptr) memset(buffer, 0, rsize); else
      if (ptr<_4GBLL) memcpy(buffer, (u8t*)(u32t)ptr + rofs, rsize); else
         if (!sys_memhicopy((u32t)buffer, ptr + rofs, rsize)) rc = E_DSK_ERRREAD;
   if (!rc) {
      buffer = (u8t*)buffer + rsize;
      pos   += rsize;
      *size -= rsize;
   }
   return rc;
}

static qserr ranges_io(file_data *fd, u64t pos, void *buffer, u32t *size, int wr) {
   u32t   sb = (u32t)(pos>>16),
         ofs = 0,
          ii = 0,
      opsize = *size;
   qserr res = 0;

   while (*size) {
      int  append = ii==fd->ranges->Count();
      u32t  rsize = append ? 0 : RSIZE((*fd->ranges)[ii]);

      if (append && !wr) {
         log_it(3, "file \"%S\" read out of range: %08LX,%u?!\n", fd->name, pos, *size);
         ranges_dump(fd);
         memset(buffer, 0, *size);
         *size = 0;
         break;
      }
      // something to do?
      if (append || sb>=ofs && sb<ofs+rsize) {
         res = wr ? rwr(fd, ii, ofs, buffer, size, pos):
                    rrd(fd, ii, ofs, buffer, size, pos);
         // update rsize since write should alter/insert into the list ...
         if (res) break; else {
            if (wr) rsize = RSIZE((*fd->ranges)[ii]);
            sb = (u32t)(pos>>16);
         }
      }
      ofs += rsize;
      ii++;
   }
   *size = opsize - *size;
   return res;
}


static qserr _exicc ramfs_read(EXI_DATA, io_handle_int file, u64t pos,
                               void *buffer, u32t *size)
{
   volinst_ret(vd, E_SYS_INVOBJECT);
   if (!buffer || !size) return E_SYS_ZEROPTR;

   file_handle *fh = (file_handle*)file;
   if (fh->sign!=RFSFH_SIGN) return E_SYS_INVPARM;
   if (!*size) return 0;
   file_data   *fd = fh->fd;
   u32t     toread = *size,
            tofill = 0;
   if (pos>=fd->size) { *size = 0; return E_SYS_EOF; }
   if (pos+toread > fd->size) toread = fd->size - pos;
   *size = toread;

   if (fd->ranges) {
      qserr res = ranges_io(fd, pos, buffer, size, 0);
      if (res) return res;
   } else {
      // take in mind unallocated space
      if (pos>fd->alloc) { tofill = toread; toread = 0; } else
         if (pos+toread > fd->alloc) {
            tofill = pos + toread - fd->alloc;
            toread-= tofill;
         }
      if (toread) memcpy(buffer, fd->data + (u32t)pos, toread);
      if (tofill) memset((u8t*)buffer + toread, 0, tofill);
   }
   io_timetoio(&fd->atime, time(0));
   return 0;
}

static qserr _exicc ramfs_write(EXI_DATA, io_handle_int file, u64t pos,
                                void *buffer, u32t *size)
{
   volinst_ret(vd, E_SYS_INVOBJECT);
   if (!buffer || !size) return E_SYS_ZEROPTR;
   if (pos>FSIZE_MAX) return E_SYS_FSLIMIT;

   file_handle *fh = (file_handle*)file;
   if (fh->sign!=RFSFH_SIGN) return E_SYS_INVPARM;
   if (!*size) return 0;
   file_data   *fd = fh->fd;
   u32t    towrite = *size;
   int     nowrite = 0;
   qserr       res = 0;
   if (pos+towrite > FSIZE_MAX) towrite = FSIZE_MAX + 1 - pos;
   // switch to sparse mode
   if (pos+towrite >= FSIZE_SPARSE && !fd->ranges) {
      res = ranges_on(fd);
      if (res) { *size = 0; return res; }
   }
   // large file i/o
   if (fd->ranges) {
      *size = towrite;
      res   = ranges_io(fd, pos, buffer, size, 1);
   } else {
      if (pos+towrite > fd->alloc) {
         // try to eliminate op at all (because realloc below is much worse)
         if (pos>=fd->alloc && *(u8t*)buffer==0)
            if (memchrnb((u8t*)buffer, 0, towrite)==0) nowrite = 1;
         if (!nowrite) {
            u32t newsize = pos+towrite;
            if (newsize>_16KB) newsize = Round2k(newsize);
            u8t *dnew = (u8t*)realloc(fd->data, newsize);
            // let it be here, even if our realloc() never returns on error
            *size     = 0;
            if (!dnew) return E_SYS_NOMEM;
            if (!fd->data) mem_shareblock(dnew);
            // zero added memory
            memset(dnew + fd->alloc, 0, newsize - fd->alloc);
            fd->data  = dnew;
            fd->alloc = newsize;
         }
      }
      if (!nowrite) memcpy(fd->data + (u32t)pos, buffer, towrite);
      *size = towrite;
   }
   // something can be written in ranges mode even if error occurs ...
   if (pos+*size > fd->size) fd->size = pos + *size;
   // always update write & access time
   io_timetoio(&fd->atime, time(0));
   fd->wtime = fd->atime;
   // log_it(3, "wr \"%S\" %X sz:%X\n", fd->name, res, fd->size);
   return res;
}

static qserr _exicc ramfs_flush(EXI_DATA, io_handle_int file) { return 0; }

static qserr _exicc ramfs_close(EXI_DATA, io_handle_int file) {
   volinst_ret(vd, E_SYS_INVOBJECT);
   file_handle *fh = (file_handle*)file;
   if (fh->sign!=RFSFH_SIGN) return E_SYS_INVPARM;
   unlink_user(fh->fd, fh);
   fh->sign = 0;
   free(fh);
   return 0;
}

static qserr _exicc ramfs_size(EXI_DATA, io_handle_int file, u64t *size) {
   if (!size) return E_SYS_ZEROPTR;
   file_handle *fh = (file_handle*)file;
   if (fh->sign!=RFSFH_SIGN) return E_SYS_INVPARM;
   *size = fh->fd->size;
   return 0;
}

static qserr _exicc ramfs_setsize(EXI_DATA, io_handle_int file, u64t newsize) {
   volinst_ret(vd, E_SYS_INVOBJECT);
   if (newsize>FSIZE_MAX) return E_SYS_FSLIMIT;
   file_handle *fh = (file_handle*)file;
   if (fh->sign!=RFSFH_SIGN) return E_SYS_INVPARM;
   file_data   *fd = fh->fd;
   u32t nsize = newsize;

   if (nsize<fd->size) {
      fd->size = nsize;
      // resize file data just here, without a DCE as is was earlier ...
      if (!nsize) file_nodata(fd); else
         if (!fd->data) ranges_shrink(fd); else
            if (fd->alloc > fd->size + 256)
               fd->data = (u8t*)realloc(fd->data, fd->alloc = fd->size + 256);
   } else
   if (nsize>fd->size) {
      // do not switch file to sparse mode here, do it only at real write!
      if (fd->data && fd->alloc>fd->size) {
         u32t diff = fd->alloc - fd->size,
              grow = nsize - fd->size;
         // zero tail of allocated block
         memset(fd->data+fd->size, 0, diff>grow?grow:diff);
      } else
      if (fd->ranges) {
         u32t cnb = Round64k(fd->size) >> 16,
              nnb = Round64k(nsize) >> 16;
         if (nnb > cnb)  {
            qserr res = ranges_expand(fd, nnb-cnb, 1);
            if (res) return res;
         }
      }
      fd->size = nsize;
   }
   return 0;
}

static void fd2dirinfo(file_data *de, io_direntry_infow *info, int name) {
   info->attrs   = de->attrs;
   info->fileno  = 0;
   info->size    = de->size;
   info->vsize   = de->alloc<de->size?de->alloc:de->size;
   info->atime   = de->atime;
   info->wtime   = de->wtime;
   info->ctime   = de->ctime;
   // io_direntry_infow (else it can be io_handle_info)
   if (name) {
      info->dir  = 0;
      wcscpy(info->name, de->name);
   }
}

static qserr _exicc ramfs_finfo(EXI_DATA, io_handle_int file, io_handle_info *info)
{
   if (!info) return E_SYS_ZEROPTR;
   volinst_ret(vd, E_SYS_INVOBJECT);
   file_handle *fh = (file_handle*)file;
   if (fh->sign!=RFSFH_SIGN) return E_SYS_INVPARM;
   // a bit faster than function below
   fd2dirinfo(fh->fd, (io_direntry_infow*)info, 0);
   return 0;
}

static qserr _exicc ramfs_pathinfo(EXI_DATA, const c16t *path, io_handle_info *info,
                                   c16t *name, c16t *altname)
{
   volinst_ret(vd, E_SYS_INVOBJECT);
   if (!path || !info) return E_SYS_ZEROPTR;
   file_data *de;
   qserr     res = follow_path(vd, path, &de);
   if (!res) {
      info->attrs   = de->attrs;
      info->fileno  = 0;
      info->size    = de->size;
      info->vsize   = de->alloc<de->size?de->alloc:de->size;
      info->atime   = de->atime;
      info->wtime   = de->wtime;
      info->ctime   = de->ctime;
      if (altname) *altname = 0;
      if (name) wcscpy(name, de->name);
   }
   return res;
}

static qserr _exicc ramfs_setinfo(EXI_DATA, const c16t *path, io_handle_info *info, u32t flags) {
   volinst_ret(vd, E_SYS_INVOBJECT);
   if (!path || !info) return E_SYS_ZEROPTR;
   file_data *de;
   qserr     res = follow_path(vd, path, &de);
   if (!res) {
      if (flags&IOSI_ATTRS) de->attrs = info->attrs;
      if (flags&IOSI_ATIME) de->atime = info->atime;
      if (flags&IOSI_CTIME) de->ctime = info->ctime;
      if (flags&IOSI_WTIME) de->wtime = info->wtime;
   }
   return res;
}

static qserr _exicc ramfs_setexattr(EXI_DATA, const c16t *path, const c16t *aname, void *buffer, u32t size) {
   volinst_ret(vd, E_SYS_INVOBJECT);
   return E_SYS_UNSUPPORTED;
}

static qserr _exicc ramfs_getexattr(EXI_DATA, const c16t *path, const c16t *aname, void *buffer, u32t *size) {
   volinst_ret(vd, E_SYS_INVOBJECT);
   return E_SYS_UNSUPPORTED;
}

static str_list* _exicc ramfs_lstexattr(EXI_DATA, const c16t *path) {
   volinst_ret(vd, 0);
   return 0;
}

// warning! it acts as rmdir too!!
static qserr _exicc ramfs_remove(EXI_DATA, const c16t *path) {
   volinst_ret(vd, E_SYS_INVOBJECT);
   if (!path) return E_SYS_ZEROPTR;
   file_data *fd;
   qserr     res = follow_path(vd, path, &fd);
   if (!res)
      if (fd->in_use || fd==vd->root) res = E_SYS_ACCESS; else
         if (fd->dir && fd->dir->Count()>1) res = E_SYS_DIRNOTEMPTY; else 
            if (fd->attrs&IOFA_READONLY) res = E_SYS_READONLY; else {
               res = file_unlink(fd);
               if (!res) file_delete(fd);
            }
   return res;
}

static qserr _exicc ramfs_renmove(EXI_DATA, const c16t *oldpath, const c16t *newpath) {
   volinst_ret(vd, E_SYS_INVOBJECT);
   if (!oldpath || !newpath) return E_SYS_ZEROPTR;

   c16t *cpatho = _wcsdup(oldpath),
        *cpathn = _wcsdup(newpath);
   qserr     rc = 0;

   if (!cpathn) rc = E_SYS_INVPARM; else
   do {
      file_data *src, *dst, *psrc;
      c16t   *cpname = wcsrchr(cpathn,'\\');
      // path _must_ be full here, so the next line is void
      if (!cpname) { rc = E_SYS_INVPARM; break; }
      // source object
      rc = follow_path(vd, cpatho, &src);
      if (rc) break;
      // check for the destination
      if (follow_path(vd, cpathn, &dst)==0) rc = dst==src?0:E_SYS_EXIST; else {
         *cpname++ = 0;
         rc = follow_path(vd, cpathn, &dst);
         if (rc) break;
         // root cannot be renamed
         psrc = src->parent;
         if (psrc==0) { rc = E_SYS_INVNAME; break; }
         // unlink it in source dir
         rc = file_unlink(src);
         if (rc) break;
         // rename it and try to move/link back
         rc = fd_setname(src, cpname);
         if (!rc) rc = file_link(dst, src);
         if (rc) rc = file_link(psrc, src);
         // file is lost if both attempts fail
         if (rc) file_delete(src);
      }
   } while (0);
   free(cpathn);
   free(cpatho);
   return rc;
}

static qserr _exicc ramfs_makepath(EXI_DATA, const c16t *path) {
   volinst_ret(vd, E_SYS_INVOBJECT);
   // follow_path() walks until last existing dir in path for us
   u32t     fpos;
   file_data *dd, *fd;
   c16t      *fp = _wcsdup(path);
   qserr     res = follow_path(vd, fp, &dd, &fpos);

   if (!res) res = E_SYS_EXIST; else
   if (res==E_SYS_NOPATH) {
      c16t *cp = fp + fpos;
      // should be executed at least once
      while (cp && *cp) {
         c16t *ce = wcschr(cp,'\\');
         if (ce) {
            *ce++ = 0;
            while (*ce=='\\') *ce++;
         }
         res = file_alloc(cp, 1, &fd);
#if RFS_DBLEVEL>0
         log_it(3, "mkdir(%S) = %08X, parent %08X, res = %X\n", cp, fd, dd, res);
#endif
         if (res) break;
         res = file_link(dd, fd);
         if (res) {
            file_delete(fd);
            break;
         }
         dd = fd;
         cp = ce;
      }
   }
   free(fp);
   return res;
}

static qserr _exicc ramfs_diropen(EXI_DATA, const c16t *path, dir_handle_int *pdh) {
   volinst_ret(vd, E_SYS_INVOBJECT);
   if (!path || !pdh) return E_SYS_ZEROPTR;
   *pdh = 0;
   if (wcslen(path)>QS_MAXPATH) return E_SYS_LONGNAME;
   // dir handle storage
   dir_enum_data *ed = (dir_enum_data*)malloc(sizeof(dir_enum_data));
   if (!ed) return E_SYS_NOMEM;
   mem_shareblock(ed);
   file_data     *dd;
   qserr         res = follow_path(vd, path, &dd);

   if (!res)
      if (dd->attrs&IOFA_DIR) {
         ed->sign = RFSDIR_SIGN;
         ed->dd   = dd;
         ed->pos  = 0;
         *pdh     = (dir_handle_int)ed;
         usage_modify(dd,1);
         dd->users.Add(ed);
      } else
         res = E_SYS_NOPATH;

   if (res && ed) free(ed);
   return res;
}

static qserr _exicc ramfs_dirnext(EXI_DATA, dir_handle_int dh,
                                  io_direntry_infow *info, int altname) {
   volinst_ret(vd, E_SYS_INVOBJECT);
   dir_enum_data *ed = (dir_enum_data*)dh;
   if (!ed || !info) return E_SYS_ZEROPTR;
   if (ed->sign!=RFSDIR_SIGN) return E_SYS_INVPARM;
   /* pos can be modified by fix_user_index() above - if file added/removed
      *before* it during enumeration process */
   if (ed->pos>=ed->dd->dir->Count()) return E_SYS_NOFILE; else
      fd2dirinfo((file_data*)((*ed->dd->dir)[ed->pos++]), info, 1);
   return 0;
}

static qserr _exicc ramfs_dirclose(EXI_DATA, dir_handle_int dh) {
   volinst_ret(vd, E_SYS_INVOBJECT);
   dir_enum_data *ed = (dir_enum_data*)dh;
   if (!ed) return E_SYS_ZEROPTR;
   if (ed->sign!=RFSDIR_SIGN) return E_SYS_INVPARM;

   ed->sign = 0;
   unlink_user(ed->dd, ed);
   free(ed);
   return 0;
}

static qserr _exicc ramfs_volinfo(EXI_DATA, disk_volume_data *info, int fast) {
   volinst_ret(vd, E_SYS_INVOBJECT);

   if (!info) return E_SYS_ZEROPTR;
   memset(info, 0, sizeof(disk_volume_data));

   if (vd->vol>=0) {
      vol_data *vdta = _extvol+vd->vol;
      // mounted?
      if (vdta->flags&VDTA_ON) {
         u64t       memsize = hlp_memavail(0,0,QSMA_PHYS) +
                              (ramfs64 ? avail64() : 0);
         info->StartSector  = FFFF64;
         info->Disk         = FFFF;
         info->DataSector   = FFFF;
         info->TotalSectors = vdta->length;
         info->SectorSize   = vdta->sectorsize;
         info->FSizeLim     = FSIZE_MAX;
         info->ClBad        = 0;
         info->SerialNum    = vdta->serial;
         info->ClSize       = 8;  // fake cluster size == page
         info->ClAvail      = (u32t)(memsize>>12 > FFFF ? FFFF : memsize>>12);
         info->ClTotal      = vdta->length;
         info->FsVer        = FST_OTHER;
         info->InfoFlags    = VIF_VFS;
         //info->Reserved     = 0;
         //info->OemData      = 0;
         strcpy(info->Label,vdta->label);
         strcpy(info->FsName,vdta->fsname);
         return 0;
      }
   }
   return E_DSK_NOTMOUNTED;
}

static qserr _exicc ramfs_setlabel(EXI_DATA, const c16t *label) {
   volinst_ret(vd, E_SYS_INVOBJECT);

   if (vd->vol>=0) {
      vol_data  *vdta = _extvol+vd->vol;
      char     cn[12];
      u32t        len = wcslen(label);
      if (len>11) return E_SYS_LONGNAME;

      if (len<11) memset(&cn, ' ', 11);
      cn[11] = 0;
      while (len--) {
         c16t cc = towupper(label[len]);
         if (cc>=128 || strchr("+.,;=[]\"*:<>/\\?|\x7F",(char)cc) || cc<' ')
            return E_SYS_INVNAME;
         else
            cn[len] = (char)cc;
      }
      // update own saved string
      strcpy(vdta->label, cn);
      trimright(vdta->label, " ");
      return 0;
   }
   return E_DSK_NOTMOUNTED;
}

static u32t _exicc ramfs_avail(EXI_DATA) { return SFAF_VFS|SFAF_NPRO; }

static qserr _exicc ramfs_service(EXI_DATA, u32t func, void *info) {
   volinst_ret(vd, E_SYS_INVOBJECT);
   return E_SYS_UNSUPPORTED;
}

static void _std ramfs_init(void *instance, void *userdata) {
   volume_data *vd = (volume_data*)userdata;
   vd->sign        = RAMFS_SIGN;
   vd->vol         = -1;
   vd->self        = (qs_sysvolume)instance;
}

static void _std ramfs_done(void *instance, void *userdata) {
   volume_data *vd = (volume_data*)userdata;
   if (vd->sign==RAMFS_SIGN)
      if (vd->vol>=0) {
         log_printf("Volume %d fini???\n", vd->vol);
         // force it to release memory
         ramfs_finalize(vd, 0);
      }
   memset(vd, 0, sizeof(volume_data));
}

static void *m_list[] = { ramfs_initialize, ramfs_finalize, ramfs_avail,
   ramfs_volinfo, ramfs_setlabel, ramfs_drive, ramfs_open, ramfs_read,
   ramfs_write, ramfs_flush, ramfs_close, ramfs_size, ramfs_setsize,
   ramfs_finfo, ramfs_setexattr, ramfs_getexattr, ramfs_lstexattr, ramfs_remove,
   ramfs_renmove, ramfs_makepath, ramfs_remove, ramfs_pathinfo, ramfs_setinfo,
   ramfs_diropen, ramfs_dirnext, ramfs_dirclose, ramfs_service };

extern "C" u32t ramfs_classid;

extern "C"
void register_ramfs(void) {
   // register private(unnamed) class
   ramfs_classid = exi_register(FS_CLASS_NAME_PFX "ramfs", m_list,
      sizeof(m_list)/sizeof(void*), sizeof(volume_data), EXIC_PRIVATE,
         ramfs_init, ramfs_done, 0);
   if (!ramfs_classid) _throwmsg_("ramfs - class reg.error");
   qserr res = io_registerfs(ramfs_classid);
   if (res) {
      log_printf("ramfs: regfs err %08X\n", res);
      _throwmsg_("ramfs - registration error");
   }
}
