//
// QSINIT "start" module
// subset of C library functions (dir related)
//
#include "qserr.h"
#include "qsio.h"
#include "errno.h"
#include "qsutil.h"
#include "qcl/qslist.h"
#include "sys/stat.h"
#include "sys/utime.h"
#include "qsshell.h"
#include "syslocal.h"
#include "limits.h"
#include "wchar.h"

char* __stdcall START_EXPORT(getcwd)(char *buf, size_t size) {
   qserr err = io_curdir(buf, size);
   if (err) set_errno_qserr(err);
   return err?0:buf;
}

char* __stdcall _getdcwd(int drive, char *buf, size_t size) {
   qserr err;
   char *dir = buf;

   if (!dir) {
      dir  = malloc_th(QS_MAXPATH+1);
      size = QS_MAXPATH+1;
   }
   err = io_diskdir(drive<0?io_curdisk():drive, dir, size);
   if (err) {
      set_errno_qserr(err);
      if (!buf) free(dir);
      dir = 0;
   } else
   if (!buf) {
      mem_localblock(dir);
      mem_cut(dir, strlen(dir)+1);
   }
   return dir;
}

c16t* __stdcall _wgetdcwd(int drive, c16t *buf, size_t size) {
   qserr err;
   c16t *dir = buf;

   if (!dir) {
      dir  = malloc_th(QS_MAXPATH+1<<1);
      size = QS_MAXPATH+1<<1;
   }
   err = io_diskdirw(drive<0?io_curdisk():drive, dir, size);
   if (err) {
      set_errno_qserr(err);
      if (!buf) free(dir);
      dir = 0;
   } else
   if (!buf) {
      mem_localblock(dir);
      mem_cut(dir, wcslen(dir)+1<<1);
   }
   return dir;
}

int __stdcall START_EXPORT(chdir)(const char *path) {
   qserr err = io_chdir(path);
   if (err) set_errno_qserr(err);
   return err?-1:0;
}

int __stdcall START_EXPORT(mkdir)(const char *path) {
   qserr err = io_mkdir(path);
   if (err) set_errno_qserr(err);
   return err?-1:0;
}

int __stdcall START_EXPORT(rmdir)(const char *path) {
   qserr err = io_rmdir(path);
   if (err) set_errno_qserr(err);
   return err?-1:0;
}

#define STAT_SIGN   (0x52494452)

typedef struct {
   u32t             sign;
   dir_handle         dh;
   u8t              call;   // call type - 1 = opendir, 2 = _dos_findfirst
   u8t              wide;   // utf call
   u16t            attrs;
   io_direntry_infow  de;
   c16t             mask[QS_MAXPATH+2];
   c16t             dirw[QS_MAXPATH+2];
   char             dira[QS_MAXPATH+2];
} StatInfo;

#define checkret_direrr(x,calltype)           \
   StatInfo *si = (StatInfo*)(fp->d_sysdata); \
   if (!si || si->sign!=STAT_SIGN || si->call!=calltype) \
      { set_errno(EINVAL); return x; }

// !!! here and below assumed that all offsets in dir_t & dirw_t are equal !!!
static void* opendir_int(const void *dpath, int wide) {
   qserr      err;
   u32t       len = sizeof(StatInfo) + sizeof(dirw_t);
   StatInfo   *si = (StatInfo*)calloc_thread(len,1);
   dirw_t     *rc = (dirw_t*)(si+1);

   if (!wide) {
      rc->d_openpath = (c16t*)&si->dira;
      err = io_fullpath (si->dira, (char*)dpath, QS_MAXPATH+1);
      if (!err) hlp_atow(si->dira, si->dirw, QS_MAXPATH+1, &err);
   } else {
      rc->d_openpath = si->dirw;
      err = io_fullpathw(si->dirw, (c16t*)dpath, QS_MAXPATH+1);
   }
   // use utf open for all cases
   if (!err) err = io_diropenw(si->dirw, &si->dh);
   if (err) {
      free(si);
      set_errno_qserr(err);
      return 0;
   }
   if (!wide)
      if (strclast(si->dira)!='\\') strcat(si->dira, "\\");

   if (si->dirw[wcslen(si->dirw)-1]!='\\') wcscat(si->dirw, L"\\");

   rc->d_sysdata = si;
   si->sign      = STAT_SIGN;
   si->call      = 1;
   si->wide      = wide?1:0;
   // set process as block owner
   mem_localblock(si);
   return rc;
}

dir_t* __stdcall START_EXPORT(opendir)(const char *dpath) {
   return (dir_t*)opendir_int(dpath, 0);
}

dirw_t* __stdcall START_EXPORT(_wopendir)(const c16t* dpath) {
   return (dirw_t*)opendir_int(dpath, 1);
}

static void time2dirent(io_direntry_infow *de, dir_t *fp) {
   switch ((u32t)mt_tlsget(QTLS_DIRENTTIME)) {
      case QDETIME_WRITE :
         fp->d_time  = io_iototimeq(&de->wtime);
         break;
      case QDETIME_CREATE:
         fp->d_time  = io_iototimeq(&de->ctime);
         break;
      case QDETIME_ACCESS:
         fp->d_time  = io_iototimeq(&de->atime);
         break;
      case QDETIME_UNIX  :
      default:
         fp->d_wtime = io_iototime(&de->wtime);
         fp->d_ctime = io_iototime(&de->ctime);
   }
}

static dirw_t* readdir_int(StatInfo *si, dirw_t *fp) {
   qserr err = io_dirnextw(si->dh, &si->de);
   if (err) { set_errno_qserr(err); return 0; }

   time2dirent(&si->de, (dir_t*)fp);

   fp->d_size   = si->de.size;
   fp->d_attr   = si->de.attrs;
   return fp;
}

dirw_t* __stdcall START_EXPORT(_wreaddir)(dirw_t *fp) {
   dirw_t *res;
   checkret_direrr(0,1);

   res = readdir_int(si,fp);
   if (res) wcscpy(fp->d_name, si->de.name);
   return res;
}

dir_t* __stdcall START_EXPORT(readdir)(dir_t *fp) {
   dir_t *res;
   checkret_direrr(0,1);

   res = (dir_t*)readdir_int(si,(dirw_t*)fp);
   if (res) {
      u32t errcnt;
      qserr  err;
      if (!hlp_wtoa_int(si->de.name, res->d_name, sizeof(res->d_name), &err,
                        '?', &errcnt))
      {
         set_errno_qserr(err);
         res = 0;
      } else
      if (errcnt) fp->d_attr|=_A_NAMEERR;
   }
   return res;
}

int __stdcall START_EXPORT(_wclosedir)(dirw_t *fp) {
   qserr      err;
   checkret_direrr(EINVAL,1);
   err = io_dirclose(si->dh);
   if (err) set_errno_qserr(err);
   fp->d_sysdata = 0;
   si->sign      = 0;
   free(si);
   // it would be nice in trace
   return err/*?1:0*/;
}

int __stdcall START_EXPORT(closedir)(dir_t *fp) {
   return START_EXPORT(_wclosedir)((dirw_t*)fp);
}


qserr __stdcall START_EXPORT(_dos_stat)(const char *path, dir_t *fp) {
   io_handle_info  hi;
   qserr          err;
   if (!path||!fp) { set_errno(EINVAL); return E_SYS_INVPARM; }

   err = io_pathinfo(path, &hi);
   // deny device info here
   if (!err && (hi.attrs&IOFA_DEVICE)) err = E_SYS_TTYDEV;
   
   if (err) set_errno_qserr(err); else {
      // warning! field positions should match!
      time2dirent((io_direntry_infow*)&hi, fp);

      fp->d_size  = hi.size;
      fp->d_attr  = hi.attrs;

      if (io_fullpath(fp->d_name, path, NAME_MAX+1)) fp->d_name[0] = 0;
      fp->d_name[NAME_MAX] = 0;
      fp->d_openpath = &fp->d_name[NAME_MAX];
      fp->d_sysdata  = 0;
      fp->d_subdir   = 0;
      fp->d_reserved = 0;
   }
   return err;
}

qserr __stdcall START_EXPORT(_dos_statw)(const wchar_t *path, dirw_t *fp) {
   io_handle_info  hi;
   qserr          err;
   if (!path||!fp) { set_errno(EINVAL); return E_SYS_INVPARM; }

   err = io_pathinfow(path, &hi);
   // deny device info here
   if (!err && (hi.attrs&IOFA_DEVICE)) err = E_SYS_TTYDEV;
   
   if (err) set_errno_qserr(err); else {
      /* warning! field positions should match!
         here and anywhere io_direntry_info and dir_t fields must be on the
         same pos in ascii and utf-16 structs */
      time2dirent((io_direntry_infow*)&hi, (dir_t*)fp);

      fp->d_size  = hi.size;
      fp->d_attr  = hi.attrs;

      if (io_fullpathw(fp->d_name, path, NAME_MAX+1)) fp->d_name[0] = 0;
      fp->d_name[NAME_MAX] = 0;
      fp->d_openpath = &fp->d_name[NAME_MAX];
      fp->d_sysdata  = 0;
      fp->d_subdir   = 0;
      fp->d_reserved = 0;
   }
   return err;
}

int __stdcall stat(const char *path, struct stat *fi) {
   io_handle_info  hi;
   qserr          err;
   if (!path||!fi) { set_errno(EINVAL); return -1; }

   err = io_pathinfo(path, &hi);
   if (err) set_errno_qserr(err); else {
      memset(fi, 0, sizeof(struct stat));
      fi->st_nlink = 1;
      fi->st_size  = hi.size;
      fi->st_atime = io_iototime(&hi.wtime);
      fi->st_mtime = fi->st_atime;
      fi->st_ctime = io_iototime(&hi.ctime);
      fi->st_dev   = hi.vol;

      fi->st_mode  = S_IRUSR|S_IRGRP|S_IROTH;
      fi->st_mode |= hi.attrs&IOFA_DEVICE?S_IFCHR:(hi.attrs&IOFA_DIR?S_IFDIR:S_IFREG);

      if ((hi.attrs&IOFA_READONLY)==0) fi->st_mode |= S_IWUSR|S_IWGRP|S_IWOTH;
   }
   return err?-1:0;
}

int __stdcall START_EXPORT(_matchpattern)(const char *pattern, const char *str) {
   static char *all="*.*";
   if (!pattern||!*pattern) pattern = all;
   if (!str) return 0;

   for (;;++str) {
      char strc     = toupper(*str);
      char patternc = toupper(*pattern++);
      switch (patternc) {
         case 0:
            return strc==0;
         case '?':
            if (strc == 0) return false;
            break;
         case '*':
            if (*pattern==0) return true;
            if (*pattern=='.') {
               char *dot;
               if (pattern[1]=='*' && pattern[2]==0) return true;
               dot=strchr(str,'.');
               if (pattern[1]==0) return dot==NULL || dot[1]==0;
               if (dot!=NULL) {
                  str=dot;
                  if (strpbrk(pattern,"*?")==NULL && strchr(str+1,'.')==NULL)
                     return stricmp(pattern+1,str+1)==0;
               }
            }
            while (*str)
               if (_matchpattern(pattern,str++)) return true;
            return false;
         default:
            if (patternc != strc)
               if (patternc=='.' && strc==0) return _matchpattern(pattern,str);
                  else return false;
            break;
      }
   }
}

int __stdcall START_EXPORT(_matchpatternw)(const wchar_t *pattern,
                                           const wchar_t *str)
{
   static wchar_t *all=L"*.*";
   if (!pattern||!*pattern) pattern = all;
   if (!str) return 0;

   for (;;++str) {
      wchar_t strc     = towupper(*str);
      wchar_t patternc = towupper(*pattern++);
      switch (patternc) {
         case 0:
            return strc==0;
         case '?':
            if (strc == 0) return false;
            break;
         case '*':
            if (*pattern==0) return true;
            if (*pattern=='.') {
               wchar_t *dot;
               if (pattern[1]=='*' && pattern[2]==0) return true;
               dot=wcschr(str,'.');
               if (pattern[1]==0) return dot==NULL || dot[1]==0;
               if (dot!=NULL) {
                  str=dot;
                  if (wcspbrk(pattern,L"*?")==NULL && wcschr(str+1,'.')==NULL)
                     return _wcsicmp(pattern+1,str+1)==0;
               }
            }
            while (*str)
               if (_matchpatternw(pattern,str++)) return true;
            return false;
         default:
            if (patternc != strc)
               if (patternc=='.' && strc==0) return _matchpatternw(pattern,str);
                  else return false;
            break;
      }
   }
}

/* this code should be wrong in many cases, but it is used as a first time
   implementation */
int __stdcall _replacepatternw(const wchar_t *frompattern, const wchar_t *topattern,
                               const wchar_t *from, wchar_t *to)
{
   if (!to) return 0;
   *to = 0;
   if (!frompattern||!topattern||!from) return 0;
   if (!_matchpatternw(frompattern, from)) return 0;

   while (true) {
      wchar_t patternc = *frompattern++,
                 reppc = *topattern++;
      // end of destination pattern
      if (!reppc) { *to=0; break; }
      // end of source pattern or file name
      if (!patternc || !*from) {
         if (wcspbrk(--topattern,L"*?")) return 0;
         wcscpy(to,topattern);
         break;
      }

      if (patternc=='?') {
         *to++ = reppc!='?'?reppc:*from;
         from++;
      } else
      if (patternc=='*') {
         wchar_t *pos = wcspbrk(frompattern,L"*?");
         if (pos==frompattern) return 0;
         // last entry?
         if (!*frompattern) {
            if (reppc=='*') wcscpy(to, from); else {
               if (wcspbrk(--topattern,L"*?")) return 0;
               wcscpy(to,topattern);
            }
            return 1;
         } else {
            int len=-1;
            if (!pos) pos = (wchar_t*)(frompattern+wcslen(frompattern));

            while (from[++len])
               if (from[len]==*frompattern)
                  if (!_wcsnicmp(frompattern,from+len,pos-frompattern)) break;

            if (reppc=='*')
               while (len--) *to++ = *from++;
            else {
               from+=len;
               len = -1; topattern--;
               while (topattern[++len])
                  if (topattern[len]==*frompattern)
                     if (!_wcsnicmp(frompattern,topattern+len,pos-frompattern)) break;
               while (len--)
                  if (wcschr(L"*?",*to++=*topattern++)) return 0;
            }
         }
      } else {
         if (reppc=='*') {
            int      len = -1;
            wchar_t *pos = wcspbrk(frompattern,L"*?");
            if (!pos) pos = (wchar_t*)(topattern+wcslen(topattern));

            while (from[++len])
               if (from[len]==*topattern)
                  if (!_wcsnicmp(topattern,from+len,pos-topattern)) break;

            while (len--) *to++ = *from++;
            frompattern += len;
         } else {
            *to++ = reppc!='?'?reppc:*from;
            from++;
         }
      }
   }
   return 1;
}

/* this code should be wrong in many cases, but it is used as a first time
   implementation */
int __stdcall _replacepattern(const char *frompattern, const char *topattern,
                              const char *from, char *to)
{
   if (!to) return 0;
   *to = 0;
   if (!frompattern||!topattern||!from) return 0;
   if (!_matchpattern(frompattern, from)) return 0;

   while (true) {
      char patternc = *frompattern++,
              reppc = *topattern++;
      // end of destination pattern
      if (!reppc) { *to=0; break; }
      // end of source pattern or file name
      if (!patternc || !*from) {
         if (strpbrk(--topattern,"*?")) return 0;
         strcpy(to,topattern);
         break;
      }

      if (patternc=='?') {
         *to++ = reppc!='?'?reppc:*from;
         from++;
      } else
      if (patternc=='*') {
         char *pos = strpbrk(frompattern,"*?");
         if (pos==frompattern) return 0;
         // last entry?
         if (!*frompattern) {
            if (reppc=='*') strcpy(to, from); else {
               if (strpbrk(--topattern,"*?")) return 0;
               strcpy(to,topattern);
            }
            return 1;
         } else {
            int len=-1;
            if (!pos) pos = (char*)(frompattern+strlen(frompattern));

            while (from[++len])
               if (from[len]==*frompattern)
                  if (!strnicmp(frompattern,from+len,pos-frompattern)) break;

            if (reppc=='*')
               while (len--) *to++ = *from++;
            else {
               from+=len;
               len = -1; topattern--;
               while (topattern[++len])
                  if (topattern[len]==*frompattern)
                     if (!strnicmp(frompattern,topattern+len,pos-frompattern)) break;
               while (len--)
                  if (strchr("*?",*to++=*topattern++)) return 0;
            }
         }
      } else {
         if (reppc=='*') {
            int   len = -1;
            char *pos = strpbrk(frompattern,"*?");
            if (!pos) pos = (char*)(topattern+strlen(topattern));

            while (from[++len])
               if (from[len]==*topattern)
                  if (!strnicmp(topattern,from+len,pos-topattern)) break;

            while (len--) *to++ = *from++;
            frompattern += len;
         } else {
            *to++ = reppc!='?'?reppc:*from;
            from++;
         }
      }
   }
   return 1;
}

/// common init for both ascii & utf _dos_findfirst() versions
static qserr findfirst_int(const void *dpath, int wide, u16t attrs, StatInfo **rsi) {
   StatInfo   *si = (StatInfo*)calloc_thread(sizeof(StatInfo),1);
   qserr      err;

   if (!wide) {
      err = io_fullpath (si->dira, (char*)dpath, QS_MAXPATH+1);
      if (!err) hlp_atow(si->dira, si->dirw, QS_MAXPATH+1, &err);
   } else
      err = io_fullpathw(si->dirw, (c16t*)dpath, QS_MAXPATH+1);

   if (!err) {
      c16t *wcp = wcsrchr(si->dirw, '\\');
      if (!wcp) {
         wcscpy(si->mask, si->dirw);
         si->dirw[0] = 0;
         si->dira[0] = 0;
      } else {
         wcscpy(si->mask, ++wcp);
         *wcp = 0;
         // make the same result for ascii string
         if (!wide) {
            char *cp = strrchr(si->dira, '\\');
            if (cp) *++cp = 0;
         }
      }
      // use utf open for all cases
      err = io_diropenw(si->dirw, &si->dh);
   }
   if (err) {
      free(si);
      // be compatible with Watcom runtime (set errno and return "OS error")
      set_errno_qserr(err);
      return err;
   }

   si->sign      = STAT_SIGN;
   si->call      = 2;
   si->attrs     = attrs;
   si->wide      = wide?1:0;
   // set process as block owner
   mem_localblock(si);
   *rsi = si;
   return 0;
}

qserr __stdcall START_EXPORT(_dos_findfirst)(const char *path, u16t attrs, dir_t *fp) {
   StatInfo *si;
   qserr    err = findfirst_int(path, 0, attrs, &si);
   if (err) return err;
   fp->d_openpath = si->dira;
   fp->d_sysdata  = si;
   fp->d_subdir   = 0;
   return _dos_findnext(fp);
}

qserr __stdcall START_EXPORT(_wdos_findfirst)(const c16t *path, u16t attrs, dirw_t *fp) {
   StatInfo *si;
   qserr    err = findfirst_int(path, 1, attrs, &si);
   if (err) return err;
   fp->d_openpath = si->dirw;
   fp->d_sysdata  = si;
   fp->d_subdir   = 0;
   return _wdos_findnext(fp);
}

qserr __stdcall START_EXPORT(_dos_findclose)(dir_t *fp) {
   qserr      err;
   checkret_direrr(E_SYS_INVPARM,2);
   err = io_dirclose(si->dh);
   if (err) set_errno_qserr(err);
   fp->d_sysdata = 0;
   si->sign      = 0;
   free(si);
   return err;
}

qserr __stdcall START_EXPORT(_wdos_findclose)(dirw_t *fp) {
   return START_EXPORT(_dos_findclose)((dir_t*)fp);
}

static qserr findnext_int(StatInfo *si, dirw_t *fp) {
   while (true) {
      qserr err = io_dirnextw(si->dh, &si->de);
      if (err) { set_errno_qserr(err); return err; }
      // filter by subdir flag only
      if (((si->attrs&_A_SUBDIR)==0 || (si->de.attrs&IOFA_DIR)!=0) &&
         _matchpatternw(si->mask,si->de.name)) 
      {
         fp->d_size     = si->de.size;
         time2dirent(&si->de, (dir_t*)fp);
         fp->d_attr     = si->de.attrs;
         break;
      }
   }
   return 0;
}

qserr __stdcall START_EXPORT(_wdos_findnext)(dirw_t *fp) {
   qserr err;
   checkret_direrr(E_SYS_INVPARM,2);
   err = findnext_int(si,fp);
   if (!err) {
      fp->d_openpath = si->dirw;
      wcscpy(fp->d_name, si->de.name);
   }
   return err;
}

qserr __stdcall START_EXPORT(_dos_findnext)(dir_t *fp) {
   qserr err;
   checkret_direrr(E_SYS_INVPARM,2);
   err = findnext_int(si,(dirw_t*)fp);
   if (!err) {
      u32t errcnt;
      qserr  err;
      if (!hlp_wtoa_int(si->de.name, fp->d_name, sizeof(fp->d_name), &err,
         '?', &errcnt)) set_errno_qserr(err);
         else
      if (errcnt) fp->d_attr|=_A_NAMEERR;

      fp->d_openpath = si->dira;
   }
   return err;
}

qserr __stdcall _dos_getdiskfree(unsigned drive, diskfree_t *dspc) {
   disk_volume_data vi;
   qserr           err;

   drive = !drive?io_curdisk():drive-1;
   err   = io_volinfo(drive, &vi);
   if (err) { set_errno_qserr(err); return err; }

   dspc->total_clusters      = vi.ClTotal;
   dspc->avail_clusters      = vi.ClAvail;
   dspc->sectors_per_cluster = vi.ClSize;
   dspc->bytes_per_sector    = vi.SectorSize;
   return 0;
}

int __stdcall unlink(const char *path) {
   qserr err = io_remove(path);
   if (err) set_errno_qserr(err);
   return err?-1:0;
}

int __stdcall START_EXPORT(rename)(const char *oldname, const char *newname) {
   qserr err = io_move(oldname, newname);
   if (err) set_errno_qserr(err);
   return err?-1:0;
}

void __stdcall _dos_getdrive(unsigned *drive) {
   if (drive) *drive = io_curdisk()+1;
}

void __stdcall _dos_setdrive(unsigned drive, unsigned *total) {
   if (drive) io_setdisk(drive-1);
   if (total) {
      u8t lmount = 1;
      io_ismounted(1, &lmount);
      *total = lmount+1;
   }
}

void __stdcall START_EXPORT(_splitpath)(const char *path, char *drive, char *dir,
   char *fname, char *ext)
{
   char *rs1,*rs2;
   if (drive) *drive=0;
   if (dir)   *dir  =0;
   if (fname) *fname=0;
   if (ext)   *ext  =0;
   if (!path||!*path) return;
   if (path[1]==':') {
      if (drive) { *drive++=path[0]; *drive++=':'; *drive++=0; }
      path+=2;
   }
   // skip duplicates
   while ((path[0]=='/'||path[0]=='\\') && (path[1]=='/'||path[1]=='\\'))
      path++;
   rs1=strrchr(path,'\\');
   rs2=strrchr(path,'/');

   if (rs1&&rs2) rs1=rs1-path>rs2-path?rs1:rs2; else
      if (rs2) rs1=rs2;
   if (rs1) {
      if (dir) strncat(dir,path,++rs1-path);
      path=rs1;
   }
   rs1=strrchr(path,'.');
   if (rs1) {
      if (ext) strcpy(ext,rs1);
      if (fname) strncat(fname,path,rs1-path);
   } else
   if (fname) strcpy(fname,path);
}

char* __stdcall START_EXPORT(_fullpath)(char *buffer, const char *path, size_t size) {
   qserr err;
   char  *rc = io_fullpath_int(buffer, path, size, &err);
   //log_it(2, "_fullpath(%08X,%s,%u) -> %08X (%s) (err:%08X)\n", buffer, path, size, rc, rc, err);
   if (err) set_errno_qserr(err);
   return err?0:rc;
}

c16t* __stdcall START_EXPORT(_wfullpath)(c16t *buffer, const c16t *path, size_t size) {
   qserr err;
   c16t  *rc = io_fullpathw_int(buffer, path, size, &err, 0, 0);
   //log_it(2, "_fullpathw(%08X,%S,%u) -> %08X (%S) (err:%08X)\n", buffer, path, size, rc, rc, err);
   if (err) set_errno_qserr(err);
   return err?0:rc;
}

#define TREE_SIGN   (0x45525452)
#define TREE_ALLOC  (_4KB-32)

typedef struct {
   u32t            sign;
   ptr_list         blk;   // list of memory blocks to free
   u32t          dircnt;
   u8t            *mblk;
   u32t           mleft;   // memory left in the current heap block
   u32t           mback;   // start of the last alloc (to revert back)
   char         *cvtbuf;   // preallocated buffer for text conversion
   u8t             wide;
} TreeState;

static int tree_cut(TreeState *rti) {
   if (rti->mblk && rti->mleft>=32)
      return mem_cut(rti->mblk - (TREE_ALLOC-rti->mleft), TREE_ALLOC-rti->mleft);
   return 0;
}

/* using 4k pools here, allocator is simple, but remainig part of the each
   block returned to heap just here */
static void* tree_alloc(TreeState *rti, u32t size) {
   void *res;
   if (rti->sign!=TREE_SIGN) return 0;
   if (!rti->blk) return 0;
   // direct alloc for huge blocks
   if (size>=TREE_ALLOC) {
      res = malloc_th(size);
      rti->blk->add(res);
      // last alloc is not from pool
      rti->mback = 0;
   } else {
      if (rti->mleft<size) {
         if (rti->mleft) tree_cut(rti);
         rti->mleft = TREE_ALLOC;
         rti->mblk  = (u8t*)malloc_th(TREE_ALLOC);
         rti->blk->add(rti->mblk);
      }
      res = rti->mblk;
      rti->mback  = rti->mleft;
      rti->mleft -= size;
      rti->mblk  += size;
   }
   return res;
}

/// "free" the last tree_alloc() op
static void tree_back(TreeState *rti) {
   if (rti->sign!=TREE_SIGN || !rti->blk || !rti->mback || rti->mback<rti->mleft)
      return;
   rti->mblk -= rti->mback - rti->mleft;
   rti->mleft = rti->mback;
   rti->mback = 0;
}

#define TREEWIDE     1   //< utf-16 output
#define TREEWIDE_U8  2   //< utf-8 output

// returned 0 means (impossible) E_SYS_LONGNAME error
static void *tree_strcopy(TreeState *rti, u32t ofs, c16t *src, int wide,
                          int wcopy, u32t *errcnt)
{
   u8t     *res;
   u32t     len,
           wlen = wcslen(src)+1<<1;
   char   *srca = 0;
   if (errcnt) *errcnt = 0;
   // it should never return 0
   if (wide && wide!=TREEWIDE_U8) {
      len   = wlen;
      wcopy = 0;
   } else {
      if (!wide) {
         if (!rti->cvtbuf)
            rti->cvtbuf = (u8t*)tree_alloc(rti, QS_MAXPATH+1);
         srca = hlp_wtoa_int(src, rti->cvtbuf, QS_MAXPATH+1, 0, '?', errcnt);
      } else {
         if (!rti->cvtbuf)
            rti->cvtbuf = (u8t*)tree_alloc(rti, QS_MAXPATH*4+1);
         srca = hlp_wto8(src, rti->cvtbuf, QS_MAXPATH*4+1, errcnt);
      }
      if (!srca) return 0;
      len = strlen(srca)+1;
   }
   res = (u8t*)tree_alloc(rti, ofs+len+(wcopy?wlen:0));
   memcpy(res+ofs, srca?(void*)srca:(void*)src, len);
   // wcopy means an utf16 name just after ascii/utf8 in the same buffer
   if (wcopy) memcpy(res+ofs+len, src, wlen);
   return res;
}

static long readtree(const c16t *Dir, const c16t *Mask, dirw_t **info,
   TreeState *rti, int subdirs, _dos_readtreew_cb callback, void *cbinfo,
   int wide)
{
   dirw_t   *rdh, *rdi, *iout, *itop;
   long       rc = 0, count, dircnt;
   int       all = !Mask;
   qserr    errv;
   wchar_t *dptr;

   /* using own heap here to minimize littering of the system heap. it
      allocs 4k heap blocks and split them for dir_t and other data. There
      is no free() but reverse of the last alloc is possible */
   if (!rti) {
      rti = (TreeState*)malloc_th(sizeof(TreeState));
      rti->sign   = TREE_SIGN;
      rti->blk    = NEW(ptr_list);
      rti->dircnt = 0;
      rti->mblk   = 0;
      rti->mleft  = 0;
      rti->mback  = 0;
      rti->cvtbuf = 0;
      rti->wide   = wide;
   }
#if 0
   log_printf("readtree(%S,%S,%08X,%08X,%i,%08X,%08X)\n", Dir, Mask, info, rti,
      subdirs, callback, cbinfo);
#endif
   dptr   = (wchar_t*)malloc_th(QS_MAXPATH+1<<1);
   errv   = io_fullpathw(dptr, Dir, QS_MAXPATH+1);
   if (!errv && dptr[wcslen(dptr)-1]!='\\') wcscat(dptr, L"\\");

   *info  = 0;
   itop   = 0;
   iout   = 0;
   count  = 0;
   dircnt = 0;
   rdh    = !errv?_wopendir(dptr):0;
   if (rdh) {
      // d_openpath value
      void *oppv = tree_strcopy(rti, 0, dptr, wide, 0, 0);

      if (!oppv) errv = E_SYS_LONGNAME; else
      do {
         rdi = _wreaddir(rdh);
         if (rdi) {
            int   isdir = rdi->d_attr&_A_SUBDIR?1:0;
            wchar_t *nm = rdi->d_name;
            // skip "."
            if (isdir && nm[0]=='.' && !nm[1]) continue;
            else {
               int match=all?1:_matchpatternw(Mask,rdi->d_name);

               if (match||isdir) {
                  u32t errcnt;
                  int  dotdot = isdir && nm[0]=='.' && nm[1]=='.' && !nm[2];
                  // we should fill d_subdir for this entry (all excpt "..")
                  int  incdir = isdir && subdirs && !dotdot;
                  // can`t be any allocs between this and tree_back() below!!
                  dirw_t  *de = (dirw_t*)tree_strcopy(rti, FIELDOFFSET(dirw_t,d_name),
                                                      nm, wide, incdir, &errcnt);
                  // should never occurs but better to be safe ...
                  if (!de) continue;
                  memcpy(de, rdi, FIELDOFFSET(dirw_t,d_name));
                  // mark .. as unmatched if "*" or "*.*" mask
                  if (isdir && (!match || all&&dotdot)) de->d_attr|=_A_DIRNMATCH;
                     else de->d_attr&=0x7F;
                  if (errcnt) de->d_attr|=_A_NAMEERR;
                  // ask callback
                  if (callback) {
                     int cbrc = callback(de,cbinfo);
                     if (cbrc<0) { rc=-1; count=0; break; } else
                     // try to free block allocated in tree_strcopy() above
                     if (!cbrc) {
                        tree_back(rti);
                        continue;
                     }
                  }
                  de->d_next     = 0;
                  de->d_sysdata  = rti;
                  de->d_openpath = oppv;
                  de->d_subdir   = 0;

                  if (incdir) dircnt++;
                  count++;
                  // link to the list
                  if (!itop) itop = de;
                  if (iout) iout->d_next = de;
                  iout = de;
               }
            }
         }
      } while (rdi);
      _wclosedir(rdh);
   }

   if (dircnt)
      for (rdi=itop; rdi; rdi=rdi->d_next)
         if ((rdi->d_attr&_A_SUBDIR)!=0 && wcscmp(rdi->d_name,L"..")) {
            long      len = wcslen(dptr), ncount;
            dirw_t *iout2 = 0;
            wchar_t   *nm = wcscat_dyn(0,dptr),
                    *nptr = rdi->d_name;
            // tree_strcopy() save a copy of utf16 name for us ...
            if (!wide || wide==TREEWIDE_U8)
               nptr = (wchar_t*)((char*)nptr + strlen((char*)nptr) + 1);

            if (nm[--len]=='\\') nm[len]='/';
            if (nm[len]!='/') nm = wcscat_dyn(nm, L"/");
            nm = wcscat_dyn(nm, nptr);

            if (wcslen(nm)<QS_MAXPATH) {
               rti->dircnt++;
               ncount = readtree(nm, Mask, &iout2, rti, 1, callback, cbinfo, wide);
               if (ncount<0) { rc=-1; break; } // stopped in child
               rc    += ncount;
               rdi->d_subdir = iout2;
            } else
               rdi->d_attr|=_A_NAMELENERR;

            free(nm);
         }
   if (rc>=0) { rc+=count; *info=itop; }
   free(dptr);
   return rc;
}

static long readtree_call(const c16t *Dir, const c16t *Mask, dirw_t **info,
   TreeState *rti, int subdirs, u32t *dircnt, _dos_readtreew_cb cbfn,
   void *cbinfo, int wide)
{
   if (dircnt) *dircnt = 0;
   if (Dir && (!info || !cbfn)) {
      dirw_t *iout = 0;
      long      rc;
      // reset it to not confuse by previous error & zero count
      set_errno(0);
      rc = readtree(Dir, Mask, &iout, 0, subdirs, cbfn, cbinfo, wide);

      if (rc>0) {
         TreeState *rti = (TreeState*)iout->d_sysdata;
         if (dircnt) *dircnt = rti->dircnt;
         tree_cut(rti);
      }
      if (!info || rc<=0) {
         _dos_freetree(iout);
         iout = 0;
      }
      if (info) *info = iout;
      return rc<0 ? 0 : rc;
   }
   if (info) *info = 0;
   set_errno(EINVAL);
   return 0;
}

u32t __stdcall START_EXPORT(_dos_readtree)(const char *dir, const char *mask,
   dir_t **info, int subdirs, int utf8, u32t *dircnt, _dos_readtree_cb cbfn,
   void *cbinfo)
{
   // filter incorrect call just here
   if (!dir || !info && !cbfn) set_errno(EINVAL); else {
      qserr err = 0;
      c16t *Dir = hlp_atow(dir, 0, 0, &err), *Mask = 0;
      u32t  res;
      // simplify the call ...
      if (mask)
         if (*mask==0 || strcmp(mask,"*")==0 || strcmp(mask,"*.*")==0)
            mask = 0;
      if (Dir) Mask = mask && *mask ? hlp_atow(mask, 0, 0, &err) : 0;
      
      if (err) set_errno_qserr(err); else
         res = readtree_call(Dir, Mask, (dirw_t**)info, 0, subdirs,
            dircnt, (_dos_readtreew_cb)cbfn, cbinfo, utf8?TREEWIDE_U8:0);

      if (Mask) free(Mask);
      if (Dir) free(Dir);

      if (!err) return res;
   }
   if (dircnt) *dircnt = 0;
   if (info) *info = 0;
   return 0;
}

u32t __stdcall START_EXPORT(_dos_readtreew)(const c16t *dir, const c16t *mask,
   dirw_t **info, int subdirs, u32t *dircnt, _dos_readtreew_cb cbfn, void *cbinfo)
{
   if (mask)
      if (*mask==0 || wcscmp(mask,L"*")==0 || wcscmp(mask,L"*.*")==0)
         mask = 0;
   return readtree_call(dir,mask,info,0,subdirs,dircnt,cbfn,cbinfo,TREEWIDE);
}

/* this function actually receives both dir_t* and dirw_t* but offsets should
   be the same, this assumed in many places */
qserr __stdcall START_EXPORT(_dos_freetree)(dir_t *info) {
   TreeState *rti;
   u32t  ii, stat;
   if (!info) return E_SYS_INVOBJECT;
   rti = (TreeState*)info[0].d_sysdata;
   if (!rti) return E_SYS_INVOBJECT;
   /* is it our`s dir_t? (to deny opendir`s dir_t and so on)
      and make it thread safe */
   if (mt_cmpxchgd(&rti->sign,0,TREE_SIGN)!=TREE_SIGN) return E_SYS_INVOBJECT;
   for (ii=0, stat=0; ii<rti->blk->count(); ii++) {
      void *blk = rti->blk->value(ii);
      stat += mem_blocksize(blk);
      free(blk);
   }
   log_it(3, "readtree(%X) - %uk used\n", info, stat>>10);

   DELETE(rti->blk);
   free(rti);
   return 0;
}

int __stdcall START_EXPORT(_access)(const char *path, int mode) {
   io_handle_info hi;
   qserr         err = io_pathinfo(path, &hi);
   if (!err)
      if ((mode&W_OK) && (hi.attrs&IOFA_READONLY) ||
         (hi.attrs&IOFA_DIR) && (mode&(W_OK|R_OK|X_OK))) err = E_SYS_ACCESS;
   if (err) set_errno_qserr(err);
   return err?-1:0;
}

char* __stdcall tmpdir(char *buffer) {
   static const char *tnames[4] = {"TMP",  "TEMP",  "TMPDIR", "TEMPDIR"};
   static char ncopy[NAME_MAX+1];
   u32t ii;

   if (!buffer) buffer = ncopy;

   for (ii=0; ii<4; ii++) {
      char *dp = getenv(tnames[ii]);
      if (dp) {
         dir_t  fd;
         int   len = strlen(dp);
         if (len>0 && len<NAME_MAX) {
            strcpy(buffer, dp);
            // remove trailing \, but not from "A:\" and from "/"
            if (len>1 && (len!=3 || buffer[1]!=':')  && (buffer[len-1]=='/' ||
               buffer[len-1]=='\\')) buffer[len-1] = 0;

            if (!_dos_stat(buffer,&fd))
               if (fd.d_attr&_A_SUBDIR) return buffer;
         }
      }
   }
   *buffer = 0;
   return 0;
}

qserr __stdcall START_EXPORT(_dos_setfileattr)(const char *path, unsigned attributes) {
   if (!path||(attributes&(_A_VOLID|_A_SUBDIR))!=0)  { 
      set_errno(EINVAL);
      return E_SYS_INVPARM;
   } else {
      io_handle_info  hi;
      qserr          err;
      hi.attrs = attributes;
      err      = io_setinfo(path, &hi, IOSI_ATTRS);

      if (err) set_errno_qserr(err);
      return err;
   }
}

qserr __stdcall START_EXPORT(_dos_getfileattr)(const char *path, unsigned *attributes) {
   io_handle_info  hi;
   qserr          err;
   if (attributes) *attributes = 0;
   if (!attributes||!path) { set_errno(EINVAL); return E_SYS_INVPARM; }

   err = io_pathinfo(path, &hi);
   if (!err && (hi.attrs&IOFA_DEVICE)) err = E_SYS_TTYDEV;
   if (err) set_errno_qserr(err); else *attributes = hi.attrs;

   return err;
}

qserr __stdcall START_EXPORT(_dos_setfiletime)(const char *path, u32t dostime, int type) {
   io_handle_info  hi;
   u32t           flv = 0;
   time_t         tmv;
   qserr          err;
   if (!path||!type||(type&~(_DT_MODIFY|_DT_CREATE)))
      { set_errno(EINVAL); return E_SYS_INVPARM; }

   tmv = dostimetotime(dostime);
   if (type&_DT_MODIFY) { io_timetoio (&hi.wtime, tmv); flv|=IOSI_WTIME; }
   if (type&_DT_CREATE) { io_timetoio (&hi.ctime, tmv); flv|=IOSI_CTIME; }
   err = io_setinfo(path, &hi, flv);

   if (err) set_errno_qserr(err);
   return err;
}

qserr __stdcall START_EXPORT(_dos_getfiletime)(const char *path, u32t *dostime, int type) {
   io_handle_info  hi;
   qserr          err;
   if (!path||!dostime||!type||(type&~(_DT_MODIFY|_DT_CREATE)))
      { set_errno(EINVAL); return E_SYS_INVPARM; }

   err = io_pathinfo(path, &hi);
   if (err) set_errno_qserr(err); else {
      time_t tmv = io_iototime(type&_DT_CREATE?&hi.ctime:&hi.wtime);
      *dostime   = timetodostime(tmv);
   }
   return err;
}

static int utime_common(const void *path, const struct utimbuf *tmv, int utf) {
   io_handle_info  fi;
   qserr          err;
   // strange but it should return EACCES for zero args (watcom help says so)
   if (!path||!tmv) { set_errno(EACCES); return -1; }

   io_timetoio(&fi.wtime, tmv->modtime);
   io_timetoio(&fi.atime, tmv->actime);
   err = utf ? io_setinfow((const c16t*)path, &fi, IOSI_ATIME|IOSI_WTIME):
               io_setinfo ((const char*)path, &fi, IOSI_ATIME|IOSI_WTIME);
   if (err) { set_errno_qserr(err); return -1; }
   return 0;
}

int __stdcall utime(const char *path, const struct utimbuf *tmv) {
   return utime_common(path, tmv, 0);
}

int __stdcall _wutime(const wchar_t *path, const struct utimbuf *tmv) {
   return utime_common(path, tmv, 1);
}
