//
// QSINIT runtime
// simple heap for the fixed size small blocks
// -------------------------------
// * 32/64 code 
//
#include "qcl/qstheap.h"
#include "qserr.h"

#define POOL_SIZE   _64KB   ///< can be increased to 128k in init
#define PTHDEBUG        0

#ifdef EFIX64
#define BMSTEP8             ///< 64-bit variant uses qwords for bit ops
#define bmi          u64t

#elif defined(__QSINIT__)
#include "qsutil.h"
#include <stdlib.h>
#define bmi          u32t

/// pool allocation
#define hpalloc(size,flags) hlp_memalloc((size), QSMA_RETERR|(flags));
/// pool free
#define hpfree(mem) hlp_memfree(mem)
/// control data allocation
static void *hmalloc(u32t size, u32t flags) {
   if (flags&QSMA_LOCAL) {
      void *res = malloc_thread(size);
      if (res) mem_localblock(res);
      return res;
   }
   return realloc_shared(0,size);
}
#define hmfree(mem) free(mem)

#else
#error ???
#endif

#define PTH_SIGN         0x49485450          // PTHI
#define PTH_DEBUG        1

typedef struct {
   u32t         sign;
   u32t         nblk,
            bperpool,  ///< # of blocks per heap pool (POOL_SIZE)
               npool,  ///< # of pools in the heap
               bsize;  ///< size of the block
   u32t         hint;
   u32t    lpoolsize;  ///< size of the last heap pool in bytes (<=POOL_SIZE)
   bmi       *bitmap;
   u8t          **mp;  ///< heap pools
   u32t        memfl;  ///< allocation flags
} pthheap;

#define instance_ret(inst,err)        \
   pthheap *ph = (pthheap*)(inst);    \
   if (ph->sign!=PTH_SIGN) return err;

#define instance_void(inst)           \
   pthheap *ph = (pthheap*)(inst);    \
   if (ph->sign!=PTH_SIGN) return;

/** initialize heap.
    Note that heap instance calls are not thread-safe, use fxszheap shared
    class to get thread safe instances.

    @param  blocksize   size of the heap block (the system memory allocation
                        size is 64k, so this size should be much smaller).
    @param  limit       maximum number of heap blocks (will be aligned to 64)
    @param  memflags    only QSMA_LOCAL and QSMA_READONLY are accepted (r/o
                        can be used for PID 1 heaps)
    @param  [out] res   heap if success
    @return error code or 0 */
qserr _std hlp_thpinit(u32t blocksize, u32t limit, u32t memflags, qs_thp *result) {
   u32t    bperp, np, bmsize;
   pthheap  *res;

   if (!result || !limit) return E_SYS_INVPARM;
   *result = 0;
#ifdef EFIX64
   if (memflags) return E_SYS_INVPARM;
#else
   if ((memflags & ~(QSMA_LOCAL|QSMA_READONLY))) return E_SYS_INVPARM;
#endif
   // at least 2 blocks per pool
   if (POOL_SIZE/2 < blocksize) return E_SYS_TOOLARGE;
   // align it for bitmap
   limit   = Round64(limit);
   // blocks in the pool
   bperp   = POOL_SIZE / blocksize;
   // if 128k can accommodate one more entry compared to 64k*2 - switch to it
   if (POOL_SIZE*2 / blocksize > bperp*2) bperp = POOL_SIZE*2 / blocksize;
   // number of pools for the heap
   np      = limit / bperp;
   // there is no "limit" aligning up since its size can be assumed by the caller
   if (limit%bperp) np++;
   // allocate heap structs (using requested memory type!)
   bmsize  = sizeof(u64t)*(limit>>6);
   // ex.: 32Mb blocks of 40 byte: 4Mb bitmap + 20k*sizeof(void*) + sizeof(pthheap)
   res     = (pthheap*)hmalloc(sizeof(pthheap)+bmsize+sizeof(u8t*)*np, memflags);
   if (!res) return E_SYS_NOMEM;

   res->sign      = PTH_SIGN;
   res->bitmap    = (bmi*)(res + 1);
   res->mp        = (u8t**)((u8t*)res->bitmap + bmsize);
   // bitmap
   memset(res->bitmap, 0, bmsize);
   // pool list
   memset(res->mp, 0, sizeof(u8t*)*np);
#if PTHDEBUG>0
   log_printf("pth %p: limit %u, %u pools, bmsize:%u, bm:%p, mp:%p\n", res,
      limit, np, bmsize, res->bitmap, res->mp);
#endif
   res->nblk      = limit;
   res->bperpool  = bperp;
   res->npool     = np;
   res->bsize     = blocksize;
   res->lpoolsize = (limit%bperp ? limit%bperp : bperp) * blocksize;

   res->hint      = 0;
   res->memfl     = memflags;
   *result        = res;
   return 0;
}

/** free heap.
    @param  hp          heap
    @return error code or 0 */
qserr _std hlp_thpfini(qs_thp hp) {
   u32t ii;
   instance_ret(hp,E_SYS_INVOBJECT);
   ph->sign = 0;

   for (ii=0; ii<ph->npool; ii++)
      if (ph->mp[ii]) { hpfree(ph->mp[ii]); ph->mp[ii] = 0; }
   hmfree(ph);
   return 0;
}

static qserr freeindex(pthheap *ph, u32t index) {
   bmi  *rp;
   if (index>=ph->nblk) return E_SYS_INVPARM;
#ifndef BMSTEP8
   if (!_bitreset(ph->bitmap + (index>>5), index&0x1F)) return E_SYS_INVPARM;
   index>>=5;
#else
   if (!_bitresetq(ph->bitmap + (index>>6), index&0x3F)) return E_SYS_INVPARM;
   index>>=6;
#endif
   if (index<ph->hint) ph->hint = index;
#if PTHDEBUG>0
   log_printf("pth %p: freeindex %u -> hint=%u\n", ph, index, ph->hint);
#endif
   return 0;
}

/** allocate block in the heap.
    @param  hp          heap
    @return pointer or 0 if error */
void* _std hlp_thpalloc(qs_thp hp) {
   u32t  idx = FFFF;
   bmi   *rp;
   instance_ret(hp,0);
   rp = ph->bitmap + ph->hint;
#ifndef BMSTEP8
   if (*rp==FFFF) rp = memchrnd(rp, FFFF, (ph->nblk>>5) - ph->hint);
   if (rp) {
      u32t bit = bsf32(~*rp);
      _bitset(rp, bit);
      idx = rp - ph->bitmap;
      if (idx>ph->hint) ph->hint = idx;
#if PTHDEBUG>1
      log_printf("pth %p: rp=%p:%08X, idx=%u, bit=%u, hint=%u\n", ph, rp, *rp,
         idx, bit, ph->hint);
#endif
      idx = (idx<<5) + bit;
   }
#else // BMSTEP8
   if (*rp==FFFF64) rp = memchrnq(rp, FFFF64, (ph->nblk>>6) - ph->hint);
   if (rp) {
      u32t bit = bsf64(~*rp);
      _bitsetq(rp, bit);
      idx = rp - ph->bitmap;
      if (idx>ph->hint) ph->hint = idx;
      idx = (idx<<6) + bit;
   }
#endif
   if (idx!=FFFF) {
      u32t   pool = idx/ph->bperpool;
      u8t   *bptr;
      if (!ph->mp[pool]) {
         // allocate exact size (it can be 18k*3 for example)
         bptr = (u8t*)hpalloc(pool==ph->npool-1 ? ph->lpoolsize :
                              ph->bperpool*ph->bsize, ph->memfl);
         if (!bptr) {
            freeindex(ph, idx);
            return 0;
         }
#if PTHDEBUG>0
         log_printf("pth %p: new heap %p\n", ph, bptr);
#endif
         ph->mp[pool] = bptr;
      }
      bptr = ph->mp[pool] + (idx - pool*ph->bperpool) * ph->bsize;
      memset(bptr, 0, ph->bsize);
#if PTHDEBUG>0
      log_printf("pth %p: alloc=%u/%p, hint=%u\n", ph, idx, bptr, ph->hint);
#endif
      return bptr;
   }
   return 0;
}

/** free heap block by pointer
    @param  hp          heap
    @param  ptr         heap block pointer
    @return error code or 0 */
qserr _std hlp_thpfree(qs_thp hp, void *ptr) {
   u32t   ii, psize;
   ptrdiff_t    src = (ptrdiff_t)ptr;
   instance_ret(hp,E_SYS_INVOBJECT);

   for (ii=0, psize=ph->bperpool*ph->bsize; ii<ph->npool; ii++) {
      ptrdiff_t pool = (ptrdiff_t)ph->mp[ii];
      if (pool) {
         if (ii==ph->npool-1) psize = ph->lpoolsize;
         // ptrdiff_t is signed now (as it should be)
         if (src-pool>=0) {
            u32t ofs = src-pool;
            if (ofs<psize) {
               if (ofs % ph->bsize) return E_SYS_INVPARM;
               ofs /= ph->bsize;
               ofs += ii*ph->bperpool;
               memset(ptr, 0, ph->bsize);
               return freeindex(ph, ofs);
            }
         }
      }
   }
   return E_SYS_INVPARM;
}

#ifdef QSINIT_EXPCLASSES

static qserr _exicc thpsc_init(EXI_DATA, u32t blocksize, u32t limit, u32t flags) {
   qs_thp hp = *(qs_thp*)data;
   if (hp) return E_SYS_INITED;
   return hlp_thpinit(blocksize, limit, flags, data);
}

static void* _exicc thpsc_alloc(EXI_DATA) {
   qs_thp hp = *(qs_thp*)data;
   if (!hp) return 0;
   return hlp_thpalloc(hp);
}
static qserr _exicc thpsc_free(EXI_DATA, void *ptr) {
   qs_thp hp = *(qs_thp*)data;
   if (!hp) return E_SYS_NONINITOBJ;
   return hlp_thpfree(hp, ptr);
}

static void _std thpsc_create(void *instance, void *data) {
   *(qs_thp*)data = 0;
}

static void _std thpsc_done(void *instance, void *data) {
   qs_thp hp = *(qs_thp*)data;
   if (hp) hlp_thpfini(hp);
}

static void *f_theap_list[3] = { thpsc_init, thpsc_alloc, thpsc_free };

void exi_register_theap(void) {
   exi_register("qs_theap", f_theap_list, 3, sizeof(qs_thp), 0,
                thpsc_create, thpsc_done, 0);
}

#endif // QSINIT_EXPCLASSES
