/*
  note that this code should be compilable as both 32/64!
*/
#include "arena.h"
#include "qserr.h"
#ifndef EFIX64
#include "qspdata.h"
#include "stdlib.h"
#include "qspage.h"
#include "qslog.h"
#define canread(x) (pag_query((u8t*)(x)-16) & PGEA_READ)
#else
#define canread(x) (1)
#endif

#define SzToFreeIdx(x) ((x)>>20-PAGESHIFT)  ///< size in page -> index in free[]

qserr _std sar_create(u64t minaddr, u32t pages, u32t aflags, mt_arena **res) {
   qs_thp        heap;
   qserr          err;
   mt_arena        *a;
   mt_arange      *b1;
   u32t          sz1m;

   if (!res) return E_SYS_ZEROPTR;
   *res = 0;
   // start address & number of pages are 4Mb aligned
   if (!pages || (minaddr&_4MB-1) || (minaddr>>PAGESHIFT>FFFF) ||
      (pages&(_4MB>>PAGESHIFT)-1)) return E_SYS_INVPARM;
   // shift to megabytes (one more for the full free entry)
   sz1m = SzToFreeIdx(pages) + 1;

   err  = hlp_thpinit(sizeof(mt_arange), pages, 0, &heap);
   if (err) return err;
   b1   = (mt_arange*)hlp_thpalloc(heap);
   a    = b1 ? (mt_arena*)malloc(sizeof(mt_arena) + sz1m*sizeof(void*)) : 0;
   if (!a) {
      hlp_thpfini(heap);
      return E_SYS_NOMEM;
   }
   a->sign         = ARENA_SIGN;
   a->minpage      = minaddr>>PAGESHIFT;
   a->npages       = pages;
   a->free         = (mt_arange**)(a+1);
   a->heap         = heap;
   a->flstsize     = sz1m;
   a->pdir         = 0;
   a->owner        = 0;
   // free block list
   memset(a->free, 0, sz1m*sizeof(void*));

   b1->sign        = ARANGE_SIGN;
   b1->startpage   = 0;
   b1->pages       = pages;

   a->free[sz1m-1] = b1;
   a->first        = b1;
   a->mid          = b1;

   *res = a;
   return 0;
}

/// delete arena
qserr _std sar_delete(mt_arena *a) {
   qserr res = 0;
   if (a->sign!=ARENA_SIGN) res = E_SYS_INVOBJECT; else {
      if (a->heap) res = hlp_thpfini(a->heap);
      a->heap = 0;
      a->sign = 0;
      free(a);
   }
   return res;
}


/* a bit paranoidal check (as always ;))
   errtitle==0 assumed as brief output */
static int verifyblock(mt_arena *a, mt_arange *r, const char *errtitle) {
   // error information
   static const char *estr[] = {0, "invalid signature", "pos/size", "prev link",
      "next link", "prev free", "free list", "free list top" };
   int  err = 0;
   u32t  ep = r->startpage + r->pages;

   if (r->sign!=ARANGE_SIGN) err = 1; else
   if (r->startpage>a->npages || ep>a->npages) err = 2; else
   if (r->prev && r->prev->startpage+r->prev->pages != r->startpage) err = 3; else
   if (r->next && ep != r->next->startpage) err = 4; else
   {
      /* check only one step back for the free block:
         it should be the same size with the previous block or in the right pos
         in free[] if it has no previous */
      if (errtitle && !r->flags) {
         u32t selfidx = SzToFreeIdx(r->pages);
         if (r->prevfree) {
            mt_arange *p = r->prevfree;
            if (p->sign!=ARANGE_SIGN) {
               // show brief dump
               verifyblock(a,p,0);
               err = 5;
            } else
            if (selfidx!=SzToFreeIdx(p->pages)) err = 6;
         } else
         if (a->free[selfidx]!=r) err = 7;
      }
   }
   if (err) {
      log_it(2, "arena %p, range %p: %s\n", a, r, estr[err]);
      log_it(2, "%p: %12lb\n", r, r);
      if (errtitle) {
         log_it(2, "caller: %s\n", errtitle);
         sar_dump(a,0);
      }
   }
   return !err;
}

/// get range with specified page (page # in arena address space)
static mt_arange* sar_findaddr(mt_arena *a, u32t page) {
   mt_arange *ar;
   if (a->sign!=ARENA_SIGN) return 0;
   if (page>=a->npages) return 0;
   // try to speed-up a bit ...
   ar = page>=a->npages>>1 ? a->mid : a->first;
   while (ar) {
      // broken range?
      if (ar->sign!=ARANGE_SIGN) { verifyblock(a,ar,"findaddr"); return 0; }
      if (page>=ar->startpage && page<ar->startpage+ar->pages) return ar;
      ar = ar->next;
   }
   return 0;
}

/// insert free block into the arena`s list
static int freelink(mt_arena *a, mt_arange *r) {
   u32t  idx1m = SzToFreeIdx(r->pages);
   // some asserts ... (is shouldn`t already be inserted)
   if (r->flags || r->sign!=ARANGE_SIGN || r->prevfree || a->free[idx1m]==r) {
      verifyblock(a,r,"freelink");
      return 0;
   }
   r->nextfree    = a->free[idx1m];
   a->free[idx1m] = r;
   if (r->nextfree) r->nextfree->prevfree = r;
   return 1;
}

/// remove a free block from the arena list
static int freeunlink(mt_arena *a, mt_arange *r) {
   u32t  idx1m = SzToFreeIdx(r->pages);
   // some asserts ...
   if (r->flags || r->sign!=ARANGE_SIGN || !r->prevfree && a->free[idx1m]!=r) {
      verifyblock(a,r,"freeunlink");
      return 0;
   }
   if (r->nextfree) r->nextfree->prevfree = r->prevfree;
   if (r->prevfree) r->prevfree->nextfree = r->nextfree;
      else a->free[idx1m] = r->nextfree;
   r->nextfree = 0;
   r->prevfree = 0;
   return 1;
}

static void updatemid(mt_arena *a, mt_arange *r) {
   if (r->startpage<=a->npages>>1 && r->startpage+r->pages>a->npages>>1)
      a->mid = r;
}

/** split range.
    Function removes the source range from free list.
    @param  a       arena
    @param  r       range to split
    @param  offset  offset from the start of the range (!), in pages
    @return second range or 0 in case of error/some data damage. */
static mt_arange* sar_split(mt_arena *a, mt_arange *r, u32t offset) {
   if (r->sign!=ARANGE_SIGN) verifyblock(a,r,"split"); else
   if (offset<r->pages) {
      mt_arange *nr = (mt_arange*)hlp_thpalloc(a->heap);

      if (nr) {
         if (!r->flags) {
            u32t idx1m = SzToFreeIdx(r->pages);
            if (r->prevfree || a->free[idx1m]==r)
               if (!freeunlink(a, r)) {
                  hlp_thpfree(a->heap,nr);
                  return 0;
               }
         }
         nr->sign        = ARANGE_SIGN;
         nr->startpage   = r->startpage + offset;
         nr->pages       = r->pages - offset;
         nr->flags       = r->flags;
         nr->next        = r->next;
         nr->prev        = r;
         r->next         = nr;
         r->pages        = offset;
         if (nr->next) nr->next->prev = nr;
         updatemid(a,nr);
      
         return nr;
      }
   }
   return 0;
}

/** merge some ranges.
    Function removes source ranges from the free list.
    @param  a       arena
    @param  r       first range
    @param  nblk    # of ranges to merge, flags must be equal!
    @return error code or 0 */
static qserr sar_merge(mt_arena *a, mt_arange *r, u32t nblk) {
   mt_arange *rp = r, *rn;
   u32t       ii;

   if (!a || !r) return E_SYS_ZEROPTR;
   if (nblk<=1) return E_SYS_INVPARM;

   for (ii=0; ii<nblk; ii++) {
      rn = rp->next;
      // min check ...
      if (rp->sign!=ARANGE_SIGN || rn &&
          rp->startpage+rp->pages != rn->startpage)
      {
         verifyblock(a,r,"merge");
         return E_SYS_SOFTFAULT;
      }
      if (ii<nblk-1) {
         if (!rn) return E_SYS_INVPARM;
         if (rn->flags!=rp->flags) return E_SYS_ACCESS;
         rp = rn;
      }
   }
   // try to remove all from the free list
   if (!r->flags)
      for (rp=r, ii=0; ii<nblk; ii++) {
         u32t idx1m = SzToFreeIdx(rp->pages);
         if (rp->prevfree || a->free[idx1m]==rp)
            if (!freeunlink(a, rp)) return E_SYS_SOFTFAULT;
         rp = rp->next;
      }
   // merge them finally ...
   for (rn=r->next, ii=1; ii<nblk; ii++) {
      r->pages += rn->pages;
      r->next   = rn->next;

      rp = rn;
      rn = rn->next;
      rp->sign = 0;
      hlp_thpfree(a->heap, rp);
   }
   if (rn) rn->prev = r;
   updatemid(a,r);

   return 0;
}

static qserr verifyloc(mt_arena *a, u32t areq, u32t *forceloc) {
   if (*forceloc==SARA_LOCBELOW) {
      if (areq>=a->npages) *forceloc = 0; else
         if (!areq) return E_SYS_ARANGE;
   } else
   if (*forceloc==SARA_LOCABOVE) {
      if (areq>=a->npages) return E_SYS_ARANGE;
   }
   return 0;
}

qserr _std sar_alloc(mt_arena *a, u32t *page, u32t pages, u32t flags,
                     void *user, const char *sign)
{
   mt_arange   *res = 0;
   u32t        sz1m, ii, offset,
           forceloc = flags & SARA_LOCMASK,
           forcemod = flags & SARA_ADDRMASK,
               areq = *page;

   if (!a || !page) return E_SYS_ZEROPTR;
   if (!pages) return E_SYS_INVPARM;
   if (a->sign!=ARENA_SIGN) return E_SYS_INVOBJECT;
   if (a->npages<pages) return E_SYS_NOMEM;
   // drop forced location flag if it out of arena range ...
   if (forceloc) {
      qserr rc = verifyloc(a, areq, &forceloc);
      if (rc) return rc;
      // wanted location
      if (forceloc==SARA_LOC) {
         if (forcemod) return E_SYS_INVPARM;
         res    = sar_findaddr(a,areq);
         if (!res) return E_SYS_ARANGE;
         // range is used?
         if (res->flags) return E_SYS_ARANGE;
         offset = areq - res->startpage;
         // lack of pages ...
         if (res->pages - offset < pages) return E_SYS_NOMEM;
      }
   }
   // common allocation
   if (!res) {
      mt_arange *sm = 0;
      // index of the smallest suitable free block list
      u32t     sz1m = SzToFreeIdx(pages), ii;

      for (ii=sz1m; ii<a->flstsize; ii++)
         if (a->free[ii]) {
            u32t    diff = FFFF;
            mt_arange *r = a->free[ii];
            while (r) {
               // use the smallest block (except MIN/MAXaddr case)
               if (r->pages>=pages && (r->pages-pages<diff || forcemod))
                  do {
                     // use highest/lowest address
                     if (forcemod) {
                        if (forcemod==SARA_ADDRMIN)
                           if (sm && sm->startpage<r->startpage) break;
                        if (forcemod==SARA_ADDRMAX)
                           if (sm && sm->startpage>r->startpage) break;
                     }
                     // verify that block fits in the requested range ...
                     if (forceloc==SARA_LOCABOVE && areq > r->startpage) {
                        u32t eofs = r->pages - pages;
                        if (r->startpage+eofs < areq) break;
                        if (!verifyblock(a,r,"alloc")) break;
                        /* at the end of this block or just at the LOCABOVE
                           address if ADDRMIN wanted */
                        sm     = r;
                        offset = forcemod==SARA_ADDRMIN ?
                                 areq - r->startpage : eofs;
                     } else {
                        u32t size = r->pages;
                        if (forceloc==SARA_LOCBELOW)
                           if (r->startpage+pages > areq) break; else
                              if (r->startpage+size > areq)
                                 size = areq - r->startpage;
                        if (!verifyblock(a,r,"alloc")) break;
                        sm     = r;
                        // take in mind combo of LOCBELOW and ADDRMAX
                        offset = forcemod==SARA_ADDRMAX && size>pages ?
                                 size - pages : 0;
                     }
                     diff = r->pages-pages;
                  } while (0);
               r = r->nextfree;
            }
            /* SARA_ADDRMIN/MAX forces entire free pool enumeration in search
               for the lowest/highest address */
            if (sm && !forcemod) break;
         }
      if (!res && sm) res = sm;
   }
   if (res) {
      // wanted address is not at the start of the block
      if (offset) {
         mt_arange *rb = sar_split(a, res, offset);
         if (!rb || !freelink(a, res)) return E_SYS_SOFTFAULT;
         // part started with offset
         res    = rb;
         offset = 0;
      } else
      if (res->pages==pages) {
         /* normally sar_split() unlinks res block from the free list, but
            in this case it just never called */
         u32t idx1m = SzToFreeIdx(res->pages);
         if (res->prevfree || a->free[idx1m]==res)
            if (!freeunlink(a, res)) return E_SYS_SOFTFAULT;
      }
      if (res->pages>pages) {
         mt_arange *rem = sar_split(a, res, pages);
         if (!rem || !freelink(a, rem)) return E_SYS_SOFTFAULT;
      }
      res->flags    = ABF_USED;
      res->userdata = user;
      res->usertext = sign ? *(ptrdiff_t*)sign : 0;

      *page = res->startpage;
      return 0;
   }
   return E_SYS_NOMEM;
}

// actual free ...
static qserr freerange(mt_arena *a, mt_arange *r, int check) {
   u32t     fcnt = 0;
   mt_arange *rp = r->prev;
   qserr     res = 0;

   if (r->sign!=ARANGE_SIGN) return E_SYS_INVOBJECT;
   if (!r->flags) return E_SYS_INVPARM;

   r->flags      = 0;
   r->userdata   = 0;
   r->usertext   = 0;

   while (rp && !rp->flags) {
      if (check && !verifyblock(a,rp,0)) return E_SYS_INVOBJECT;
      fcnt++;
      if (rp->prev && !rp->prev->flags) rp = rp->prev; else break;
   }
   if (!fcnt) rp = r;
   // count self too ...
   while (r && !r->flags) {
      if (check && !verifyblock(a,r,0)) return E_SYS_INVOBJECT;
      r = r->next;
      fcnt++;
   }
   //log_printf("[[%X %u]]\n", rp->startpage, fcnt);
   if (fcnt>1) res = sar_merge(a, rp, fcnt);
   if (!res && !freelink(a,rp)) res = E_SYS_SOFTFAULT;

   return res;
}

/** update flags of the address range (or free it).
    @param  a       arena
    @param  page    start page of the allocated range (offset in arena)
    @param  flags   new flags value (SARA_*) or 0 to free the range
    @return error code or 0 */
qserr _std sar_update(mt_arena *a, u32t page, u32t flags, void *user) {
   mt_arange *r;
   if (!a) return E_SYS_ZEROPTR;
   if (a->sign!=ARENA_SIGN) return E_SYS_INVOBJECT;
   if (a->npages<=page) return E_SYS_INVPARM;

   r = sar_findaddr(a,page);
   if (!r || r->startpage!=page) return E_SYS_INVPARM;

   if (flags) {
      r->flags    = flags|ABF_USED;
      r->userdata = user;
   } else
      return freerange(a,r,1);

   return 0;
}

qserr _std sar_ufree(mt_arena *a, void *user) {
   mt_arange *rp, *head;
   qserr      res, fres;

   if (!a || !user) return E_SYS_ZEROPTR;
   if (a->sign!=ARENA_SIGN) return E_SYS_INVOBJECT;

   /* make a list at first pass since we cannot use ->next of the releasing
      block because next can be merged with this one */
   rp   = a->first;
   head = 0;
   while (rp) {
      // make a check for every block here!
      if (!verifyblock(a,rp,"ufree")) return E_SYS_SOFTFAULT;

      if ((rp->flags & ABF_USED) && rp->userdata==user) {
         rp->userdata = head;
         head = rp;
      }
      rp = rp->next;
   }
   if (!head) return E_SYS_NOTFOUND;
   // free all entries in the list created above
   res = 0;
   while (head) {
      // save next
      rp   = (mt_arange*)head->userdata;
      // no verify here since we done it above for each block!
      fres = freerange(a,head,0);
      if (fres) res = fres;
      // still loop even after errors!
      head = rp;
   }
   return res;
}

qserr _std sar_query(mt_arena *a, u32t *page, u32t *pages, u32t *flags, void **user) {
   mt_arange *r;
   if (!a || !page) return E_SYS_ZEROPTR;
   if (a->sign!=ARENA_SIGN) return E_SYS_INVOBJECT;
   if (a->npages<=*page) return E_SYS_INVPARM;

   r = sar_findaddr(a,*page);
   if (!r) return E_SYS_SOFTFAULT;
   *page = r->startpage;

   if (user)
      *user = r->flags ? r->userdata : 0;
   if (flags) *flags = r->flags;
   if (pages) *pages = r->pages;
   return 0;
}

/** query free space.
    Note that SARA_LOC, SARA_LOCABOVE and SARA_LOCBELOW will filter returned
    "total" value (for SARA_LOC it will be equal to smax - a length of while
    block available at this address)

    @param          a       arena
    @param [out]    total   # of available pages (ptr can be 0)
    @param [out]    smax    the longest available block size (in pages, ptr can be 0)
    @param          flags   SARA_LOC, SARA_LOCABOVE and SARA_LOCBELOW accepted here
    @param          page    wanted page (offset from the start of arena, only with flags)
    @return error code or 0 */
qserr _std sar_getfree(mt_arena *a, u32t *total, u32t *smax, u32t flags, u32t page) {
   mt_arange  *r;
   u32t      max = 0,
             all = 0;

   if (!total && !smax) return E_SYS_INVPARM;
   if (total) *total = 0;
   if (smax)  *smax = 0;

   if (!a) return E_SYS_ZEROPTR;
   if (a->sign!=ARENA_SIGN) return E_SYS_INVOBJECT;
   if (flags&~(SARA_LOC|SARA_LOCABOVE|SARA_LOCBELOW)) return E_SYS_INVPARM;
   // drop forced location flag if it out of arena range ...
   if (flags) {
      qserr rc = verifyloc(a, page, &flags);
      if (rc) return rc;
   }
   // wanted location
   if (flags==SARA_LOC) {
      r = sar_findaddr(a,page);
      if (!r || r->flags) return E_SYS_ARANGE;
      all = max = r->pages - (page - r->startpage);
   } else {
      int ii;
      // walk from the max free size to 0
      for (ii=a->flstsize-1; ii>=0; ii--)
         if (a->free[ii]) {
            mt_arange *r = a->free[ii];
            while (r) {
               u32t size;
               if (r->sign!=ARANGE_SIGN || r->flags) {
                  verifyblock(a,r,"getfree");
                  return E_SYS_SOFTFAULT;
               }
               size = r->pages;
               // filter by address range
               if (flags) {
                  u32t ep = r->startpage + size;
                  if (flags==SARA_LOCABOVE) {
                     if (ep<=page) size = 0; else
                        if (r->startpage<page && ep>page) size = ep - page;
                  } else
                  if (flags==SARA_LOCBELOW) {
                     if (r->startpage>=page) size = 0; else
                        if (r->startpage<page && ep>page) size = page - r->startpage;
                  }
               }
               // blocks can be cut by address, so check all of them for max
               if (max<size) max = size;
               all += size;
               r    = r->nextfree;
            }
         }
   }
   if (total) *total = all;
   if (smax)  *smax = max;
   return 0;
}

void _std sar_dump(mt_arena *a, printf_function pfn) {
   u8t* dumpaddr = 0;
   if (!pfn) pfn = log_printf;
   
   if (a->sign!=ARENA_SIGN) {
      pfn("Arena %08X: Broken signature!\n", a);
      dumpaddr = (u8t*)a;
   } else {
      mt_arange *rp;
      char       ln[88];

      pfn("  Arena : %08X/%08X\n", a, a->heap);
      pfn("  Range : %09LX-%09LX\n", (u64t)a->minpage<<PAGESHIFT,
         ((u64t)a->minpage+a->npages<<PAGESHIFT)-1);
      pfn("  Size  : %uMb (%X pages)\n", a->flstsize-1, a->npages);

      rp = a->first;
      while (rp) {
         char sign[9];
         u64t   a1 = ((u64t)rp->startpage<<PAGESHIFT) + ((u64t)a->minpage<<PAGESHIFT),
                a2 = a1 + ((u64t)rp->pages<<PAGESHIFT) - 1;
         int  llen;

         if (rp->sign!=ARANGE_SIGN) {
            pfn("!Broken record signature!\n");
            dumpaddr = (u8t*)rp;
            break;
         }
         if (rp->flags&ABF_USED) {
            *(ptrdiff_t*)&sign = rp->usertext;
            sign[sizeof(ptrdiff_t)] = 0;
         } else
            sign[0] = 0;

         llen = sprintf(ln, "%10LX..%10LX : %-8s ", a1, a2, sign);

         if (rp->flags & ABF_USED) {
#ifndef EFIX64
            if (rp->userdata && canread(rp->userdata)) {
               mt_prcdata *pd = (mt_prcdata*)rp->userdata;
               mt_thrdata *td = (mt_thrdata*)pd;

               if (pd->piSign==PROCINFO_SIGN)
                  sprintf(ln+llen, "pid %u", pd->piPID); else
               if (td->tiSign==THREADINFO_SIGN)
                  sprintf(ln+llen, "pid %u tid %u", td->tiPID, td->tiTID);
               else
                  sprintf(ln+llen, "%X ?", rp->userdata);
            } else
#endif
            if (!rp->userdata) strcpy(ln+llen, "global"); else
               sprintf(ln+llen, "%X", rp->userdata);
         } else
            strcpy(ln+llen, "free");

         if (rp==a->mid) strcat(ln, " mid");
         strcat(ln, "\n");
         pfn(ln);

         rp = rp->next;
      }
      if (!dumpaddr) {
         u32t ii;
         for (ii=0; ii<a->flstsize; ii++)
            if (a->free[ii]) {
               rp = a->free[ii];
               sprintf(ln, "%uMb:", ii);

               while (rp) {
                  u32t len = strlen(ln);
                  // commit the line
                  if (len>sizeof(ln)-14) {
                     pfn("%s\n", ln);
                     len = 0;
                  }
                  sprintf(ln+len, " %LX", ((u64t)rp->startpage<<PAGESHIFT) +
                     ((u64t)a->minpage<<PAGESHIFT));
                  rp = rp->nextfree;
               }
               pfn("%s\n", ln);
            }
      }
   }
   if (dumpaddr) {
      // try to step back 1 line
      if (canread(dumpaddr-16)) dumpaddr-=16;

      pfn("%4lb\n", dumpaddr);
      pfn("%4lb\n", dumpaddr+16);
      pfn("%4lb\n", dumpaddr+32);
   }
}
