//
// QSINIT
// address space management
//

#ifndef QSINIT_ADDRSPACE
#define QSINIT_ADDRSPACE

#ifdef __cplusplus
extern "C" {
#endif

// this should be the single include to be compiled both as 32/64
#include "qcl/qstheap.h"

#pragma pack(1)

#define ASPACE_SIGN    0x43505341  ///< ASPC
#define ARANGE_SIGN    0x474E5241  ///< ARNG
#define ARENA_SIGN     0x414E5241  ///< ARNA

typedef struct mt_arange_s {
   u32t                sign;       ///< header signature (ARNG)
   u32t           startpage;       ///< start address
   u32t               pages;       ///< # of pages
   u32t               flags;       ///< block flags
   struct mt_arange_s *prev,       ///< prev range
                      *next;       ///< next range
   union {
      struct mt_arange_s 
                  *prevfree;       ///< prev free block (only free ranges)
      ptrdiff_t    usertext;       ///< user`s text signature for the block
   };
   union {
      struct mt_arange_s 
                  *nextfree;       ///< next free block (only free ranges)
      void        *userdata;       ///< user data (free block has no userdata)
   };
} mt_arange;

/// @name mt_arange flags
//@{
#define ABF_USED         0x00001   ///< block is used
//@}


typedef struct mt_arena_s {
   u32t                sign;       ///< header signature (ARNA)
   u32t             minpage;       ///< start of arena in the parent`s space
   u32t              npages;       ///< # of pages in arena
   u32t            flstsize;       ///< # of entires in free[] (1 per 1Mb + 1)

   mt_arange         *first,       ///< first block
                       *mid,       ///< block with npages/2 addr
                     **free;       ///< free lists (mt_arange*[flstsize])
   qs_thp              heap;       ///< heap handle
   void               *pdir;       ///< page directory (if present)
   void              *owner;       ///< ...
} mt_arena;

/** create the arena.
    Start address and number of pages must be aligned to 4Mb.

    @return error code or 0. */
qserr _std sar_create (u64t minaddr, u32t pages, u32t aflags, mt_arena **res);

/// delete arena
qserr _std sar_delete (mt_arena *a);

/// @name sar_alloc() flags
//@{
#define SARA_LOC         0x00001   ///< allocate at specified page
#define SARA_LOCABOVE    0x00002   ///< allocate at >= page
#define SARA_LOCBELOW    0x00003   ///< allocate at < page
#define SARA_LOCMASK     0x00007
#define SARA_ADDRMASK    0x00018
#define SARA_ADDRMIN     0x00008   ///< allocate with min possible address
#define SARA_ADDRMAX     0x00010   ///< allocate with max possible address
//@}

/** allocate an address range.
    Both SARA_LOCABOVE and SARA_LOCBELOW can be combined with SARA_ADDRMIN or
    SARA_ADDRMAX.

    @param          a       arena
    @param [in,out] page    start (wanted) page (offset from the start of arena)
    @param          pages   # of pages to allocate
    @param          flags   SARA_*
    @param          user    user data pointer
    @param          sign    text signature for the log (4 chars)
    @return error code or 0 */
qserr _std sar_alloc (mt_arena *a, u32t *page, u32t pages, u32t flags,
                      void *user, const char *sign);

/** update flags of the address range (or free it).
    @param  a       arena
    @param  page    start page of the allocated range (offset in arena)
    @param  flags   new flags value (SARA_*) or 0 to free the range
    @param  user    new user data pointer (flags==0 - ignored, else
                    not changed if user==0)
    @return error code or 0 */
qserr _std sar_update (mt_arena *a, u32t page, u32t flags, void *user);

#define sar_free(a,page) sar_update(a,page,0,0)

/** free by "user" value.
    @param  a       arena
    @param  user    user data pointer to search, CANNOT be 0.
    @return error code, E_SYS_NOTFOUND if none such blocks found or 0 if at
            least one block was freed */
qserr _std sar_ufree  (mt_arena *a, void *user);

/** query address range data.
    @param          a       arena
    @param [in,out] page    in - page to query (offset in arena), out - start
                            of this range of pages
    @param [out]    pages   # of pages in this range (ptr can be 0)
    @param [out]    flags   flags (ptr can be 0)
    @param [out]    user    user data pointer, (ptr can be 0)
    @return error code or 0 */
qserr _std sar_query  (mt_arena *a, u32t *page, u32t *pages, u32t *flags,
                       void **user);

/** query free space.
    Note that SARA_LOC, SARA_LOCABOVE and SARA_LOCBELOW will filter returned
    "total" value (for SARA_LOC it will be equal to smax - length of a whole
    block available at this address)

    @param          a       arena
    @param [out]    total   # of available pages (ptr can be 0)
    @param [out]    smax    the longest available block size (in pages, ptr can be 0)
    @param          flags   SARA_LOC, SARA_LOCABOVE and SARA_LOCBELOW accepted here
    @param          page    wanted page (offset from the start of arena, only with flags)
    @return error code or 0 */
qserr _std sar_getfree(mt_arena *a, u32t *total, u32t *smax, u32t flags, u32t page);

void  _std sar_dump   (mt_arena *a, printf_function pfn);

#pragma pack()

#ifdef __cplusplus
}
#endif

#endif // QSINIT_ADDRSPACE
