//
// QSINIT
// memory management for physical memory above 4Gb ....
//
#include "syslocal.h"
#include "meminit.h"
#include "arena.h"
#include "efnlist.h"
#include "qecall.h"

extern TSysMemList  pcmem;  ///< the original memory list, must be r/o
static TSysMemList    mem;  ///< local copy ...

/// high memory arena (initialized only when memory is really available)
static mt_arena      *hma = 0;
static u64t        hibase = 0;
extern u32t       ramfs64;

#define LOWBORDER  _4GBLL

static void _std efiexit_cb(sys_eventinfo *info) {
   // free memory only if reason is "exit to efi code"
   if (info->info==SHUT_EFICODE) {
      log_it(2, "releasing high memory ...\n");

      for (u32t ii=0; ii<mem.Count(); ii++) {
         mem_entry &me = mem[ii];
         if (me.start>=LOWBORDER && me.owner==PCMEM_FREE) {
            u64t res = call64l(EFN_FREEPAGES, 1, 2, me.start, me.res[0]);
            if (res)
               log_it(2, "Error %LX in free address %LX (%X pages)\n", res,
                  me.start, me.res[0]);
         }
      }
   }
}

// assumed to be called in MT lock or before MT mode
qserr hlp_meminit64(void) {
   u32t     ii;
   if (hma) return E_SYS_INITED;
   // use actual UEFI memory map ...
   if (hlp_hosttype()==QSHT_EFI) {
      int once = 0;
      setup_getmem(mem);
      /* GET memory from the firmware ...
         only blocks >=1Mb;
         all entries will be freed in exit callback above */
      for (ii=0; ii<mem.Count(); ii++) {
         mem_entry &me = mem[ii];
         if (me.start>=LOWBORDER && me.owner==PCMEM_FREE) {
            if (me.len>=_1MB) {
               // who knows!?
               if (me.start & PAGEMASK) {
                  me.len  -= PAGESIZE - (me.start & PAGEMASK);
                  me.start = PAGEROUNDLL(me.start);
               }
               me.res[0] = me.len - (me.len&PAGEMASK) >> PAGESHIFT; // # of pages

               u64t addr = me.start,
                     res = call64l(EFN_GETPAGES, 0, 3, PGALLOCEX_ADDR, &addr,
                                   me.res[0]);
               if (res) {
                  log_it(2, "Error %LX reserving address %LX (%X pages)\n", res,
                     addr, me.res[0]);
                  me.owner = PCMEM_RESERVED;
               } else
                  once++;
            } else
               me.owner = PCMEM_RESERVED;
         }
      }
      if (once) 
         sys_notifyevent(SECB_QSEXIT|SECB_GLOBAL, efiexit_cb);
   } else
      mem = pcmem;

   u64t start = 0,
          end = 0;
   /* everything can be unaligned here - in theory ...
      so, the code below just aligns it all - arena size to 4Mb and block
      sizes to 4kb */
   for (ii=0; ii<mem.Count(); ii++) {
      u64t bend = mem[ii].start + mem[ii].len;

      if (!end && mem[ii].start>=LOWBORDER) {
         start = mem[ii].start & ~(u64t)(_4MB-1);
         end   = bend;
      } else
      if (end && end<bend) end = bend;
   }
   end = end+_4MB-1 & ~(u64t)(_4MB-1);

   if (!end) {
      mem.Clear();
      return E_SYS_NOMEM;
   }
   log_printf("init high arena ... %LX..%LX\n", start, end-1);

   mt_arena  *a = 0;
   u32t  npages = end - start >> PAGESHIFT, page;
   qserr    res = sar_create(start, npages, 0, &a);

   if (!res) {
      // list is sorted, so alloc all holes between entries as a global memory ...
      u64t  pos = start, bs;

      for (ii=0; ii<mem.Count(); ii++)
         if (mem[ii].start>=pos && mem[ii].owner==PCMEM_FREE) {
            bs = PAGEROUNDLL(mem[ii].start);
            if (bs>pos) {
               page = pos - start >> PAGESHIFT;
               res  = sar_alloc(a, &page, bs-pos>>PAGESHIFT, SARA_LOC, 0, "resv");
               if (res) break;
            }
            pos = mem[ii].start + mem[ii].len & ~(u64t)PAGEMASK;
            // mark it as owned ...
            sys_markmem(bs, pos-bs>>PAGESHIFT, PCMEM_QSINIT);
         }
      if (!res && pos<end) {
         page = pos - start >> PAGESHIFT;
         res  = sar_alloc(a, &page, end-pos>>PAGESHIFT, SARA_LOC, 0, "resv");
      }
         
      if (res) {
         log_it(2, "Error %X adding addr %LX!\n", res, bs);
         pcmem_print(0, mem);
      }
   }

   if (!res) {
      hibase = start;
      hma    = a;
      hlp_memprint64(0);
      // allow high memory for drive b:
      if (!hlp_insafemode()) ramfs64 = 1;
   } else {
      // free arena!!
      sar_delete(a);
   }
   return res;
}

/// verify the requsted location and fill arena flags (OR them actually)
static qserr check_wanted(u64t &wanted, u64t &size, u32t flags, u32t &aflags) {
   switch (flags&QSMA_LOCEXACT) {
      case QSMA_LOCABOVE:
         if (wanted>hibase) aflags|=SARA_LOCABOVE; else wanted = 0;
         break;
      case QSMA_LOCBELOW:
         if (wanted<=hibase) return E_SYS_ARANGE;
         aflags|= SARA_LOCBELOW;
         break;
      case QSMA_LOCEXACT:
         if (wanted<hibase) return E_SYS_ARANGE;
         if (wanted&PAGEMASK) {
            u32t diff = wanted&PAGEMASK;
            size     += diff;
            wanted   -= diff;
         }
         aflags|= SARA_LOC;
         break;
   }
   return 0;
}

u64t  _std hlp_memalloc64(u64t size, const char *sign, u32t flags,
                          u64t wanted, qserr *error)
{
   u64t blk = 0;
   if (error) *error = E_SYS_NOMEM;
   // above 4Gb
   if (flags & QSMA_HIGH) {
      if (!hma) return 0;
      mt_prcdata *owner = 0;
      u32t         page, npages;
      u32t           fl = flags&QSMA_MAXADDR ? SARA_ADDRMAX :
                          flags&QSMA_MINADDR ? SARA_ADDRMIN : 0;
      qserr          rc = 0;
      // wanted, size and fl can be updated here ...
      if (flags&QSMA_LOCEXACT) rc = check_wanted(wanted, size, flags, fl);
      if (!rc) {
         page   = fl&SARA_LOCMASK ? wanted-hibase>>PAGESHIFT : 0;
         npages = PAGEROUNDLL(size) >> PAGESHIFT;
         if (!npages) rc = E_SYS_INVPARM;
      }
      if (!rc) {
         // lock only mt_getdata() and arena op
         mt_swunlock();
         if (flags&QSMA_LOCAL) mt_getdata(&owner, 0);

         rc = sar_alloc(hma, &page, npages, fl, owner, sign);
         mt_swunlock();
      }
      if (rc) {
         if (error) *error = rc;
         return 0;
      }
      blk = hibase + ((u64t)page<<PAGESHIFT);

      if ((flags&QSMA_NOCLEAR)==0)
         if (hlp_hosttype()==QSHT_EFI)
            sys_memhiset(blk, 0, (u64t)npages<<PAGESHIFT-3, 8);
         else
            sys_memhiset(blk, 0, (u64t)npages<<PAGESHIFT-2, 4);
   } else {
      if (size>FFFF) return 0;
      // deny QSMA_LOCABOVE/BELOW (need to rewrite allocator)
      if ((flags & QSMA_LOCEXACT)==0)
         blk = (u32t)hlp_memallocsig(size, sign, flags|QSMA_PHYS|QSMA_RETERR);
            else
      if ((flags&QSMA_LOCEXACT)!=QSMA_LOCEXACT || wanted>FFFF) return 0; else {
         u32t reason;
         blk = (u32t)hlp_memreserve((u32t)wanted, size, &reason);
         /* hlp_memreserve() used as a regular alloc here, so deny truncated
            block and zero-fill the result */
         if (blk && reason==QSMR_TRUNCATED) {
             hlp_memfreephys(blk);
             blk = 0;
         }
         if (blk && (flags&QSMA_NOCLEAR)==0) {
            u32t sz = hlp_memgetsize((void*)(u32t)blk);
            memset((void*)(u32t)blk, 0, sz);
         }
         if (!blk && error) *error = E_SYS_ARANGE;
      }
   }
   if (blk && error) *error = 0;
   return blk;
}


qserr _std hlp_memavail64(u64t *maxblock, u64t *avail, u64t *total, u32t flags,
                          u64t addr)
{
   if (maxblock) *maxblock = 0;
   if (avail) *avail = 0;
   if (total) *total = 0;

   if ((flags&QSMA_HIGH)==0) {
      u32t asz;
      /* fix me! this need a separate call to browse base mem tables
         to calc it */
      if (flags&QSMA_LOCEXACT) return E_SYS_UNSUPPORTED;

      asz = hlp_memavail((u32t*)maxblock, (u32t*)total, QSMA_PHYS);
      if (avail) *avail = asz;
      return 0;
   }
   if (!hma) return E_SYS_NOMEM;

   u32t  afl = 0;
   qserr  rc = 0;
   // convert flags & verify wanted address
   if (flags&QSMA_LOCEXACT) {
      u64t wsize = 1;
      rc = check_wanted(addr, wsize, flags, afl);
   }
   if (!rc) {
      mt_swlock();
      rc = sar_getfree(hma, (u32t*)avail, (u32t*)maxblock, afl,
                       (u32t)(addr?addr-hibase>>PAGESHIFT:0));
      mt_swunlock();
   }
   if (!rc) {
      if (total) *total = (u64t)hma->npages<<PAGESHIFT;
      if (avail) *avail<<=PAGESHIFT;
      if (maxblock) *maxblock<<=PAGESHIFT;
   }
   return rc;
}


qserr _std hlp_freephys64(u64t addr) {
   // should never occurs
   if (addr<_4GBLL) return E_SYS_SOFTFAULT;
   if (!hma || addr<hibase) return E_SYS_INVPARM;
   u64t page = addr - hibase >> PAGESHIFT;
   if (page>FFFF) return E_SYS_INVPARM;

   mt_swlock();
   qserr rc = sar_free(hma,(u32t)page);
   mt_swunlock();

   return rc;
}

/// query physical memory block size by an address
u64t _std hlp_memgetsize64(u64t addr) {
   if (addr<_4GBLL) return hlp_memgetsize((void*)(u32t)addr);
   if (!hma || addr<hibase) return E_SYS_INVPARM;
   u64t   page = addr - hibase >> PAGESHIFT;
   if (page>FFFF) return E_SYS_INVPARM;
   u32t inpage = (u32t)page,
         pages = 0,
         flags = 0;

   mt_swlock();
   qserr   rc = sar_query(hma, &inpage, &pages, &flags, 0);
   mt_swunlock();
   /* there is no panic here on a wrong address like in hlp_memgetsize(), but
      at least a log message */
   if (!rc && inpage!=page)
      log_it(2, "Wrong address in hlp_memgetsize64(%LX)!\n", addr);

   if (rc || !flags || inpage!=page) return 0;
   return (u64t)pages<<PAGESHIFT;
}


void mem64_pid_done(mt_prcdata *pd) {
   if (hma && pd) {
      mt_swlock();
      qserr  rc = sar_ufree(hma, pd);
      mt_swunlock();
      if (rc && rc!=E_SYS_NOTFOUND) 
         log_it(2, "PID %u high mem free error %X!\n", pd->piPID, rc);
   }
}

void _std hlp_memprint64(printf_function pfn) {
   if (hma) {
      if (!pfn) pfn = log_printf;
      pfn("<====High memory dump====>\n");
      mt_swlock();
      sar_dump(hma,pfn);
      mt_swunlock();
   }
}
