//
// QSINIT
// fill memory info
//
#include "qsint.h"
#include "qsutil.h"
#include "qsstor.h"
#include "qslog.h"
#include "meminit.h"
#include "qspage.h"
#include "qsmodext.h"
#include "syslocal.h"
#include "pagedesc.h"
#include "qspdata.h"
#include "qssys.h"
#include "qsinit.h"
#include "qsxcpt.h"
#include "errno.h"

#define LASTPAGE   0xFFFFF000    // limit phys maps to this page

typedef u32t (_std *pfmemisfree)(u32t physaddr, u32t length, u32t reserve);

typedef struct {
   u64t      start;
   u64t        len;
   u32t       type;
} acpi_entry;

static physmem_block  *physmem = 0; // exported physmem array
u32t                  maxmap4g = 0; // top address to be mapped in 1st 4GBs
static u32t           pages_lo = 0, // total free pages below 4Gb
                      pages_hi = 0; // total free pages above 4Gb
static u64t             hi_min = 0,
                        hi_max = 0;
extern "C" {
extern mt_proc_cb mt_exechooks;     // MT service hooks
#pragma aux mt_exechooks  "_*";
}

#define VMBASE   ((u32t)-INIT_SYS_AREA)  ///< base address of VM area

#define IS_FLAT_MODE() \
   (mt_exechooks.mtcb_pgmode==QPGM_FLAT || mt_exechooks.mtcb_pgmode==QPGM_EFI)

static void setup_memlist(void);
static void hlp_pgprint(printf_function pfn);
static u32t _std memisfree(u32t physaddr, u32t length, u32t reserve);

// can use static class, because List init doesn`t call to any heap alloc
TSysMemList              pcmem;
static TSysPGAllocList    memx;

void pcmem_print(printf_function pfn, TSysMemList &ml) {
   if (!pfn) pfn = log_printf;
   // lock pager mutex!
   PAGEMUX_THIS_FUNC lk;

   for (u32t ii=0; ii<ml.Count(); ii++) {
      u64t start = ml[ii].start,
             len = ml[ii].len;
      pfn("%2d. %010LX-%010LX, size: %8u kb, flags %02X %04X\n", ii+1, start,
         start+len-1, (u32t)(len>>10), ml[ii].flags, ml[ii].owner);
   }
}

void _std hlp_memcprint(printf_function pfn) {
   if (!pfn) pfn = log_printf;
   pfn("<=====PC memory dump=====>\n");
   pcmem_print(pfn,pcmem);
   // print page allocator table
   hlp_pgprint(pfn);
}

// sort callback
static int __stdcall mem_addr_compare(const void *b1, const void *b2) {
   register mem_entry *m1 = (mem_entry*)b1,
                      *m2 = (mem_entry*)b2;
   if (m1->start > m2->start) return 1;
   if (m1->start == m2->start) return 0;
   return -1;
}

void setup_getmem(TSysMemList &ml) {
   u32t    ii, acnt;
   // parse full E820 list
   AcpiMemInfo *tbl = hlp_int15mem();

   acnt = 0;
   while (tbl[acnt].LengthLow || tbl[acnt].LengthHigh) acnt++;
   if (acnt) {
      u32t        mcnt = 0;
      // allocate array (one for possible 4Gb border-cross and one for 1st Mb)
      ml.SetCount(acnt+2);
      // 640 kb
      ml[0].start = 0;
      ml[0].len   = (u32t)int12mem()<<10;
      ml[0].flags = MEM_DIRECT;
      ml[0].owner = PCMEM_RESERVED;
      mcnt = 1;
      // ACPI list
      for (ii=0; ii<acnt; ii++) {
         acpi_entry *ci = (acpi_entry*)(tbl + ii);
         mem_entry  &me = ml[mcnt];
         if (ci->start<_1MB) continue;   // below 1Mb

         if (ci->len>=_4KB) {
            // high bits are now used for type, so trunc it here
            u8t type = ci->type;
            me.start = ci->start;
            me.len   = ci->len;
            if (!me.len) continue;

            me.flags = 0;
            me.owner = type==1 ? PCMEM_FREE : PCMEM_HIDE|PCMEM_RESERVED;
            mcnt++;
            // split hypotetic block on the 4Gb border (never see the one)
            if (me.start<=FFFF && me.start+me.len>_4GBLL) {
               mem_entry &nme = ml[mcnt];
               ml[mcnt].start = _4GBLL;
               ml[mcnt].len   = nme.start + nme.len - _4GBLL;
               nme.len           = _4GBLL - nme.start;
               mcnt++;
            }
         }
      }
      ml.SetCount(mcnt);
   } else {
      // no E820, get physmem from initial parsing (here must be int15 88/E8)
      physmem_block *phm = (physmem_block*)sto_data(STOKEY_PHYSMEM);
      u32t     phm_count = sto_size(STOKEY_PHYSMEM)/sizeof(physmem_block);

      ml.SetCount(phm_count);
      for (ii=0; ii<phm_count; ii++) {
         ml[ii].start = phm[ii].startaddr;
         ml[ii].len   = phm[ii].blocklen;
         ml[ii].flags = 0;
         ml[ii].owner = PCMEM_FREE;
      }
   }
   free(tbl);
   // sort by address
   qsort((void*)ml.Value(), ml.Count(), sizeof(mem_entry), mem_addr_compare);
}

extern "C"
void setup_memory(void) {
   u32t  availmem, ii,
          membase = sto_dword(STOKEY_BASEMEM);
   /* query range, used by common memory manager: membase..membase+availmem-1,
      hlp_memavail makes an additional check of memory structs */
   hlp_memavail(0, &availmem, QSMA_PHYS);

   setup_getmem(pcmem);

   // print it sorted, but unaligned
   log_printf("<=====Original table=====>\n");
   pcmem_print(0,pcmem);
   // align free entries to 64k
   for (ii = 0; ii<pcmem.Count(); ) {
      mem_entry  &me = pcmem[ii];
   
      if (me.start>=_1MB)
         if (me.owner==PCMEM_FREE) {
            /* align free blocks down to 64k and delete all smaller than 64k.
               this alignment should restore a match between free blocks and
               memory holes, reseved in memmgr (and aligned UP to 64k) */
            if (me.len<_64KB) { pcmem.Delete(ii); continue; }
            if (me.start & 0xFFFF) {
               u32t diff = 0x10000 - (me.start & 0xFFFF);
               me.start += diff;
               me.len   -= diff;
            }
            if (me.len & 0xFFFF) me.len &= ~(u64t)0xFFFF;
            if (me.len<_64KB) { pcmem.Delete(ii); continue; }
         }
      ii++;
   }
   hi_min = 0;
   hi_max = 0;
   // search for lowest border of usable memory in first 4Gbs
   maxmap4g = _1MB;
   for (ii = 0; ii<pcmem.Count(); ii++) {
      mem_entry  &me = pcmem[ii];
      if (me.owner==PCMEM_FREE)
         if (me.start<_4GBLL) {
            u32t eaddr = me.start + me.len;
            if (eaddr > maxmap4g) maxmap4g = eaddr;
         
            if (me.start>=_1MB) pages_lo += me.len>>PAGESHIFT;
         } else {
            if (!hi_min || hi_min>me.start) hi_min = me.start;
            if (!hi_max || hi_max<me.start+me.len) hi_max = me.start+me.len;
            pages_hi += me.len>>PAGESHIFT;
         }
   }
   /* mark blocks:
      * 1st Mb            - DIRECT
      * QSINIT mem        - DIRECT + QSINIT
      * other free in 4Gb - DIRECT              */
   for (ii=0; ii<pcmem.Count(); ii++) {
      u64t start = pcmem[ii].start;
      if (start<_1MB) pcmem[ii].flags |= MEM_DIRECT;
         else
      if (start<=membase && start+pcmem[ii].len>=membase+availmem) {
         pcmem[ii].flags|= MEM_DIRECT;
         pcmem[ii].owner = PCMEM_QSINIT;
      } else
      if (start<maxmap4g && pcmem[ii].owner==PCMEM_FREE)
         pcmem[ii].flags |= MEM_DIRECT;
   }
   // dump to log
   if (hi_min)
      log_it(2, "phys top: %08X, hi top: %09LX\n", maxmap4g, hi_max);
   else
      log_it(2, "phys top: %08X\n", maxmap4g);
   // update physmem to actual values
   hlp_setupmem(0, 0, 0, SETM_SPLIT16M);
   /* leave pcmem array for multiple hlp_setupmem() calls (it contain pure
      PC memory list) and make a copy of it for page allocator */
   setup_memlist();
   /* print it to log (except safe mode, where it will use too many lines
      on screen) */
   if (!hlp_insafemode()) hlp_memcprint(0);

   if (hlp_hosttype()==QSHT_EFI) {
      u64t ptlen = sys_efiptlen();
      log_it(2, "pt ends at: %LX\n", ptlen);
   }
}

pcmem_entry* _std sys_getpcmem(int freeonly) {
   PAGEMUX_THIS_FUNC lk;

   if (!pcmem.Count()) return 0;
   pcmem_entry *rc = (pcmem_entry*)malloc_local((pcmem.Count()+1)*sizeof(pcmem_entry));
   u32t ii=0, ti=0;
   for (; ii<pcmem.Count(); ii++) {
      u64t start = pcmem[ii].start;
      u32t flags = pcmem[ii].owner;

      if (start>=_1MB && (!freeonly || (flags&PCMEM_TYPEMASK)==PCMEM_FREE)) {
         rc[ti].start = start;
         rc[ti].pages = pcmem[ii].len>>PAGESHIFT;
         rc[ti].flags = flags;
         ti++;
      }
   }
   rc[ti].start = 0;
   rc[ti].pages = 0;
   return rc;
}

u32t _std sys_ramtotal(u32t *above4gb) {
   if (above4gb) *above4gb = pages_hi;
   return pages_lo;
}

u32t _std sys_endofram(void) {
   return maxmap4g;
}

int _std sys_gethirange(u64t *minaddr, u64t *maxaddr) {
   if (minaddr) *minaddr = hi_min;
   if (maxaddr) *maxaddr = hi_max;
   return hi_min?1:0;
}

// insert block into the right place
static void insert_block_a(mem_entry &nb) {
   for (u32t ii=0; ii<pcmem.Count(); ii++)
      if (pcmem[ii].start>nb.start) {
         pcmem.Insert(ii,nb);
         return;
      }
   pcmem.Add(nb);
}

#define IS_INCLUDED()  (start<=address && epos>start && epos<=end)

#define IS_INCLUDE()   (start>=address && epos>start && epos>=end)

#define IS_INTERSECT() (start>address && start<epos && end>epos || \
                        start<address && end>address && end<epos)

// search blocks, interferenced with addr range
static u32t find_block(u64t address, u32t pages, u32t &bfirst, u32t &blast) {
   u32t   rc = 0, ii;
   u64t epos = address + ((u64t)pages<<PAGESHIFT) - 1;
   bfirst    = 0;
   blast     = 0;
   for (ii=0; ii<pcmem.Count(); ii++) {
      u64t start = pcmem[ii].start,
             end = pcmem[ii].start + pcmem[ii].len;
      // search for first..last touched blocks
      if (IS_INCLUDED() || IS_INCLUDE() || IS_INTERSECT()) {
         if (!bfirst) bfirst = ii;
         blast = ii;
         rc++;
      } else
      if (rc) break;
   }
   return rc;
}

static u32t _std sys_markprocess(u64t address, u32t pages, u32t owner) {
   u32t bf, bl, ii, cnt;
   if ((address&PAGEMASK) || !pages) return EINVAL;
   if ((owner & ~(PCMEM_HIDE|PCMEM_TYPEMASK))!=0) return EINVAL;
   // lock pager mutex!
   PAGEMUX_THIS_FUNC lk;

   cnt = find_block(address, pages, bf, bl);
   if (!cnt) return ENOENT;
#if 0
   hlp_memcprint(0);
   log_printf("cnt=%d, bf=%d, bl=%d\n", cnt, bf, bl);
#endif
   // check - all requested blocks have compatible type?
   for (ii=bf; ii<=bl; ii++) {
      u32t n_ow = pcmem[ii].owner&PCMEM_TYPEMASK,
         set_ow = owner&PCMEM_TYPEMASK;

      if (set_ow==PCMEM_FREE) {
         if (n_ow<PCMEM_RAMDISK) return EACCES;
      } else {
         // not a free block and not changing of PCMEM_HIDE flag
         if (!(set_ow==n_ow && (owner^PCMEM_HIDE)==pcmem[ii].owner) &&
            n_ow!=PCMEM_FREE) return EACCES;
      }
   }
   for (ii=bf; ii<=bl; ii++) {
      int  included = ii>bf && ii<bl,
             theone = ii==bf && ii==bl;
      u32t oneowner = pcmem[ii].owner;
      if (ii==bf)   // block can be splitted on start addr
         if (address>pcmem[ii].start) {
            mem_entry sb;
            sb.start = address;
            sb.len   = pcmem[ii].len - (address - pcmem[ii].start);
            sb.flags = pcmem[ii].flags;
            sb.owner = owner;

            pcmem[ii].len = address - pcmem[ii].start;
            insert_block_a(sb);
            bl++; ii++;
         } else
            included = 1;
      if (ii==bl) {   // block can be splitted at the end addr
         u64t bend = pcmem[ii].start+pcmem[ii].len,
             rbend = address+((u64t)pages<<PAGESHIFT);

         if (rbend<bend) {
            mem_entry sb;
            sb.start = rbend;
            sb.len   = bend - rbend;
            sb.flags = pcmem[ii].flags;
            sb.owner = theone ? oneowner : pcmem[ii].owner;

            pcmem[ii].len   = sb.start - pcmem[ii].start;
            pcmem[ii].owner = owner;
            insert_block_a(sb);
            included = 0;
         } else
         if (!included) included = !theone;
      }
      if (included) // block included into requested range
         pcmem[ii].owner = owner & (PCMEM_HIDE|PCMEM_TYPEMASK);
   }
   // merge blocks with the same owner/flags
   ii = 0;
   while (ii+1<pcmem.Count())
      if (pcmem[ii].owner==pcmem[ii+1].owner && pcmem[ii].start+pcmem[ii].len==
         pcmem[ii+1].start && pcmem[ii].flags==pcmem[ii+1].flags)
      {
         pcmem[ii].len+=pcmem[ii+1].len;
         pcmem.Delete(ii+1);
      } else ii++;
#if 1
   hlp_memcprint(0);
#endif
   return 0;
}

u32t _std sys_markmem(u64t address, u32t pages, u32t owner) {
   return sys_markprocess(address, pages, owner);
}

u32t _std sys_unmarkmem(u64t address, u32t pages) {
   return sys_markprocess(address, pages, PCMEM_FREE);
}

u32t _std hlp_setupmem(u32t fakemem, u32t *logsize, u32t *removemem, u32t flags) {
   PAGEMUX_THIS_FUNC lk;   // lock entire func

   u32t lphysaddr = 0, ii, jj, icnt = pcmem.Count(), ccount, csize,
         lactsize = logsize? *logsize : 0;
   int    pdalign = flags & SETM_NOPDALIGN ? 0 : 1;
   // search for first block above 4Gb
   for (ii=0; ii<pcmem.Count(); ii++)
      if (pcmem[ii].start>=_4GBLL) { icnt = ii; break; }
   // update physmem pointer
   if (physmem) free(physmem);
   physmem = (physmem_block*)malloc(sizeof(physmem_block)*(icnt+1));
   register physmem_block *pm = physmem;
   // MEMLIMIT in 64k blocks
   if (fakemem) fakemem = (fakemem<16?0:fakemem-1) << 4;
   // fill physmem array
   for (ii=0, ccount=0, csize=0; ii<icnt; ii++) {
      // block, hidden from OS/2
      if ((pcmem[ii].owner&PCMEM_HIDE)!=0) continue;
      // remove manually specified blocks
      if (removemem) {
         for (jj=0; removemem[jj]; jj++)
            if (removemem[jj]==pcmem[ii].start) break;
         if (removemem[jj]) continue;
      }
      // skip block < one page dir entry
      if (pdalign && pcmem[ii].len<_4MB) continue;

      u32t nbstart = pcmem[ii].start,
             nblen = pcmem[ii].len;
      // check hole size
      if (pdalign && ccount) {
         physmem_block *cb = pm + ccount;
         u32t         pend = cb[-1].startaddr + cb[-1].blocklen;
         // hole on the same page dir entry - truncate both blocks
         if ((nbstart & PD4M_ADDRMASK) == (pend & PD4M_ADDRMASK)) {
            u32t  diff1 = pend & PD4M_MASK,
                  diff2 = PD4M_MASK + 1 - (nbstart & PD4M_MASK);

            if (diff1)
               if (diff1 < cb[-1].blocklen) cb[-1].blocklen -= diff1;
            if (diff2)
               if (diff2 >= nblen) continue; else {
                  nblen   -= diff2;
                  nbstart += diff2;
               }
         }
      }
      // copy block
      pm[ccount].startaddr = nbstart;
      pm[ccount].blocklen  = nblen;
      // MEMLIMIT: only blocks above 1st Mb have used in calc
      if (pm[ccount].startaddr >= _1MB) {
         u32t blocklen = pm[ccount].blocklen >> 16;
         // truncate block to fit MEMLIMIT
         if (fakemem) {
            if (csize + blocklen > fakemem) {
               blocklen = fakemem - csize;
               if (!blocklen) break;
               pm[ccount].blocklen = blocklen << 16;
               ii = icnt; // break cycle!
            }
         }
         // update total size in 64k blocks
         csize += blocklen;
      }
      // merge bounded blocks
      if (ccount && pm[ccount].startaddr==pm[ccount-1].startaddr+
         pm[ccount-1].blocklen)
      {
         pm[ccount-1].blocklen += pm[ccount].blocklen;
      } else
         ccount++;
   }
   // fix too large log size (leave at least 15Mb for the system)
   if (lactsize && lactsize + (15 << 4) > csize)
      lactsize = csize <= (15 << 4) ? 0 : csize - (15 << 4);
   // split block on 16M border (we always must fit to icnt+1)
   if (flags&SETM_SPLIT16M) {
      ii = 0;
      while (ii<ccount) {
         register physmem_block *ppm = pm+ii;
         if (ppm->startaddr<_16MB && ppm->startaddr+ppm->blocklen>_16MB) {
            memmove(ppm+1, ppm, (ccount-ii)*sizeof(physmem_block));
            ppm[1].startaddr = _16MB;
            ppm[1].blocklen  = ppm->startaddr + ppm->blocklen - _16MB;
            ppm->blocklen    = _16MB - ppm->startaddr;
            ccount++;
            break;
         } else ii++;
      }
   }
   // search for suitable place to place log buffer
   if (lactsize) {
      ii = ccount-1;
      while (ii > 0) { // we can skip 1st entry, because it contain 1Meg block
         physmem_block *pb = pm+ii--;
         u32t len = pb->blocklen >> 16;
         if (len >= lactsize) {
            pb->blocklen = len - lactsize << 16;
            lphysaddr = pb->startaddr + pb->blocklen;
            *logsize  = lactsize;
            break;
         }
      }
      if (!lphysaddr) *logsize = 0;
   }
   // update storage key
   sto_save(STOKEY_PHYSMEM, physmem, ccount * sizeof(physmem_block), 0);

   return lphysaddr;
}

// -----------------------------------------------------------------------
// -----------------------------------------------------------------------
void setup_mem64(void) {
//   if (hi_min && hi_min<hi_max) {
   if (1) {
      qserr err = hlp_meminit64();
      if (err) log_printf("meminit64() error %X\n", err);
   }
}

void pag_inittables(int vmmode) {
   u32t pgm = sys_pagemode();
   // a small assert
   if (Xor(vmmode,pgm==QPGM_VPAE||pgm==QPGM_V32))
      _throwmsg_("wrong inittables call!");

   setup_mem64();

   if (!vmmode) {
      // map 1 Mb
      pag_mappages(0, _1MB, 0, MEM_DIRECT);

      for (u32t ii=0; ii<memx.Count(); ii++) {
         u32t start = memx[ii].start,
                len = memx[ii].len,
                 fl = memx[ii].flags;
         if (start>=_1MB && (!vmmode || (fl&MEM_MAPPED)))
            pag_mappages(start, len, fl&MEM_MAPPED ? memx[ii].paddr : start, fl);
      }
   } else {
      _throwmsg_("not finished!");
      /* round end down to 1/2 Mb
         unmap all big pages between end and VMBASE
         map unaligned end via pag_mappages
      */

      // MEM_MAPPED should not exist in this case!
   }
}

static void setup_memlist(void) {
   if (memx.Count()) return;

   u32t pgm = sys_pagemode(),
        top = maxmap4g;
   // page mode already set?
   int pgon = pgm==QPGM_VPAE || pgm==QPGM_V32 ? 1 : 0;
   if (pgon && top>VMBASE) top = VMBASE;

   /* the entire area should be accessible (as expected by all code).
      Some "BIOS table" pointers can use "reserved" blocks, this causes
      a page fault on access.
      Probably "reserved" areas should be mapped as r/o, but they are writable
      in non-paged mode anyway, so there is no special reason for this. */
   memx.SetCount(1+pgon);
   memx[0].start = _1MB;
   memx[0].len   = top - _1MB;
   memx[0].flags = MEM_DIRECT|MEM_PHYSEND;
   memx[0].usage = 1;
   memx[0].paddr = 0;

   if (pgon) {
      memx[1].start = VMBASE;
      memx[1].len   = INIT_SYS_AREA;
      memx[1].flags = MEM_RESERVED;
      memx[1].usage = x7FFF;
      memx[1].paddr = FFFF;

      pag_inittables(1);
      pag_setpage0();
   } else
   if (pgm==QPGM_EFI) setup_mem64();
}

/// find block index in memx array
static int pag_findidx(u32t physaddr, u32t len) {
   if (!memx.Count()) return -1;
   // len with overflow?
   if (physaddr+len < physaddr) return -1;
   // searching block
   for (u32t ii=0; ii<memx.Count(); ii++)
      if (memx[ii].start<=physaddr && memx[ii].start+memx[ii].len>=
         physaddr+len) return ii;
   return -1;
}

#if 0
int pag_findblock(void *address, u32t flags, vmem_entry *rc) {
   int idx = pag_findidx(address, 1, flags);
   if (rc && idx>=0) *rc = memx[idx];
   return idx<0 ? 0 : 1;
}
#endif

/* searching for intersection with one or more existing map entries.
   We must unite it, else page dir creation code will be confused -
   it assume, what mapped entries, coming from non-paged mode not intersect
   each other.
   Check intersection with allocating block only, because NO intersection
   before allocation, so all of it we produce by new address. */
static void merge_nonpaged(u32t address, u32t size) {
   u32t icnt, ii;
   do {
      icnt = memx.Count();
      ii   = 0;
      while (ii<memx.Count()) {
         if ((memx[ii].flags&MEM_MAPPED)!=0) {
            u32t pa = memx[ii].start,
                 pe = pa + memx[ii].len,
               aend = address + size;
            // asked address include entire mapped entry
            if (pa>=address && pa<aend && pe<=aend) {
               memx[ii].usage++;
               memx[ii].start = address;
               memx[ii].len   = size;
               memx[ii].paddr = address;
               // exit inner cycle
               ii = FFFF; continue;
            } else
            // not included and not include and not equal - only intersection!
            if (pa>address && pa<aend && pe>aend ||
                pa<address && pe>address && pe<aend)
            {
               memx[ii].usage++;
               if (pa>address) memx[ii].start = address;
               if (aend>pe) memx[ii].len = aend - memx[ii].start;
               memx[ii].paddr = memx[ii].start;
               ii = FFFF; continue;
            }
         }
         ii++;
      }
   // repeat while inner cycle can remove entries
   } while (icnt > memx.Count());
}

// insert block into the right place
static void insert_block_b(vmem_entry &nb) {
   for (u32t ii=0; ii<memx.Count(); ii++)
      if (memx[ii].start>nb.start) {
         memx.Insert(ii,nb);
         return;
      }
   memx.Add(nb);
}

void* pag_physmap(u64t address, u32t size, u32t flags) {
   // deny all before init!
   if (!maxmap4g) {
      log_printf("map before meminit!!! (%LX,%u,%X)\n", address, size, flags);
      return 0;
   }
   // fixme!
   if (mt_exechooks.mtcb_pgmode==QPGM_V32) {
      if (address>(u64t)((u32t)-INIT_SYS_AREA)) return 0;

      // !!!!!
      return 0;
   }
   if (IS_FLAT_MODE()) {
      if (address>(u64t)FFFF) return 0;
      if (address+size>(u64t)FFFF+1) return 0;
   }
   // align request to nearest page boundary
   u32t pgdiff = address & PAGEMASK;
   address -= pgdiff;
   size    += pgdiff;
   if (size&PAGEMASK) size = PAGEROUND(size); // size is 32bit
   // map on border of physical memory
   if (address<maxmap4g && address+size>maxmap4g) return 0;

   if (address<maxmap4g && (flags&PHMAP_FORCE)==0) {
      // do nothing, just return offset

      // fixme! for VMADDR!
      return (u8t*)(u32t)address + pgdiff;
   } else
   if ((flags&PHMAP_FORCE) && IS_FLAT_MODE()) {
      log_it(3,"Force mapping in non-paged mode! %LX, %X\n", address, size);
      return 0;
   } else {
      u32t  ii;
      // lock pager mutex from this point!
      PAGEMUX_THIS_FUNC lk;

      // called before structs init?
      if (!memx.Count()) return 0;
      /* in plain non-paged mode we MUST bound this request to partially
         intersected existing map (to avoid dupes), so catch any kind of
         such entries and merge it */
      if (IS_FLAT_MODE()) merge_nonpaged(address, size);
      // search for mapped block (no diff here between paged/unpaged mode)
      for (ii=0; ii<memx.Count(); ii++)
         if ((memx[ii].flags&MEM_MAPPED)!=0)
            if (memx[ii].paddr<=address && memx[ii].paddr+memx[ii].len>=
               address+size)
            {
               memx[ii].usage++;
               return (u8t*)memx[ii].start + (u32t)(address-memx[ii].paddr) +
                  pgdiff;
            }
      if (IS_FLAT_MODE()) {
         // just create new map here, code above guarantee it to be unique
         vmem_entry  ne;
         ne.start = address;
         ne.len   = size;
         ne.flags = MEM_MAPPED;
         ne.usage = 1;
         ne.paddr = address;
         insert_block_b(ne);
         return (u8t*)ne.start + pgdiff;
      } else {
         u32t p_end = maxmap4g,
              saddr = 0, slen = 0, once = 0;
         // catch only _start_ address
         int  big_aligned = (address&PD2M_MASK)==0 && size>=1<<PD2M_ADDRSHL ?1:0,
              pass;

         for (pass = 0; pass<1+big_aligned; pass++) {
            // dumb searching for smallest hole above maxmap4g
            for (ii=0; ii<=memx.Count(); ii++) {
               u32t cstart = ii==memx.Count() ? LASTPAGE : memx[ii].start;
               if (cstart >= p_end) {
                  // align block to 2Mb
                  if (big_aligned)
                     p_end = p_end + PD2M_MASK & ~PD2M_MASK;

                  if (p_end && cstart > p_end) {
                      u32t  len = cstart - p_end;

                      if (len >= size && (!slen || len<slen)) {
                         saddr = p_end;
                         slen  = len;
                      }
                  }
                  if (ii<memx.Count()) p_end = cstart + memx[ii].len;
                  once++;
               }
            }
            if (saddr)
               if (pag_mappages(saddr, size, address, MEM_MAPPED)) {
                  vmem_entry  ne;
                  ne.start = saddr;
                  ne.len   = size;
                  ne.flags = MEM_MAPPED;
                  ne.usage = 1;
                  ne.paddr = address;
                  insert_block_b(ne);
                  return (u8t*)ne.start + pgdiff;
               }
            // drop "aligned" flag and try to use any hole
            big_aligned = 0;
         }
      }
   }
   return 0;
}

u32t pag_physunmap(void *address) {
   u32t addr = (u32t)address;
   // fake unmap of common memory
   if (addr<maxmap4g) return 1;
   // lock pager mutex!
   PAGEMUX_THIS_FUNC lk;
   // real unmap, with decrementing and address match
   if (!memx.Count()) return 0;
   int idx = pag_findidx(addr, 1);
   if (idx<0) return 0;

   if (memx[idx].usage<=1) {
      if ((memx[idx].flags&MEM_MAPPED)==0) return 0;
      // update page dir
      if (!IS_FLAT_MODE()) pag_unmappages(memx[idx].start, memx[idx].len);
      // delete entry
      memx.Delete(idx);
   } else
      memx[idx].usage--;
   return 1;
}

static void hlp_pgprint(printf_function pfn) {
   if (!pfn) pfn = log_printf;
   pfn("<=====PG memory dump=====>\n");
   // lock pager mutex!
   PAGEMUX_THIS_FUNC lk;

   for (u32t ii=0; ii<memx.Count(); ii++) {
      u32t start = memx[ii].start,
             len = memx[ii].len,
              fl = memx[ii].flags;
      static char buf[128];
      int outlen = sprintf(buf, "%2d. %08X-%08X, size: %8u kb, flags %04X",
         ii+1, start, start+len-1, len>>10, memx[ii].flags);

      if (fl&MEM_MAPPED) outlen += sprintf(buf+outlen, ", cnt %d, phys %010LX",
         memx[ii].usage, memx[ii].paddr);
      strcpy(buf+outlen, "\n");

      pfn(buf);
   }
}
