#include "seldesc.h"
#include "cpudef.h"
#include "qssys.h"
#include "qshm.h"
#include "syslocal.h"
#include "qecall.h"
#include "pagedesc.h"

static const char *pattype[8] = { "UC", "WC", "???", "???", "WT",
   "WP", "WB", "UC-" };

static void fillinfo(u32t lowdd, char *tgt, u64t patmsr, int big) {
   // "W U A D G UC-"
   // "R S A D G WT"
   memset(tgt, ' ', 14);
   tgt[0] = lowdd&PT_WRITE ?'W':'R';
   tgt[2] = lowdd&PT_R3ACCESS ?'U':'S';

   if (lowdd&PT_ACCESSED) tgt[4] = 'A';
   if (lowdd&PT_DIRTY)    tgt[6] = 'D';
   if (big && (lowdd&PT_BIGPAGE) || !big)
      if (lowdd&PT_GLOBAL) tgt[8] = 'G';

   if (patmsr==FFFF64) {
      if (lowdd&PT_PWT)  { tgt[10] = 'W'; tgt[11] = 'T'; }
      if (lowdd&PT_CDIS) { tgt[10] = 'C'; tgt[11] = 'D'; }
   } else {
      u32t idx = (lowdd&PT_PWT?1:0) + (lowdd&PT_CDIS?2:0);
      if (!big) idx+=lowdd&PT_PAT4K?4:0; else
      if ((lowdd&PT_BIGPAGE)) idx+=lowdd&PT_PATBIG?4:0;

      strcpy(tgt+10, pattype[(u32t)(patmsr>>idx)&7]);
   }
   tgt[14] = 0;
}

static void dump_pagedir4(printf_function prf, u32t *pd, u64t patmsr, int pse) {
   u32t ii, lnp = 0;

   for (ii=0; ii<1024; ii++) {
      u32t *pde = pd + ii,
             pv = *pde,
          paddr = pv & PT_ADDRMASK,
           addr = ii*_4MB;
      char astr[48], va_str[32];

      snprintf(va_str, 32, "%08X:", addr);

      if (pv&PT_PRESENT) {
         fillinfo(pv, astr, patmsr, 1);
         // 4Mb pages are available only in PSE mode
         if (pse && (pv&PD_BIGPAGE)) {
            prf("%s big page (%010LX) %s\n", va_str,
               (u64t)(paddr&PD4M_HIADDRMASK) << 32-PD4M_HIADDRSHL |
                  paddr & PD4M_ADDRMASK, astr);
         } else {
            u32t pti;
            u32t *pt = (u32t*)paddr;
            prf("%s page tab (%08X) %s\n", va_str, paddr, astr);
            for (pti=0; pti<1024; pti++) {
               u32t ptv = pt[pti];
               paddr    = ptv & PT_ADDRMASK;
               snprintf(va_str, 32, "%08X:", addr+pti*PAGESIZE);

               if (ptv&PT_PRESENT) {
                  fillinfo(ptv, astr, patmsr, 0);
                  prf("%s          (%08X) %s\n", va_str, paddr, astr);
               } else
                  prf("%s np\n", va_str);
            }
         }
         lnp = 0;
      } else
      if (!lnp) { prf("%s not present\n", va_str, addr); lnp = 1; }
   }
}

static void dump_pagedir(printf_function prf, u64t *pd, u64t vbase,
                         u64t patmsr, int is64)
{
   u32t ii, lnp = 0;

   for (ii=0; ii<512; ii++) {
      u64t *pde = pd + ii,
             pv = *pde,
          paddr = pv & ((u64t)1<<PT64_MAXADDR)-PAGESIZE,
           addr = vbase + ii*2*_1MB;
      char astr[48], va_str[32];

      snprintf(va_str, 32, is64?"%012LX:":"%08LX:", addr);

      if (pv&PT_PRESENT) {
         fillinfo(pv, astr, patmsr, 1);

         if (pv&PD_BIGPAGE) {
            prf("%s big page (%012LX) %s\n", va_str, paddr&~(u64t)((1<<PD2M_ADDRSHL)-1),
               astr);
         } else
         if (paddr>=_4GBLL) {
            prf("%s page table above 4Gb! (%012LX) %s\n", va_str, paddr, astr);
         } else {
            u32t pti;
            u64t *pt = (u64t*)paddr;
            prf("%s page tab (%012LX) %s\n", va_str, paddr, astr);
            for (pti=0; pti<512; pti++) {
               u64t ptv = pt[pti];
               paddr    = ptv & ((u64t)1<<PT64_MAXADDR)-PAGESIZE;
               snprintf(va_str, 32, is64?"%012LX:":"%08LX:", addr+pti*PAGESIZE);

               if (ptv&PT_PRESENT) {
                  fillinfo(ptv, astr, patmsr, 0);
                  prf("%s          (%012LX) %s\n", va_str, paddr, astr);
               } else
                  prf("%s np\n", va_str);
            }
         }
         lnp = 0;
      } else
      if (!lnp) { prf("%s not present\n", va_str, addr); lnp = 1; }

   }
}

void _std pag_dumptable(printf_function prf, int mode, u32t cr3, u64t patmsr) {
   u32t ii;
   if (prf==0) prf = log_printf;
   switch (mode) {
      case 0:   // 32-bit
      case 4:   // pse
         dump_pagedir4(prf, (u32t*)(cr3&~PAGEMASK), patmsr, mode?1:0);
         break;
      case 1: { // PAE table
         u64t *pdpt = (u64t*)(cr3 & ~(1<<CR3PAE_PDPSHL-1));
         for (ii=0; ii<4; ii++) {
            u32t lnp = 0;
            prf("PDPTE[%d]: %010LX\n", ii, pdpt[ii]);
            dump_pagedir(prf, (u64t*)(pdpt[ii]&~(u64t)PAGEMASK), ii*_1GB, patmsr, 0);
         }
         break;
      }
      case 2: { // 64-bit table
         u64t *pmlt = (u64t*)(cr3 & PT_ADDRMASK);
         u32t   lnp = 0, pmli;
         
         for (pmli=0; pmli<512; pmli++) {
            u64t   pv = pmlt[pmli],
               staddr = (u64t)pmli<<39;
            char astr[48];
         
            if (pv&PT_PRESENT) {
               u64t pdpt = pv & ((u64t)1<<PT64_MAXADDR)-PAGESIZE;
               fillinfo(pv, astr, patmsr, 1);
               lnp = 0;
         
               prf("PML4E[%d]: %012LX %s\n", pmli, pdpt, astr);
         
               if (pdpt>=_4GBLL) continue;
         
               for (ii=0; ii<512; ii++, staddr+=_1GB) {
                  u64t  paddr;
                  pv    = ((u64t*)pdpt)[ii];
                  paddr = pv & ((u64t)1<<PT64_MAXADDR)-PAGESIZE;
                  fillinfo(pv, astr, patmsr, 1);
         
                  if (pv&PT_PRESENT) {
                     lnp = 0;
                     if (pv&PD_BIGPAGE) {
                        prf("%012LX: big page (%012LX) %s\n", staddr,
                           paddr&~(u64t)((1<<PD1G_ADDRSHL)-1), astr);
                     } else
                     if (paddr>=_4GBLL) {
                        prf("%012LX: PDPTE above 4Gb! (%012LX) %s\n", staddr,
                           paddr, astr);
                     } else {
                        prf("%012LX: PDPTE (%012LX) %s\n", staddr, paddr, astr);
                        dump_pagedir(prf, (u64t*)paddr, staddr, patmsr, 1);
                     }
                  } else
                  if (!lnp) { prf("%012LX: not present\n", staddr); lnp = 1; }
               }
            } else
            if (!lnp) { prf("%012LX: not present\n", staddr); lnp = 1; }
         }
         break;
      }
   }
}

// it is not locked!
void _std pag_printall(printf_function prf) {
   u32t  ii, pi, cr0v, cr3v, cr4v;
   u64t  patmsr;
   if (prf==0) prf = log_printf;
   prf("<====== Page table ======>\n");
   cr0v = getcr0();
   cr3v = getcr3();
   if ((cr0v&CPU_CR0_PG)==0) return;
   cr4v = cr4warn?0:getcr4();
   if ((cr4v&(CPU_CR4_PAE|CPU_CR4_PSE))==0) return;

   if (sys_isavail(SFEA_PAT)) {
      hlp_readmsr(MSR_IA32_PAT, (u32t*)&patmsr, (u32t*)&patmsr + 1);
      prf("PAT (0..7): %8b\n", &patmsr);
   } else
      patmsr = FFFF64;

   if (sys_is64mode()) {
      u64t  cr3x, *pmlt;
      prf("64-bit mode!\n");
      cr3x = call64l(EFN_HLPSETREG, 2, 2, SETREG64_CR3, FFFF64);
      if (cr3x==FFFF64) { prf("Version mismatch!\n"); return; }
      if (cr3x>=_4GBLL) { prf("Page table above 4GB!\n"); return; }

      pag_dumptable(prf, 2, cr3x, patmsr);
   } else
      pag_dumptable(prf, sys_pagemode() & QPGM_PAE?1:4, cr3v, patmsr);
}

u64t _std sys_efiptlen(void) {
   static u64t ptlen = 0;
   u32t   ii, pdi, pti;
   u64t      cr3x;
   if (!sys_is64mode()) return 0;
   // return cached value (firmware page table should be static)
   if (ptlen) return ptlen;

   cr3x = call64l(EFN_HLPSETREG, 2, 2, SETREG64_CR3, FFFF64);
   log_it(2, "cr3=%LX\n", cr3x);
   // should never occurs
   if (cr3x!=FFFF64) {
      int  himem = cr3x>=_4GBLL;
      u64t *pmlt = 0,
           *pdpb = 0;
      
      if (himem) {
         pmlt = malloc_th(PAGESIZE);
         // exception will be printed in log
         if (!sys_memhicopy((u32t)pmlt, cr3x & ~(u64t)PAGEMASK, PAGESIZE)) {
            free(pmlt);
            pmlt = 0;
         }
      } else
         pmlt = (u64t*)((u32t)cr3x & PT_ADDRMASK);
      
      for (ii=0; pmlt && ii<512 && !ptlen; ii++) {
         u64t   pv = pmlt[ii],
            staddr = (u64t)ii<<39;
      
         if (pv&PT_PRESENT) {
            u64t pdptv = pv & ((u64t)1<<PT64_MAXADDR)-PAGESIZE, *pdpt;
      
            if (pdptv>=_4GBLL) {
               log_it(2, "pml[%u]=%LX\n", ii, pdptv);
               if (!pdpb) pdpb = malloc_th(PAGESIZE);
               if (!sys_memhicopy((u32t)pdpb, pdptv, PAGESIZE)) continue;
               pdpt = pdpb;
            } else
               pdpt = (u64t*)(u32t)pdptv;
            // round it to 1Gb
            for (pdi=0; pdi<512 && !ptlen; pdi++, staddr+=_1GB)
               if (!(pdpt[pdi]&PT_PRESENT)) { ptlen = staddr; break; }
      
         } else { ptlen = staddr; break; }
      }
      if (himem && pmlt) free(pmlt);
      if (pdpb) free(pdpb);
   }
   return ptlen;
}
