#define Uses_TApplication
#define Uses_TDeskTop
#include <tv.h>
#include "helpwin.h"

#define cHelpViewer     "\x06\x07\x08"
#define cHelpWindow     "\x80\x81\x82\x83\x84\x85\x86\x87"

#define HEADER_LINES    2
#define HEADER_BOTTOM   '\xDF'
#define HEADER_RIGHT    " \xDC"

struct helpfileinfo {
   const char  *name;
   const char  *link;
   qsh_file       fh;
   u32t       topics;
   u32t     tvscheme;
   u32t     minwidth;
   int         index;
   char       *title;
};

struct historyentry {
   int       hfindex;
   int       topicid;
   int          line;
   int       sellink;
};

#define N_HELP_FILES    3
// known help files
static helpfileinfo LHF[N_HELP_FILES] = {
// HELPFILE_SYSVIEW
#ifdef __QSINIT__
   {"SYSVIEW.QSH",
#else
   {"msg\\sysview.txt",
#endif
    "SYSVIEW",0,0,0,0,-1,0},
// HELPFILE_QSSYS
#ifdef __QSINIT__
   {"SYS.QSH",
#else
   {"..\\help\\help\\sys.hlp",
#endif
    "SYS",0,0,0,0,-1,0},
// reserved
   {"",0,0,0,0,0,-1,0}
};

// global topic history
static TCollection *GTH = 0;

static void hf_close(void) {
   for (u32t ii=0; ii<N_HELP_FILES; ii++) {
      if (LHF[ii].title) {
         free(LHF[ii].title);
         LHF[ii].title = 0;
      }
      if (LHF[ii].fh) {
         help_close(LHF[ii].fh);
         LHF[ii].fh = 0;
      }
   }
   if (GTH) delete GTH;
}

static helpfileinfo& hf_get(int index) {
   // return zeroed last entry if index is invalid
   if (index<0 || index>=N_HELP_FILES) return LHF[N_HELP_FILES - 1];
   return LHF[index];
}

static void hf_getminsize(int index, TPoint &min) {
   if (min.y < 8) min.y = 8;
   helpfileinfo &fi = hf_get(index);
   if (min.x < fi.minwidth + 4) min.x = fi.minwidth + 4;
}

qserr TQH_LoadHelp(int index) {
   static int exitset = 0;
   if (index<0 || index>=N_HELP_FILES) return E_SYS_INVPARM;
   if (LHF[index].fh) return 0;

   qserr res = help_open(LHF[index].name, &LHF[index].fh);
   if (!res) {
      qsh_info *inf = help_info(LHF[index].fh);
      if (!inf) {
         help_close(LHF[index].fh);
         LHF[index].fh = 0;
         return E_SYS_SOFTFAULT;
      }
      // used as helf window title
      LHF[index].title    = strdup(inf->title ? inf->title : "Help");
      LHF[index].topics   = inf->ntopics;
      LHF[index].index    = inf->index;
      LHF[index].minwidth = inf->minwidth;
      LHF[index].tvscheme = 0;
      // try find color scheme named "tv"
      for (int ii=0; ii<inf->nschemes; ii++)
         if (stricmp(inf->scheme[ii],"tv")==0) {
            LHF[index].tvscheme = ii;
            break;
         }
      free(inf);
   }

   if (!res && !exitset) {
      exitset = 1;
      atexit(hf_close);
   }
   return res;
}

TQHViewer::TQHViewer(const TRect &bounds, TScrollBar *aVScrollBar,
                     int HFIndex, const char *context, Boolean histLocal) :
                     TScroller(bounds, 0, aVScrollBar)
{
   options  = (options | ofSelectable);
   growMode = gfGrowHiX | gfGrowHiY;
   selected = -1;
   lth      = histLocal ? new TCollection(0,10) : 0;
   topic    = 0;
   topicid  = 0;
   hfIndex  = HFIndex;
   ldrawerr = TQH_LoadHelp(hfIndex);

   if (!ldrawerr) {
      helpfileinfo &hf = hf_get(hfIndex);
      topicid  = help_topicid(hf.fh, context);

      if (!topicid) ldrawerr = E_HLP_NOTOPIC; else
         topic = help_draw(hf.fh, hf.tvscheme, topicid, size.x, &ldrawerr);
   }
   if (topic) {
      setLimit(78, topic->lines+HEADER_LINES);
      // find top link
      selected = findLink(HLINK_ANY);
   }
}

void TQHViewer::updateTopic() {
   if (topicid && topic && topic->width!=size.x) {
      if (topic) free(topic);
      helpfileinfo &hf = hf_get(hfIndex);
      topic = help_draw(hf.fh, hf.tvscheme, topicid, size.x, &ldrawerr);
   }
   // it`s probably unsafe to call setLimit here because it called from draw()
}

TQHViewer::~TQHViewer() {
   if (topic) free(topic);
   if (lth) delete lth;
}

void TQHViewer::changeBounds(const TRect &bounds) {
   TScroller::changeBounds(bounds);
   updateTopic();
   if (topic) setLimit(limit.x, topic->lines+HEADER_LINES);
}

void TQHViewer::pushHistory(int HFIndex, int topicid, int line, int sellink) {
   historyentry *he = new historyentry;
   he->hfindex = HFIndex;
   he->topicid = topicid;
   he->line    = line<0 ? 0 : line;
   he->sellink = sellink;
   // first time creation
   if (!lth && !GTH) GTH = new TCollection(0,10);

   (lth?lth:GTH)->insert(he);
}

Boolean TQHViewer::goBack() {
   TCollection *src = lth?lth:GTH;
   if (!src) return False;
   ccIndex    count = src->getCount();
   Boolean      res = False;

   if (count>0) {
      historyentry *he = (historyentry*)src->at(count-1);
      src->atRemove(count-1);
      // show topic, but skip history push
      res = gotoToTopic(he->hfindex, he->topicid, he->line, 0, he->sellink, True);
      delete he; 
   }
   return res;
}

void TQHViewer::checkLink() {
   if (topic) {
      int lidx = selected,
           vis = 0,
          rtop = delta.y - HEADER_LINES; // can be < 0
      
      while (lidx>=0 && lidx<topic->nlinks) {
         qsh_link &lnk = topic->llist[lidx];
         // next group?
         if (lidx!=selected && lnk.back!=0xFFFF) break;
         lidx++;
         if ((int)lnk.ypos<rtop || (int)lnk.ypos>=rtop+size.y) continue;
         vis++;
      }
      if (!vis) selected = findLink(HLINK_ANY);
   }
}

void TQHViewer::draw() {
   TDrawBuffer   b;
   int      hdrlen = -1;
   ushort   normal = getColor(1);

   updateTopic();
   checkLink();

   for (int ii = 0; ii<size.y; ii++) {
      int rline = ii + delta.y - HEADER_LINES;

      if (!topic || size.x > topic->width || rline<0 || rline>=topic->lines)
         b.moveChar(0, ' ', normal, size.x);

      if (topic && rline>=0 && rline < topic->lines) {
         b.moveBuf(0, topic->text + topic->width * rline, 0, topic->width);
         // draw selected link in this line
         if (selected>=0)
            for (int lidx = selected; lidx<topic->nlinks; lidx++) {
               qsh_link &lnk = topic->llist[lidx];
               if (lidx!=selected && lnk.back!=0xFFFF) break; // next group
               if (lnk.ypos==rline) {
                  for (int xx = 0; xx<lnk.len; xx++)
                     b.putAttribute(lnk.xpos+xx, topic->selcol);
                  break;
               }
            }
      }
      // print error text
      if (!topic && !rline) {
         char *err = SysApp.GetPTErr(ldrawerr);
         b.moveStr(2, err, normal);
         free(err);
      } else
      if (topic && rline<0) {
         char str[maxViewWidth]; // 512
         if (rline==-2 || rline==-1 && hdrlen<0) {
            int   len = strlen(topic->topic),
                limit = size.x*4/5;
            // cut too long topic name
            if (len > limit) {
               memcpy(str, topic->topic, limit);
               strcpy(str + limit, "...");
            } else
               strcpy(str, topic->topic);
            strcat(str, HEADER_RIGHT);
            hdrlen = strlen(str);
            // this can be a calc for "delta.y = 1" case
            if (rline==-2) b.moveStr(1, str, normal);
         } else
            b.moveChar(2, HEADER_BOTTOM, normal, hdrlen-1);
      }
      writeLine(0, ii, size.x, 1, b);
   }
}

TPalette &TQHViewer::getPalette() const {
   static TPalette palette(cHelpViewer, sizeof(cHelpViewer)-1);
   return palette;
}

int TQHViewer::findLink(int dir) {
   if (!topic || dir>HLINK_LEFT || dir!=HLINK_ANY && selected<0) return -1;
   int  rtop = delta.y - HEADER_LINES,          // can be < 0
        vhgt = rtop<0 ? size.y + rtop : size.y; // visible help height
   return help_findlink(topic, selected, dir, rtop<0?0:rtop, vhgt);
}

void TQHViewer::makeSelectVisible(int newsel) {
   if (topic && newsel<topic->nlinks) {
      int y1 = 0xFFFF, y2 = -1;

      for (int lidx = newsel; lidx<topic->nlinks; lidx++) {
         qsh_link &lnk = topic->llist[lidx];
         // next group
         if (lidx!=newsel && lnk.back!=0xFFFF) break;

         if (lnk.ypos<y1) y1 = lnk.ypos;
         if (lnk.ypos>y2) y2 = lnk.ypos;
      }
      int  rtop = delta.y - HEADER_LINES,          // can be < 0
           vhgt = rtop<0 ? size.y + rtop : size.y; // visible help height
      
      int newy = delta.y;
      if (y2 >= rtop + vhgt) newy = y2 + HEADER_LINES - size.y + 1;
      if (y1 <= rtop) newy = y1 > 1 ? y1 - 1 + HEADER_LINES : 0;

      selected = newsel;
      
      if (newy!=delta.y) scrollTo(0, newy);
   }
}

Boolean TQHViewer::gotoToTopic(int hfindex, int newid, int line,
                               const char *mark, int sellink, Boolean noPush)
{
   helpfileinfo  &hf = hf_get(hfindex);
   // use minwidth and try to resize window below
   qsh_topic *ntopic = help_draw(hf.fh, hf.tvscheme, newid, hf.minwidth >
                                 size.x ? hf.minwidth : size.x, &ldrawerr);
   if (!ntopic) return False;
   // "mark" has priority above "line" arg
   if (mark)
      for (u32t mm=0; mm<ntopic->nmarks; mm++)
         if (!stricmp(mark, ntopic->mark[mm])) {
            line = ntopic->mlist[mm];
            if (line) line += HEADER_LINES;
            break;
         }
   // push current topic to the history
   if (!noPush && topic)
      pushHistory(hfIndex, topicid, delta.y, selected);
   // delete topic and reset pointer to skip draw if window must be resized!
   if (topic) { delete topic; topic = 0; }
   // resize window to the new file`s minwidth
   int newfile = hfindex!=hfIndex;
   if (newfile)
      if (owner->size.x < hf.minwidth + 4) {
         TRect r(owner->getBounds());
         r.a.x = SysApp.deskTop->size.x - hf.minwidth - 4 >> 1;
         r.b.x = r.a.x + hf.minwidth + 4;
         owner->locate(r);
         growTo(hf.minwidth, size.y);
      }

   hfIndex  = hfindex;
   topic    = ntopic;
   topicid  = newid;
   // pos to specified line in the topic
   if (line>0)
      if (line>=topic->lines+HEADER_LINES) line = 0; else
         if (line+size.y > topic->lines+HEADER_LINES)
            line = topic->lines + HEADER_LINES < size.y ? 0 :
                   topic->lines + HEADER_LINES - size.y;
   // disable draw while both pos/size update
   drawLock++;
   setLimit(limit.x, topic->lines+HEADER_LINES);
   scrollTo(0, line);
   drawLock--;

   if (sellink<0) selected = findLink(HLINK_ANY); else {
      selected = sellink;
      checkLink();
   }
   drawView();
   if (newfile)
      message(owner, evBroadcast, cmUpdateTitle, 0);
   return True;
}

/* accepts modes: switchToTopic(id, "mark");
                  switchToTopic(0,  "topic/mark"); 
                  switchToTopic(0,  "?file/topic/mark"); */
Boolean TQHViewer::switchToTopic(int newid, const char *mark) {
   if (!newid && !mark) return False;

   int     hfi = hfIndex;
   char *tlink = 0,
        *mlink = 0;

   if (!newid) {
      // try to change file
      if (mark[0]=='?') {
         char *flink = opts_getword(mark, 1, "/");
         hfi = -1;
         // search for a "short" file name
         for (int ii=0; ii<N_HELP_FILES; ii++)
            if (LHF[ii].link && !stricmp(LHF[ii].link, flink+1)) {
               hfi = ii;
               break;
            }
         delete flink;
         if (hfi<0) return False;
         // check that file is ready
         if (hfi!=hfIndex)
            if (TQH_LoadHelp(hfi)) return False;

         tlink = opts_getword(mark, 2, "/"),
         mlink = opts_getword(mark, 3, "/");
      } else {
         tlink = opts_getword(mark, 1, "/"),
         mlink = opts_getword(mark, 2, "/");
      }
      mark = *mlink ? mlink : 0;
   }
   helpfileinfo  &hf = hf_get(hfi);
   Boolean       res = False;
   if (!newid) newid = *tlink ? help_topicid(hf.fh, tlink) : hf.index;
   if (tlink) delete tlink;
   if (newid)    res = gotoToTopic(hfi, newid, 0, mark, -1, False);
   if (mlink) delete mlink;

   return res;
}

void TQHViewer::handleEvent(TEvent &event) {
   // filter navigation keys before TScroller!
   if (topic && selected>=0 && event.what==evKeyDown) {
      int nlink = -1;
      switch (event.keyDown.keyCode) {
         case kbUp   :
         case kbCtrlE: nlink = findLink(HLINK_UP);  break;
         case kbLeft :
         case kbCtrlS: nlink = findLink(HLINK_LEFT); break;
         case kbRight:
         case kbCtrlD: nlink = findLink(HLINK_RIGHT); break;
         case kbDown :
         case kbCtrlX: nlink = findLink(HLINK_DOWN); break;
      }
      if (nlink>=0) {
         selected = nlink;
         drawView();
         clearEvent(event);
         return;
      }
   }

   TScroller::handleEvent(event);
   switch (event.what) {
      case evKeyDown: {
         Boolean upd = False;
         switch (event.keyDown.keyCode) {
            case kbTab:
               if (topic && topic->nlinks>0) {
                  int ll = 0;
                  if (selected>=0) {
                     for (ll = selected; ll<topic->nlinks; ll++)
                        if (ll!=selected && topic->llist[ll].back!=0xFFFF) break;
                     if (ll==topic->nlinks) ll = 0;
                  }
                  // can found self, actually 
                  if (ll!=selected) {
                     makeSelectVisible(ll);
                     upd = True;
                  }
               }
               break;
            case kbShiftTab:
               if (topic && topic->nlinks>0) {
                  int dst = topic->llist[selected<0?0:selected].back;
                  // a single invisible link
                  if (dst==0xFFFF && selected<0) dst = 0;
                  if (dst!=0xFFFF) {
                     makeSelectVisible(dst);
                     upd = True;
                  }
               }
               break;
            case kbAltF1:
            case kbBack:
               upd = goBack();
               break;
            case kbEnter:
               if (topic && selected>=0 && selected <= topic->nlinks)
                  upd = switchToTopic(0, topic->link[selected]);
               break;
            case kbEsc:
               event.what = evCommand;
               event.message.command = cmClose;
               putEvent(event);
               break;

            default:
               return;
         }
         if (upd) drawView();
         clearEvent(event);
         break;
      }
      case evMouseDown:
         if (topic) {
            TPoint mouse = makeLocal(event.mouse.where);
            mouse.x += delta.x;
            mouse.y += delta.y - HEADER_LINES;

            if (mouse.y>=0)
               for (int lidx=0, group=0; lidx<topic->nlinks; lidx++) {
                  qsh_link &lnk = topic->llist[lidx];
                  if (lidx && lnk.back!=0xFFFF) group = lidx;

                  if (lnk.ypos==mouse.y && mouse.x>=lnk.xpos && 
                     mouse.x<lnk.xpos+lnk.len)
                  {
                     selected = lidx;
                     drawView();
                     if (event.mouse.eventFlags & meDoubleClick)
                        switchToTopic(0, topic->link[selected]);
                     clearEvent(event);
                     break;
                  }
               }
         }
         break;
      case evCommand:
         if (event.message.command == cmClose) {
            if (topic)
               pushHistory(hfIndex, topicid, delta.y, selected);

            if ((owner->state & sfModal) != 0) {
               endModal(cmClose);
               clearEvent(event);
            }
         }
         if (event.message.command == cmHelpBack)
            if (goBack()) {
               drawView();
               clearEvent(event);
            }
         break;
   }
}

TAppQHWindow::TAppQHWindow(int HFIndex, const char *context) :
   TAppWindow (SysApp.deskTop->getExtent(),
               HFIndex<0? "Help" : hf_get(HFIndex).title, AppW_Help),
   TWindowInit(initFrame)
{
   TRect r(SysApp.deskTop->getExtent());
   options = (options | ofCentered);
   r.grow(-2,-1);

   if (HFIndex<0)
      if (!GTH || !GTH->getCount()) {
         // goto SysView index topic if no history
         HFIndex = HELPFILE_SYSVIEW;
         context = 0;
      }

   hwc = new TQHViewer(r, standardScrollBar(sbVertical | sbHandleKeyboard),
                       HFIndex<0 ? N_HELP_FILES-1 : HFIndex, context, False);
   insert(hwc);
   if (HFIndex<0) hwc->goBack();
}

void TAppQHWindow::goTopic(int HFIndex, const char *topic) {
   if (HFIndex<0) hwc->goBack(); else {
      helpfileinfo &hf = hf_get(HFIndex);
      int topicid  = help_topicid(hf.fh, topic);
      if (topicid>0) hwc->gotoToTopic(HFIndex, topicid, 0, 0, -1, False);
   }
}

void TAppQHWindow::sizeLimits(TPoint &min, TPoint &max) {
   TWindow::sizeLimits(min, max);
   hf_getminsize(hwc->hfNum(), min);
}

const char *TAppQHWindow::getTitle(short) {
   return hf_get(hwc->hfNum()).title;
}

void TAppQHWindow::handleEvent(TEvent &event) {
   if (event.what == evBroadcast && event.message.command == cmUpdateTitle) {
      if (frame) frame->drawView();
      clearEvent(event);
   }
   TAppWindow::handleEvent(event);
}

TPalette &TAppQHWindow::getPalette() const {
   static TPalette palette(cHelpWindow, sizeof(cHelpWindow)-1);
   return palette;
}

TQHWindow::TQHWindow(int HFIndex, const char *context):
   TWindow(TRect(0,0,50,18), "Help", wnNoNumber),
   TWindowInit(TQHWindow::initFrame) {
   TRect r(0, 0, 50, 18);
   options = (options | ofCentered);
   r.grow(-2,-1);
   hwc = new TQHViewer(r, standardScrollBar(sbVertical | sbHandleKeyboard),
                       HFIndex, context, True);
   insert(hwc);
}

void TQHWindow::sizeLimits(TPoint &min, TPoint &max) {
   TWindow::sizeLimits(min, max);
   hf_getminsize(hwc->hfNum(), min);
}

TPalette &TQHWindow::getPalette() const {
   static TPalette palette(cHelpWindow, sizeof(cHelpWindow)-1);
   return palette;
}
