//
// QSINIT
// ISO file images
// ------------------------------------------------------------------
//
#include "vhdd.h"
#include "sys/fs/isofs.h"

#define CACHE_LEN     13     /* number of sectors to cache in raw mode (<=17)
                                both 2352*13 & 2448*13 fit into 32k default
                                i/o buffer size */
#define RAW_SECTOR  2352
#define MDF_SECTOR  2448

/// "qs_isodisk" class data
typedef struct {
   u32t           sign;
   qs_dyndisk  selfptr;    // pointer to own instance
   s32t         qsdisk;
   io_handle        df;    // disk file
   u64t       startpos;
   char        *dfname;    // disk file name
   disk_geo_data  info;
   qs_extdisk    einfo;    // attached to this disk ext.info class
   int         rawmode;
   u32t      rawsector;    // RAW_SECTOR or MDF_SECTOR
   u32t       cachepos;
   u8t          *cache;
} diskdata;

static qserr _exicc dsk_close(EXI_DATA);

/// create new image file
static qserr _exicc dsk_make(EXI_DATA, const char *fname, u32t sectorsize, u64t sectors) {
   return E_SYS_UNSUPPORTED;
}

static qserr _exicc dsk_open(EXI_DATA, const char *fname, const char *options) {
   qserr         rc;
   str_list   *opts = 0;
   u32t    sectorsz = 0,
             imglen = 0;     // length in sectors
   u64t      offset = 0;

   instance_ret(diskdata,di,E_SYS_INVOBJECT);
   // already opened?
   if (di->df) return E_SYS_INITED;

   if (options) {
      char *key;
      opts = str_split(options,";");
      key  = str_findkey(opts, "sector", 0);
      if (key) {
         sectorsz = str2ulong(key);
         if (sectorsz!=ISO_SECTOR_SIZE && sectorsz!=RAW_SECTOR &&
            sectorsz!=MDF_SECTOR) return E_DSK_SSIZE;
      }
      key = str_findkey(opts, "offset", 0);
      if (key) offset = str2uint64(key);
      key = str_findkey(opts, "size", 0);
      if (key) {
         imglen = str2ulong(key);
         // force this to simplify code below (16 + primary + term)
         if (imglen<18) return E_SYS_INVPARM;
      }
      // just deny them
      if (str_findkey(opts, "spt", 0) || str_findkey(opts, "heads", 0))
         return E_SYS_INVPARM;
   }
   // open a file
   rc = io_open(fname, IOFM_OPEN_EXISTING|IOFM_READ, &di->df, 0);

   if (!rc) {
      u64t sz;
      rc = io_size(di->df, &sz);

      if (!rc) {
         u32t assumesize = sectorsz ? sectorsz : ISO_SECTOR_SIZE;
         u8t        *buf = 0;
         // at least one sector after offset
         if (imglen && offset + (u64t)imglen * assumesize > sz ||
            offset + assumesize > sz) rc = E_SYS_EOF;
         else 
         do {
            int   raw = -1;
            if (sectorsz==ISO_SECTOR_SIZE) raw = 0;
            // need to detect ISO mode?
            if (!sectorsz || sectorsz==RAW_SECTOR || sectorsz==MDF_SECTOR) {
               /* read 17 RAW sectors, CACHE_LEN of them will be reused for a
                  cache below */
               buf = malloc_th(17 * MDF_SECTOR);
               if (io_read(di->df, buf, 17*MDF_SECTOR) != 17*MDF_SECTOR)
                  if (!(rc = io_lasterror(di->df))) rc = E_DSK_ERRREAD;
               if (rc) break;
            }
            // try to detect sector size
            if (!sectorsz) {
               if (!memcmp(buf + 0x8001, ISO_STANDARD_ID, 5) ||
                   !memcmp(buf + 0x8009, ISO_SIERRA_ID, 5))
               {
                  sectorsz = ISO_SECTOR_SIZE;
                  raw      = 0;
               } else
               if (!memcmp(buf + 16*RAW_SECTOR + 17, ISO_STANDARD_ID, 5) ||
                   !memcmp(buf + 16*RAW_SECTOR + 25, ISO_SIERRA_ID, 5))
               {
                  sectorsz = RAW_SECTOR;
                  raw      = 1;          // mode 1
               } else
               if (!memcmp(buf + 16*RAW_SECTOR + 25, ISO_STANDARD_ID, 5) ||
                   !memcmp(buf + 16*RAW_SECTOR + 33, ISO_SIERRA_ID, 5))
               {
                  sectorsz = RAW_SECTOR;
                  raw      = 2;          // mode 2
               } else
               if (!memcmp(buf + 16*MDF_SECTOR + 17, ISO_STANDARD_ID, 5) ||
                   !memcmp(buf + 16*MDF_SECTOR + 25, ISO_SIERRA_ID, 5))
               {
                  sectorsz = MDF_SECTOR;
                  raw      = 1;          // mode 1
               } else
               if (!memcmp(buf + 16*MDF_SECTOR + 25, ISO_STANDARD_ID, 5) ||
                   !memcmp(buf + 16*MDF_SECTOR + 33, ISO_SIERRA_ID, 5))
               {
                  sectorsz = MDF_SECTOR;
                  raw      = 2;          // mode 2
               }
            }
            // try to detect sector size and RAW sector mode
            if (!sectorsz || (sectorsz==RAW_SECTOR || sectorsz==MDF_SECTOR) &&
                raw<=0)
            {
               static u8t syncdata[12] = {
                  0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0
               };
               int syncmatch = !memcmp(buf, &syncdata, 12) && (buf[15]==1 || buf[15]==2);
               if (syncmatch && !memcmp(buf + 16*RAW_SECTOR, &syncdata, 12) &&
                   buf[16*RAW_SECTOR + 15] == buf[15])
               {
                  sectorsz = RAW_SECTOR;
                  raw      = buf[15];
               } else 
               if (syncmatch && !memcmp(buf + 16*MDF_SECTOR, &syncdata, 12) &&
                   buf[16*MDF_SECTOR + 15] == buf[15])
               {
                  sectorsz = MDF_SECTOR;
                  raw      = buf[15];
               } else {
                  // there is no sync data or mode byte is wrong
                  if (sectorsz==RAW_SECTOR) {
                     log_printf("ISO mount - sync data or mode byte (%u) mismatch\n", buf[15]);
                     rc = E_SYS_BADFMT;
                     break;
                  }
                  sectorsz = ISO_SECTOR_SIZE;
                  raw      = 0;
               }
            }
            // check size again, with known sector size
            if (!imglen) imglen = (sz - offset) / sectorsz; else
               if (offset + (u64t)imglen * sectorsz > sz) { rc = E_SYS_EOF; break; }

            di->info.TotalSectors = imglen;
            di->info.SectorSize   = ISO_SECTOR_SIZE;
            di->info.Heads        = 2;
            di->info.SectOnTrack  = 128;
            di->info.Cylinders    = 0;
            // fill missing chs values
            rc = dsk_fakechs(&di->info);
            if (rc) break;
            
            di->startpos  = offset;
            di->rawmode   = raw;
            di->rawsector = sectorsz;
            // save full file name
            di->dfname    = _fullpath(0, fname, 0);
            mem_modblock(di->dfname);
            // detach file to make it process-independent
            io_setstate (di->df, IOFS_DETACHED, 1);

            di->cachepos = FFFF;
            if (di->rawmode>0) {
               // block is module-owned
               di->cache = malloc(sectorsz * CACHE_LEN);
               // save ready data from the buffer here
               if (buf) {
                  memcpy(di->cache, buf, sectorsz * CACHE_LEN);
                  di->cachepos = 0;
               }
            }
            log_printf("\"%s\" - mode %u, sec %u\n", di->dfname, di->rawmode,
               di->rawsector);
         } while (0);
         if (buf) free(buf);
      }
   }
   if (rc) dsk_close(data, 0);
   return rc;
}

/// query "disk" info
static qserr _exicc dsk_query(EXI_DATA, disk_geo_data *info, char *fname,
                              u32t *sectors, u32t *used, u32t *state)
{
   instance_ret(diskdata,di,E_SYS_INVOBJECT);
   if (info)
      if (di->df) memcpy(info, &di->info, sizeof(disk_geo_data));
         else memset(info, 0, sizeof(disk_geo_data));
   if (fname)
      if (di->dfname) strcpy(fname, di->dfname); else *fname = 0;
   if (sectors) *sectors = di->info.TotalSectors;
   if (used)    *used    = di->info.TotalSectors;
   if (state)   *state   = EDSTATE_NOCACHE | EDSTATE_RO;
   return 0;
}

static u32t _exicc dsk_read(EXI_DATA, u64t sector, u32t count, void *usrdata) {
   instance_ret(diskdata,di,0);

   // no file?
   if (!di->df) return 0;
   // check limits
   if (sector>=di->info.TotalSectors) return 0;
   if (sector+count>di->info.TotalSectors) count = di->info.TotalSectors - sector;

   if (count) {
      if (di->rawmode) {
         // 8Tb ISO is impossible (no such volume support), so dword is enough
         u32t  sno = sector,
              epos = sno + count;

         for (; sno<epos; sno++) {
            // read raw data into the buffer
            if (sno < di->cachepos || sno >= di->cachepos + CACHE_LEN) {
               u32t ncpos;
               // special end of disk handling
               if (sno > di->info.TotalSectors - CACHE_LEN)
                  ncpos = di->info.TotalSectors - CACHE_LEN;
               else
                  ncpos = sno;
               // old cache data still valid if seek error
               if (io_seek(di->df, di->startpos + (u64t)ncpos * di->rawsector,
                  IO_SEEK_SET)==FFFF64) return count - (epos - sno);

               if (io_read(di->df, di->cache, di->rawsector * CACHE_LEN) !=
                  di->rawsector * CACHE_LEN)
               {
                  di->cachepos = FFFF;
                  return count - (epos - sno);
               }
               di->cachepos = ncpos;
            }
            memcpy(usrdata, di->cache + (sno - di->cachepos) * di->rawsector +
               (di->rawmode==1 ? 16 : 24), ISO_SECTOR_SIZE);
            usrdata = (u8t*)usrdata + ISO_SECTOR_SIZE;
         }
         return count;
      } else {
         if (io_seek(di->df, di->startpos + ((u64t)sector<<ISO_SECTOR_BITS),
            IO_SEEK_SET)==FFFF64) return 0;
         return io_read(di->df, usrdata, count<<ISO_SECTOR_BITS) >> ISO_SECTOR_BITS;
      }
   }
   return 0;
}

static u32t _exicc dsk_write(EXI_DATA, u64t sector, u32t count, void *usrdata) {
   instance_ret(diskdata,di,0);
   return 0;
}

/** mount as QSINIT disk.
    @return disk number or negative value on error */
static s32t _exicc dsk_mount(EXI_DATA) {
   struct qs_diskinfo dp;
   instance_ret(diskdata,di,-1);

   memset(&dp, 0, sizeof(dp));
   dp.qd_sectors     = di->info.TotalSectors;
   dp.qd_cyls        = di->info.Cylinders;
   dp.qd_heads       = di->info.Heads;
   dp.qd_spt         = di->info.SectOnTrack;
   dp.qd_extread     = diskio_read;
   dp.qd_extwrite    = diskio_write;
   dp.qd_extremove   = diskio_remove;
   dp.qd_sectorsize  = ISO_SECTOR_SIZE;
   dp.qd_sectorshift = ISO_SECTOR_BITS;

   if (cid_ext) {
      qs_extdisk  eptr = (qs_extdisk)exi_createid(cid_ext, EXIF_SHARED);
      if (eptr) {
         de_data *de  = (de_data*)exi_instdata(eptr);
         de->self     = di->selfptr;
         dp.qd_extptr = eptr;
         di->einfo    = eptr;
      }
   }
   // install new "hdd"
   di->qsdisk = hlp_diskadd(&dp, 0);
   if (di->qsdisk<0 || di->qsdisk>QDSK_DISKMASK) log_printf("disk add error!\n");
      else mounted[di->qsdisk] = di->selfptr;

   return di->qsdisk;
}

/** query QSINIT disk number.
    @return disk number or negative value if not mounted */
static s32t _exicc dsk_disk(EXI_DATA) {
   instance_ret(diskdata,di,-1);
   return di->qsdisk;
}

/// unmount "disk" from QSINIT
static qserr _exicc dsk_umount(EXI_DATA) {
   instance_ret(diskdata,di,E_SYS_INVOBJECT);

   if (di->qsdisk>=0) {
      if (di->qsdisk<=QDSK_DISKMASK) mounted[di->qsdisk] = 0;
      hlp_diskremove(di->qsdisk);
      di->qsdisk = -1;
      if (di->einfo) { DELETE(di->einfo); di->einfo = 0; }
      return 0;
   }
   return E_DSK_NOTMOUNTED;
}

static u32t _exicc dsk_compact(EXI_DATA, int freeF6) { return 0; }

/// close image file
static qserr _exicc dsk_close(EXI_DATA) {
   instance_ret(diskdata,di,E_SYS_INVOBJECT);
   if (!di->df) return E_SYS_NONINITOBJ;
   // unmount it first
   if (di->qsdisk>=0) dsk_umount(data, 0);
   // and close file
   io_close(di->df);
   if (di->dfname) free(di->dfname);
   if (di->cache) free(di->cache);
   di->df     = 0;
   di->dfname = 0;
   di->cache  = 0;
   return 0;
}

static void _exicc dsk_trace(EXI_DATA, int self, int bitmaps) {
   instance_void(diskdata,di);
   // self - query it, because user able to change it by exi_chainset() too ;)
   if (self>=0) {
      int cstate = exi_chainset(di->selfptr,-1);
      if (cstate>=0)
         if (Xor(self,cstate)) exi_chainset(di->selfptr, self);
   }
}

static qserr _exicc dsk_setgeo(EXI_DATA, disk_geo_data *geo) {
   return E_SYS_UNSUPPORTED;
}

static void *qs_isoimg_list[] = { dsk_make, dsk_open, dsk_query, dsk_read,
   dsk_write, dsk_compact, dsk_mount, dsk_disk, dsk_umount, dsk_close,
   dsk_trace, dsk_setgeo };

static void _std dsk_init(void *instance, void *data) {
   diskdata  *di = (diskdata*)data;
   di->sign      = VHDD_SIGN;
   di->qsdisk    = -1;
   di->df        = 0;
   di->dfname    = 0;
   di->selfptr   = instance;
   di->einfo     = 0;
   di->startpos  = 0;
   di->cache     = 0;
   /// force mutex for every instance!
   exi_mtsafe(instance, 1);
}

static void _std dsk_done(void *instance, void *data) {
   instance_void(diskdata,di);

   if (di->qsdisk>=0) dsk_umount(data, 0);
   if (di->einfo) DELETE(di->einfo);
   if (di->df) dsk_close(data, 0);

   memset(di, 0, sizeof(diskdata));
}

u32t init_cddisk(void) {
   // something forgotten! interface part is not match to implementation
   if (sizeof(_qs_dyndisk)!=sizeof(qs_isoimg_list)) {
      log_printf(VHDD_VHD_I ": function list mismatch!\n");
      return 0;
   }
   return exi_register(VHDD_VHD_I, qs_isoimg_list, sizeof(qs_isoimg_list)/
      sizeof(void*), sizeof(diskdata), 0, dsk_init, dsk_done, 0);
}
