/*++

Copyright (c) 2004 - 2010, Intel Corporation. All rights reserved.<BR>
This program and the accompanying materials
are licensed and made available under the terms and conditions of the BSD License
which accompanies this distribution.  The full text of the license may be found at
http://opensource.org/licenses/bsd-license.php

THE PROGRAM IS DISTRIBUTED UNDER THE BSD LICENSE ON AN "AS IS" BASIS,
WITHOUT WARRANTIES OR REPRESENTATIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED.

Module Name:

  fwimage.c

Abstract:

  Converts a pe32/pe32+ image to an FW image type

--*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "qspefmt.h"

#define UTILITY_NAME      "FwImage"
#define UTILITY_VERSION   "v1.0f1"

#define STATUS_SUCCESS    0
#define STATUS_WARNING    1
#define STATUS_ERROR      2

typedef union {
   IMAGE_NT_HEADERS32 PeHeader32;
   IMAGE_NT_HEADERS64 PeHeader64;
} PE_HEADER;

void Error(const char *FName, const char *Text) {
   if (FName)
      printf("%s : %s\n", FName, Text);
   else
      printf("Error : %s\n", Text);
}


void Usage(void) {
   int         Index;
   const char  *Str[] = {
      UTILITY_NAME" "UTILITY_VERSION" - Intel Firmware Image Utility",
      "  Copyright (C), 2004 - 2008 Intel Corporation",
      "",
      "Usage:",
      "  "UTILITY_NAME" [OPTION]... FWTYPE SOURCE [DEST]",
      "Description:",
      "  Converts a pe32/pe32+ SOURCE to DEST with FWTYPE image type.",
      "Options:",
      "  FWTYPE        Can be one of APPLICATION, BS_DRIVER, RT_DRIVER, SAL_RT_DRIVER,",
      "                COMBINED_PEIM_DRIVER, SECURITY_CORE, PEI_CORE, PE32_PEIM and",
      "                RELOCATABLE_PEIM",
      "  -t time-date  Add Time Stamp for output image",
      "  -e            Not clear ExceptionTable for output image",
      "  -i            Not clear ImportTable for output image",
      "  -r            Not strip zero pending of .reloc for output image",
      NULL
   };
   for (Index = 0; Str[Index] != NULL; Index++)
      fprintf (stdout, "%s\n", Str[Index]);
}

static int FReadFile(FILE *in, void **Buffer, u32t *Length) {
   fseek (in, 0, SEEK_END);
   *Length = ftell(in);
   *Buffer = malloc(*Length);
   fseek (in, 0, SEEK_SET);
   fread (*Buffer, *Length, 1, in);
   return STATUS_SUCCESS;
}

static int FWriteFile(FILE *out, void *Buffer, u32t Length) {
   fseek (out, 0, SEEK_SET);
   fwrite (Buffer, Length, 1, out);
   if ((u32t)ftell(out) != Length) {
      Error(0, "write error");
      return STATUS_ERROR;
   }
   free (Buffer);
   return STATUS_SUCCESS;
}

void ZeroTable(u8t *FileBuf, IMAGE_DOS_HEADER *DosHdr, PE_HEADER *PeHdr,
               u32t Table, const char *SectioName, int ZeroData)
{
   u32t                   PdataSize;
   u32t                 PdataOffset;
   u32t                PdataRVASize;
   u32t                    PdataRVA;
   u32t                   SecOffset;
   u16t                   SecNumber;
   u32t                 SecNameSize;
   IMAGE_SECTION_HEADER    *Section;

   PdataSize    = 0;
   PdataOffset  = 0;
   PdataRVASize = 0;
   PdataRVA     = 0;
   SecOffset    = 0;

   // Search .pdata section
   if (PeHdr->PeHeader32.OptHeader.Magic == IMAGE_NT_OPTIONAL_HDR32_MAGIC) {
      if ((PeHdr->PeHeader32.OptHeader.NumberOfRvaAndSizes > Table) &&
          (PeHdr->PeHeader32.OptHeader.DataDirectory[Table].VirtualAddress != 0) &&
          (PeHdr->PeHeader32.OptHeader.DataDirectory[Table].Size != 0)) {

         PdataRVA     = PeHdr->PeHeader32.OptHeader.DataDirectory[Table].VirtualAddress;
         PdataRVASize = PeHdr->PeHeader32.OptHeader.DataDirectory[Table].Size;

         PeHdr->PeHeader32.OptHeader.DataDirectory[Table].VirtualAddress = 0;
         PeHdr->PeHeader32.OptHeader.DataDirectory[Table].Size = 0;

         SecOffset = sizeof(PeHdr->PeHeader32);
      }
   } else {
      if ((PeHdr->PeHeader64.OptHeader.NumberOfRvaAndSizes > Table) &&
          (PeHdr->PeHeader64.OptHeader.DataDirectory[Table].VirtualAddress != 0) &&
          (PeHdr->PeHeader64.OptHeader.DataDirectory[Table].Size != 0)) {

         PdataRVA     = PeHdr->PeHeader64.OptHeader.DataDirectory[Table].VirtualAddress;
         PdataRVASize = PeHdr->PeHeader64.OptHeader.DataDirectory[Table].Size;

         PeHdr->PeHeader64.OptHeader.DataDirectory[Table].VirtualAddress = 0;
         PeHdr->PeHeader64.OptHeader.DataDirectory[Table].Size = 0;

         SecOffset = sizeof(PeHdr->PeHeader64);
      }
   }

   if ((PdataRVASize != 0) && (PdataRVA != 0)) {
      SecNumber   = PeHdr->PeHeader32.FileHeader.NumberOfSections;
      SecNameSize = sizeof(Section->Name);
      while (SecNumber > 0) {
         Section = (IMAGE_SECTION_HEADER*) &FileBuf[DosHdr->e_lfanew + SecOffset];
         if (strcmp(Section->Name, SectioName) == 0) {
           // Zero .pdata Section Header Name
           memset(FileBuf + DosHdr->e_lfanew + SecOffset, 0,
                  SecNameSize);
           // Zero .pdata Secton raw data
           if (ZeroData) {
              PdataOffset = Section->PointerToRawData;
              PdataSize   = Section->SizeOfRawData;
              memset(FileBuf + PdataOffset, 0, PdataSize);
           }
           break;
         }
         SecNumber--;
         SecOffset += sizeof(IMAGE_SECTION_HEADER);
      }
   }
   return;
}

void StripZeroPendingReloc(u8t *FileBuf, u32t *FileLength,
                           IMAGE_DOS_HEADER *DosHdr, PE_HEADER *PeHdr)
{
   IMAGE_OPTIONAL_HEADER32    *Opt32;
   IMAGE_OPTIONAL_HEADER64    *Opt64;
   IMAGE_SECTION_HEADER      *SecHdr;
   u32t                    AlRelSize;
   u32t                        Index;

   if (PeHdr->PeHeader32.OptHeader.Magic == IMAGE_NT_OPTIONAL_HDR32_MAGIC) {
      Opt32 = &PeHdr->PeHeader32.OptHeader;
      if ((Opt32->NumberOfRvaAndSizes > IMAGE_DIRECTORY_ENTRY_BASERELOC) &&
          (Opt32->DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size != 0))
      {
         SecHdr = (IMAGE_SECTION_HEADER*)(FileBuf + DosHdr->e_lfanew +
                     sizeof(u32t) + sizeof(IMAGE_FILE_HEADER) +
                     PeHdr->PeHeader32.FileHeader.SizeOfOptionalHeader);

         for (Index = 0; Index < PeHdr->PeHeader32.FileHeader.NumberOfSections; Index++, SecHdr++) {
            /* Look for the Section Header that starts as the same virtual
               address as the Base Relocation Data Directory */
            if (strcmp(SecHdr->Name, ".reloc") == 0) {
               SecHdr->Misc.VirtualSize = Opt32->DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size;

               AlRelSize = (Opt32->DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size +
                            Opt32->FileAlignment - 1) & (~(Opt32->FileAlignment - 1));
               /* Check to see if there is zero padding at the end of the base
                  relocations */
               if (AlRelSize < SecHdr->SizeOfRawData) {
                  /* Check to see if the base relocations are at the end of
                     the file */
                  if (SecHdr->PointerToRawData + SecHdr->SizeOfRawData == Opt32->SizeOfImage) {
                     /* All the required conditions are met to strip the zero
                        padding of the end of the base relocations section */
                     Opt32->SizeOfImage           -= (SecHdr->SizeOfRawData - AlRelSize);
                     Opt32->SizeOfInitializedData -= (SecHdr->SizeOfRawData - AlRelSize);
                     SecHdr->SizeOfRawData         = AlRelSize;
                     *FileLength                   = Opt32->SizeOfImage;
                  }
               }
            }
         }
      }
   } else {
      Opt64 = (IMAGE_OPTIONAL_HEADER64*)&PeHdr->PeHeader64.OptHeader;
      if ((Opt64->NumberOfRvaAndSizes > IMAGE_DIRECTORY_ENTRY_BASERELOC) &&
           (Opt64->DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size != 0))
      {
         SecHdr = (IMAGE_SECTION_HEADER*)(FileBuf + DosHdr->e_lfanew +
                     sizeof(u32t) + sizeof(IMAGE_FILE_HEADER) +
                     PeHdr->PeHeader64.FileHeader.SizeOfOptionalHeader);

         for (Index = 0; Index < PeHdr->PeHeader64.FileHeader.NumberOfSections; Index++, SecHdr++) {
            /* Look for the Section Header that starts as the same virtual
               address as the Base Relocation Data Directory */
            if (strcmp(SecHdr->Name, ".reloc") == 0) {
               SecHdr->Misc.VirtualSize = Opt64->DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size;

               AlRelSize = (Opt64->DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size +
                            Opt64->FileAlignment - 1) & (~(Opt64->FileAlignment - 1));
               /* Check to see if there is zero padding at the end of the base
                  relocations */
               if (AlRelSize < SecHdr->SizeOfRawData) {
                  /* Check to see if the base relocations are at the end of
                     the file */
                  if (SecHdr->PointerToRawData + SecHdr->SizeOfRawData == Opt64->SizeOfImage) {
                     /* All the required conditions are met to strip the zero
                        padding of the end of the base relocations section */
                     Opt64->SizeOfImage           -= (SecHdr->SizeOfRawData - AlRelSize);
                     Opt64->SizeOfInitializedData -= (SecHdr->SizeOfRawData - AlRelSize);
                     SecHdr->SizeOfRawData         = AlRelSize;
                     *FileLength                   = Opt64->SizeOfImage;
                  }
               }
            }
         }
      }
   }
}

int main(int argc, char *argv[])  {
   u32t                                    Type = 0;
   u8t*                                     Ext = 0;
   u8t*                                       p;
   u8t*                                      pe;
   u8t*                            OutImageName = NULL;
   u8t                             outname[500];
   FILE                                   *fpIn;
   FILE                                  *fpOut;
   IMAGE_DOS_HEADER                     *DosHdr;
   PE_HEADER                             *PeHdr;
   time_t                             TimeStamp = 0;
   struct tm                         TimeStruct;
//   IMAGE_DOS_HEADER                BackupDosHdr;
   u32t                                   Index;
   int                         TimeStampPresent = 0,
                            ClearExceptionTable = 1,
                               ClearImportTable = 1,
                                StripZeroRelocs = 1;
   u8t                              *FileBuf;
   u32t                              FileLength;
   IMAGE_OPTIONAL_HEADER32               *Opt32;
   IMAGE_OPTIONAL_HEADER64               *Opt64;

   /* dixie: well, w/o loop, order of options was forced to be -t -e -r,
             well done Intel! */
   while (1) {
      /* Look for -t time-date option first. If the time is "0", then
         skip it */
      if (argc>2 && !strcmp (argv[1], "-t")) {
         TimeStampPresent = 1;
         if (strcmp (argv[2], "0") != 0) {
            // Convert the string to a value
            memset ((char *) &TimeStruct, 0, sizeof (TimeStruct));
            if (sscanf(
                argv[2], "%d/%d/%d,%d:%d:%d",
                &TimeStruct.tm_mon,   /* months since January - [0,11] */
                &TimeStruct.tm_mday,  /* day of the month - [1,31] */
                &TimeStruct.tm_year,  /* years since 1900 */
                &TimeStruct.tm_hour,  /* hours since midnight - [0,23] */
                &TimeStruct.tm_min,   /* minutes after the hour - [0,59] */
                &TimeStruct.tm_sec    /* seconds after the minute - [0,59] */
                  ) != 6) {
              Error(argv[2], "failed to convert to mm/dd/yyyy,hh:mm:ss format");
              return STATUS_ERROR;
            }
            // Now fixup some of the fields
            TimeStruct.tm_mon--;
            TimeStruct.tm_year -= 1900;
            // Sanity-check values?
            // Convert
            TimeStamp = mktime (&TimeStruct);
            if (TimeStamp == (time_t) - 1) {
               Error(argv[2], "failed to convert time");
               return STATUS_ERROR;
            }
         }
         // Skip over the args
         argc -= 2;
         argv += 2;
      } else
      // Look for -e option.
      if (argc>1 && !strcmp (argv[1], "-e")) {
         ClearExceptionTable = 0;
         // Skip over the args
         argc -= 1;
         argv += 1;
      } else
      // Look for -e option.
      if (argc>1 && !strcmp (argv[1], "-i")) {
         ClearImportTable = 0;
         // Skip over the args
         argc -= 1;
         argv += 1;
      } else
      // Look for -r option
      if (argc>1 && !strcmp (argv[1], "-r")) {
         StripZeroRelocs = 0;
         // Skip over the args
         argc -= 1;
         argv += 1;
      } else
         break;
   }
   // Check for enough args
   if (argc < 3) {
      Usage ();
      return STATUS_ERROR;
   }
   if (argc == 4) OutImageName = argv[3];
   // Get new image type
   p = argv[1];
   if (*p == '/' || *p == '\\') {
     p += 1;
   }

   if (_stricmp (p, "app") == 0 || _stricmp (p, "APPLICATION") == 0) {
      Type  = IMAGE_SUBSYSTEM_EFI_APPLICATION;
      Ext   = ".efi";
   } else if (_stricmp (p, "bsdrv") == 0 || _stricmp (p, "BS_DRIVER") == 0) {
      Type  = IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER;
      Ext   = ".efi";
   } else if (_stricmp (p, "rtdrv") == 0 || _stricmp (p, "RT_DRIVER") == 0) {
      Type  = IMAGE_SUBSYSTEM_EFI_RUNTIME_DRIVER;
      Ext   = ".efi";
   } else if (_stricmp (p, "rtdrv") == 0 || _stricmp (p, "SAL_RT_DRIVER") == 0) {
      Type  = IMAGE_SUBSYSTEM_SAL_RUNTIME_DRIVER;
      Ext   = ".efi";
   } else if (_stricmp (p, "SECURITY_CORE") == 0) {
      Type  = IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER;
      Ext   = ".sec";
   } else if (_stricmp (p, "peim") == 0 ||
             _stricmp (p, "PEI_CORE") == 0 ||
             _stricmp (p, "PE32_PEIM") == 0 ||
             _stricmp (p, "RELOCATABLE_PEIM") == 0 ||
             _stricmp (p, "combined_peim_driver") == 0
            ) {
      Type  = IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER;
      Ext   = ".pei";
   } else {
      Usage ();
      return STATUS_ERROR;
   }
   // open source file
   fpIn = fopen (argv[2], "rb");
   if (!fpIn) {
      Error(argv[2], "failed to open input file for reading");
      return STATUS_ERROR;
   }
   FReadFile(fpIn, (void **)&FileBuf, &FileLength);
   // Read the dos & pe hdrs of the image
   DosHdr = (IMAGE_DOS_HEADER*) FileBuf;
   if (DosHdr->e_magic != IMAGE_DOS_SIGNATURE) {
      Error(argv[2], "DOS header signature not found in source image");
      fclose (fpIn);
      return STATUS_ERROR;
   }

   PeHdr = (PE_HEADER *)(FileBuf + DosHdr->e_lfanew);
   if (PeHdr->PeHeader32.Signature != IMAGE_NT_SIGNATURE) {
      Error(argv[2], "PE header signature not found in source image");
      fclose (fpIn);
      return STATUS_ERROR;
   }
   // open output file
   strcpy (outname, argv[2]);
   pe = NULL;
   for (p = outname; *p; p++)
      if (*p == '.') pe = p;
   if (!pe) pe = p;

   strcpy (pe, Ext);

   if (!OutImageName) OutImageName = outname;

   fpOut = fopen (OutImageName, "w+b");
   if (!fpOut) {
      Error(OutImageName, "could not open output file for writing");
      fclose (fpIn);
      return STATUS_ERROR;
   }
#if 0
   // Zero all unused fields of the DOS header
   memcpy (&BackupDosHdr, DosHdr, sizeof(IMAGE_DOS_HEADER));
   memset (DosHdr, 0, sizeof(IMAGE_DOS_HEADER));
   DosHdr->e_magic  = BackupDosHdr.e_magic;
   DosHdr->e_lfanew = BackupDosHdr.e_lfanew;
   for (Index = sizeof(IMAGE_DOS_HEADER); Index<DosHdr->e_lfanew; Index++)
      FileBuf[Index] = (u8t) DosHdr->e_cp;
#endif

   // Modify some fields in the PE header

   // TimeDateStamp's offset is fixed for PE32/32+
   if (TimeStampPresent)
      PeHdr->PeHeader32.FileHeader.TimeDateStamp = (u32t)TimeStamp;

   // PE32/32+ has different optional header layout
   // Determine format is PE32 or PE32+ before modification
   if (PeHdr->PeHeader32.OptHeader.Magic == IMAGE_NT_OPTIONAL_HDR32_MAGIC) {
      // PE32 image
      Opt32 = &PeHdr->PeHeader32.OptHeader;

      Opt32->MajorLinkerVersion          = 0;
      Opt32->MinorLinkerVersion          = 0;
      Opt32->MajorOperatingSystemVersion = 0;
      Opt32->MinorOperatingSystemVersion = 0;
      Opt32->MajorImageVersion           = 0;
      Opt32->MinorImageVersion           = 0;
      Opt32->MajorSubsystemVersion       = 0;
      Opt32->MinorSubsystemVersion       = 0;
      Opt32->Win32VersionValue           = 0;
      Opt32->CheckSum                    = 0;
      Opt32->SizeOfStackReserve          = 0;
      Opt32->SizeOfStackCommit           = 0;
      Opt32->SizeOfHeapReserve           = 0;
      Opt32->SizeOfHeapCommit            = 0;
      Opt32->Subsystem                   = (u16t)Type;

      // Strip zero padding at the end of the .reloc section
      if (StripZeroRelocs)
        StripZeroPendingReloc(FileBuf, &FileLength, DosHdr, PeHdr);
   } else if (PeHdr->PeHeader32.OptHeader.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC) {
      // PE32+ image
      Opt64 = &PeHdr->PeHeader64.OptHeader;

      Opt64->MajorLinkerVersion          = 0;
      Opt64->MinorLinkerVersion          = 0;
      Opt64->MajorOperatingSystemVersion = 0;
      Opt64->MinorOperatingSystemVersion = 0;
      Opt64->MajorImageVersion           = 0;
      Opt64->MinorImageVersion           = 0;
      Opt64->MajorSubsystemVersion       = 0;
      Opt64->MinorSubsystemVersion       = 0;
      Opt64->Win32VersionValue           = 0;
      Opt64->CheckSum                    = 0;
      Opt64->SizeOfStackReserve          = 0;
      Opt64->SizeOfStackCommit           = 0;
      Opt64->SizeOfHeapReserve           = 0;
      Opt64->SizeOfHeapCommit            = 0;
      Opt64->Subsystem                   = (u16t)Type;

      // Strip zero padding at the end of the .reloc section
      if (StripZeroRelocs)
         StripZeroPendingReloc(FileBuf, &FileLength, DosHdr, PeHdr);
   } else {
      Error(argv[2], "Unsupported PE image");
      fclose (fpIn);
      fclose (fpOut);
      return STATUS_ERROR;
   }

   // Zero PDATA section for smaller binary size after compression
   if (ClearExceptionTable)
      ZeroTable (FileBuf, DosHdr, PeHdr, IMAGE_DIRECTORY_ENTRY_EXCEPTION, ".pdata", 1);
   // Remove import section (but leave raw data as is)
   if (ClearImportTable)
      ZeroTable (FileBuf, DosHdr, PeHdr, IMAGE_DIRECTORY_ENTRY_IMPORT, ".idata", 0);

   FWriteFile (fpOut, FileBuf, FileLength);

   // Done
   fclose (fpIn);
   fclose (fpOut);

   return STATUS_SUCCESS;
}
