//
// QSINIT EFI loader
// log & minor runtime
//
#include "qsloader.h"
#include "efilib.h"

typedef void (*print_outf)(int ch, void *stream);
int _prt_common(void *fp, const char *fmt, ptrdiff_t *arg, print_outf outc);
int _snprint(char *buf, u32t count, const char *fmt, ptrdiff_t *argp);

#define LOGIF_LEVELMASK 0x0003 ///< mask for priority value
#define LOGIF_SECOND    0x0004 ///< 1 second bit for dostime
#define LOGIF_DELAY     0x0010 ///< entry was delayed
#define LOGIF_REALMODE  0x0020 ///< entry was created in real mode
#define LOGIF_UEFICODE  0x0040 ///< entry was created in 64-bit code

int log_it(int level, const char *fmt, ...) {
   if (!Header32) return 0; else {
      char      *cp = (char*)(u64t)Header32->logrmbuf,
             buffer[128],
                *cb = buffer;
      u32t      pos = Header32->logrmpos;
      if (!cp || pos>=LOGBUF_SIZE-5-4) return 0;

      _snprint(buffer, sizeof(buffer), fmt, (ptrdiff_t*)&fmt + 1);

      // level 0..3 and LOGIF_DELAY+LOGIF_UEFICODE flags
      level = level&LOGIF_LEVELMASK|LOGIF_DELAY+LOGIF_UEFICODE;
      /* if there is no EOL at the end of last entry - appends this line to it,
         else starts a new entry */
      if (!pos || cp[pos-1]=='\n') {
         u64t time = tm_getdate();
         if (pos) pos++;
         if (time&1) level|=LOGIF_SECOND;
         cp[pos++] = (u8t)level;
         /* save dos time.
            this is mandatory action, else log_flush() on exiting from 64-bit
            part will ask for current time */
         *(u32t*)(cp+pos) = (u32t)(time>>1); pos+=4;
      }
      while (pos<LOGBUF_SIZE-1)
         if (*cb=='\r') cb++; else
            if ((cp[pos] = *cb++)==0) break; else pos++;
      // logrmpos must point to the term NULL in the last string!
      Header32->logrmpos = (u16t)pos;
   }
   return 1;
}

/// stricmp for A..Z chars
int RtStrICmp(const CHAR16 *s1, const CHAR16 *s2) {
   if (!s1&&!s2) return 0;
   if (!s1) return -1;
   if (!s2) return 1;
   while (1) {
      CHAR16 ch1 = *s1++,
             ch2 = *s2++;
      if (!ch1 && !ch2) return 0;
      if (!ch1) return -1;
      if (!ch2) return 1;
      if (ch1>='a' && ch1<='z') ch1-=0x20;
      if (ch2>='a' && ch2<='z') ch2-=0x20;
      if (ch1<ch2) return -1;
      if (ch1>ch2) return 1;
   }
}

/// strnicmp for A..Z chars
int RtStrNICmp(const CHAR16 *s1, const CHAR16 *s2, UINTN len) {
   if (!s1&&!s2) return 0;
   if (!s1) return -1;
   if (!s2) return 1;
   while (len--) {
      CHAR16 ch1 = *s1++,
             ch2 = *s2++;
      if (!ch1 && !ch2) return 0;
      if (!ch1) return -1;
      if (!ch2) return 1;
      if (ch1>='a' && ch1<='z') ch1-=0x20;
      if (ch2>='a' && ch2<='z') ch2-=0x20;
      if (ch1<ch2) return -1;
      if (ch1>ch2) return 1;
   }
   return 0;
}

int strnicmp(const char *s1, const char *s2, u32t len) {
   if (!s1&&!s2) return 0;
   if (!s1) return -1;
   if (!s2) return 1;
   while (len--) {
      char ch1 = *s1++,
           ch2 = *s2++;
      if (!ch1 && !ch2) return 0;
      if (!ch1) return -1;
      if (!ch2) return 1;
      if (ch1>='a' && ch1<='z') ch1-=0x20;
      if (ch2>='a' && ch2<='z') ch2-=0x20;
      if (ch1<ch2) return -1;
      if (ch1>ch2) return 1;
   }
   return 0;
}

CHAR16* wcscat(CHAR16 *dst, const CHAR16 *src) {
   wcscpy(dst+wcslen(dst), src);
   return dst;
}

void atow(const char *src, CHAR16 *dst, UINTN maxlen) {
   UINTN  len = strlen(src), ii;
   if (maxlen && len>=maxlen) len = maxlen - 1;

   for (ii=0; ii<len; ii++) dst[ii] = src[ii];
   dst[len] = 0;
}

CHAR16 *strto16(const char *src, UINTN maxlen) {
   CHAR16 *res;
   UINTN   len = strlen(src), ii;
   if (maxlen && len>maxlen) len = maxlen;

   res = (CHAR16*)AllocatePool(len+1);
   if (!res) return 0;

   for (ii=0; ii<len; ii++) res[ii] = src[ii];
   res[len] = 0;
   return res;
}

u16t ff_convert(u16t chr, unsigned int dir) {
   if (chr>=0x20 && chr<0x7F) return chr;
   return 0;
}

static void vioprn(int ch, void *stream) {
   vio_charout(ch);
}

int printf(const char *fmt, ...) {
   return _prt_common(0, fmt, (ptrdiff_t*)&fmt + 1, vioprn);
}

void GuidToStr(char *buf, EFI_GUID *g) {
   snprintf(buf, 38, "%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X",
      g->Data1, g->Data2, g->Data3, g->Data4[0], g->Data4[1], g->Data4[2],
         g->Data4[3], g->Data4[4], g->Data4[5], g->Data4[6], g->Data4[7]);
}

void DumpHandleProtocols(const char *info, EFI_HANDLE *eh) {
   EFI_STATUS    rc;
   EFI_GUID    **ga;
   UINTN      count;

   rc = BS->ProtocolsPerHandle(eh, &ga, &count);
   if (!EFI_ERROR(rc)) {
      EFI_DEVICE_PATH  *dp;
      u32t            pidx;

      log_it(3, "%s -> handle %LX: %u protocols\n", info, eh, count);
      dp = DevicePathFromHandle(eh);
      if (dp) {
         char buf[1024];
         DevPathToStr(buf, sizeof(buf), dp);
         log_it(3, "device path: %s\n", buf);
      }
      for (pidx=0; pidx<count; pidx++) {
         char  gstr[64];
         GuidToStr(gstr, ga[pidx]);
         log_it(3, "%u: %s\n", pidx, gstr);
      }

      BS->FreePool(ga);
   } else {
      log_it(3, "%s -> handle %LX: %LX\n", info, eh, rc);
   }
}
