#include <qsbase.h>
#include <qsdm.h>
#include "vioext.h"
#include "stdlib.h"

#define HEADER    "Init GPT"

struct {
   u32t       disk;
   int       empty;
   u64t       size;
} dsk[QDSK_DISKMASK+1];

void showhelp(void) {
   printf("Initialize empty disk to GPT\n"
          "   -all  - allow all disks (no 2Tb limit)\n");
   exit(1);
}

int main(int argc, char *argv[]) {
   u32t  ndsk, ii, fcnt,
         anysize = 0;
   qserr  err;

   for (ii=1; ii<argc; ii++) {
      if (argv[ii][0]=='/') argv[ii][0] = '-';
      if (!strcmp(argv[ii],"-?")) showhelp(); else
         if (!stricmp(argv[ii],"-all")) anysize = 1;
            else showhelp();
   }

   ndsk = hlp_diskcount(0);

   // enumerate disks
   for (ii=0, fcnt=0; ii<ndsk; ii++) {
      u32t sector;
      dsk[fcnt].size  = hlp_disksize64(ii, &sector);
      dsk[fcnt].disk  = ii;
      dsk[fcnt].empty = 0;
      // deny non-512 (CD/DVD/other strange)
      if (sector!=512) continue;
      /* force data rescan
         big floppies will be denied (error E_PTE_FLOPPY) */
      err = dsk_ptrescan(ii,1);
      if (err==E_PTE_EMPTY) dsk[fcnt].empty = 1; else
         if (err) continue;
      // number of existing partitions (empty extended is not counted)
      if (dsk_partcnt(ii)>0) continue;
      // this can be empty GPT disk - deny it too
      if (dsk_isgpt(ii,-1)) continue;
      // size limit?
      if (!anysize && dsk[fcnt].size<_4GBLL) continue;
      // free to use
      fcnt++;
   }
   if (!fcnt) {
      vio_msgbox(HEADER, "No free disks found.", MSG_GRAY|MSG_OK, 0);
      return 2;
   } else {
      vio_listref  ref;
      char        *str = strdup("<10|>10|<15^"); // listbox format string
      ref.size  = sizeof(ref);
      ref.items = fcnt;
      ref.id    = 0;
      ref.subm  = 0;
      
      // build item list for the listbox
      for (ii=0; ii<fcnt; ii++) {
         char *line = sprintf_dyn("Disk %u|%s|%s", dsk[ii].disk,
            dsk_formatsize(512, dsk[ii].size, 8, 0),
               dsk[ii].empty ? "clean" : "MBR (empty)");

         str = strcat_dyn(str,line);
         if (ii<fcnt-1) str = strcat_dyn(str,"^");
         free(line);
      }
      ref.text = str_split(str, "^");
      free(str);

      ii = vio_showlist(HEADER, &ref, MSG_BLUE, 0);
      free(ref.text);
      // action!
      if (ii--) {
         u32t disk = dsk[ii].disk, rc;

         str = sprintf_dyn("Create GPT tables on the empty disk (%s)?",
            dsk_formatsize(512, dsk[ii].size, 8, 0));
         rc  = vio_msgbox(HEADER, str, MSG_GRAY|MSG_YESNO, 0);
         free(str);

         if (rc==MRES_YES) {
            // wipe MBR (empty extended can be here, it will stop dsk_gptinit())
            dsk_newmbr(disk, DSKBR_CLEARALL);
            // wipe MBR area (to delete any possible LVM data)
            err = dsk_emptysector(disk, 1, 255);
            // create it finally
            if (!err) err = dsk_gptinit(disk);

            if (err) {
               if (err==E_PTE_EMPTY)
                  // should never occur
                  str = strdup("Disk is not empty (internal error).");
               else
                  str = cmd_shellerrmsg(EMSG_QS, err);

               vio_msgbox(HEADER, str, MSG_RED|MSG_OK, 0);
               free(str);
               return 3;
            }  else {
               // check that it really GPT now (must be)
               if (dsk_isgpt(disk,-1) != 1) {
                  vio_msgbox(HEADER, "Disk initialization error!", MSG_RED|MSG_OK, 0);
                  return 4;
               } else
                  vio_msgbox(HEADER, "Disk initialized.", MSG_GRAY|MSG_OK, 0);
            }
         }
      }
   }
   return 0;
}
